/*
 * CollabModalAlgorithm.java
 * 
 * Version 1.0
 *
 * 06/04/2004
 * 
 * Copyright (c) 2004 Federal Universtity of Pernambuco. All rights reserved.
 */
package feers.recommender.algorithms.symbolic;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.recommender.algorithms.CFAlgorithm;
import org.recommender.algorithms.CFNotImplementedException;
import org.recommender.algorithms.CompPrediction;
import org.recommender.algorithms.SimilarityWeight;
import org.recommender.server.DataManager;
import org.recommender.server.ItemPrediction;
import org.recommender.server.ItemRating;
import org.recommender.server.Stats;
import org.recommender.server.SystemProperties;
import org.recommender.server.SystemUtils;
import org.recommender.server.UserInfo;

import feers.recommender.server.DataItemManager;
import feers.recommender.server.ItemContent;

/**
 * Describes an implementation of the CFAlgorithm interface
 * to make hybrid filtering method. This method combines the 
 * Collaborative Simple Pearson algorithm with the Content 
 * based Filtering based on the Modal Symbolic method 
 * implemented in the ModalSymbolicAlgorithm.
 * 
 * @see feers.recommender.algortihms.symbolic.ModalSymbolicAlgorithm
 * @author <a href="mailto:bldb@cin.ufpe.br">Byron L. D. Bezerra</a>
 */
public class CollabModalAlgorithm implements CFAlgorithm {

	/**
	 * Database tools for this algorithm
	 */
	private DataManager dataMgr;
	private DataItemManager dataItemMgr;

	/**
	 * This attribute store the users profiles in cache in order
	 * to optimize the algorithm.
	 */
	private UserProfile[] cacheOfUserProfiles;

	/**
	 * Represents the type of each sub-profile.
	 */
	private ModalSymbolicObjectRelevance[] relevances;

	///////////// Parameters /////////////////
	// DL Some explanations are ambiguous or incomplete.  Some may change...

	/**
	 * Minimum number of neighbors required.  If we have fewer than
	 * this number of neighbors, return ERR_RATING rather than a
	 * prediction.
	 */
	private int MIN_NEIGHBORHOOD_SIZE;

	/**
	 * Maximum number of neighbors to use for generating a prediction.
	 */
	private int MAX_NEIGHBORHOOD_SIZE;

	/**
	 * Maximum ratio of possible neighbors to use.  i.e., if this is set
	 * to 0.1, then we'll use no more than 10% of the users with which
	 * we can compute a correlation.
	 */
	private float NEIGHBORHOOD_SIZE_RATIO;

	/**
	 * Minimum number of items in common for a correlation to be valid.
	 *
	 * (Generally left at 2.)
	 */
	//private int MIN_COMMON_ITEMS;

	/**
	 * Ratio of common items required for a correlation to be valid.
	 * Superceded by MIN_COMMON_ITEMS.
	 *
	 * (Generally unused; not found to be helpful.)
	 */
	//private float HORT_RATIO;

	/**
	 * Maximum number of items to consider for weighting a correlation
	 * based on the number of items.  Correlations benefit from being
	 * based on a larger number of items up to this number.
	 */
	//private int MAX_COMMON_ITEMS;

	/**
	 * Neighbor weight threshold.  Neighbors with weights below this value
	 * will not be used.
	 */
	//private float MIN_CORRELATION;

	/**
	 * If false, we'll compute the active user's similarity to each neighbor
	 * who has rated the target item whenever a rating is requested.  This
	 * is the default behavior.  When this option is true, we instead
	 * compute the active user's similarity to *all* users.  In most cases,
	 * this is slower.  However, it also allows us to reuse the similarity
	 * computation if we're requesting a lot of predictions in a row for
	 * the same user, as is the case for most experiments.
	 */
	private boolean CACHE_LAST_NEIGHBORS;

	/**
	 * If is true we load all user profiles in cache.
	 */
	private boolean CACHE_ALL_USER_PROFILES;

	private boolean USE_REPLICATION_OR_WEIGHT;

	/* Keep track of the last user whose neighbors were requested.
	 * If the same user is requested twice in a row, we can save time
	 * with this computation.
	 */
	private static int lastActiveUser = -1;
	private static List lastNeighbors = null;

	private long Time = 0;
	private long NumPredictions = 0;
	private Stats timeStats = new Stats();

	public CollabModalAlgorithm(DataManager dataMgr) {
		this(dataMgr, new DataItemManager(0, dataMgr.getMaxItem()));
	}

	public CollabModalAlgorithm(
		DataManager dataMgr,
		DataItemManager dataItemMgr) {
		this.dataMgr = dataMgr;
		this.dataItemMgr = dataItemMgr;
		this.cacheOfUserProfiles = new UserProfile[dataMgr.getMaxUser() + 1];

		// Get parameters
		String prefix = "CFServer.alg.collabmodal.";
		MIN_NEIGHBORHOOD_SIZE =
			SystemProperties.getInt(prefix + "minNeighborhoodSize", 1);
		MAX_NEIGHBORHOOD_SIZE =
			SystemProperties.getInt(prefix + "maxNeighborhoodSize", 50);
		NEIGHBORHOOD_SIZE_RATIO =
			SystemProperties.getFloat(prefix + "neighborhoodSizeRatio", 1);
		// HORT_RATIO = SystemProperties.getFloat(prefix + "hortRatio", 1);
		//		MIN_COMMON_ITEMS =
		//			SystemProperties.getInt(prefix + "minCommonItems", 2);
		//		MAX_COMMON_ITEMS =
		//			SystemProperties.getInt(prefix + "maxCommonItems", 60);
		//		MIN_CORRELATION =
		//			SystemProperties.getFloat(prefix + "minCorrelation", 0);
		CACHE_LAST_NEIGHBORS =
			SystemProperties.getBoolean(prefix + "cacheLastNeighbors", true);

		if (dataMgr.getMaxRating() <= 1.0)
			throw new RuntimeException("This algorithm is not prepared to work with grades beetwen 0 and 1.");

		CACHE_ALL_USER_PROFILES =
			SystemProperties.getBoolean(prefix + "cacheAllUserProfiles", false);

		USE_REPLICATION_OR_WEIGHT =
			SystemProperties.getBoolean(prefix + "useReplication", false);

		// create the subprofiles keys
		// create the subprofiles keys
		// supondo que as notas sao entre 1 e 5
		if (USE_REPLICATION_OR_WEIGHT == false) {
			int numberOfClasses =
				(int) (this.dataMgr.getMaxRating()
					- this.dataMgr.getMinRating()
					+ 1);
			relevances = new ModalSymbolicObjectRelevance[numberOfClasses];
			// versao 2.2
			relevances[0] = new ModalSymbolicObjectRelevance(3, -1, 0);
			relevances[1] = new ModalSymbolicObjectRelevance(2, -1, 1);
			relevances[2] = new ModalSymbolicObjectRelevance(1, -1, 2);
			relevances[3] = new ModalSymbolicObjectRelevance(4, 1, 3);
			relevances[4] = new ModalSymbolicObjectRelevance(5, 1, 4);
		} else {
			relevances = new ModalSymbolicObjectRelevance[2];
			relevances[0] = new ModalSymbolicObjectRelevance(1, 1, 0);
			relevances[1] = new ModalSymbolicObjectRelevance(1, -1, 1);
		}

		if (CACHE_ALL_USER_PROFILES)
			loadUserProfiles();
	}

	/**
	 * Build user profiles and store in cache.
	 */
	private void loadUserProfiles() {
		// first we check if the user profiles file exists
		long startTime = System.currentTimeMillis();
		String installdir = System.getProperty("cf.path", ".");
		String filename = installdir + "/lib/" + "user_profiles.dat";
		File file = new File(filename);
		if (file.exists()) {
			try {
				// load profiles from file
				FileInputStream fin = new FileInputStream(file);
				ObjectInputStream din = new ObjectInputStream(fin);
				this.cacheOfUserProfiles = (UserProfile[]) din.readObject();
				din.close();
				fin.close();
			} catch (Exception e) {
				// could not be possible
				e.printStackTrace();
			}
		} else {
			for (int i = 0; i < this.dataMgr.getMaxUser(); i++) {
				UserInfo activeUserInfo = dataMgr.getUserInfo(i);
				if (activeUserInfo != null)
					if (activeUserInfo.getReadersLock(true))
						getUserProfile(i);
					else
						throw new RuntimeException(
							"Cannot get the readers lock for user " + i);
			}
			try {
				// save file
				file.createNewFile();
				FileOutputStream fout = new FileOutputStream(file);
				ObjectOutputStream dout = new ObjectOutputStream(fout);
				dout.writeObject(this.cacheOfUserProfiles);
				dout.flush();
				dout.close();
				fout.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		long endTime = System.currentTimeMillis();
		SystemUtils.printDebug(
			"Time to cache all user profiles = "
				+ (endTime - startTime)
				+ " ms");
	}

	/* (non-Javadoc)
	 * @see org.recommender.algorithms.CFAlgorithm#predictRating(int, int)
	 */
	public ItemPrediction predictRating(int activeUser, int item)
		throws CFNotImplementedException {
		float numerator = 0;
		float denominator = 0;

		long startTime = System.currentTimeMillis();

		List neighborCoeffs;

		UserInfo activeUserInfo = dataMgr.getUserInfo(activeUser);
		//get the lock here, so you don't have to worry about waiting inbetween
		//your get weight calls and you insure your average matches what you weights
		// are calculated at.
		if (!activeUserInfo.getReadersLock(true))
			return new ItemPrediction(activeUser, item, dataMgr.getErrRating());

		float activeUserMean = dataMgr.getMeanRatingCached(activeUser);
		if (CACHE_LAST_NEIGHBORS) {
			neighborCoeffs = getNeighborsCached(activeUserInfo);
			activeUserInfo.releaseReadersLock();
		} else {
			/* get list of neighbors, calculate weights, and sort in
			 * reverse order, so we use the best ones first.
			 */
			//Iterator neighbors = getNeighborhood(item);
			//neighborCoeffs = getSimilarityWeights(activeUser, neighbors, item);

			neighborCoeffs =
				getSimilarityWeightsTopN(activeUserInfo, dataMgr.getAllUsers());
			activeUserInfo.releaseReadersLock();
			Collections.sort(neighborCoeffs, Collections.reverseOrder());
		}

		int numNeighbors =
			(int) (neighborCoeffs.size() * NEIGHBORHOOD_SIZE_RATIO);

		if (numNeighbors < MIN_NEIGHBORHOOD_SIZE) {
			return new ItemPrediction(activeUser, item, dataMgr.getErrRating());
		} else if (
			MAX_NEIGHBORHOOD_SIZE > 0
				&& numNeighbors > MAX_NEIGHBORHOOD_SIZE) {
			numNeighbors = MAX_NEIGHBORHOOD_SIZE;

			// now, calculate the prediction based on each neighbor
		}
		Iterator iter = neighborCoeffs.iterator();
		int neighborsUsed = 0;

		while (iter.hasNext() && neighborsUsed < numNeighbors) {
			SimilarityWeight w = (SimilarityWeight) iter.next();
			int neighbor = w.getNeighborID();
			// HACK -- Avoid correlations with ourself. (DLowd)
			if (neighbor == activeUser) {
				continue;
			}

			float rating = dataMgr.getRatingCached(neighbor, item);
			float neighborMean = dataMgr.getMeanRatingCached(neighbor);
			float weight = w.getWeight();

			if (rating == dataMgr.getErrRating()) {
				continue;
			}

			numerator += (weight * (rating - neighborMean));
			denominator += Math.abs(weight);

			// Select only the best neighbors.
			neighborsUsed++;
		}

		// start the final calc!

		float retval;
		if (denominator != 0 && neighborsUsed >= MIN_NEIGHBORHOOD_SIZE) {

			retval = activeUserMean + (numerator / denominator);

			if (retval > dataMgr.getMaxRating()) {
				retval = dataMgr.getMaxRating();
			} else if (retval < dataMgr.getMinRating()) {
				retval = dataMgr.getMinRating();
			}
		} else {
			retval = dataMgr.getErrRating();

		}

		long stopTime = System.currentTimeMillis();

		NumPredictions++;
		Time += (stopTime - startTime);
		this.timeStats.addNum(stopTime - startTime);

		ItemPrediction predictedRating =
			new ItemPrediction(activeUser, item, retval);
		return predictedRating; // return predicted rating

	}

	/** 
	 * This method was just copied from SimplePearsonAlgorithm
	 * @see org.recommender.algorithms.CFAlgorithm#getRecommendations(int, int)
	 */
	public ItemPrediction[] getRecommendations(int activeUser, int n)
		throws SQLException, CFNotImplementedException {
		ItemPrediction[] itemPredList;
		ItemPrediction[] tempPredList;
		int maxItem = dataMgr.getMaxItem();
		tempPredList = new ItemPrediction[maxItem + 1];
		float[] num = new float[maxItem + 1];
		float[] den = new float[maxItem + 1];

		for (int u = 0; u <= maxItem; u++) {
			num[u] = 0;
			den[u] = 0;
		}

		List neighborCoeffs;
		UserInfo activeUserInfo = dataMgr.getUserInfo(activeUser);

		//get the lock here, so you don't have to worry about waiting inbetween
		//your get weight calls and you insure your average matches what you weights
		// are calculated at.
		float activeUserMean = dataMgr.getMeanRatingCached(activeUser);

		if (!activeUserInfo.getReadersLock(true))
			return new ItemPrediction[0];

		if (CACHE_LAST_NEIGHBORS) {
			neighborCoeffs = getNeighborsCached(activeUserInfo);
			activeUserInfo.releaseReadersLock();
		} else {
			/* get list of neighbors, calculate weights, and sort in
			 * reverse order, so we use the best ones first.
			 */
			//Iterator neighbors = getNeighborhood(item);
			//neighborCoeffs = getSimilarityWeights(activeUser, neighbors, item);

			//    neighborCoeffs = getSimilarityWeightsTopN(activeUser, dataMgr.getAllUsers());

			neighborCoeffs =
				getSimilarityWeightsTopN(activeUserInfo, dataMgr.getAllUsers());

			activeUserInfo.releaseReadersLock();
			Collections.sort(neighborCoeffs, Collections.reverseOrder());
		}

		int numNeighbors =
			(int) (neighborCoeffs.size() * NEIGHBORHOOD_SIZE_RATIO);

		if (numNeighbors < MIN_NEIGHBORHOOD_SIZE) {
			return new ItemPrediction[0];
		}

		Iterator iter = neighborCoeffs.iterator();
		ItemRating r = null;
		float rating = 0;
		float mean = 0;

		// While there are neighbors in topNCoeff
		while (iter.hasNext()) {
			//Get new neighbor
			SimilarityWeight w = (SimilarityWeight) iter.next();
			float weight = w.getWeight();
			if (w.getNeighborID() == activeUser) {
				continue;
			}

			UserInfo neighborInfo = dataMgr.getUserInfo(w.getNeighborID());
			if (neighborInfo == null)
				continue;
			if (!neighborInfo.getReadersLock(false))
				continue;

			// Get new neighbor's average rating
			mean = neighborInfo.getUserMean();

			int[] itemIDs = neighborInfo.getRatings().ids;
			float[] ratings = neighborInfo.getRatings().ratings;
			for (int i = 0; i < neighborInfo.getRatings().getLength(); i++) {
				int index = itemIDs[i];
				num[index] += ((ratings[i] - mean) * weight);
				den[index] += weight;
			}
			neighborInfo.releaseReadersLock();

			// Get neighbor's item list
			/*Iterator i_index = dataMgr.getItemList(w.getNeighborID());
			
			while (i_index.hasNext()) { //While there are items in list
				r = (ItemRating) i_index.next();
				rating = r.getRating();
				if (rating > 0) {
					// Get neighbor's weight
					float weight = w.getWeight();
					num[r.getItemID()] += ((rating - mean) * weight);
					den[r.getItemID()] += Math.abs(weight);
				}
			}*/
		}

		tempPredList[0] = new ItemPrediction(0, 0, 0);
		int predIndex = 0;

		for (int i = 1; i < num.length; i++) {
			//Make sure you are not dividing by zero and activeUser has not already rated item
			if (num[i] > 0
				&& den[i] > 0
				&& dataMgr.getRatingCached(activeUser, i)
					== dataMgr.getErrRating()) {
				// Since the user's mean rating is constant for every prediction, it doesn't help
				// us to differentiate between items. So we use the meanItemRating instead.
				// Unfortunately, this means it could be inconsistent with the normal getPrediction()
				//tempPredList[predIndex++] = new ItemPrediction(0, i,
				//dataMgr.getMeanItemRatingCached(i) + num[i] / den[i]);
				tempPredList[predIndex++] =
					new ItemPrediction(0, i, activeUserMean + num[i] / den[i]);
			}
		}

		Arrays.sort(tempPredList, 0, predIndex, new CompPrediction(dataMgr));

		if (predIndex > n) {
			predIndex = n;
		}
		itemPredList = new ItemPrediction[predIndex];
		for (int i = 0; i < itemPredList.length; i++) {
			if (tempPredList[i].getPrediction() > dataMgr.getMaxRating()) {
				tempPredList[i] =
					new ItemPrediction(
						tempPredList[i].getUserID(),
						tempPredList[i].getItemID(),
						dataMgr.getMaxRating());
			}
			if (tempPredList[i].getPrediction() < dataMgr.getMinRating()) {
				tempPredList[i] =
					new ItemPrediction(
						tempPredList[i].getUserID(),
						tempPredList[i].getItemID(),
						dataMgr.getMinRating());

			}
			itemPredList[i] = tempPredList[i];
		}
		return itemPredList;
	}

	/* (non-Javadoc)
	 * @see org.recommender.algorithms.CFAlgorithm#getAveragePredictionTime()
	 */
	public long getAveragePredictionTime() throws CFNotImplementedException {
		return (Time / NumPredictions);
	}

	/**
	 * This method was just copied from the ModalSymbolicAlgorithm.
	 * In fact, it needs to update the cache of user neighboors too.
	 * But, it does not do at now.
	 * 
	 * @see org.recommender.algorithms.CFAlgorithm#updateUser(int)
	 */
	public void updateUser(int userID) throws CFNotImplementedException {
		UserInfo activeUserInfo = dataMgr.getUserInfo(userID);
		// get the lock here, so you don't have to worry about waiting inbetween
		//		while (!activeUserInfo.getWritersLock())
		//			try {
		//				wait(1000);
		//			} catch (InterruptedException e) {
		//			}
		// remove the user profile of cache
		cacheOfUserProfiles[userID] = null;
		// create the user profile and put it in cache
		UserProfile userProfile = getUserProfile(userID);
		//		activeUserInfo.releaseWritersLock();
	}

	/**
	 * This method was just copied from SimplePearsonAlgorithm.
	
	 * @see org.recommender.algorithms.CFAlgorithm#getRecommendationsByType(int, int, int)
	 */
	public ItemPrediction[] getRecommendationsByType(
		int activeUser,
		int n,
		int type)
		throws SQLException, CFNotImplementedException {
		ItemPrediction[] itemPredList;
		int maxItem = dataMgr.getMaxItem();

		//    ItemPrediction[] tempPredList = new ItemPrediction[maxItem + 1];
		float[] num = new float[maxItem + 1];
		float[] den = new float[maxItem + 1];

		int[] typeArr = new int[maxItem + 1]; // for storing type

		for (int u = 0; u <= maxItem; u++) {
			num[u] = 0;
			den[u] = 0;
			typeArr[u] = 0;
		}

		UserInfo activeUserInfo = dataMgr.getUserInfo(activeUser);

		Iterator typedIt = dataMgr.getCachedTypedItemLists(type);
		int size = dataMgr.getSizeOfTypedItemList(type);

		ItemPrediction[] tempPredList = new ItemPrediction[size + 1];

		while (typedIt.hasNext()) {
			int i = ((Integer) typedIt.next()).intValue();

			typeArr[i] = 1;
		}

		List neighborCoeffs;
		//get the lock here, so you don't have to worry about waiting in between
		//your get weight calls and you insure your average matches what you weights
		// are calculated at.
		if (!activeUserInfo.getReadersLock(true))
			return new ItemPrediction[0];

		//long start = System.currentTimeMillis();
		if (CACHE_LAST_NEIGHBORS) {
			neighborCoeffs = getNeighborsCached(activeUserInfo);
			activeUserInfo.releaseReadersLock();
		} else {
			/* get list of neighbors, calculate weights, and sort in
			 * reverse order, so we use the best ones first.
			 */
			//Iterator neighbors = getNeighborhood(item);
			//neighborCoeffs = getSimilarityWeights(activeUser, neighbors, item);

			neighborCoeffs =
				getSimilarityWeightsTopN(activeUserInfo, dataMgr.getAllUsers());
			activeUserInfo.releaseReadersLock();

			Collections.sort(neighborCoeffs, Collections.reverseOrder());
		}

		int numNeighbors =
			(int) (neighborCoeffs.size() * NEIGHBORHOOD_SIZE_RATIO);

		if (numNeighbors < MIN_NEIGHBORHOOD_SIZE) {
			return new ItemPrediction[0];
		}

		//long stop = System.currentTimeMillis();
		//    System.out.println("Time to select best neighbors= " + (stop - start));

		Iterator iter = neighborCoeffs.iterator();
		//start = System.currentTimeMillis();

		ItemRating r = null;
		float rating = 0;
		float mean = 0;

		// While there are neighbors less than max number
		while (iter.hasNext()) {
			//Get new neighbor
			SimilarityWeight w = (SimilarityWeight) iter.next();
			// Get new neighbor's average rating

			if (w.getNeighborID() == activeUser) {
				continue;
			}

			UserInfo neighborInfo = dataMgr.getUserInfo(w.getNeighborID());
			if (neighborInfo == null)
				continue;
			if (!neighborInfo.getReadersLock(false))
				continue;

			// mean = dataMgr.getMeanRatingCached(w.getNeighborID());
			mean = neighborInfo.getUserMean();

			// Get neighbor's item list
			/*  Iterator i_index = dataMgr.getItemList(w.getNeighborID());
			
				// Get typed item list
				//  Iterator typed_index = typedItemList.iterator();
			
				while (i_index.hasNext()) { //While there are items in list
			
					r = (ItemRating) i_index.next();
			
					rating = r.getRating();
					int item = r.getItemID();
			
					if (rating > 0 && typeArr[item] > 0) {
						// Get neighbor's weight
						float weight = w.getWeight();
						num[item] += ((rating - mean) * weight);
						den[item] += Math.abs(weight);
			
					}
			
				} // while (i_index)*/
			float weight = w.getWeight();
			int[] itemIDs = neighborInfo.getRatings().ids;
			float[] ratings = neighborInfo.getRatings().ratings;
			for (int i = 0; i < neighborInfo.getRatings().getLength(); i++) {
				int index = itemIDs[i];
				if (typeArr[index] > 0) {
					num[index] += ((ratings[i] - mean) * weight);
					den[index] += weight;
				}
			}
			neighborInfo.releaseReadersLock();

		} // while (iter)

		//stop = System.currentTimeMillis();
		//    System.out.println("Time to compute predictions=" + (stop - start));

		//start = System.currentTimeMillis();
		tempPredList[0] = new ItemPrediction(0, 0, 0);

		int predIndex = 0;
		float activeUserMean = dataMgr.getMeanRatingCached(activeUser);
		for (int i = 1; i < num.length; i++) {
			// int i = ((Integer) typedItemList.get(k)).intValue();

			//check that activeUser has not already rated item
			if (typeArr[i] > 0
				&& num[i] > 0
				&& den[i] > 0
				&& dataMgr.getRatingCached(activeUser, i)
					== dataMgr.getErrRating()) {

				/*itemPredList[predIndex++] = new ItemPrediction(0, i,
						active_user_avr + num[i]/den[i]);*/

				//Updated: now uses item mean for better performance
				//tempPredList[predIndex++] = new ItemPrediction(0, i,
				//        dataMgr.getMeanItemRatingCached(i) + num[i] / den[i]);

				tempPredList[predIndex++] =
					new ItemPrediction(0, i, activeUserMean + num[i] / den[i]);

			}

		}

		//stop = System.currentTimeMillis();
		//		  System.out.println("Time to fill pred array=" + (stop - start));

		//start = System.currentTimeMillis();
		Arrays.sort(tempPredList, 0, predIndex, new CompPrediction(dataMgr));
		//stop = System.currentTimeMillis();
		//		  System.out.println("Time to sort the array= " + (stop - start));

		if (predIndex > n) {
			predIndex = n;
		}
		itemPredList = new ItemPrediction[predIndex];

		//fill array to be returned
		for (int i = 0; i < itemPredList.length; i++) {
			if (tempPredList[i].getPrediction() > dataMgr.getMaxRating()) {
				tempPredList[i] =
					new ItemPrediction(
						tempPredList[i].getUserID(),
						tempPredList[i].getItemID(),
						dataMgr.getMaxRating());
			}
			if (tempPredList[i].getPrediction() < dataMgr.getMinRating()) {
				tempPredList[i] =
					new ItemPrediction(
						tempPredList[i].getUserID(),
						tempPredList[i].getItemID(),
						dataMgr.getMinRating());

			}

			itemPredList[i] = tempPredList[i];
		}

		return itemPredList;
	}

	/**
	 * This method was just copied from SimplePearsonAlgorithm.
	 * 
	 * @param activeUserInfo
	 * @return
	 */
	private List getNeighborsCached(UserInfo activeUserInfo) {

		if (activeUserInfo.userID == lastActiveUser) {
			return lastNeighbors;
		}

		// Compute similarity with all available users
		//			  lastNeighbors = getSimilarityWeightsTopN(activeUser,  dataMgr.getAllUsers());
		lastNeighbors =
			getSimilarityWeightsTopN(activeUserInfo, dataMgr.getAllUsers());

		// Sort similarities in reverse order, so we use the best ones first.
		Collections.sort(lastNeighbors, Collections.reverseOrder());

		lastActiveUser = activeUserInfo.userID;
		return lastNeighbors;
	}

	/**
	 * This method was just copied from SimplePearsonAlgorithm and a bit modified.
	 * 
	 * @param activeUserInfo
	 * @param neighbors
	 * @return
	 */
	private List getSimilarityWeightsTopN(
		UserInfo activeUserInfo,
		Iterator neighbors) {
		LinkedList vals = new LinkedList();
		Integer n;
		//UserInfo activeUserInfo = dataMgr.getUserInfo(activeUser);

		//long st = System.currentTimeMillis();
		while (neighbors.hasNext()) {
			n = (Integer) neighbors.next();
			if (n.intValue() == activeUserInfo.userID)
				continue;
			// SimilarityWeight w = getWeight(activeUserInfo, n.intValue(), -1);
			SimilarityWeight w = getWeight(activeUserInfo, n.intValue());
			if (w.getWeight() > 0) {
				vals.add(w);
			}
		}
		//		long et = System.currentTimeMillis();
		//		System.out.println("Time to get SimilarityWeightsTopN = " + (et-st));
		return vals;
	}

	/**
	 * This method returns the similarity between the activeUserInfo and 
	 * otherUserId based on the modal symbolic approach.
	 * 
	 * @param activeUserInfo
	 * @param otherUserId
	 * @return
	 */
	private SimilarityWeight getWeight(
		UserInfo activeUserInfo,
		int otherUserId) {
		// get the active user profile
		//long st = System.currentTimeMillis();
		UserProfile activeUserProfile = getUserProfile(activeUserInfo.userID);
		//long et = System.currentTimeMillis();
		//System.out.println("Time to get user " + activeUserInfo.userID + " = " + (et-st));
		// get the symbolic profile of other user
		//st = System.currentTimeMillis();
		UserProfile otherUserProfile = getUserProfile(otherUserId);
		//et = System.currentTimeMillis();
		//System.out.println("Time to get user " + otherUserId + " = " + (et-st));
		// calculate the similarity between both user profiles
		//st = System.currentTimeMillis();
		float similarity =
			(float) activeUserProfile.similarityTo(otherUserProfile);
		//et = System.currentTimeMillis();
		//System.out.println("Time to calculate the similarity = " + (et-st));

		// Theoretically, the weight should never be NaN, nor more than 1;
		// however, floating point error sometimes leads to strange situations.
		if (Float.isNaN(similarity)) {
			similarity = 0;
		} else if (similarity > 1) {
			similarity = 1;
		}

		return new SimilarityWeight(otherUserId, similarity);
	}

	/**
	 * This method was just copied from ModalSymbolicAlgorithm.
	 * 
	 * @param activeUserInfo
	 * @return
	 */
	private UserProfile getUserProfile(int userId) {
		// trying to get the user profile in cache
		UserProfile userProfile = null;
		if (cacheOfUserProfiles[userId] != null)
			userProfile = cacheOfUserProfiles[userId];
		else {
			UserInfo activeUserInfo = dataMgr.getUserInfo(userId);
			if (activeUserInfo != null)
				if (activeUserInfo.getReadersLock(true)) {
					int maxRating = (int) dataMgr.getMaxRating();
					int minRating = (int) dataMgr.getMinRating();
					int numberOfRates = maxRating - minRating + 1;
					int mediumRate = numberOfRates / 2;
					// creating the keys of modal symbolic objects
					// in this algorithm there is two modal objects (one positive and one negative)
					ItemToModalMapping[] itemsMapping =
						(USE_REPLICATION_OR_WEIGHT)
							? getMappingConfigurationBasedOnReplication(
								activeUserInfo,
								minRating,
								numberOfRates,
								mediumRate,
								relevances)
							: getMappingConfigurationBasedOnWeight(
								activeUserInfo,
								relevances);
					// creating the user profile
					userProfile =
						new UserProfile(userId, itemsMapping, relevances);
					// putting the user profile in cache
					cacheOfUserProfiles[userId] = userProfile;
				} else {
					throw new RuntimeException(
						"Cannot get the readers lock for user " + userId);
				}
		}
		return userProfile;
	}

	private ItemToModalMapping[] getMappingConfigurationBasedOnWeight(
		UserInfo activeUserInfo,
		ModalSymbolicObjectRelevance[] relevances) {
		// creating the mapping configuration for evaluated items
		int[] itemIds = activeUserInfo.getRatings().getIds();
		float[] ratings = activeUserInfo.getRatings().getRatings();
		ItemToModalMapping[] itemsMapping =
			new ItemToModalMapping[itemIds.length];
		for (int i = 0; i < itemIds.length; i++) {
			ItemContent itemContent = dataItemMgr.getItemContent(itemIds[i]);
			int rate = (int) ratings[i];
			itemsMapping[i] =
				new ItemToModalMapping(itemContent, relevances[rate - 1], 1);
		}
		return itemsMapping;
	}

	/*
	 * It is useful when there is just two modalSymbolicObjects 
	 * per user profile.
	 */
	private ItemToModalMapping[] getMappingConfigurationBasedOnReplication(
		UserInfo activeUserInfo,
		int minRating,
		int numberOfRates,
		int mediumRate,
		ModalSymbolicObjectRelevance[] relevances) {
		// creating the mapping configuration for evaluated items
		int[] itemIds = activeUserInfo.getRatings().getIds();
		float[] ratings = activeUserInfo.getRatings().getRatings();
		ItemToModalMapping[] itemsMapping =
			new ItemToModalMapping[itemIds.length];
		for (int i = 0; i < itemIds.length; i++) {
			ItemContent itemContent = dataItemMgr.getItemContent(itemIds[i]);
			int rate = (int) ratings[i];
			if (rate < (mediumRate + minRating)) {
				// add this item in the negative object
				int times = mediumRate + minRating - rate;
				itemsMapping[i] =
					new ItemToModalMapping(itemContent, relevances[1], times);
			} else {
				// add this item in the positive object
				if (numberOfRates % 2 == 0) {
					// quantidade par
					int times = rate + 1 - (mediumRate + minRating);
					itemsMapping[i] =
						new ItemToModalMapping(
							itemContent,
							relevances[0],
							times);
				} else {
					// quantidade impar
					if (rate > mediumRate + minRating) {
						// ignore items evaluated with medium grade
						int times = rate - (mediumRate + minRating);
						itemsMapping[i] =
							new ItemToModalMapping(
								itemContent,
								relevances[0],
								times);
					} else {
						// add items evaluated with a middle grade in negative subprofile
						itemsMapping[i] =
							new ItemToModalMapping(
								itemContent,
								relevances[1],
								1);
					}
				}
			}
		}
		return itemsMapping;
	}

	/**
	 * Return the name of this algorithm.
	 * @return String
	 */
	public String toString() {
		return "CollabModalAlgorithm.java\n";
	}
	/**
	 * @return
	 */
	public Stats getTimeStats() {
		return timeStats;
	}

}
