/*
 * ContentBoostedCollabAlgorithm.java
 * 
 * Version 1.0
 *
 * 01/05/2004
 * 
 * Copyright (c) 2004 Federal Universtity of Pernambuco. All rights reserved.
 */
package feers.recommender.algorithms.symbolic;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.lang.reflect.Constructor;
import java.sql.SQLException;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;

import org.recommender.algorithms.CFAlgorithm;
import org.recommender.algorithms.CFNotImplementedException;
import org.recommender.server.DataManager;
import org.recommender.server.ItemPrediction;
import org.recommender.server.Stats;
import org.recommender.server.SystemProperties;
import org.recommender.server.SystemUtils;
import org.recommender.server.UserInfo;

import feers.recommender.server.DataItemManager;

/**
 * This class implements a CFAlgortihm based on the hybrid filtering
 * method proposed in:
 * "Content-Boosted Collaborative Filtering for Improved Recommendations",
 * Prem Melville, Raymond J. Mooney and Ramadass Nagarajan. AAAI-2002.
 * 
 * @author <a href="mailto:bldb@cin.ufpe.br>Byron L. D. Bezerra</a>
 */
public class ContentBoostedCollabAlgorithm implements CFAlgorithm {

	/**
	 * Database tools for this algorithm
	 */
	private DataManager dataMgr;
	private DataItemManager dataItemMgr;

	/**
	 * Matrix of user evaluations and estimated ratings.
	 * The key is an user id and the values are treesets
	 * of ItemRating.
	 */
	private Hashtable ratingsMatrix;

	///////// Parameters /////////////////

	//	private long Time = 0;
	//	private long NumPredictions = 0;
	//	private Stats timeStats = new Stats();

	private String CONTENT_BASED_ALGORITHM;
	private String COLLAB_BASED_ALGORITHM;
	private String PSEUDO_RATING_MATRIX_FILENAME;

	private CFAlgorithm collabAlgorithm;
	private CFAlgorithm contentAlgorithm;

	public ContentBoostedCollabAlgorithm(DataManager dataMgr) {
		this(dataMgr, new DataItemManager(0, dataMgr.getMaxItem()));
	}

	public ContentBoostedCollabAlgorithm(
		DataManager dataMgr,
		DataItemManager dataItemMgr) {
		this.dataMgr = dataMgr;
		this.dataItemMgr = dataItemMgr;

		// Get parameters
		String prefix = "CFServer.alg.contentboostedcollab.";

		CONTENT_BASED_ALGORITHM =
			SystemProperties.getString(
				prefix + "contentalgorithm.name",
				"feers.recommender.algorithms.symbolic.ModalSymbolicAlgorithm");
		System.out.println("Content algorithm => " + CONTENT_BASED_ALGORITHM);

		COLLAB_BASED_ALGORITHM =
			SystemProperties.getString(
				prefix + "collabalgorithm.name",
				"feers.recommender.algorithms.symbolic.BoostedPearsonAlgorithm");
		System.out.println("Collab algorithm => " + COLLAB_BASED_ALGORITHM);

		PSEUDO_RATING_MATRIX_FILENAME =
			SystemProperties.getString(
				prefix + "pseudoratingmatrix.filename",
				"pseudoRatingMatrix.dat");

		if (dataMgr.getMaxRating() <= 1.0)
			throw new RuntimeException("This algorithm is not prepared to work with grades beetwen 0 and 1.");

		// creating the collaborative algorithm
		this.collabAlgorithm = this.getCollaborativeBasedAlgorithm();

		// creating the content algorithm
		this.contentAlgorithm = this.getContentBasedAlgorithm();

		// load the pseudo ratings matrix
		this.loadPseudoRatingMatrix();

	}

	/**
		 * 
		 */
	private void loadPseudoRatingMatrix() {

		SystemUtils.printDebug("Loading the pseudo rating matrix");

		// loading file
		String installdir = System.getProperty("cf.path", ".");
		String filename = installdir + "/lib/" + PSEUDO_RATING_MATRIX_FILENAME;
		File file = new File(filename);
		if (file.exists()) {
			try {
				FileInputStream fin = new FileInputStream(filename);
				ObjectInputStream in = new ObjectInputStream(fin);
				this.ratingsMatrix = (Hashtable) in.readObject();
				in.close();
				fin.close();
			} catch (Exception e) {
				e.printStackTrace();
				System.out.println("ERROR: Could not construct algorithm.");
				System.exit(0);
			}
		} else {
			computePseudoRatingMatrix(filename);
		}

		// update dataMgr with pseudo ratings
		for (Iterator iter = ratingsMatrix.keySet().iterator();
			iter.hasNext();
			) {
			Integer userId = (Integer) iter.next();
			Set items = (Set) ratingsMatrix.get(userId);
			for (Iterator iterator = items.iterator(); iterator.hasNext();) {
				ItemRatingComparable itemRating =
					(ItemRatingComparable) iterator.next();
				if (itemRating.isEstimated())
					this.dataMgr.addCachedRating(itemRating);
			}
		}

		SystemUtils.printDebug("Pseudo rating matrix loaded.");
	}

	private void computePseudoRatingMatrix(String filename) {
		this.ratingsMatrix = new Hashtable();

		int maxUserId = this.dataMgr.getMaxUser();
		for (int userId = 1; userId < maxUserId; userId++) {
			try {
				UserInfo user = this.dataMgr.loadUser(userId, false);
				computePseudoVector(user);
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}

		if (this.ratingsMatrix.size() > 0) {
			// store the ratingsMatrix in File
			try {
				FileOutputStream fout = new FileOutputStream(filename);
				ObjectOutputStream out = new ObjectOutputStream(fout);
				out.writeObject(this.ratingsMatrix);
				out.flush();
				fout.flush();
				out.close();
				fout.close();
			} catch (Exception e) {
				e.printStackTrace();
			}
			System.out.println("Pseudo rating matrix stored in file.");
		}
	}

	private void computePseudoVector(UserInfo user) {
		if (user != null) {
			HashSet ratings = new HashSet();
			this.ratingsMatrix.put(new Integer(user.userID), ratings);
			int maxItemId = this.dataMgr.getMaxItem();
			// for each user compute the pseudo vector based
			// on content based algorithm 
			for (int itemId = 1; itemId < maxItemId; itemId++) {
				ItemRatingComparable rating = null;
				if (user.getRatings().containsKey(itemId)) {
					rating =
						new ItemRatingComparable(
							user.userID,
							itemId,
							user.getRatings().getRating(itemId),
							false);
				} else {
					try {
						// compute the pseudo rating for this item
						ItemPrediction prediction =
							contentAlgorithm.predictRating(user.userID, itemId);
						rating =
							new ItemRatingComparable(
								user.userID,
								itemId,
								prediction.getPrediction(),
								true);
					} catch (CFNotImplementedException e1) {
						e1.printStackTrace();
					}
				}
				// add the rating in the users vector
				if (rating != null)
					ratings.add(rating);
			}
			// System.out.println("Computed pseudo vector for user " + user.userID + ".");
		} else {
			System.out.println(
				"User " + user.userID + " does not exists in database.");
		}
	}

	/**
	 * @return
	 */
	private CFAlgorithm getContentBasedAlgorithm() {
		CFAlgorithm newAlgObj = null;
		try {
			Class algorithm = Class.forName(this.CONTENT_BASED_ALGORITHM);
			// SystemUtils.printDebug("Using algorithm: " + algorithm.getName());

			Constructor[] cons = algorithm.getConstructors();

			SystemUtils.printDebug(
				"Constructing a new algorithm object, please wait...");

			long start = System.currentTimeMillis();

			for (int i = 0; i < cons.length; i++) {
				try {
					newAlgObj =
						(CFAlgorithm) cons[i].newInstance(
							new Object[] { dataMgr, dataItemMgr });
					break;
				} catch (Exception e10) {
				}
			}

			long stop = System.currentTimeMillis();

			SystemUtils.printDebug(
				"Finish constructing a new algorithm object, time = "
					+ (stop - start)
					+ " ms");

		} catch (Exception e) {
			// Various exceptions can be thrown by the reflect API, so we catch them all here
			//System.out.println(" class not found Exception" + e);
			e.printStackTrace();
		}

		if (newAlgObj == null) {
			// Unknown algorithm name.  Print an error and quit.
			System.out.println("ERROR: Could not construct algorithm.");
			System.exit(0);
		}
		return newAlgObj;
	}

	/**
		 * @return
		 */
	private CFAlgorithm getCollaborativeBasedAlgorithm() {
		CFAlgorithm newAlgObj = null;
		try {
			Class algorithm = Class.forName(this.COLLAB_BASED_ALGORITHM);
			// SystemUtils.printDebug("Using algorithm: " + algorithm.getName());

			Constructor[] cons = algorithm.getConstructors();

			SystemUtils.printDebug(
				"Constructing a new algorithm object, please wait...");

			long start = System.currentTimeMillis();

			for (int i = 0; i < cons.length; i++) {
				try {
					newAlgObj =
						(CFAlgorithm) cons[i].newInstance(
							new Object[] { dataMgr });
					break;
				} catch (Exception e10) {
				}
			}

			long stop = System.currentTimeMillis();

			SystemUtils.printDebug(
				"Finish constructing a new algorithm object, time = "
					+ (stop - start)
					+ " ms");

		} catch (Exception e) {
			// Various exceptions can be thrown by the reflect API, so we catch them all here
			//System.out.println(" class not found Exception" + e);
			e.printStackTrace();
		}

		if (newAlgObj == null) {
			// Unknown algorithm name.  Print an error and quit.
			System.out.println("ERROR: Could not construct algorithm.");
			System.exit(0);
		}
		return newAlgObj;
	}

	/* (non-Javadoc)
	 * @see org.recommender.algorithms.CFAlgorithm#predictRating(int, int)
	 */
	public ItemPrediction predictRating(int userID, int itemID)
		throws CFNotImplementedException {
		return this.collabAlgorithm.predictRating(userID, itemID);
	}

	/* (non-Javadoc)
	 * @see org.recommender.algorithms.CFAlgorithm#getRecommendations(int, int)
	 */
	public ItemPrediction[] getRecommendations(int activeUser, int n)
		throws SQLException, CFNotImplementedException {
		return this.collabAlgorithm.getRecommendations(activeUser, n);
	}

	/* (non-Javadoc)
	 * @see org.recommender.algorithms.CFAlgorithm#getAveragePredictionTime()
	 */
	public long getAveragePredictionTime() throws CFNotImplementedException {
		return this.collabAlgorithm.getAveragePredictionTime();
	}

	/* (non-Javadoc)
	 * @see org.recommender.algorithms.CFAlgorithm#updateUser(int)
	 */
	public void updateUser(int userID) throws CFNotImplementedException {
		UserInfo user = this.dataMgr.getUserInfo(userID);
		this.ratingsMatrix.remove(new Integer(userID));
		this.computePseudoVector(user);
		this.collabAlgorithm.updateUser(userID);
	}

	/* (non-Javadoc)
	 * @see org.recommender.algorithms.CFAlgorithm#getRecommendationsByType(int, int, int)
	 */
	public ItemPrediction[] getRecommendationsByType(
		int activeUser,
		int n,
		int type)
		throws SQLException, CFNotImplementedException {
		return this.collabAlgorithm.getRecommendationsByType(
			activeUser,
			n,
			type);
	}

	/* (non-Javadoc)
	 * @see org.recommender.algorithms.CFAlgorithm#getTimeStats()
	 */
	public Stats getTimeStats() {
		return this.collabAlgorithm.getTimeStats();
	}
	
	public String toString() {
		return "ContentBoostedCollabAlgorithm\n";
	}

	public static void main(String[] args) {
		String installdir = System.getProperty("cf.path", ".");
		String filename = installdir + "/lib/" + "CFServer.properties";
		// Load properties
		SystemProperties.loadProperties(filename);

		DataManager dataMgr = new DataManager(true, false);
		ContentBoostedCollabAlgorithm algorithm =
			new ContentBoostedCollabAlgorithm(dataMgr);
	}

}
