/*
 * ModalSymbolicAlgorith.java
 * 
 * Version 1.0
 * 
 * 28/03/2004
 *
 * Copyright (c) 2004 Federal Universtity of Pernambuco. All rights reserved.
 */
package feers.recommender.algorithms.symbolic;

import java.sql.SQLException;
import java.util.Arrays;
import java.util.HashMap;

import org.recommender.algorithms.CFAlgorithm;
import org.recommender.algorithms.CFNotImplementedException;
import org.recommender.algorithms.CompPrediction;
import org.recommender.server.DataManager;
import org.recommender.server.ItemPrediction;
import org.recommender.server.Stats;
import org.recommender.server.SystemProperties;
import org.recommender.server.UserInfo;

import feers.recommender.server.DataItemManager;
import feers.recommender.server.ItemContent;

/**
 * Describes an implementation of the CFAlgorithm interface
 * to make ContentBased predictions using the Modal Symbolic
 * method described in paper:
 * "Information Filtering Based on Modal Symbolic Objects"
 * Bezerra, B. L. D. and De Carvalho, F. A. T.
 * Accepted in Information Processing Letters in 2003, to be
 * published in 2004.
 * 
 * @author <a href="mailto:bldb@cin.ufpe.br">Byron L. D. Bezerra</a>
 */
public class ModalSymbolicAlgorithm implements CFAlgorithm {

	/**
	 * Database tools for this algorithm
	 */
	private DataManager dataMgr;
	private DataItemManager dataItemMgr;

	/**
	 * This attribute store the users profiles in cache in order
	 * to optimize the algorithm.
	 */
	private HashMap cacheOfUserProfiles;

	/**
	 * Represents the type of each sub-profile.
	 */
	private ModalSymbolicObjectRelevance[] relevances;

	private boolean USE_REPLICATION_OR_WEIGHT;

	/////////// Parameters /////////////////

	private long Time = 0;
	private long NumPredictions = 0;
	private Stats timeStats = new Stats();

	public ModalSymbolicAlgorithm(DataManager dataMgr) {
		this(dataMgr, new DataItemManager(0, dataMgr.getMaxItem()));
	}

	public ModalSymbolicAlgorithm(
		DataManager dataMgr,
		DataItemManager dataItemMgr) {
		this.dataMgr = dataMgr;
		this.dataItemMgr = dataItemMgr;
		this.cacheOfUserProfiles = new HashMap();

		// Get parameters
		String prefix = "CFServer.alg.modal.";

		USE_REPLICATION_OR_WEIGHT =
			SystemProperties.getBoolean(prefix + "useReplication", false);

		System.out.println(
			"Using replication in ModalSymbolicAlgorithm ? "
				+ USE_REPLICATION_OR_WEIGHT);

		if (dataMgr.getMaxRating() <= 1.0)
			throw new RuntimeException("This algorithm is not prepared to work with grades beetwen 0 and 1.");

		// create the subprofiles keys
		// supondo que as notas sao entre 1 e 5
		if (USE_REPLICATION_OR_WEIGHT == false) {
			int numberOfClasses =
				(int) (this.dataMgr.getMaxRating()
					- this.dataMgr.getMinRating()
					+ 1);
			relevances = new ModalSymbolicObjectRelevance[numberOfClasses];
			relevances[0] = new ModalSymbolicObjectRelevance(3, -1, 0);
			relevances[1] = new ModalSymbolicObjectRelevance(2, -1, 1);
			relevances[2] = new ModalSymbolicObjectRelevance(1, -1, 2);
			relevances[3] = new ModalSymbolicObjectRelevance(2, 1, 3);
			relevances[4] = new ModalSymbolicObjectRelevance(3, 1, 4);
		} else {
			relevances = new ModalSymbolicObjectRelevance[2];
			relevances[0] = new ModalSymbolicObjectRelevance(1, 1, 0);
			relevances[1] = new ModalSymbolicObjectRelevance(1, -1, 1);
		}
	}

	/* (non-Javadoc)
	 * @see org.recommender.algorithms.CFAlgorithm#predictRating(int, int)
	 */
	public ItemPrediction predictRating(int activeUser, int item)
		throws CFNotImplementedException {

		long startTime = System.currentTimeMillis();

		// ler o usuario e os itens que ele avaliou
		UserInfo activeUserInfo = dataMgr.getUserInfo(activeUser);
		//get the lock here, so you don't have to worry about waiting inbetween
		//your get weight calls and you insure your average matches what you weights
		// are calculated at.
		if (!activeUserInfo.getReadersLock(true))
			return new ItemPrediction(activeUser, item, dataMgr.getErrRating());

		ItemPrediction prediction =
			(USE_REPLICATION_OR_WEIGHT)
				? getPredictionBasedOnReplication(item, activeUserInfo)
				: getPredictionBasedOnWeight(item, activeUserInfo);

		long stopTime = System.currentTimeMillis();

		NumPredictions++;
		Time += (stopTime - startTime);
		timeStats.addNum(stopTime - startTime);

		return prediction;
	}

	private ItemPrediction getPredictionBasedOnReplication(
		int item,
		UserInfo activeUserInfo) {
		// build the user profile based on the modal symbolic description
		int maxRating = (int) dataMgr.getMaxRating();
		int minRating = (int) dataMgr.getMinRating();
		UserProfile userProfile = getUserProfile(activeUserInfo);
		// read the item content
		ItemContent targetItem = dataItemMgr.getItemContent(item);
		// get the similarity beetwen 0 and 1
		double similarity = userProfile.similarityTo(targetItem);
		// calculate the grade for the target item
		float grade = (float) ((maxRating - minRating) * similarity);
		ItemPrediction prediction =
			new ItemPrediction(activeUserInfo.userID, item, grade);
		return prediction;
	}

	private ItemPrediction getPredictionBasedOnWeight(
		int item,
		UserInfo activeUserInfo) {
		// build the user profile based on the modal symbolic description
		UserProfile userProfile = getUserProfile(activeUserInfo);
		// read the item content
		ItemContent targetItem = dataItemMgr.getItemContent(item);
		// predict the class for this item
		float grade = userProfile.predictClass(targetItem);
		ItemPrediction prediction =
			new ItemPrediction(activeUserInfo.userID, item, grade);
		return prediction;
	}

	private ItemPrediction getPredictionBasedOnReplication(
		int item,
		UserProfile userProfile) {
		int maxRating = (int) dataMgr.getMaxRating();
		int minRating = (int) dataMgr.getMinRating();
		// read the item content
		ItemContent targetItem = dataItemMgr.getItemContent(item);
		// get the similarity beetwen 0 and 1
		double similarity = userProfile.similarityTo(targetItem);
		// calculate the grade for the target item
		if (similarity <= 0.000001 || Double.isNaN(similarity))
			similarity = 0;
		float grade = (float) ((maxRating - minRating) * similarity);
		ItemPrediction prediction =
			new ItemPrediction(userProfile.getUserId(), item, grade);
		return prediction;
	}

	private ItemPrediction getPredictionBasedOnWeight(
		int item,
		UserProfile userProfile) {
		// read the item content
		ItemContent targetItem = dataItemMgr.getItemContent(item);
		// predict the class for this item
		float grade = userProfile.predictClass(targetItem);
		ItemPrediction prediction =
			new ItemPrediction(userProfile.getUserId(), item, grade);
		return prediction;
	}

	private UserProfile getUserProfile(UserInfo activeUserInfo) {
		// trying to get the user profile in cache
		UserProfile userProfile;
		Integer key = new Integer(activeUserInfo.userID);
		if (cacheOfUserProfiles.containsKey(key))
			userProfile = (UserProfile) cacheOfUserProfiles.get(key);
		else {
			int maxRating = (int) dataMgr.getMaxRating();
			int minRating = (int) dataMgr.getMinRating();
			int numberOfRates = maxRating - minRating + 1;
			int mediumRate = numberOfRates / 2;
			// creating the keys of modal symbolic objects
			// in this algorithm there is one modal object for each class
			ItemToModalMapping[] itemsMapping =
				(USE_REPLICATION_OR_WEIGHT)
					? getMappingConfigurationBasedOnReplication(
						activeUserInfo,
						minRating,
						numberOfRates,
						mediumRate,
						relevances)
					: getMappingConfigurationBasedOnWeight(
						activeUserInfo,
						relevances);
			// creating the user profile
			userProfile =
				new UserProfile(
					activeUserInfo.userID,
					itemsMapping,
					relevances);
			// putting the user profile in cache
			cacheOfUserProfiles.put(key, userProfile);
		}
		return userProfile;
	}

	/*
	 * It is useful when there is just two modalSymbolicObjects 
	 * per user profile.
	 */
	private ItemToModalMapping[] getMappingConfigurationBasedOnReplication(
		UserInfo activeUserInfo,
		int minRating,
		int numberOfRates,
		int mediumRate,
		ModalSymbolicObjectRelevance[] relevances) {
		// creating the mapping configuration for evaluated items
		int[] itemIds = activeUserInfo.getRatings().getIds();
		float[] ratings = activeUserInfo.getRatings().getRatings();
		ItemToModalMapping[] itemsMapping =
			new ItemToModalMapping[itemIds.length];
		for (int i = 0; i < itemIds.length; i++) {
			ItemContent itemContent = dataItemMgr.getItemContent(itemIds[i]);
			int rate = (int) ratings[i];
			if (rate < (mediumRate + minRating)) {
				// add this item in the negative object
				int times = mediumRate + minRating - rate;
				itemsMapping[i] =
					new ItemToModalMapping(itemContent, relevances[1], times);
			} else {
				// add this item in the positive object
				if (numberOfRates % 2 == 0) {
					// quantidade par
					int times = rate + 1 - (mediumRate + minRating);
					itemsMapping[i] =
						new ItemToModalMapping(
							itemContent,
							relevances[0],
							times);
				} else {
					// quantidade impar
					if (rate > mediumRate + minRating) {
						// ignore items evaluated with medium grade
						int times = rate - (mediumRate + minRating);
						itemsMapping[i] =
							new ItemToModalMapping(
								itemContent,
								relevances[0],
								times);
					} else {
						// add items evaluated with a middle grade in negative subprofile
						itemsMapping[i] =
							new ItemToModalMapping(
								itemContent,
								relevances[1],
								1);
					}
				}
			}
		}
		return itemsMapping;
	}

	private ItemToModalMapping[] getMappingConfigurationBasedOnWeight(
		UserInfo activeUserInfo,
		ModalSymbolicObjectRelevance[] relevances) {
		// creating the mapping configuration for evaluated items
		int[] itemIds = activeUserInfo.getRatings().getIds();
		float[] ratings = activeUserInfo.getRatings().getRatings();
		ItemToModalMapping[] itemsMapping =
			new ItemToModalMapping[itemIds.length];
		for (int i = 0; i < itemIds.length; i++) {
			ItemContent itemContent = dataItemMgr.getItemContent(itemIds[i]);
			int rate = (int) ratings[i];
			itemsMapping[i] =
				new ItemToModalMapping(itemContent, relevances[rate - 1], 1);
		}
		return itemsMapping;
	}

	/* (non-Javadoc)
	 * @see org.recommender.algorithms.CFAlgorithm#getRecommendations(int, int)
	 */
	public ItemPrediction[] getRecommendations(int activeUser, int n)
		throws SQLException, CFNotImplementedException {

		UserInfo activeUserInfo = dataMgr.getUserInfo(activeUser);
		//get the lock here, so you don't have to worry about waiting inbetween
		//your get weight calls and you insure your average matches what you weights
		// are calculated at.
		if (!activeUserInfo.getReadersLock(true))
			return new ItemPrediction[0];

		UserProfile userProfile = getUserProfile(activeUserInfo);

		ItemPrediction[] tempPredList;
		int maxItem = dataMgr.getMaxItem();
		tempPredList = new ItemPrediction[maxItem + 1];
		tempPredList[0] = new ItemPrediction(0, 0, 0);
		int predIndex = 1;

		while (predIndex < tempPredList.length) {
			// read the item content
			ItemContent targetItem = dataItemMgr.getItemContent(predIndex);
			if (targetItem != null) {
				// get the similarity beetwen 0 and 1
				float similarity = (float) userProfile.similarityTo(targetItem);
				if (similarity <= 0.000001 || Double.isNaN(similarity))
					similarity = 0;
				tempPredList[predIndex] =
					new ItemPrediction(
						userProfile.getUserId(),
						predIndex,
						similarity);
			}
			predIndex++;
		}

		Arrays.sort(tempPredList, 0, predIndex, new CompPrediction(dataMgr));

		if (predIndex > n) {
			predIndex = n;
		}
		ItemPrediction[] itemPredList = new ItemPrediction[predIndex];
		for (int i = 0; i < itemPredList.length; i++) {
			itemPredList[i] = tempPredList[i];
		}

		return itemPredList;
	}

	/* (non-Javadoc)
	 * @see org.recommender.algorithms.CFAlgorithm#getAveragePredictionTime()
	 */
	public long getAveragePredictionTime() throws CFNotImplementedException {
		return (Time / NumPredictions);
	}

	/* (non-Javadoc)
	 * @see org.recommender.algorithms.CFAlgorithm#updateUser(int)
	 */
	public void updateUser(int userID) throws CFNotImplementedException {
		UserInfo activeUserInfo = dataMgr.getUserInfo(userID);
		// get the lock here, so you don't have to worry about waiting inbetween
		//		int time = 0;
		//		while (!activeUserInfo.getWritersLock())
		//			try {
		//				time++;
		//				wait(1000);
		//			} catch (InterruptedException e) {
		//			}
		//		System.out.println(
		//			"Waited about " + time + " seconds to get writers lock.");
		// remove the user profile of cache
		cacheOfUserProfiles.remove(new Integer(userID));
		// create the user profile and put it in cache
		UserProfile userProfile = getUserProfile(activeUserInfo);
	}

	/* (non-Javadoc)
	 * @see org.recommender.algorithms.CFAlgorithm#getRecommendationsByType(int, int, int)
	 */
	public ItemPrediction[] getRecommendationsByType(
		int activeUser,
		int n,
		int type)
		throws SQLException, CFNotImplementedException {
		// TODO Auto-generated method stub
		return null;
	}

	public static boolean hasIDF(String variableType) {
		// TODO implementar
		return false;
	}

	public static double getWeigh(String variableType) {
		// TODO implementar
		return 1.0;
	}

	/**
	 * @return
	 */
	public long getNumPredictions() {
		return NumPredictions;
	}

	/**
	 * @return
	 */
	public long getTime() {
		return Time;
	}

	/**
	 * Return the name of this algorithm.
	 * @return String
	 */
	public String toString() {
		return "ModalSymbolicAlgorithm.java\n";
	}

	/**
	 * @return
	 */
	public Stats getTimeStats() {
		return timeStats;
	}

}
