/*
 * ModalSymbolicDescriptor.java
 * 
 * Version 1.0
 *
 * 28/03/2004
 * 
 * Copyright (c) 2004 Federal Universtity of Pernambuco. All rights reserved.
 */
package feers.recommender.algorithms.symbolic;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import feers.recommender.server.DataItemManager;
import feers.recommender.server.descriptor.CathegoricalMultivaluedDescriptor;
import feers.recommender.server.descriptor.CathegoricalUnivaluedDescriptor;
import feers.recommender.server.descriptor.Descriptor;
import feers.recommender.server.descriptor.IntervalDescriptor;
import feers.recommender.server.descriptor.NumericDescriptor;
import feers.recommender.server.descriptor.TextualDescriptor;
import feers.recommender.util.StringProcessing;

/**
 * Represents a modal symbolic variable.
 * 
 * @author <a href="mailto:bldb@cin.ufpe.br>Byron L. D. Bezerra</a>
 */
public class ModalSymbolicDescriptor extends Descriptor {

	/*
	 * the support (key) of this variable and 
	 * the weight associated with each 
	 * value in the support
	 */
	private TreeSet distribution;

	/*
	 * Contains the number of descriptors used to build this one.
	 * It must be incremented everytime the add method is called.
	 */
	private int numberOfDescriptors;

	/*
	 * It is true if the values of this variable are numeric.
	 */
	private boolean ordinalVariable;

	/**
	 * Constructs a modal symbolic variable through a given descriptor.
	 *  
	 * @param d
	 */
	public ModalSymbolicDescriptor(Descriptor d) {
		super(d.getName());
		distribution = new TreeSet(new ComparatorPairValueWeight());
		ordinalVariable = false;

		if (d instanceof NumericDescriptor) {
			PairValueWeight pair = new PairValueWeight(d.getValue(), 1);
			distribution.add(pair);
			ordinalVariable = true;
		} else if (d instanceof IntervalDescriptor) {
			throw new UnsupportedOperationException(
				"The variable "
					+ d.getClass().getName()
					+ " cannot be transformed in a modal variable.");
		} else if (d instanceof CathegoricalUnivaluedDescriptor) {
			PairValueWeight pair = new PairValueWeight(d.getValue(), 1);
			distribution.add(pair);
		} else if (d instanceof CathegoricalMultivaluedDescriptor) {
			List collection = ((List) d.getValue());
			for (Iterator iter = collection.iterator(); iter.hasNext();) {
				Object element = iter.next();
				PairValueWeight newPair =
					new PairValueWeight(element, 1.0 / collection.size());
				if (distribution.contains(newPair)) {
					PairValueWeight old =
						(PairValueWeight) distribution.tailSet(newPair).first();
					newPair.setWeight(newPair.getWeight() + old.getWeight());
					distribution.remove(old);
				}
				distribution.add(newPair);
			}
		} else if (d instanceof TextualDescriptor) {
			String text = (String) d.getValue();
			HashMap map = StringProcessing.getCentroid(text);
			for (Iterator iter = map.keySet().iterator(); iter.hasNext();) {
				Object element = iter.next();
				Double value = (Double) map.get(element);
				PairValueWeight newPair =
					new PairValueWeight(element, value.doubleValue());
				distribution.add(newPair);
			}
		} else {
			throw new UnsupportedOperationException(
				"The variable "
					+ d.getClass().getName()
					+ " cannot be transformed in a modal variable.");
		}
		numberOfDescriptors = 1;
	}

	/**
	 * Returns the value of this modal symbolic variable in a 
	 * HashMap object. The weights are not normalized.
	 *
	 * @see feers.recommender.server.descriptor.Descriptor#getValue()
	 * @return Object
	 */
	public Object getValue() {
		return distribution;
	}

	public String toString() {
		StringBuffer res = new StringBuffer(super.toString());
		res.append('{');
		for (Iterator iter = distribution.iterator(); iter.hasNext();) {
			PairValueWeight element = (PairValueWeight) iter.next();
			res.append(element);
			if (iter.hasNext())
				res.append(" , ");
		}
		res.append("} *obs: non-normalized and without idf");
		return res.toString();
	}

	public void add(Descriptor d) {
		// the variable d must be compatible with the base variable
		if (!d.getName().equals(this.getName()))
			throw new IllegalArgumentException(
				"The variables are incompatibles. Base variable: "
					+ this.getName()
					+ ". Argument variable: "
					+ d.getName());

		if (d instanceof NumericDescriptor) {
			PairValueWeight newPair = new PairValueWeight(d.getValue(), 1);
			if (distribution.contains(newPair)) {
				PairValueWeight old =
					(PairValueWeight) distribution.tailSet(newPair).first();
				newPair.setWeight(newPair.getWeight() + old.getWeight());
				distribution.remove(old);
			}
			distribution.add(newPair);
		} else if (d instanceof IntervalDescriptor) {
			throw new UnsupportedOperationException(
				"The variable "
					+ d.getClass().getName()
					+ " cannot be transformed in a modal variable.");
		} else if (d instanceof CathegoricalUnivaluedDescriptor) {
			PairValueWeight newPair = new PairValueWeight(d.getValue(), 1);
			if (distribution.contains(newPair)) {
				PairValueWeight old =
					(PairValueWeight) distribution.tailSet(newPair).first();
				newPair.setWeight(newPair.getWeight() + old.getWeight());
				distribution.remove(old);
			}
			distribution.add(newPair);
		} else if (d instanceof CathegoricalMultivaluedDescriptor) {
			List collection = ((List) d.getValue());
			for (Iterator iter = collection.iterator(); iter.hasNext();) {
				Object element = iter.next();
				PairValueWeight newPair =
					new PairValueWeight(element, 1.0 / collection.size());
				if (distribution.contains(newPair)) {
					PairValueWeight old =
						(PairValueWeight) distribution.tailSet(newPair).first();
					newPair.setWeight(newPair.getWeight() + old.getWeight());
					distribution.remove(old);
				}
				distribution.add(newPair);
			}
		} else if (d instanceof TextualDescriptor) {
			String text = (String) d.getValue();
			HashMap map = StringProcessing.getCentroid(text);
			for (Iterator iter = map.keySet().iterator(); iter.hasNext();) {
				Object element = iter.next();
				Double value = (Double) map.get(element);
				PairValueWeight newPair =
					new PairValueWeight(element, value.doubleValue());
				distribution.add(newPair);
			}
		} else {
			throw new UnsupportedOperationException(
				"The variable "
					+ d.getClass().getName()
					+ " cannot be transformed in a modal variable.");
		}
		numberOfDescriptors++;
	}

	public double similarityTo(ModalSymbolicDescriptor other) {
		double res = 0.0;

		// the variables must be compatibles
		if (!this.getName().equals(other.getName()))
			throw new IllegalArgumentException(
				"The variables are incompatibles. Base variable: "
					+ this.getName()
					+ ". Argument variable: "
					+ other.getName());
		// TODO verificar
		if (ordinalVariable)
			res =
				(ordinalSimilarity(other) + nonOrdinalSimilarity(other)) / 2.0;
		else
			res = nonOrdinalSimilarity(other);

		if (Double.isNaN(res) || res < ModalSymbolicObject.ZERO) {
			res = ModalSymbolicObject.ZERO;
		}

		if (res > 1.0) {
			throw new RuntimeException(
				"Something is wrong because the similarity between modal symbolic variables must be lower or equals to 1.0. It is: "
					+ res);
		}

		return res;
	}

	private double ordinalSimilarity(ModalSymbolicDescriptor other) {
		double similarity = 0.0;
		TreeSet distA = this.distribution;
		TreeSet distB = other.distribution;
		int numSamplesA = this.numberOfDescriptors;
		int numSamplesB = other.numberOfDescriptors;

		if ((numSamplesA > 0)
			&& (numSamplesB > 0)
			&& (distA.size() > 0)
			&& (distB.size() > 0)) {
			Set a = distA;
			Set b = distB;

			Integer max =
				new Integer(((PairValueWeight) Collections.max(a)).getValue());
			Integer min =
				new Integer(((PairValueWeight) Collections.min(a)).getValue());
			{
				Integer temp =
					new Integer(
						((PairValueWeight) Collections.max(b)).getValue());
				if (max.intValue() < temp.intValue()) {
					max = temp;
				}
				temp =
					new Integer(
						((PairValueWeight) Collections.min(b)).getValue());
				if (min.intValue() > temp.intValue()) {
					min = temp;
				}
			}

			TreeSet universe = new TreeSet();
			for (int i = min.intValue(); i <= max.intValue(); i++) {
				// TODO verificar !!!
				universe.add(new PairValueWeight("" + i, 1.0));
			}

			TreeSet compA = ((TreeSet) universe.clone());
			compA.removeAll(a);
			TreeSet compB = ((TreeSet) universe.clone());
			compB.removeAll(b);

			TreeSet aINTERb = new TreeSet(a);
			aINTERb.retainAll(b);
			TreeSet aINTERcompB = new TreeSet(a);
			aINTERcompB.retainAll(compB);
			TreeSet compAINTERb = new TreeSet(compA);
			compAINTERb.retainAll(b);

			{
				double alfaA = 0.0;
				double alfaB = 0.0;
				double beta = 0.0;
				double gama = 0.0;

				for (Iterator iter = aINTERb.iterator(); iter.hasNext();) {
					Object element = iter.next();
					// calculando alfaA
					if (a.contains(element)) {
						// obtem o peso do elemento na distribuicao A
						double peso =
							((PairValueWeight) distA.tailSet(element).first())
								.getWeight();
						alfaA += peso;
					}
					// calculando alfaB
					if (b.contains(element)) {
						// obtem o peso do elemento na distribuicao B
						double peso =
							((PairValueWeight) distB.tailSet(element).first())
								.getWeight();
						alfaB += peso;
					}
				}
				alfaA /= numSamplesA;
				alfaB /= numSamplesB;

				for (Iterator iter = aINTERcompB.iterator(); iter.hasNext();) {
					Object element = iter.next();
					// calculando beta
					if (a.contains(element)) {
						// obtem o peso do elemento na distribuicao A
						double peso =
							((PairValueWeight) distA.tailSet(element).first())
								.getWeight();
						beta += peso;
					}

				}
				beta /= numSamplesA;

				for (Iterator iter = compAINTERb.iterator(); iter.hasNext();) {
					Object element = iter.next();
					// calculando gama
					if (b.contains(element)) {
						// obtem o peso do elemento na distribuicao B
						double peso =
							((PairValueWeight) distB.tailSet(element).first())
								.getWeight();
						gama += peso;
					}

				}
				gama /= numSamplesB;

				// calculando a dissimilaridade
				similarity =
					0.5
						* ((beta + gama) / (alfaA + beta + gama)
							+ (beta + gama) / (alfaB + beta + gama));
			}

			similarity = 1.0 - similarity;
		}

		if (similarity > 1.0)
			throw new RuntimeException(
				"Something is wrong because the similarity should be "
					+ "lower or equals to 1.0 and it is "
					+ similarity);

		return similarity;
	}

	private double nonOrdinalSimilarity(ModalSymbolicDescriptor other) {
		double similarity = 0.0, sumThis = 0.0, sumOther = 0.0;

		Object[] arrThis = this.distribution.toArray();
		Object[] arrOther = other.distribution.toArray();
		int contThis = 0, contOther = 0;
		while (contThis < arrThis.length && contOther < arrOther.length) {
			PairValueWeight thisPair = (PairValueWeight) arrThis[contThis];
			PairValueWeight otherPair = (PairValueWeight) arrOther[contOther];

			if (thisPair.getValue().compareTo(otherPair.getValue()) < 0) {
				// this < other => the current value in this is not in other!
				sumThis
					+= (thisPair.getWeight()
						* getNormalizedIDF(thisPair.getValue()));
				contThis++;
			} else if (
				thisPair.getValue().compareTo(otherPair.getValue()) > 0) {
				// this > other => the current value in other is not in this!
				sumOther
					+= (otherPair.getWeight()
						* getNormalizedIDF(otherPair.getValue()));
				contOther++;
			} else { // equal
				contThis++;
				contOther++;
			}
		}

		// adding the remaining weights of this
		for (int i = contThis; i < arrThis.length; i++) {
			PairValueWeight thisPair = (PairValueWeight) arrThis[contThis];
			sumThis
				+= (thisPair.getWeight() * getNormalizedIDF(thisPair.getValue()));
		}

		// adding the remaining weights of other
		for (int i = contOther; i < arrOther.length; i++) {
			PairValueWeight otherPair = (PairValueWeight) arrOther[contOther];
			sumOther
				+= (otherPair.getWeight()
					* getNormalizedIDF(otherPair.getValue()));
		}

		double dissThisOther =
			(contThis == 0) ? 1.0 : (sumThis / this.numberOfDescriptors);
		double dissOtherThis =
			(contOther == 0) ? 1.0 : (sumOther / other.numberOfDescriptors);

		similarity = (2.0 - dissThisOther - dissOtherThis) / 2.0;

		if (similarity > 1.0)
			throw new RuntimeException(
				"Something is wrong because the similarity should be "
					+ "lower or equals to 1.0 and it is "
					+ similarity);

		return similarity;
	}

	/**
	 * @param key
	 * @return
	 */
	private double getNormalizedIDF(Object key) {
		double res = 1.0;
		if (ModalSymbolicAlgorithm.hasIDF(this.getName())) {
			res = DataItemManager.getIdf(this.getName(), key.toString());
		}
		return res;
	}

}
