/*
 * ModalSymbolicObject.java
 * 
 * Version 1.0
 *
 * 28/03/2004
 * 
 * Copyright (c) 2004 Federal Universtity of Pernambuco. All rights reserved.
 */
package feers.recommender.algorithms.symbolic;

import java.io.Serializable;
import java.util.Collections;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.TreeSet;

import org.recommender.server.SystemProperties;

import feers.recommender.server.DataItemManager;
import feers.recommender.server.ItemContent;
import feers.recommender.server.descriptor.Descriptor;

/**
 * Represents a modal symbolic object, which is a vector of
 * modal symbolic variables.
 * 
 * @author <a href="mailto:bldb@cin.ufpe.br>Byron L. D. Bezerra</a>
 */
public class ModalSymbolicObject implements Serializable {

	/*
	 * A map of variables. The key is the variable name. 
	 * Each value in this map is a type of ModalSymbolicDescriptor. 
	 */
	private Hashtable modalVariables;
	public static final double ZERO = 0.00000001;

	public ModalSymbolicObject() {
		this.modalVariables = new Hashtable();
	}

	public ModalSymbolicObject(ItemContent item) {
		this();
		add(item);
	}

	/**
	 * Transform the item variables to modal symbolic variables 
	 * (pre-processing step) and add it in this modal symbolic object
	 * (generalization step).
	 *  
	 * @param item
	 */
	public void add(ItemContent item) {
		if (item != null) {
			for (Iterator iter = item.variables(); iter.hasNext();) {
				Descriptor variable = (Descriptor) iter.next();
				ModalSymbolicDescriptor modalVariable;
				if (this.modalVariables.containsKey(variable.getName())) {
					// getting the associated modal variable in this object
					modalVariable =
						(ModalSymbolicDescriptor) modalVariables.get(
							variable.getName());
					modalVariable.add(variable);
				} else {
					// creating a new modal variable
					modalVariable = new ModalSymbolicDescriptor(variable);
					modalVariables.put(variable.getName(), modalVariable);
				}
			}
		} else {
			throw new IllegalArgumentException("The item cannot be null.");
		}
	}

	public double similarityTo(ModalSymbolicObject other) {
		double similarity = 0.0;
		double weightSum = 0.0;

		Iterator i = this.modalVariables.keySet().iterator();
		while (i.hasNext()) {
			Object key = i.next();
			ModalSymbolicDescriptor source =
				(ModalSymbolicDescriptor) this.modalVariables.get(key);
			ModalSymbolicDescriptor target =
				(ModalSymbolicDescriptor) other.modalVariables.get(key);
			double weigh = ModalSymbolicAlgorithm.getWeigh((String) key);
			if (source != null && target != null) {
				double aux = source.similarityTo(target);
//				System.out.println("Similarity between " + source.getName()
//					+ " and " + target.getName() + " = " + aux);
				similarity += (aux * weigh);
				weightSum += weigh;
			}
		}

		similarity = (weightSum > 0) ? (similarity / weightSum) : similarity;

		if (Double.isNaN(similarity) || similarity < ZERO) {
			similarity = ZERO;
		}

		if (similarity > 1.0) {
			throw new RuntimeException(
				"Something is wrong because the similarity between modal symbolic objects must be lower or equals to 1.0. It is: "
					+ similarity);
		}

		return similarity;
	}

	public static void main(String[] args) {
		/*
		TreeSet a = new TreeSet();
		PairValueWeight p = new PairValueWeight(new Integer(103), 10);
		PairValueWeight p2 = new PairValueWeight(new Integer(109), 10);
		a.add(p);
		a.add(p2);
		PairValueWeight m = (PairValueWeight) Collections.max(a);
		System.out.println(m.getValue());
		*/
		
		String installdir = System.getProperty("cf.path", ".");
		String filename = installdir + "/lib/" + "CFServer.properties";
		// Load properties
		SystemProperties.loadProperties(filename);
		DataItemManager dataItemMgr = new DataItemManager(1, 100);

		System.out.println("Loading item ...");
		ItemContent item1 = dataItemMgr.getItemContent(20);
		System.out.println(item1);
		System.out.println("Creating the modal symbolic object ...");
		ModalSymbolicObject ob1 = new ModalSymbolicObject(item1);
		System.out.println(ob1);

		System.out.println("Loading item ...");
		ItemContent item2 = dataItemMgr.getItemContent(16);
		System.out.println(item2);
		System.out.println("Creating the modal symbolic object ...");
		ModalSymbolicObject ob2 = new ModalSymbolicObject(item2);
		System.out.println(ob2);
		
		System.out.println("Similarity beetwen ob1 e ob1 = " +
			ob1.similarityTo(ob1));
		System.out.println("Similarity beetwen ob2 e ob2 = " +
					ob2.similarityTo(ob2));
		System.out.println("Similarity beetwen ob1 e ob2 = " +
					ob1.similarityTo(ob2));
		System.out.println("Similarity beetwen ob2 e ob1 = " +
					ob2.similarityTo(ob1));
		
	}
}
