/*
 * UserProfile.java
 * 
 * Version 1.0
 *
 * 28/03/2004
 * 
 * Copyright (c) 2004 Federal Universtity of Pernambuco. All rights reserved.
 */
package feers.recommender.algorithms.symbolic;

import java.io.Serializable;

import feers.recommender.server.ItemContent;

/**
 * Represents a user profile based on modal symbolic objects.
 * 
 * @author <a href="mailto:bldb@cin.ufpe.br>Byron L. D. Bezerra</a>
 * 
 * @see feers.recommender.algorithms.symbolic.ModalSymbolicAlgorithm
 */
public class UserProfile implements Serializable {

	private int userId;

	private ModalSymbolicObjectRelevance[] modalSymbolicObjectKeys;
	private ModalSymbolicObject[] modalSymbolicObjects;

	public static final int UNKNOWN_CLASS = -1000;

	/**
	 * Creates a user profile based on symbolic data. 
	 * 
	 * @param userId the user id
	 * @param itemsMapping the content of all items evaluated by the user with the mapping
	 * @param modalSymbolicObjectKeys indicates how many modal symbolic objects must be created
	 */
	public UserProfile(
		int userId,
		ItemToModalMapping[] itemsMapping,
		ModalSymbolicObjectRelevance[] modalSymbolicObjectKeys) {
		super();
		this.userId = userId;
		// creating the modal symbolic objects
		this.modalSymbolicObjectKeys = modalSymbolicObjectKeys;
		this.modalSymbolicObjects =
			new ModalSymbolicObject[modalSymbolicObjectKeys.length];
		for (int i = 0; i < modalSymbolicObjectKeys.length; i++) {
			this.modalSymbolicObjects[i] = new ModalSymbolicObject();
		}
		add(itemsMapping);
	}

	public void add(ItemToModalMapping[] itemsMapping) {
		// adding evaluated items to this profile
		for (int i = 0; i < itemsMapping.length; i++) {
			if (itemsMapping[i] != null)
				add(itemsMapping[i]);
		}
	}

	public void add(ItemToModalMapping itemMapping) {
		ItemToModalMapping schema = itemMapping;
		ModalSymbolicObject modalObject =
			this
				.modalSymbolicObjects[schema
				.getModalSymbolicObjectKey()
				.hashCode()];
		for (int j = 0; j < schema.getTimes(); j++) {
			modalObject.add(schema.getItem());
		}
	}

	/**
	 * Calculate the similarity between this user profile and 
	 * otherUserProfile based on weighted mean of the similarity 
	 * of each modal symbolic object pair.
	 * 
	 * @param otherUserProfile
	 * @return
	 */
	public double similarityTo(UserProfile otherUserProfile) {
		double res = 0.0;
		double denominador = 0.0;

		for (int i = 0; i < modalSymbolicObjectKeys.length; i++) {
			ModalSymbolicObjectRelevance config = modalSymbolicObjectKeys[i];
			ModalSymbolicObject objectThis = modalSymbolicObjects[i];
			ModalSymbolicObject objectOther =
				otherUserProfile.modalSymbolicObjects[i];
			if (objectOther != null) {
				double similarity = objectThis.similarityTo(objectOther);
				res += config.getWeigh() * similarity;
				denominador += config.getWeigh();
			}
		}
		if (denominador != 0.0) {
			res /= denominador;
		}

		if (Double.isNaN(res) || res < ModalSymbolicObject.ZERO) {
			res = ModalSymbolicObject.ZERO;
		}

		if (res > 1.0) {
			throw new RuntimeException(
				"Something is wrong because the similarity between user profiles must be lower or equals to 1.0. It is: "
					+ res);
		}

		return res;
	}

	public double similarityTo(ItemContent other) {
		ModalSymbolicObject object = new ModalSymbolicObject(other);
		return this.similarityTo(object);
	}

	/**
	 * Compute the similarity beetwen 0 and 1 based on the weighted
	 * similarity of "other" with positive modalSymbolicObjects of 
	 * "this" plus the weighted dissimilarity of "other" with negative
	 * ones.
	 *  
	 * @param other
	 * @return
	 */
	public double similarityTo(ModalSymbolicObject other) {
		double res = 0.0, denominador = 0.0;

		for (int i = 0; i < modalSymbolicObjectKeys.length; i++) {
			ModalSymbolicObjectRelevance config = modalSymbolicObjectKeys[i];
			ModalSymbolicObject object = modalSymbolicObjects[i];
			double similarity = object.similarityTo(other);
			if (config.getPolarity() == 1)
				res += config.getWeigh() * similarity;
			else if (config.getPolarity() == -1)
				res += config.getWeigh() * (1 - similarity);
			else
				throw new RuntimeException("The polarity should be 1 or -1.");
			denominador += config.getWeigh();
		}
		if (denominador != 0.0) {
			res /= denominador;
		}

		if (Double.isNaN(res) || res < ModalSymbolicObject.ZERO) {
			res = ModalSymbolicObject.ZERO;
		}

		if (res > 1.0) {
			throw new RuntimeException(
				"Something is wrong because the similarity between modal symbolic object and user profile must be lower or equals to 1.0. It is: "
					+ res);
		}

		return res;
	}

	public int predictClass(ModalSymbolicObject other) {
		int classe = UNKNOWN_CLASS;
		double maxSim = 0.0;
		for (int i = 0; i < modalSymbolicObjectKeys.length; i++) {
			ModalSymbolicObjectRelevance config = modalSymbolicObjectKeys[i];
			ModalSymbolicObject object = modalSymbolicObjects[i];
			double similarity = object.similarityTo(other);
			if (similarity > maxSim) {
				classe = config.hashCode();
				maxSim = similarity;
			}
		}
		return classe;
	}

	/**
	 * @return
	 */
	public int getUserId() {
		return userId;
	}

	/**
	 * @param targetItem
	 * @return
	 */
	public int predictClass(ItemContent targetItem) {
		ModalSymbolicObject o = new ModalSymbolicObject(targetItem);
		return this.predictClass(o);
	}

}
