/*
 * DataItemManager.java
 * 
 * Version 1.0
 *
 * 28/03/2004
 * 
 * Copyright (c) 2004 Federal Universtity of Pernambuco. All rights reserved.
 */
package feers.recommender.server;

import java.io.DataInputStream;
import java.io.FileInputStream;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

import org.recommender.server.SystemProperties;
import org.recommender.server.SystemUtils;

import feers.recommender.server.descriptor.CathegoricalMultivaluedDescriptor;
import feers.recommender.server.descriptor.CathegoricalUnivaluedDescriptor;
import feers.recommender.server.descriptor.Descriptor;
import feers.recommender.server.descriptor.NumericDescriptor;

/**
 * This class provides all methods to manipulate the content 
 * of items stored in database. 
 * 
 * @author <a href="mailto:bldb@cin.ufpe.br>Byron L. D. Bezerra</a>
 */
public class DataItemManager {

	private static Connection connection;
	private static String classeDriver;
	private static String url;
	private static String username;
	private static String password;

	private ItemContent[] cacheOfItems;
	private int minItemIdInCache;
	private int maxItemIdInCache;

	// variable names
	private static final String[] UNIVALUED_VARIABLES = { /*"name",*/ "runtime", "year" };
	// 0 - cathegorical , 1 - numeric
	private static final byte[] UNIVALUED_VARIABLES_TYPE = { 1, 1 };
	private static final String[] MULTIVALUE_VARIABLES =
		{
			"cast",
			"certification",
			"color",
			"country",
			"director",
			"genre",
			"language",
			"sound",
			"writer" };

	/////////////// SQL statements ///////////////
	private static final String SQL_LOAD_MULTIVALUE =
		"select t_###.name from t_###, t_movie, r_###_movie "
			+ "where t_###.id = r_###_movie.id_### and t_movie.id = r_###_movie.id_movie "
			+ "and t_movie.id = ?";

	public DataItemManager(int minItemId, int maxItemId) {
		SystemUtils.printDebug("Starting Data Item Manager...");
		classeDriver = SystemProperties.getString("CFServer.jdbc.drivers");

		url =
			SystemProperties.getString("CFServer.jdbc.itemdb.url")
				+ "?autoReconnect=true";
		//	   System.out.println(url);       	
		username = SystemProperties.getString("CFServer.jdbc.itemdb.username");
		password = SystemProperties.getString("CFServer.jdbc.itemdb.password");
		try {
			Class.forName(classeDriver);
			connection = DriverManager.getConnection(url, username, password);
		} catch (SQLException sqle) {
			System.out.println(
				"SQLException => ConnectionManager: " + sqle.getMessage());
		} catch (ClassNotFoundException cnfe) {
			System.out.println(
				"Driver nao encontrado => ConnectionManager: "
					+ cnfe.getMessage());
		}

		long startTime = System.currentTimeMillis();
		this.minItemIdInCache = minItemId;
		this.maxItemIdInCache = maxItemId;
		this.cacheOfItems = new ItemContent[maxItemId - minItemId + 1];
		for (int i = minItemId; i < maxItemId; i++) {
			ItemContent item = this.getItemContent(i);
			this.cacheOfItems[i - minItemId] = item;
		}
		long endTime = System.currentTimeMillis();
		SystemUtils.printDebug(
			"Time to caching elements from id "
				+ minItemId
				+ " to id "
				+ maxItemId
				+ " = "
				+ (endTime - startTime)
				+ " ms");

		SystemUtils.printDebug("Data Item Manager Started");
	}

	private PreparedStatement reservaStatement(String sql) {
		PreparedStatement statement = null;
		try {
			statement = connection.prepareStatement(sql);
		} catch (SQLException sqle) {
			System.out.println(
				"SQLException => ConnectionManager: " + sqle.getMessage());
		}
		return statement;
	}

	public ItemContent getItemContent(int itemId) {
		ItemContent item;
		if (itemId >= minItemIdInCache
			&& itemId <= maxItemIdInCache
			&& this.cacheOfItems[itemId - minItemIdInCache] != null)
			item = this.cacheOfItems[itemId - minItemIdInCache];
		else {
			item = new ItemContent(itemId);
			try {
				// read the main table		
				PreparedStatement pstmt =
					reservaStatement("select * from t_movie where id = ?");
				pstmt.setInt(1, itemId);
				ResultSet rset = pstmt.executeQuery();
				if (rset.next()) {
					// get the univalued variables
					for (int i = 0; i < UNIVALUED_VARIABLES.length; i++) {
						Descriptor variable = null;
						if (UNIVALUED_VARIABLES_TYPE[i] == 0) {
							String value =
								rset.getString(UNIVALUED_VARIABLES[i]);
							if (value != null)
								variable =
									new CathegoricalUnivaluedDescriptor(
										UNIVALUED_VARIABLES[i],
										value);
						} else if (UNIVALUED_VARIABLES_TYPE[i] == 1) {
							int value = rset.getInt(UNIVALUED_VARIABLES[i]);
							if (value > 0)
								variable =
									new NumericDescriptor(
										UNIVALUED_VARIABLES[i],
										value);
						} else {
							throw new RuntimeException("Invalid variable type.");
						}
						if (variable != null)
							item.addVariable(variable);
					}
					rset.close();
					// read multivalued variables
					for (int i = 0; i < MULTIVALUE_VARIABLES.length; i++) {
						pstmt.close();
						String sql =
							SQL_LOAD_MULTIVALUE.replaceAll(
								"###",
								MULTIVALUE_VARIABLES[i]);
						pstmt = reservaStatement(sql);
						pstmt.setInt(1, itemId);
						rset = pstmt.executeQuery();
						if (rset.next()) {
							List values = new LinkedList();
							do {
								values.add(rset.getString("name"));
							} while (rset.next());
							Descriptor variable =
								new CathegoricalMultivaluedDescriptor(
									MULTIVALUE_VARIABLES[i],
									values);
							item.addVariable(variable);
						}
					}
				}
				// putItemInCache(item);
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
		return item;
	}

	/**
	 * @param item
	 */
	private void putItemInCache(ItemContent item) {
		// arraycopy(Object src, int srcPos, Object dest, int destPos, int length)
		if (item.getItemId() < this.minItemIdInCache) {
			ItemContent[] temp = new ItemContent[this.cacheOfItems.length+this.minItemIdInCache-item.getItemId()];
			System.arraycopy(this.cacheOfItems, 0, temp, (this.minItemIdInCache-item.getItemId()), this.cacheOfItems.length);
			temp[0] = item;
			this.minItemIdInCache = item.getItemId();
			this.cacheOfItems = temp;
		} else if (item.getItemId() > this.minItemIdInCache) {
			ItemContent[] temp = new ItemContent[this.cacheOfItems.length+item.getItemId()-this.maxItemIdInCache];
			System.arraycopy(this.cacheOfItems, 0, temp, 0, this.cacheOfItems.length);
			temp[temp.length-1] = item;
			this.maxItemIdInCache = item.getItemId();
			this.cacheOfItems = temp;
		} else 
			this.cacheOfItems[item.getItemId()] = item;
	}

	public static double getIdf(String variableName, String value) {
		// TODO implementar
		return 1;
	}

	private void importRatingTableFile(String file) {
		try {
			FileInputStream fin = new FileInputStream(file);
			DataInputStream din = new DataInputStream(fin);
			PreparedStatement pstmt =
				reservaStatement("insert into rating_table values (?,?,?)");
			while (din.available() > 0) {
				String line = din.readLine();
				StringTokenizer tokens = new StringTokenizer(line, ",");
				int userId = Integer.parseInt(tokens.nextToken());
				int itemId = Integer.parseInt(tokens.nextToken());
				float rate = Float.parseFloat(tokens.nextToken());
				pstmt.clearParameters();
				pstmt.setInt(1, userId);
				pstmt.setInt(2, itemId);
				pstmt.setFloat(3, rate);
				pstmt.execute();
			}
			pstmt.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public static void main(String args[]) {
		String installdir = System.getProperty("cf.path", ".");
		String filename = installdir + "/lib/" + "CFServer.properties";
		// Load properties
		SystemProperties.loadProperties(filename);
		DataItemManager dataItemMgr = new DataItemManager(1, 10);
		//ItemContent item = dataItemMgr.getItemContent(20);
		//System.out.println(item);
		//dataItemMgr.importRatingTableFile("C:/Documents and Settings/Byron/Meus documentos/UFPE/datasets/rating_table.txt");
	}

}
