/*
 * CoreAlgorithmTest.java
 * 
 * Version 1.0
 *
 * 08/04/2004
 * 
 * Copyright (c) 2004 Federal Universtity of Pernambuco. All rights reserved.
 */
package feers.recommender.test;

import java.sql.SQLException;
import java.util.Arrays;
import java.util.Iterator;

import org.recommender.algorithms.CFAlgorithm;
import org.recommender.algorithms.CFNotImplementedException;
import org.recommender.algorithms.SimplePearsonAlgorithm;
import org.recommender.algorithms.TestStats;
import org.recommender.algorithms.Experimental.Comp;
import org.recommender.server.DataManager;
import org.recommender.server.ItemPrediction;
import org.recommender.server.ItemRating;
import org.recommender.server.SystemProperties;
import org.recommender.server.SystemUtils;

import feers.recommender.algorithms.symbolic.CollabModalAlgorithm;
import feers.recommender.algorithms.symbolic.ContentBoostedCollabAlgorithm;
import feers.recommender.algorithms.symbolic.ModalSymbolicAlgorithm;
import feers.recommender.server.DataItemManager;

/**
 * Describe this classe here ...
 * 
 * @author <a href="mailto:bldb@cin.ufpe.br>Byron L. D. Bezerra</a>
 */
public class CoreAlgorithmTest {

	private CFAlgorithm[] algorithms;
	private ExperimentalDataset datasetForPredictionExperiment;
	private ExperimentalDataset datasetForRankExperiment;
	private DataManager dataMgr;

	public CoreAlgorithmTest(
		String experimentalDatasetForPredictionFilename,
		String experimentalDatasetForRankFilename,
		CFAlgorithm[] algorithms,
		DataManager dataMgr) {
		this.algorithms = algorithms;
		this.datasetForPredictionExperiment =
			ExperimentalDataset.loadDataset(
				experimentalDatasetForPredictionFilename);
		if (datasetForPredictionExperiment == null)
			throw new RuntimeException(
				"Cannot load the experimental dataset "
					+ experimentalDatasetForPredictionFilename);
		this.datasetForRankExperiment =
			ExperimentalDataset.loadDataset(experimentalDatasetForRankFilename);
		if (datasetForRankExperiment == null)
			throw new RuntimeException(
				"Cannot load the experimental dataset "
					+ experimentalDatasetForRankFilename);
		this.dataMgr = dataMgr;
		this.datasetForPredictionExperiment.setAlgorithms(this.algorithms);
		this.datasetForRankExperiment.setAlgorithms(this.algorithms);
	}

	public void startPredictionExperiment() throws Exception {
		SystemUtils.printDebug("Starting Prediction Experiments ...");
		TestStats[][] testStats =
			new TestStats[this
				.algorithms
				.length][datasetForPredictionExperiment
				.getNumberOfExperiments()];
		for (int i = 0; i < algorithms.length; i++) {
			for (int j = 0;
				j < datasetForPredictionExperiment.getNumberOfExperiments();
				j++)
				testStats[i][j] = new TestStats(dataMgr);
			this.datasetForPredictionExperiment.setResults(i, testStats[i]);
		}

		// load users in cache (it is not needed if all users are previously loaded)
		for (Iterator iterUserSet =
			datasetForPredictionExperiment.getUserSets(0);
			iterUserSet.hasNext();
			) {
			UserSets userSets = (UserSets) iterUserSet.next();
			this.dataMgr.loadUser(userSets.getUserId(), false);
		}

		for (int i = 0;
			i < datasetForPredictionExperiment.getNumberOfExperiments();
			i++) {
			// for each experiment 
			System.out.println("");
			for (Iterator iterUserSet =
				datasetForPredictionExperiment.getUserSets(i);
				iterUserSet.hasNext();
				) {
				// for each user ...
				System.out.print(".");

				UserSets userSets = (UserSets) iterUserSet.next();

				ItemRating[] userRatings =
					new ItemRating[userSets.getTestItems().length];
				// getting ratings of items of test set from cache
				for (int j = 0; j < userSets.getTestItems().length; j++) {
					int itemId = userSets.getTestItems()[j];
					userRatings[j] =
						this.dataMgr.getItemRatingCached(
							userSets.getUserId(),
							itemId);
					if (userRatings[j] == null)
						System.out.println(itemId + " is null!");
					if (userRatings[j].getRating()
						== this.dataMgr.getErrRating())
						System.out.println(
							itemId
								+ " not rated in cache for user "
								+ userSets.getUserId());
				}
				// remove all items of this user from cache different of those in training set
				Arrays.sort(userSets.getTrainingItems());
				int[] trainingItems = userSets.getTrainingItems();
				for (int itemId = 1;
					itemId <= dataMgr.getMaxItem();
					itemId++) {
					ItemRating itemRat =
						this.dataMgr.getItemRatingCached(
							userSets.getUserId(),
							itemId);
					if (itemRat.getRating() != this.dataMgr.getErrRating()
						&& Arrays.binarySearch(trainingItems, itemId) < 0)
						this.dataMgr.removeCachedRating(itemRat);
				}

				// say the user was updated
				for (int k = 0; k < algorithms.length; k++) {
					algorithms[k].updateUser(userSets.getUserId());
				}

				// predict items of test set for this user using all algorithms
				ItemPrediction[][] predictions =
					new ItemPrediction[algorithms.length][];
				for (int k = 0; k < algorithms.length; k++) {
					predictions[k] =
						new ItemPrediction[userSets.getTestItems().length];
					for (int j = 0; j < userSets.getTestItems().length; j++) {
						int itemId = userSets.getTestItems()[j];
						predictions[k][j] =
							predictRate(
								userSets.getUserId(),
								itemId,
								algorithms[k]);
						testStats[k][i].addRating(
							userSets.getUserId(),
							itemId,
							userRatings[j].getRating(),
							predictions[k][j].getPrediction());
					}
				}
				// reload the user
				this.dataMgr.loadUser(userSets.getUserId(), false);
				// say the user was updated
				for (int k = 0; k < algorithms.length; k++) {
					algorithms[k].updateUser(userSets.getUserId());
				}
			}

			System.out.println(
				"\n*** Partial Results for experiment " + (i + 1) + " ***");
			datasetForPredictionExperiment.printResults();
		}

		// print results
		System.out.println("\n****** Final Results ******");
		datasetForPredictionExperiment.printResults();

		// print time stats for each algorithm
		for (int i = 0; i < algorithms.length; i++) {
			System.out.println("\n" + algorithms[i]);
			System.out.println(
				"Average time for predict rate = "
					+ algorithms[i].getTimeStats().calcMean());
			System.out.println(
				"Std time for predict rate = "
					+ algorithms[i].getTimeStats().calcSD());
		}

		SystemUtils.printDebug("Finished Prediction Experiments ...");
	}

	public void startRankExperiment(int targetAlgorithmInComparison) throws Exception {

		TestStats[][] testStats =
			new TestStats[this
				.algorithms
				.length][datasetForRankExperiment
				.getNumberOfExperiments()];
		for (int i = 0; i < algorithms.length; i++) {
			for (int j = 0;
				j < datasetForRankExperiment.getNumberOfExperiments();
				j++)
				testStats[i][j] = new TestStats(dataMgr);
			this.datasetForRankExperiment.setResults(i, testStats[i]);
		}

		// load users in cache (it is not needed if all users are previously loaded)
		for (Iterator iterUserSet = datasetForRankExperiment.getUserSets(0);
			iterUserSet.hasNext();
			) {
			UserSets userSets = (UserSets) iterUserSet.next();
			this.dataMgr.loadUser(userSets.getUserId(), false);
		}

		for (int i = 0;
			i < datasetForRankExperiment.getNumberOfExperiments();
			i++) {
			// for each experiment 
			System.out.println("");
			for (Iterator iterUserSet =
				datasetForRankExperiment.getUserSets(i);
				iterUserSet.hasNext();
				) {
				// for each user ...
				System.out.print(".");

				UserSets userSets = (UserSets) iterUserSet.next();

				ItemRating[] userRatings =
					new ItemRating[userSets.getTestItems().length];
				ItemRating[] userRatingsSorted =
					new ItemRating[userSets.getTestItems().length];
				// getting ratings of items of test set from cache
				for (int j = 0; j < userSets.getTestItems().length; j++) {
					int itemId = userSets.getTestItems()[j];
					userRatings[j] =
						this.dataMgr.getItemRatingCached(
							userSets.getUserId(),
							itemId);
					if (userRatings[j] == null)
						System.out.println(itemId + " is null!");
					if (userRatings[j].getRating()
						== this.dataMgr.getErrRating())
						System.out.println(
							itemId
								+ " not rated in cache for user "
								+ userSets.getUserId());
					userRatingsSorted[j] = userRatings[j];
				}
				// remove all items of this user from cache different of those in training set
				Arrays.sort(userSets.getTrainingItems());
				int[] trainingItems = userSets.getTrainingItems();
				for (int itemId = 1;
					itemId <= dataMgr.getMaxItem();
					itemId++) {
					ItemRating itemRat =
						this.dataMgr.getItemRatingCached(
							userSets.getUserId(),
							itemId);
					if (itemRat.getRating() != this.dataMgr.getErrRating()
						&& Arrays.binarySearch(trainingItems, itemId) < 0)
						this.dataMgr.removeCachedRating(itemRat);
				}

				// say the user was updated
				for (int k = 0; k < algorithms.length; k++) {
					algorithms[k].updateUser(userSets.getUserId());
				}

				// sort the copy of user ratings to use in breese metric
				Arrays.sort(
					userRatingsSorted,
					0,
					userRatingsSorted.length,
					new Comp());
				// predict items of test set for this user using all algorithms
				ItemPrediction[][] predictions =
					new ItemPrediction[algorithms.length][];
				for (int k = 0; k < algorithms.length; k++) {
					predictions[k] =
						new ItemPrediction[userSets.getTestItems().length];
					int[] testItems = userSets.getTestItems();
					predictions[k] =
						findGoodItems(
							userSets.getUserId(),
							algorithms[k],
							testItems);
					// measure breese metric for this algorithm
					float[] breese =
						computeBreeseValue(
							userSets.getUserId(),
							predictions[k],
							userRatingsSorted);
					testStats[k][i].addBreeseValues(breese[0], // rMax
					breese[1]); // rVal
				}
				// reload the user
				this.dataMgr.loadUser(userSets.getUserId(), false);
				// say the user was updated
				for (int k = 0; k < algorithms.length; k++) {
					algorithms[k].updateUser(userSets.getUserId());
				}
			}

//			System.out.println(
//				"\n*** Partial Results for experiment " + (i + 1) + " ***");
//			datasetForRankExperiment.printResults();
		}

		// print results
		System.out.println("\n****** Final Results ******");
		datasetForRankExperiment.printResults();
		datasetForRankExperiment.printStatisticalTest(targetAlgorithmInComparison);

		// print time stats for each algorithm
		for (int i = 0; i < algorithms.length; i++) {
			System.out.println("\n" + algorithms[i]);
			System.out.println(
				"Average time for predict rate = "
					+ algorithms[i].getTimeStats().calcMean());
			System.out.println(
				"Std time for predict rate = "
					+ algorithms[i].getTimeStats().calcSD());
		}
	}

	/**
	 * @param userId
	 * @param algorithm
	 * @param testItems
	 * @return
	 */
	private ItemPrediction[] findGoodItems(
		int userId,
		CFAlgorithm algorithm,
		int[] testItems)
		throws Exception {
		Arrays.sort(testItems);
		ItemPrediction[] allItems =
			algorithm.getRecommendations(userId, this.dataMgr.getMaxItem()+1);
		ItemPrediction[] res = new ItemPrediction[testItems.length];
		int qtd = 0;
		for (int i = 0; i < allItems.length; i++) {
			int index = Arrays.binarySearch(testItems, allItems[i].getItemID()); 
			if (index >= 0) {
				res[qtd++] = allItems[i];
			}
		}
		return res;
	}

	/**
	 * @param predictions
	 * @param userRatings
	 * @return
	 */
	private float[] computeBreeseValue(
		int userID,
		ItemPrediction[] predictions,
		ItemRating[] userRatingsSorted) {
		float[] result = new float[2];

		Arrays.sort(
			predictions,
			0,
			predictions.length,
			new ComparatorItemPrediction());
		float rMax =
			computeBestValue(userID, userRatingsSorted, predictions.length);
		float rVal =
			computeUtility(
				userID,
				predictions,
				userRatingsSorted,
				predictions.length);
		result[0] = rMax;
		result[1] = rVal;

		return result;
	}

	// compute the maximum utility for userID.
	private float computeBestValue(
		int userID,
		ItemRating[] ratingsSorted,
		int len) {
		float DEFAULT_RATING = dataMgr.getMeanRatingCached(userID);
		float ALPHA = 5; // Rating list half-life
		float total = 0;
		for (int j = 0; j < len; j++) {
			float rating = ratingsSorted[j].getRating();
			// Ignore ratings below the default
			if (rating <= DEFAULT_RATING)
				continue;
			// DEBUG
			//System.out.println("" + j + "\t" + rating);
			total += (rating - DEFAULT_RATING) * Math.pow(0.5, j / (ALPHA - 1));
		}
		return total;
	}

	// compute the maximum utility for userID.
	private float computeUtility(
		int userID,
		ItemPrediction[] predictions,
		ItemRating[] userRatingsSorted,
		int len) {
		float DEFAULT_RATING = dataMgr.getMeanRatingCached(userID);
		float ALPHA = 5; // Rating list half-life
		float total = 0;
		for (int j = 0; j < len; j++) {
			if (predictions[j] != null) {
				int itemId = predictions[j].getItemID();
				float rating = getRating(itemId, userRatingsSorted);
				// Ignore ratings below the default
				if (rating <= DEFAULT_RATING)
					continue;
				// DEBUG
				//System.out.println("" + j + "\t" + rating);
				total += (rating - DEFAULT_RATING) * Math.pow(0.5, j / (ALPHA - 1));
			}
		}
		return total;
	}

	/**
	 * @param itemId
	 * @param userRatingsSorted
	 * @return
	 */
	private float getRating(int itemID, ItemRating[] userRatingsSorted) {
		float res = -1000;
		for (int i = 0; i < userRatingsSorted.length; i++) {
			if (itemID == userRatingsSorted[i].getItemID()) {
				res = userRatingsSorted[i].getRating();
				break;
			}
		}
		if (res == -1000)
			throw new RuntimeException("There is something wrong!");
		return res;
	}

	/**
	 * @param i
	 * @param itemId
	 * @param algorithm
	 * @return
	 */
	private ItemPrediction predictRate(
		int userId,
		int itemId,
		CFAlgorithm algorithm) {
		try {
			return algorithm.predictRating(userId, itemId);
		} catch (CFNotImplementedException e) {
			// we hope never happen!
			e.printStackTrace();
		}
		return null;
	}

	public static void main(String[] args) {
		String installdir = "D:/ufpe/projetos/FEERS/cfengine-dist"; //System.getProperty("cf.path", ".");
		String filename = installdir + "/lib/" + "CFServer.properties";
		// Load properties
		SystemProperties.loadProperties(filename);

		DataManager dataMgr = new DataManager(true, false);
		DataItemManager dataItemMgr = new DataItemManager(1, dataMgr.getMaxItem());

		CFAlgorithm algorithms[] = new CFAlgorithm[1];
		algorithms[0] = new CollabModalAlgorithm(dataMgr, dataItemMgr);
		//algorithms[0] = new SimplePearsonAlgorithm(dataMgr);
		//algorithms[0] = new ContentBoostedCollabAlgorithm(dataMgr, dataItemMgr);
		//algorithms[3] = new ModalSymbolicAlgorithm(dataMgr, dataItemMgr);
		
		String experimentalDataSetForPrediction =
			installdir
				+ "/lib/"
				+ DatasetSetup.EXPERIMENTAL_DATASET_PREDICTION_FILENAME;
		String experimentalDataSetForRank =
			installdir
				+ "/lib/"
				// + DatasetSetup.EXPERIMENTAL_DATASET_RANK_FILENAME;
				+ "tenFoldCrossValidationExtratifiedDatasetForRankExperimentWith5ItemsInUserProfile.dat";
		CoreAlgorithmTest teste =
			new CoreAlgorithmTest(
				experimentalDataSetForPrediction,
				experimentalDataSetForRank,
				algorithms,
				dataMgr);
		try {
			//teste.startPredictionExperiment();
			teste.startRankExperiment(0);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

}
