/*
 * DatasetSetup.java
 * 
 * Version 1.0
 *
 * 06/04/2004
 * 
 * Copyright (c) 2004 Federal Universtity of Pernambuco. All rights reserved.
 */
package feers.recommender.test;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collections;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.recommender.server.SystemProperties;
import org.recommender.server.SystemUtils;

/**
 * This class is responsible to prepare the database in order
 * to do some experiments.
 * 
 * @author <a href="mailto:bldb@cin.ufpe.br>Byron L. D. Bezerra</a>
 */
public class DatasetSetup {

	private static String classeDriver;

	private static Connection ratingsDBConnection;
	private static String ratingsDBUrl;
	private static String ratingsDBUsername;
	private static String ratingsDBPassword;

	private static Connection itemsDBConnection;
	private static String itemsDBUrl;
	private static String itemsDBUsername;
	private static String itemsDBPassword;

	public static final String EXPERIMENTAL_DATASET_PREDICTION_FILENAME =
		"tenFoldCrossValidationExtratifiedDataset.dat";

	public static final String EXPERIMENTAL_DATASET_RANK_FILENAME = 
		"tenFoldCrossValidationDatasetForRankExperimentWith5ItemsInUserProfile.dat";

	public DatasetSetup() {
		SystemUtils.printDebug("Starting Dataset Setup ...");
		String installdir = System.getProperty("cf.path", ".");
		String filename = installdir + "/lib/" + "CFServer.properties";
		SystemProperties.loadProperties(filename);

		classeDriver = SystemProperties.getString("CFServer.jdbc.drivers");

		itemsDBUrl =
			SystemProperties.getString("CFServer.jdbc.itemdb.url")
				+ "?autoReconnect=true";
		itemsDBUsername =
			SystemProperties.getString("CFServer.jdbc.itemdb.username");
		itemsDBPassword =
			SystemProperties.getString("CFServer.jdbc.itemdb.password");

		ratingsDBUrl =
			SystemProperties.getString("CFServer.jdbc.url")
				+ "?autoReconnect=true";
		ratingsDBUsername =
			SystemProperties.getString("CFServer.jdbc.username");
		ratingsDBPassword =
			SystemProperties.getString("CFServer.jdbc.password");

		try {
			Class.forName(classeDriver);
			itemsDBConnection =
				DriverManager.getConnection(
					itemsDBUrl,
					itemsDBUsername,
					itemsDBPassword);
			ratingsDBConnection =
				DriverManager.getConnection(
					ratingsDBUrl,
					ratingsDBUsername,
					ratingsDBPassword);
		} catch (SQLException sqle) {
			System.out.println(
				"SQLException => ConnectionManager: " + sqle.getMessage());
		} catch (ClassNotFoundException cnfe) {
			System.out.println(
				"Driver nao encontrado => ConnectionManager: "
					+ cnfe.getMessage());
		}
		SystemUtils.printDebug("Dataset Started");
	}

	private PreparedStatement reservaStatement(String sql, Connection con) {
		PreparedStatement statement = null;
		try {
			statement = con.prepareStatement(sql);
		} catch (SQLException sqle) {
			System.out.println(
				"SQLException => ConnectionManager: " + sqle.getMessage());
		}
		return statement;
	}

	/*
	 * This method must remove the items of the ratings database which are not  
	 * in the content database.
	 */
	public void matchDatabases() throws SQLException {
		SystemUtils.printDebug("Match databases started ...");
		long startTime = System.currentTimeMillis();
		// first select all ids in content database
		LinkedList ids = new LinkedList();
		PreparedStatement pstmt =
			reservaStatement("select id from t_movie", itemsDBConnection);
		ResultSet rset = pstmt.executeQuery();
		while (rset.next()) {
			ids.addFirst(new Integer(rset.getInt(1)));
		}
		rset.close();
		pstmt.close();
		// now select all ids in ratings database
		LinkedList idsRatings = new LinkedList();
		pstmt =
			reservaStatement(
				"select ItemId from item_table",
				ratingsDBConnection);
		rset = pstmt.executeQuery();
		while (rset.next()) {
			idsRatings.addFirst(new Integer(rset.getInt(1)));
		}
		rset.close();
		pstmt.close();
		// removing all ids from idsRatings that are in ids
		idsRatings.removeAll(ids);
		// now I can clean the ratings database
		String[] tables = { "item_type_table", "rating_table", "item_table" };
		for (int i = 0; i < tables.length; i++) {
			pstmt =
				reservaStatement(
					"delete from " + tables[i] + " where ItemId = ?",
					ratingsDBConnection);
			for (Iterator iter = idsRatings.iterator(); iter.hasNext();) {
				Integer id = (Integer) iter.next();
				pstmt.clearParameters();
				pstmt.setInt(1, id.intValue());
				try {
					pstmt.execute();
				} catch (Exception e2) {
					e2.printStackTrace();
				}
			}
			pstmt.close();
		}
		long endTime = System.currentTimeMillis();
		SystemUtils.printDebug(
			"Time to match the databases = " + (endTime - startTime) + " ms");
	}

	/**
	 * This method defines the test set with all users having evaluated at least
	 * minEvalQuantOfUserInTestSet items. Additionaly, it randomly chooses 
	 * a numberOfItemsForTrainingEachUser for training set of the user. 
	 * In the RatingSet column, the training data is defined with 0 (zero). 
	 * 
	 * @param minEvalQuantOfUserInTestSet
	 * @param numberOfItemsForTrainingEachUser
	 */
	public void setupTestsetBasedOnNumberOfUserEvaluation(
		int minEvalQuantOfUserInTestSet,
		int numberOfItemsForTrainingEachUser,
		int numberOfItemsForTestingEachUser,
		int numberOfExperiments)
		throws SQLException, IOException {
		SystemUtils.printDebug("Setup testset started ...");
		long startTime = System.currentTimeMillis();

		// first we select all users with the minimum evaluation number
		List users = new LinkedList();
		PreparedStatement pstmt =
			reservaStatement(
				"select userid, count(itemid) as qtd from rating_table group by userid having qtd >= ?",
				ratingsDBConnection);
		pstmt.setInt(1, minEvalQuantOfUserInTestSet);
		ResultSet rset = pstmt.executeQuery();
		while (rset.next()) {
			users.add(new Integer(rset.getInt(1)));
		}
		rset.close();
		pstmt.close();

		// select the test set for each user
		ExperimentalDataset dataset =
			new ExperimentalDataset(
				minEvalQuantOfUserInTestSet,
				numberOfItemsForTrainingEachUser,
				numberOfExperiments);
		pstmt =
			reservaStatement(
				"select ItemId from rating_table where UserId = ?",
				ratingsDBConnection);
		for (Iterator iter = users.iterator(); iter.hasNext();) {
			Integer userId = (Integer) iter.next();
			pstmt.clearParameters();
			pstmt.setInt(1, userId.intValue());
			rset = pstmt.executeQuery();
			List items = new LinkedList();
			while (rset.next()) {
				items.add(new Integer(rset.getInt(1)));
			}
			for (int i = 0; i < numberOfExperiments; i++) {
				UserSets userSets =
					UserSets.generateUserSets(
						userId.intValue(),
						numberOfItemsForTrainingEachUser,
						numberOfItemsForTestingEachUser,
						items);
				dataset.addUserSets(userSets, i);
			}
		}
		rset.close();
		pstmt.close();

		// write the dataset to file
		String installdir = System.getProperty("cf.path", ".");
		String filename = installdir + "/lib/" + EXPERIMENTAL_DATASET_PREDICTION_FILENAME;
		FileOutputStream fout = new FileOutputStream(filename);
		ObjectOutputStream out = new ObjectOutputStream(fout);
		out.writeObject(dataset);
		out.flush();
		out.close();
		fout.close();

		long endTime = System.currentTimeMillis();
		System.out.println("Number of users in test set = " + users.size());
		SystemUtils.printDebug(
			"Time to setup testset = " + (endTime - startTime) + " ms");
	}

	/**
	 * This method defines the testset based on the percentual values specified
	 * in parameter. So, it randomly choose percUsersInTestSet% of all users
	 * in database to build the test set. Additionaly, it randomly chooses for 
	 * each of the previous users a percEvalOfEachUserInTestSet% of items evaluated
	 * by the user to compose the test set of him/her. In the RatingSet column,
	 * the training data is defined with 0 (zero).
	 *    
	 * @param percUsersInTestSet
	 * @param percEvalOfEachUserInTestSet
	 * @throws SQLException
	 */
	public void setupTestsetBasedOnPercentualValues(
		int percUsersInTestSet,
		int percEvalOfEachUserInTestSet)
		throws SQLException {
		SystemUtils.printDebug("Setup testset started ...");
		long startTime = System.currentTimeMillis();
		// first we get the user ids
		List users = new LinkedList();
		PreparedStatement pstmt =
			reservaStatement(
				"select UserId from user_table",
				ratingsDBConnection);
		ResultSet rset = pstmt.executeQuery();
		while (rset.next()) {
			users.add(new Integer(rset.getInt(1)));
		}
		rset.close();
		pstmt.close();
		// randomly select users for test set
		Collections.shuffle(users);
		int offset = (int) ((double) users.size() * percUsersInTestSet / 100.0);
		users = users.subList(0, offset);
		// reset the testset
		PreparedStatement pstmtUpdate =
			reservaStatement(
				"UPDATE rating_table SET ratingset = 0",
				ratingsDBConnection);
		pstmtUpdate.execute();
		pstmtUpdate.close();
		// select the test set for each user
		int testSet = 1;
		pstmt =
			reservaStatement(
				"select ItemId from rating_table where UserId = ?",
				ratingsDBConnection);
		pstmtUpdate =
			reservaStatement(
				"update rating_table SET RatingSet = ? where UserId = ? and ItemId = ?",
				ratingsDBConnection);
		for (Iterator iter = users.iterator(); iter.hasNext(); testSet++) {
			Integer userId = (Integer) iter.next();
			pstmt.clearParameters();
			pstmt.setInt(1, userId.intValue());
			rset = pstmt.executeQuery();
			List items = new LinkedList();
			while (rset.next()) {
				items.add(new Integer(rset.getInt(1)));
			}
			// randomly select items for test set
			Collections.shuffle(items);
			offset =
				(int) ((double) items.size()
					* percEvalOfEachUserInTestSet
					/ 100.0);
			items = items.subList(0, offset);
			// set this test set
			for (Iterator iterator = items.iterator(); iterator.hasNext();) {
				Integer itemId = (Integer) iterator.next();
				pstmtUpdate.clearParameters();
				pstmtUpdate.setInt(1, testSet);
				pstmtUpdate.setInt(2, userId.intValue());
				pstmtUpdate.setInt(3, itemId.intValue());
				pstmtUpdate.execute();
			}
		}
		rset.close();
		pstmt.close();
		pstmtUpdate.close();
		long endTime = System.currentTimeMillis();
		SystemUtils.printDebug(
			"Time to setup testset = " + (endTime - startTime) + " ms");
	}

	/**
	 * This method sets the RatingSet column with code value for userId and all
	 * items in array.
	 * 
	 * @param userId
	 * @param items
	 * @param code
	 */
	public void updateRatingSetColumn(int userId, int items[], int code)
		throws SQLException {
		// reset the testset
		PreparedStatement pstmtUpdate =
			reservaStatement(
				"UPDATE rating_table SET ratingset = 0",
				ratingsDBConnection);
		pstmtUpdate.execute();
		pstmtUpdate.close();
		// select the test set for each user
		pstmtUpdate =
			reservaStatement(
				"update rating_table SET RatingSet = ? where UserId = ? and ItemId = ?",
				ratingsDBConnection);
		for (int i = 0; i < items.length; i++) {
			int itemId = items[i];
			pstmtUpdate.clearParameters();
			pstmtUpdate.setInt(1, code);
			pstmtUpdate.setInt(2, userId);
			pstmtUpdate.setInt(3, itemId);
			pstmtUpdate.execute();
		}
		pstmtUpdate.close();

	}

	public void createNFoldCrossValidationExtratifiedDataset(
		int n,
		int minEvaluationPerUser)
		throws Exception {
		SystemUtils.printDebug("Setup testset started ...");
		long startTime = System.currentTimeMillis();
		int folderSize = minEvaluationPerUser / n; // equals to testset size
		// first we get the user ids
		List users = new LinkedList();
		// first we select all users with the minimum evaluation number
		PreparedStatement pstmt =
			reservaStatement(
				"select userid, count(itemid) as qtd from rating_table group by userid having qtd >= ?",
				ratingsDBConnection);
		pstmt.setInt(1, minEvaluationPerUser);
		ResultSet rset = pstmt.executeQuery();
		while (rset.next()) {
			users.add(new Integer(rset.getInt(1)));
		}
		rset.close();
		pstmt.close();

		// create this dataset using N fold cross validation process
		int itemsInTestSet = folderSize;
		int itemsInTrainingSet = minEvaluationPerUser - folderSize;
		ExperimentalDataset dataset =
			new ExperimentalDataset(minEvaluationPerUser, itemsInTrainingSet, n);
		pstmt =
			reservaStatement(
				"select ItemId, Rating from rating_table where UserId = ?",
				ratingsDBConnection);
		for (Iterator iter = users.iterator(); iter.hasNext();) {
			Integer userId = (Integer) iter.next();
			pstmt.clearParameters();
			pstmt.setInt(1, userId.intValue());
			rset = pstmt.executeQuery();
			// get results
			int total = 0;
			Hashtable extractedItems = new Hashtable();
			while (rset.next()) {
				Double rate = new Double(rset.getDouble(2));
				int itemId = rset.getInt(1);
				List items = null;
				if (extractedItems.containsKey(rate))
					items = (List) extractedItems.get(rate);
				else {
					items = new LinkedList();
					extractedItems.put(rate, items);
				}
				items.add(new Integer(itemId));
				total++;
			}
			// shuffle results
			for (Iterator iterator = extractedItems.keySet().iterator();
				iterator.hasNext();
				) {
				Object rate = iterator.next();
				List items = (List) extractedItems.get(rate);
				Collections.shuffle(items);
			}
			// put items in folders
			List[] folders = new LinkedList[n];
			for (int i = 0; i < folders.length; i++) {
				folders[i] = new LinkedList();
			}
			boolean finished = false;
			int indexFolder = 0;
			int indexRate = 0;
			Object[] rates = extractedItems.keySet().toArray();
			while (!finished) {
				Object rate = rates[indexRate];
				List items = (List) extractedItems.get(rate);
				int oldIndex = indexRate;
				while (items.size() == 0) {
					indexRate = (indexRate + 1) % rates.length;
					rate = rates[indexRate];
					items = (List) extractedItems.get(rate);
					if (indexRate == oldIndex) {
						// all extracts are empty
						finished = true;
						break;
					}
				}
				if (!finished) {
					oldIndex = indexFolder;
					while (folders[indexFolder].size() == folderSize) {
						indexFolder = (indexFolder + 1) % folders.length;
						if (indexFolder == oldIndex) {
							// all folders are filled
							finished = true;
							break;
						}
					}
					if (!finished) {
						Object itemId = items.remove(0);
						folders[indexFolder].add(itemId);
						indexFolder = (indexFolder + 1) % folders.length;
						indexRate = (indexRate + 1) % rates.length;
					}
				}
			}

			// create UserSets and put them in dataset
			for (int i = 0; i < n; i++) {
				int[] trainingItems = new int[itemsInTrainingSet];
				int indexTraining = 0;
				int[] testItems = new int[itemsInTestSet];
				int indexTest = 0;
				for (int index = 0; index < n; index++) {
					if (index == i) {
						// put in test set
						for (int k = 0; k < folders[index].size(); k++) {
							Integer item = (Integer) folders[index].get(k);
							testItems[indexTest++] = item.intValue();
						}
					} else {
						// put in training set
						for (int k = 0; k < folders[index].size(); k++) {
							Integer item = (Integer) folders[index].get(k);
							trainingItems[indexTraining++] = item.intValue();
						}
					}
				}
				UserSets userSets =
					new UserSets(userId.intValue(), trainingItems, testItems);
				dataset.addUserSets(userSets, i);
			}
		}
		rset.close();
		pstmt.close();

		// write the dataset to file
		String installdir = System.getProperty("cf.path", ".");
		String filename = installdir + "/lib/" + EXPERIMENTAL_DATASET_PREDICTION_FILENAME;
		FileOutputStream fout = new FileOutputStream(filename);
		ObjectOutputStream out = new ObjectOutputStream(fout);
		out.writeObject(dataset);
		out.flush();
		out.close();
		fout.close();

		long endTime = System.currentTimeMillis();
		System.out.println("Number of users in test set = " + users.size());
		SystemUtils.printDebug(
			"Time to setup testset = " + (endTime - startTime) + " ms");
	}

	public void createNFoldCrossValidationExtratifiedForRankExperimentDatasetWithMItemsInUserProfile(
		int n, int m,
		int minEvaluationPerUser)
		throws Exception {
		SystemUtils.printDebug("Setup testset started ...");
		long startTime = System.currentTimeMillis();
		int folderSize = m;
		// first we get the user ids
		List users = new LinkedList();
		// first we select all users with the minimum evaluation number
		PreparedStatement pstmt =
			reservaStatement(
				"select userid, count(itemid) as qtd from rating_table group by userid having qtd >= ?",
				ratingsDBConnection);
		pstmt.setInt(1, minEvaluationPerUser);
		ResultSet rset = pstmt.executeQuery();
		while (rset.next()) {
			users.add(new Integer(rset.getInt(1)));
		}
		rset.close();
		pstmt.close();

		// create this dataset using N fold cross validation process
		int itemsInTestSet = folderSize * (n-1);
		int itemsInTrainingSet = folderSize;
		ExperimentalDataset dataset =
			new ExperimentalDataset(itemsInTestSet, itemsInTrainingSet, n);
		pstmt =
			reservaStatement(
				"select ItemId, Rating from rating_table where UserId = ?",
				ratingsDBConnection);
		for (Iterator iter = users.iterator(); iter.hasNext();) {
			Integer userId = (Integer) iter.next();
			pstmt.clearParameters();
			pstmt.setInt(1, userId.intValue());
			rset = pstmt.executeQuery();
			// get results
			int total = 0;
			Hashtable extractedItems = new Hashtable();
			while (rset.next()) {
				Double rate = new Double(rset.getDouble(2));
				int itemId = rset.getInt(1);
				List items = null;
				if (extractedItems.containsKey(rate))
					items = (List) extractedItems.get(rate);
				else {
					items = new LinkedList();
					extractedItems.put(rate, items);
				}
				items.add(new Integer(itemId));
				total++;
			}
			// shuffle results
			for (Iterator iterator = extractedItems.keySet().iterator();
				iterator.hasNext();
				) {
				Object rate = iterator.next();
				List items = (List) extractedItems.get(rate);
				Collections.shuffle(items);
			}
			// put items in folders
			List[] folders = new LinkedList[n];
			for (int i = 0; i < folders.length; i++) {
				folders[i] = new LinkedList();
			}
			boolean finished = false;
			int indexFolder = 0;
			int indexRate = 0;
			Object[] rates = extractedItems.keySet().toArray();
			while (!finished) {
				Object rate = rates[indexRate];
				List items = (List) extractedItems.get(rate);
				int oldIndex = indexRate;
				while (items.size() == 0) {
					indexRate = (indexRate + 1) % rates.length;
					rate = rates[indexRate];
					items = (List) extractedItems.get(rate);
					if (indexRate == oldIndex) {
						// all extracts are empty
						finished = true;
						break;
					}
				}
				if (!finished) {
					oldIndex = indexFolder;
					while (folders[indexFolder].size() == folderSize) {
						indexFolder = (indexFolder + 1) % folders.length;
						if (indexFolder == oldIndex) {
							// all folders are filled
							finished = true;
							break;
						}
					}
					if (!finished) {
						Object itemId = items.remove(0);
						folders[indexFolder].add(itemId);
						indexFolder = (indexFolder + 1) % folders.length;
						indexRate = (indexRate + 1) % rates.length;
					}
				}
			}

			// create UserSets and put them in dataset
			for (int i = 0; i < n; i++) {
				int[] trainingItems = new int[itemsInTrainingSet];
				int indexTraining = 0;
				int[] testItems = new int[itemsInTestSet];
				int indexTest = 0;
				for (int index = 0; index < n; index++) {
					if (index != i) {
						// put in test set
						for (int k = 0; k < folders[index].size(); k++) {
							Integer item = (Integer) folders[index].get(k);
							testItems[indexTest++] = item.intValue();
						}
					} else {
						// put in training set
						for (int k = 0; k < folders[index].size(); k++) {
							Integer item = (Integer) folders[index].get(k);
							trainingItems[indexTraining++] = item.intValue();
						}
					}
				}
				UserSets userSets =
					new UserSets(userId.intValue(), trainingItems, testItems);
				dataset.addUserSets(userSets, i);
			}
		}
		rset.close();
		pstmt.close();

		// write the dataset to file
		String installdir = System.getProperty("cf.path", ".");
		String filename = installdir + "/lib/" + EXPERIMENTAL_DATASET_RANK_FILENAME;
		FileOutputStream fout = new FileOutputStream(filename);
		ObjectOutputStream out = new ObjectOutputStream(fout);
		out.writeObject(dataset);
		out.flush();
		out.close();
		fout.close();

		long endTime = System.currentTimeMillis();
		System.out.println("Number of users in test set = " + users.size());
		SystemUtils.printDebug(
			"Time to setup testset = " + (endTime - startTime) + " ms");
	}
	
	public void createNFoldCrossValidationForRankExperimentDatasetWithMItemsInUserProfile(
			int n, int m,
			int minEvaluationPerUser)
			throws Exception {
			SystemUtils.printDebug("Setup testset started ...");
			long startTime = System.currentTimeMillis();
			int folderSize = m;
			// first we get the user ids
			List users = new LinkedList();
			// first we select all users with the minimum evaluation number
			PreparedStatement pstmt =
				reservaStatement(
					"select userid, count(itemid) as qtd from rating_table group by userid having qtd >= ?",
					ratingsDBConnection);
			pstmt.setInt(1, minEvaluationPerUser);
			ResultSet rset = pstmt.executeQuery();
			while (rset.next()) {
				users.add(new Integer(rset.getInt(1)));
			}
			rset.close();
			pstmt.close();

			// create this dataset using N fold cross validation process
			int itemsInTestSet = folderSize * (n-1);
			int itemsInTrainingSet = folderSize;
			ExperimentalDataset dataset =
				new ExperimentalDataset(itemsInTestSet, itemsInTrainingSet, n);
			pstmt =
				reservaStatement(
					"select ItemId, Rating from rating_table where UserId = ?",
					ratingsDBConnection);
			for (Iterator iter = users.iterator(); iter.hasNext();) {
				Integer userId = (Integer) iter.next();
				pstmt.clearParameters();
				pstmt.setInt(1, userId.intValue());
				rset = pstmt.executeQuery();
				// get results
				List items = new LinkedList();
				while (rset.next()) {
					int itemId = rset.getInt(1);
					items.add(new Integer(itemId));
				}
				Collections.shuffle(items);
				
				// put items in folders
				List[] folders = new List[n];
				for (int i = 0; i < folders.length; i++) {
					int lI = i*m;
					int lF = (i+1)*m;
					folders[i] = items.subList(lI, lF);
				}
				
				// create UserSets and put them in dataset
				for (int i = 0; i < n; i++) {
					int[] trainingItems = new int[itemsInTrainingSet];
					int indexTraining = 0;
					int[] testItems = new int[itemsInTestSet];
					int indexTest = 0;
					for (int index = 0; index < n; index++) {
						if (index != i) {
							// put in test set
							for (int k = 0; k < folders[index].size(); k++) {
								Integer item = (Integer) folders[index].get(k);
								testItems[indexTest++] = item.intValue();
							}
						} else {
							// put in training set
							for (int k = 0; k < folders[index].size(); k++) {
								Integer item = (Integer) folders[index].get(k);
								trainingItems[indexTraining++] = item.intValue();
							}
						}
					}
					UserSets userSets =
						new UserSets(userId.intValue(), trainingItems, testItems);
					dataset.addUserSets(userSets, i);
				}
			}
			rset.close();
			pstmt.close();

			// write the dataset to file
			String installdir = System.getProperty("cf.path", ".");
			String filename = installdir + "/lib/" + EXPERIMENTAL_DATASET_RANK_FILENAME;
			FileOutputStream fout = new FileOutputStream(filename);
			ObjectOutputStream out = new ObjectOutputStream(fout);
			out.writeObject(dataset);
			out.flush();
			out.close();
			fout.close();

			long endTime = System.currentTimeMillis();
			System.out.println("Number of users in test set = " + users.size());
			SystemUtils.printDebug(
				"Time to setup testset = " + (endTime - startTime) + " ms");
		}

	public static void main(String args[]) {
		DatasetSetup dataSetup = new DatasetSetup();
		try {
			//dataSetup.matchDatabases();
			// used for annotation in context experiment
			// dataSetup.createNFoldCrossValidationExtratifiedDataset(10, 100);
			// used for find good items experiment
			dataSetup.createNFoldCrossValidationForRankExperimentDatasetWithMItemsInUserProfile(
				10,
				5,
				100);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

}
