/*
 * ExperimentalDataset.java
 * 
 * Version 1.0
 *
 * 08/04/2004
 * 
 * Copyright (c) 2004 Federal Universtity of Pernambuco. All rights reserved.
 */
package feers.recommender.test;

import java.io.FileInputStream;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.recommender.algorithms.CFAlgorithm;
import org.recommender.algorithms.TestStats;
import org.recommender.server.Stats;

/**
 * Encapsulates the user information (UserSets) to produce the experiments.
 * 
 * @author <a href="mailto:bldb@cin.ufpe.br>Byron L. D. Bezerra</a>
 */
public class ExperimentalDataset implements Serializable {

	/*
	 * Each position of this vector represents the list of users
	 */
	private List[] usersSets;

	/*
	 * Each column i of this matrix is associated with the results for experimental
	 * dataset defined by usersSets[i]. Each line represents the results for one algorithm.
	 */
	private transient TestStats[][] results;

	private transient CFAlgorithm[] algorithms;

	private int numberOfItemsInTestSet,
		numberOfItemsForTrainingEachUser,
		numberOfExperiments;

	public ExperimentalDataset(
		int numberOfItemsInTestSet,
		int numberOfItemsForTrainingEachUser,
		int numberOfExperiments) {
		this.numberOfItemsInTestSet = numberOfItemsInTestSet;
		this.numberOfItemsForTrainingEachUser =
			numberOfItemsForTrainingEachUser;
		this.numberOfExperiments = numberOfExperiments;
		this.usersSets = new List[numberOfExperiments];
		for (int i = 0; i < usersSets.length; i++) {
			usersSets[i] = new LinkedList();
		}
	}

	public void addUserSets(UserSets u, int experiment) {
		this.usersSets[experiment].add(u);
	}

	public Iterator getUserSets(int experiment) {
		return usersSets[experiment].iterator();
	}

	public String toString() {
		String r =
			"\nNumber of items in test set = "
				+ numberOfItemsInTestSet
				+ "\nNumber of items used for training each user = "
				+ numberOfItemsForTrainingEachUser
				+ "\nNumber of (training,test) sets generated per user = "
				+ this.numberOfExperiments;
		return r;
	}

	public static ExperimentalDataset loadDataset(String filename) {
		ExperimentalDataset dataset = null;
		try {
			FileInputStream fin = new FileInputStream(filename);
			ObjectInputStream in = new ObjectInputStream(fin);
			dataset = (ExperimentalDataset) in.readObject();
			in.close();
			fin.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
		return dataset;
	}

	/**
	 * @return
	 */
	public int getNumberOfExperiments() {
		return numberOfExperiments;
	}

	public void setResults(int algorithm, TestStats[] res) {
		this.results[algorithm] = res;
	}

	/**
	 * @param algorithms
	 */
	public void setAlgorithms(CFAlgorithm[] algorithms) {
		this.algorithms = algorithms;
		this.results = new TestStats[algorithms.length][];
	}

	public void printResults() {
		System.out.println(this);
		for (int i = 0; i < this.algorithms.length; i++) {
			System.out.println(
				"\nPrint results for algorithm " + algorithms[i]);
			// calculate the average and std values for mae and breese
			Stats mae = new Stats();
			Stats mse = new Stats();
			Stats breese = new Stats();
			System.out.println(
				"\nPartial results for breese = ");
			for (int j = 0; j < this.results[i].length; j++) {
				if (results[i][j].getMAE() != -1)
					mae.addNum(results[i][j].getMAE());
				if (results[i][j].getMSE() != -1)
					mse.addNum(results[i][j].getMSE());
				if (results[i][j].getBreese() != -1) {
					float b = results[i][j].getBreese();
					breese.addNum(b);
					System.out.print(b + " ");
				}
			}
			System.out.println("MAE mean = " + mae.calcMean());
			System.out.println("MAE std = " + mae.calcSD());
			System.out.println("MSE mean = " + mse.calcMean());
			System.out.println("MSE std = " + mse.calcSD());
			System.out.println("Breese mean = " + breese.calcMean());
			System.out.println("Breese std = " + breese.calcSD());
		}
	}

	public void printStatisticalTest(int targetAlg) {
		for (int i = 0; i < this.algorithms.length; i++) {
			if (i != targetAlg) {
				Stats diffMae = new Stats();
				Stats diffMse = new Stats();
				Stats diffBreese = new Stats();
				for (int j = 0; j < this.results[i].length; j++) {
					if (results[i][j].getMAE() != -1) {
						float d = results[targetAlg][j].getMAE() - results[i][j].getMAE(); 
						diffMae.addNum(d);
					}
					if (results[i][j].getMSE() != -1) {
						float d = results[targetAlg][j].getMSE() - results[i][j].getMSE(); 
						diffMse.addNum(d);
					}
					if (results[i][j].getBreese() != -1) {
						float d = results[targetAlg][j].getBreese() - results[i][j].getBreese(); 
						diffBreese.addNum(d);
					}
				}
				System.out.println(
					"\n\nComparison results between " + algorithms[targetAlg] + " and " + algorithms[i]);
				System.out.println("\tMAE diff mean = " + diffMae.calcMean());
				System.out.println("\tMAE diff std  = " + diffMae.calcSD());
				System.out.println("\tMAE t value   = " + diffMae.calcTValue());
				System.out.println("\tMSE diff mean = " + diffMse.calcMean());
				System.out.println("\tMSE diff std  = " + diffMse.calcSD());
				System.out.println("\tMSE t value   = " + diffMse.calcTValue());
				System.out.println("\tBreese diff mean = " + diffBreese.calcMean());
				System.out.println("\tBreese diff std  = " + diffBreese.calcSD());
				System.out.println("\tBreese t value   = " + diffBreese.calcTValue());
			}
		}
	}

}
