/*
 * AutoTester.java
 *
 * Version 1.1
 *
 * 23/08/2002
 *
 * Copyright (c) 2002 Oregon State Universtity. All rights reserved.
 */

package org.recommender.algorithms;

//import edu.orst.recommender.server.DataBank.*;

import org.recommender.algorithms.Experimental.Comp;
import org.recommender.server.*;

import java.util.*;
import java.io.*;
import java.sql.*;
import java.util.Date;

import java.lang.reflect.*;

//import ojb.broker.PersistenceBroker;
//import ojb.broker.PersistenceBrokerFactory;
//import ojb.broker.PersistenceBrokerException;

/**
 * A class containing test code to be run immediately upon server loading.
 * Evaluates the effectiveness of an implemented algorithm using
 * metrics defined in TestStats.  Includes a rating test for generic
 * training/test splits and an all-but-one test, for comparing to
 * published papers that use all-but-one tests.  Testing for top-N
 * recommendations is in progress, but doesn't yet work very well so
 * it has been commented out.
 *
 * Note: to run the autotest mode, we need a different table format from
 * the server mode: autotest mode need four columns in the RATING_TABLE:
 * UserID, ItemID, Rating and RatingSet.
 *
 * @author Daniel Lowd &lt;dlowd@cs.orst.edu&gt;
 * @author Shuzhen Nong &lt;nong@cs.orst.edu&gt;
 */

public class AutoTester {
	private DataManager dataMgr;
	private CFAlgorithm algorithm;

	// HACK: allow an alternate algorithm to augment the main algorithm
	// by supplying a prediction when the first cannot make one.
	private CFAlgorithm altAlgorithm;

	private int MIN_RATING;
	private int MAX_RATING;

	/**
	 * Construct an AutoTester object to test the provided algorithm
	 * using data from the provided data manager.
	 * @param dataMgr data manager to provide ratings and other support
	 *     services.
	 * @param algorithm algorithm to test
	 */
	public AutoTester(DataManager dataMgr, CFAlgorithm algorithm) {
		this.dataMgr = dataMgr;
		this.algorithm = algorithm;

		altAlgorithm =
			constructAlgorithm(
				SystemProperties.getString("CFServer.alg.altname"),
				dataMgr);

		System.out.println("Alternate algorithm: " + altAlgorithm);

		MIN_RATING = (int) dataMgr.getMinRating();
		MAX_RATING = (int) dataMgr.getMaxRating();
	}

	/**
	 * Given the name of an algorithm and a DataManager, construct an instance
	 * of the specified algorithm.
	 * @param algName
	 * @param dataMgr
	 */
	static CFAlgorithm constructAlgorithm(
		String algName,
		DataManager dataMgr) {
		CFAlgorithm retVal = null;

		if (algName != null && algName.length() != 0) {
			try {
				Class algorithm = Class.forName(algName);
				Constructor[] cons = algorithm.getConstructors();
				Object newAlgObj =
					cons[0].newInstance(new Object[] { dataMgr });
				retVal = (CFAlgorithm) newAlgObj;

			} catch (Exception e) {
				System.out.println(
					"Error loading alternate algorithm: " + algName);
				retVal = null;
			}
		}

		return retVal;
	}

	/**
	 * Run automated tests and return the results using the provided
	 * algorithm, data manager.
	 * @param dataMgr data manager to provide ratings and other support
	 *     services.
	 * @param algorithm algorithm to test
	 * @param whereClause place to load the testset
	 */
	public static void runTests(
		DataManager dataMgr,
		CFAlgorithm algorithm,
		String whereClause) {
		AutoTester at = new AutoTester(dataMgr, algorithm);

		/*
		             try {
		        at.allButOne(dataMgr.getRatingList(whereClause));
		             } catch( SQLException e ) {
		        System.out.println("Error: could not load test set \""
		            + whereClause + "\"");
		             }
		*/     

		try {
			at.testRatings(dataMgr.getRatingList(whereClause));
		} catch (SQLException e) {
			System.out.println(
				"Error: could not load test set \"" + whereClause + "\"");
		}

		
		try {
		    at.testCoherence();
		} catch (Exception e) {
		   e.printStackTrace();
		}    
		
		try {
		    at.testRecommendations(dataMgr.getRatingList(whereClause));
		} catch( SQLException e ) {
		    System.out.println("Error: could not load test set \""
		           + whereClause + "\"");
		}

		try {
			at.testRecommendations2(dataMgr.getRatingList(whereClause));
		} catch (SQLException e) {
			System.out.println(
				"Error: could not load test set \"" + whereClause + "\"");
		}

	}

	/**
	 * Run an all-but-one test, given an iterator of ratings representing
	 * the test set.  All ratings for any given user must be in a single
	 * block for this to work properly.  That is, ratings must be grouped
	 * by UserID.
	 *
	 * In testing predictions, this method loads all ratings in the test set
	 * from a given user, then removes each one temporarily for predicting it.
	 * Ratings for that user that are already cached may be used by the
	 * algorithms as training data, but are not removed or predicted.
	 * When all the user's ratings in the test set have been predicted,
	 * these ratings are removed and the next user is tested.
	 *
	 * The overall Mean Average Error (MAE) is printed out after every 1000
	 * predictions.  When testing is complete, more thorough results are
	 * displayed to System.out in a format determined by TestStats.
	 *
	 * @param ratings Iterator of all ItemRatings in the test set, grouped
	 *      by UserID.
	 */
	public void allButOne(Iterator ratings) {
		System.out.println("Running an all-but-one test...");

		long ratingTestBegin = System.currentTimeMillis();
		TestStats testStats = new TestStats(dataMgr);
		List currUserRatings = new LinkedList();

		int lastUserID = 0;
		int i = 0;
		while (ratings.hasNext()) {
			ItemRating currRating = (ItemRating) ratings.next();
			int userID = currRating.getUserID();

			if (userID != lastUserID) {
				addCachedRatings(currUserRatings.iterator());
				testCachedRatings(currUserRatings.iterator(), testStats);
				removeCachedRatings(currUserRatings.iterator());

				currUserRatings = new LinkedList();
				lastUserID = userID;
			}
			currUserRatings.add(currRating);

			if (++i % 1000 == 0) {
				System.out.println(testStats.getMAE());
			}
		}

		long ratingTestEnd = System.currentTimeMillis();
		System.out.println(
			"Rating test time: "
				+ (ratingTestEnd - ratingTestBegin) / 1000
				+ "s");
		System.out.println(testStats);
	}

	/**
	 * Add ratings to cache.
	 * @param ratings the tested user's iter of ItemRating
	 */
	protected void addCachedRatings(Iterator ratings) {
		if (!ratings.hasNext()) {
			return;
		}

		ItemRating ir = null;
		while (ratings.hasNext()) {
			ir = (ItemRating) ratings.next();
			dataMgr.addCachedRating(ir);
		}

		try {
			algorithm.updateUser(ir.getUserID());
		} catch (CFNotImplementedException e) {
			//put debug here
		}
		dataMgr.updateStats(ir.getUserID());
	}

	/**
	 * Remove ratings from cache
	 * @param ratings the tested user's iter of ItemRating
	 */
	protected void removeCachedRatings(Iterator ratings) {
		if (!ratings.hasNext()) {
			return;
		}

		ItemRating ir = null;
		while (ratings.hasNext()) {
			ir = (ItemRating) ratings.next();
			dataMgr.removeCachedRating(ir);
		}
		try {
			algorithm.updateUser(ir.getUserID());
		} catch (CFNotImplementedException e) {
			//put debug here
		}
		dataMgr.updateStats(ir.getUserID());
	}

	/**
	 * Run test for one user - the one just added in cache,
	 * and add results to testStats.
	 * @param ratings the tested user's ItemRating
	 * @param testStats place to put test results
	 */
	protected void testCachedRatings(Iterator ratings, TestStats testStats) {
		while (ratings.hasNext()) {
			ItemRating rating = (ItemRating) ratings.next();
			dataMgr.removeCachedRating(rating);

			int userID = rating.getUserID();
			int itemID = rating.getItemID();
			float actualRating = rating.getRating();

			// Take into account the removed rating
			try {
				algorithm.updateUser(rating.getUserID());
			} catch (CFNotImplementedException e) {
				//put debug here
			}

			//dataMgr.updateStats(userID);
			ItemPrediction predictedRating = null;
			try {
				predictedRating = algorithm.predictRating(userID, itemID);
			} catch (CFNotImplementedException e) {
				System.out.println(e);
			}
			testStats.addRating(
				predictedRating.getUserID(),
				predictedRating.getItemID(),
				actualRating,
				predictedRating.getPrediction());

			/* Restore the rating
			 * (This method isn't terribly efficient, but it's
			 * cleaner for it to be encapsulated in DataManager.)
			 */
			dataMgr.addCachedRating(rating);
		}
	}

	/**
	 * Run a training/test split rating test, given an iterator of ratings
	 * representing the test set.  It is presumed that all ratings in the
	 * training set are already loaded into the data manager.  All ratings for
	 * any given user should be in a single block for some metrics to work
	 * properly.  That is, ratings ought to be grouped by UserID.  This allows
	 * data to be collected on the average mean error per user, in addition to
	 * the overall mean rating error.
	 *
	 * This method tests ratings from the test set one at a time.  If a
	 * rating in the test set is for a user with fewer than 5 ratings loaded
	 * by the data manager, then the rating is skipped.  None of the
	 * ratings in the test set are ever added to the database to be used
	 * as training data.  Thus, it is important that the test set consist
	 * of additional ratings for users who are also in the training set.
	 *
	 * The overall Mean Average Error (MAE) is printed out after every 1000
	 * predictions.  When testing is complete, more thorough results are
	 * displayed to System.out in a format determined by TestStats.
	 *
	 * @param iter Iterator of all ItemRatings in the test set, grouped by
	 * UserID.
	 */
	public void testRatings(Iterator iter) {
		long ratingTestBegin = System.currentTimeMillis();

		float[] MAEnums = new float[21];
		float[] MAEdens = new float[21];

		int n_ratings = 0;

		TestStats testStats = new TestStats(dataMgr);
		TestStats augStats = new TestStats(dataMgr);

		int i = 0;
		while (iter.hasNext()) {
			++n_ratings;

			ItemRating ir = (ItemRating) iter.next();
			int userID = ir.getUserID();
			int itemID = ir.getItemID();
			float actualRating = ir.getRating();

			if (dataMgr.getNumRatings(userID) < 5) {
				continue;
			}

			ItemPrediction returnRating = null;
			ItemPrediction augmentedRating = null;
			try {
				returnRating = algorithm.predictRating(userID, itemID);
				if (returnRating.getPrediction() == dataMgr.getErrRating()
					&& altAlgorithm != null) {
					augmentedRating =
						altAlgorithm.predictRating(userID, itemID);
				} else {
					augmentedRating = returnRating;
				}
			} catch (Exception e) {
				System.out.println(
					" exception in invoking predictRating method ");
				e.printStackTrace();
			}

			testStats.addRating(
				userID,
				itemID,
				actualRating,
				returnRating.getPrediction());
			augStats.addRating(
				userID,
				itemID,
				actualRating,
				augmentedRating.getPrediction());
			i++;

			if (i % 1000 == 0) {
				System.out.println(
					"the " + i + "th rating, MAE= " + testStats.getMAE());
			}
		}

		System.out.println(
			" number of ratings tested in AutoTester = " + n_ratings);

		long ratingTestEnd = System.currentTimeMillis();
		System.out.println(
			"Rating test time: "
				+ (ratingTestEnd - ratingTestBegin) / 1000
				+ "s");

		System.out.println(testStats);

		if (altAlgorithm != null) {
			System.out.println(
				"Augmented stats -- MAE: "
					+ augStats.getMAE()
					+ "; MSE: "
					+ augStats.getMSE()
					+ "; Coverage: "
					+ augStats.getCoverage());
		}

		//////////////Shuzhen add: store the new experiment result ojbect//////////////

		/*     System.out.println(" try to create a new experiment ojbect...");
		      PersistenceBroker broker;
		      broker = null;
		     try
		     {
		     broker = PersistenceBrokerFactory.createPersistenceBroker("repository.xml");
		     }
		     catch (Throwable t)
		     {
		         t.printStackTrace();
		     }
		     Date exp_time = new Date();
		     Experiment newExperiment = new Experiment(0,exp_time.toString(),algorithm.toString(),testStats.getMAE(), testStats.getCoverage());
		     try
		     {
		         broker.beginTransaction();
		         broker.store(newExperiment);
		         broker.commitTransaction();
		     }
		     catch (Exception ex)
		     {
		         broker.abortTransaction();
		         System.out.println(ex.getMessage());
		         ex.printStackTrace();
		     }
		     System.out.println(" finish storing the new experiment...");
		 */

		/////////////////////End of OJB actions //////////////////////////////////

	}

	/* Test the quality of an algorithm's Top N recommendations.  In
	 * progress... currently, this doesn't work very well.  Rates
	 * algorithms by looking at the average displacement of a rating,
	 * i.e. if an item known to be a 4 is ranked before 2 items known
	 * to be 5's, then that item has an error of 2.  We average over
	 * all item ratings in the test set that appeared in the top-N
	 * list requested from the algorithm.  Note that only returning
	 * items we've already rated (and thus are not in the test set)
	 * would result in an error of 0.  Thus, this metric is imperfect
	 * and should be revised.  The implementation may also be buggy.
	     public void testRecommendations(Iterator iter)
	     {
	    long ratingTestBegin = System.currentTimeMillis();
	    int lastUserID = 0;
	    int usersTested = 0;
	    float distance = 0;
	    // Cache each user's ratings as an array
	    int maxItem = dataMgr.getMaxItem();
	    int[] userRatings = new int[maxItem+1];
	    Arrays.fill(userRatings, 0);
	    while (iter.hasNext())
	    {
	        ItemRating currRating = (ItemRating)iter.next();
	        int userID = currRating.getUserID();
	        int itemID = currRating.getItemID();
	        int rating = (int)currRating.getRating();
	        if (dataMgr.getNumRatings(userID) < 5)
	            continue;
	        if ( userID != lastUserID && lastUserID != 0 )
	        {
	            // Add a rating for the last user
	            distance += rateRecommendation(lastUserID, userRatings);
	            usersTested++;
	            // Clear out ratings for the next user
	            for (int i = 0; i <= maxItem; i++)
	                userRatings[i] = 0;
	            // Print out some stats along the way, to satisfy curiosity
	            if (usersTested % 100 == 0)
	                System.out.println((float)distance/usersTested);
	        }
	        userRatings[itemID] = rating;
	        lastUserID = userID;
	    }
	    // Output statistics
	    System.out.println("\n" + distance/usersTested + " avg error");
	    System.out.println("Users: " + usersTested);
	    long ratingTestEnd = System.currentTimeMillis();
	    System.out.println("Rating test time: "
	            + (ratingTestEnd - ratingTestBegin)/1000 + "s");
	     }
	     // Helper function -- rates the algorithm's recommendation
	     // by comparing it to the test set ratings for the
	     // user, which are passed in the realRatings array.
	     final static int NUM_RECS = 100;
	     private float rateRecommendation(int userID, int[] realRatings)
	     {
	    ItemRating[] ratings = null;
	    // Get the recommendations using the current algorithm
	    try {
	        ratings = algorithm.getRecommendations(userID, NUM_RECS);
	    } catch (SQLException e) {
	        System.out.println("Error: " + e);
	    }
	    int numRecs = ratings.length < NUM_RECS ? ratings.length : NUM_RECS;
	    // Keep track of how many (known) items have each rating
	    int[] numRatings = new int[MAX_RATING + 1];
	    Arrays.fill(numRatings, 0);
	    // Count up all the real ratings
	    for (int i = 0; i < realRatings.length; i++)
	    {
	        if (realRatings[i] < 0)
	            continue;
	        numRatings[realRatings[i]]++;
	    }
	    // Keep track of total error and number of items that's based on,
	    // so we can compute the average number of items that should have
	    // been placed before an item.
	    int error = 0;
	    int numRated = 0;
	    for (int currRating = 0; currRating < numRecs; currRating++)
	    {
	        int itemID = ratings[currRating].getItemID();
	        int realVal = realRatings[itemID];
	        if (realVal > 0)
	        {
	            numRated++;
	            numRatings[realVal]--;
	            // If this item has come before N others with better ratings,
	            // increase the error by N.
	            for (int i = realVal + 1; i <= MAX_RATING; i++)
	                error += numRatings[i];
	        }
	    }
	    if (numRated == 0)
	        return 0;
	    return ((float)error)/numRated;
	     }
	 */

	/* An alternate top-N metric.  This is an attempt to implement the
	 * top-N metric described by Breese, et al. in the paper,
	 * "An Empirical Analysis of Predictive Algorithms for Collaborative
	 * Filtering".  Also may be buggy, since I have not yet been able to
	 * reproduce the results of that paper.
	 */
	public void testRecommendations(Iterator iter) {
		long ratingTestBegin = System.currentTimeMillis();
		int lastUserID = -1;
		int usersTested = 0;
		float numerator = 0;
		float denominator = 0;
		// Cache each user's ratings as an array
		int maxItem = dataMgr.getMaxItem();
		ItemRating[] userRatings = new ItemRating[maxItem + 1];
		for (int i = 0; i <= maxItem; i++)
			userRatings[i] = null;
		int numRatings = 0;
		while (iter.hasNext()) {
			ItemRating currRating = (ItemRating) iter.next();
			int userID = currRating.getUserID();
			if (dataMgr.getNumRatings(userID) < 5)
				continue;
			if (userID != lastUserID && lastUserID != -1) {
				// Sort it by rating value and get its rating
				Arrays.sort(userRatings, 0, numRatings, new Comp());
				// Get the algorithm's best top-N list, using the
				// ratings it has available to it.
				ItemPrediction[] predictions = null;
				try {
					predictions = algorithm.getRecommendations(userID, 50);
				} catch (SQLException e) {
					e.printStackTrace();
				} catch (CFNotImplementedException e) {
					e.printStackTrace();
				}
				// Add ratings...
				for (int i = 0; i < numRatings; i++)
					dataMgr.addCachedRating(userRatings[i]);
				// Compute quality
				float bestVal =
					computeBestValue(userID, userRatings, numRatings);
				float algVal = computeUtility(userID, predictions, 50);
				// Remove ratings...
				for (int i = 0; i < numRatings; i++) {
					dataMgr.removeCachedRating(userRatings[i]);
					userRatings[i] = null;
				}
				// Add up the ratio, so we can calc the mean later
				if (bestVal > 0) {
					numerator += algVal;
					denominator += bestVal;
					usersTested++;
					// Print out some stats along the way, to satisfy curiosity
					if (usersTested % 100 == 0)
						System.out.println(100.0 * numerator / denominator);
				}
				lastUserID = userID;
				numRatings = 0;
			}
			userRatings[numRatings++] = currRating;
			lastUserID = userID;
		}
		// Output statistics
		System.out.println("\n" + (100.0 * numerator / denominator) + " avg (breese)");
		System.out.println("Users: " + usersTested);
		long ratingTestEnd = System.currentTimeMillis();
		System.out.println(
			"Rating test time: "
				+ (ratingTestEnd - ratingTestBegin) / 1000
				+ "s");
	}

	// compute the maximum utility for userID.
	float computeBestValue(int userID, ItemRating[] ratings, int len) {
		float DEFAULT_RATING = dataMgr.getMeanRatingCached(userID);
		float ALPHA = 5; // Rating list half-life
		float total = 0;
		for (int j = 0; j < len; j++) {
			int itemID = ratings[j].getItemID();
			float rating = dataMgr.getRatingCached(userID, itemID);
			// Ignore ratings below the default
			if (rating <= DEFAULT_RATING)
				continue;
			// DEBUG
			//System.out.println("" + j + "\t" + rating);
			total += (rating - DEFAULT_RATING) * Math.pow(0.5, j / (ALPHA - 1));
		}
		return total;
	}

	// compute the maximum utility for userID.
	float computeUtility(int userID, ItemPrediction[] predictions, int len) {
		float DEFAULT_RATING = dataMgr.getMeanRatingCached(userID);
		float ALPHA = 5; // Rating list half-life
		float total = 0;
		for (int j = 0; j < len; j++) {
			int itemID = predictions[j].getItemID();
			float rating = dataMgr.getRatingCached(userID, itemID);
			// Ignore ratings below the default
			if (rating <= DEFAULT_RATING)
				continue;
			// DEBUG
			//System.out.println("" + j + "\t" + rating);
			total += (rating - DEFAULT_RATING) * Math.pow(0.5, j / (ALPHA - 1));
		}
		return total;
	}

	/*************************************************************
	 * Test the quality of an algorithm's Top N recommendations.
	 *
	 *************************************************************/
	final static int numParam = 12;
	final static int reduceSetOpt = 1;
	final static int givenNparam = 30;

	public void testRecommendations2(Iterator iter1) {
		System.out.println("#:" + numParam + " opt:" + reduceSetOpt);

		long ratingTestBegin = System.currentTimeMillis();

		int totalUsers = 0;
		float usersTested[] = new float[10];
		TestResults t = null;
		TestResults total = new TestResults();
		total.userID = -999;

		Arrays.fill(usersTested, 0);

		LinkedList ll = dataMgr.sparcifyData(iter1, numParam, reduceSetOpt);
		Iterator iter = ll.iterator();
		while (iter.hasNext()) {
			t = rateRecommendation2((LinkedList) iter.next());
			total.totalCached = total.totalCached + t.totalCached;
			total.totalRemoved = total.totalRemoved + t.totalRemoved;
			total.totalTried = total.totalTried + t.totalTried;
			for (int i = 0; i < 5; i++) {
				if (t.topNPrecision[i] >= 0) {
					total.topNPrecision[i] =
						total.topNPrecision[i] + t.topNPrecision[i];
					usersTested[i]++;
				}
				if (t.topNHitRate[i] >= 0) {
					total.topNHitRate[i] =
						total.topNHitRate[i] + t.topNHitRate[i];
					usersTested[i + 5]++;
				}
				total.iNCRC[i] = total.iNCRC[i] + t.iNCRC[i];
			}
			totalUsers++;

		}

		//Tally finally results
		for (int i = 0; i < 5; i++) {
			total.topNPrecision[i] = total.topNPrecision[i] / usersTested[i];
			total.topNHitRate[i] = total.topNHitRate[i] / usersTested[i + 5];
			total.iNCRC[i] = total.iNCRC[i] / totalUsers;
		}
		total.totalCached = total.totalCached / totalUsers;
		total.totalRemoved = total.totalRemoved / totalUsers;
		total.totalTried = total.totalTried / totalUsers;

		System.out.println("\n***********Final test results**********");

		if (reduceSetOpt == 0) {
			System.out.println(
				"\n**Data set was reduced to each user containing only "
					+ numParam
					+ " ratings\n");
		} else {
			System.out.println(
				"\n**Data set was reduced by reducing each user to "
					+ numParam
					+ "% of there total ratings\n");
		}
		if (givenNparam > 0) {
			System.out.println(
				"**Each item was then seperately reduced to having "
					+ givenNparam
					+ " ratings while generating their reccomendations");
		}
		System.out.println("Total users:" + totalUsers);
		System.out.println(
			"Average Total Ratings:"
				+ (total.totalCached + total.totalRemoved));
		System.out.println("Average Cached Ratings:" + total.totalCached);
		System.out.println("Average Non Cached Ratings:" + total.totalRemoved);
		System.out.println(
			"\n*******Top N Precision (4 or better) || Users tested at that level *******");
		System.out.println(
			"1:" + total.topNPrecision[0] + "  ||  " + usersTested[0]);
		System.out.println(
			"5:" + total.topNPrecision[1] + "  ||  " + usersTested[1]);
		System.out.println(
			"10:" + total.topNPrecision[2] + "  ||  " + usersTested[2]);
		System.out.println(
			"15:" + total.topNPrecision[3] + "  ||  " + usersTested[3]);
		System.out.println(
			"20:" + total.topNPrecision[4] + "  ||  " + usersTested[4]);
		System.out.println(
			"\n*******Top N Hit Rate  || Users tested at that level *******");
		System.out.println(
			"1:" + total.topNHitRate[0] + "  ||  " + usersTested[5]);
		System.out.println(
			"5:" + total.topNHitRate[1] + "  ||  " + usersTested[6]);
		System.out.println(
			"10:" + total.topNHitRate[2] + "  ||  " + usersTested[7]);
		System.out.println(
			"15:" + total.topNHitRate[3] + "  ||  " + usersTested[8]);
		System.out.println(
			"20:" + total.topNHitRate[4] + "  ||  " + usersTested[9]);

		long ratingTestEnd = System.currentTimeMillis();
		System.out.println(
			"Rating test time: "
				+ (ratingTestEnd - ratingTestBegin) / 1000
				+ "s");
	}

	// Helper function -- rates the algorithm's recommendation
	// by comparing it to the test set ratings for the
	// user, which are passed in the realRatings array.
	final static int NUM_RECS = 10000;

	private TestResults rateRecommendation2(LinkedList ll) {
		TestResults t = new TestResults();

		ItemPrediction[] ratings = null;
		ItemRating rate = (ItemRating) ll.getFirst();
		int userID = rate.getUserID();
		// Get the recommendations using the current algorithm
		t.userID = userID;
		LinkedList givenNlist = null;
		if (givenNparam > 0) {
			givenNlist = dataMgr.givenNHide(userID, givenNparam, 0);
		}

		//print cached movies
		Iterator it = dataMgr.getItemList(userID);
		//System.out.println("Currently cashed ratings for user:"+userID);
		while (it.hasNext()) {
			t.totalCached++;
			ItemRating tr = (ItemRating) it.next();
			//System.out.println("Movie:"+tr.getItemID()+"  Rating:"+tr.getRating());
		}

		try {
			ratings = algorithm.getRecommendations(userID, NUM_RECS);
			dataMgr.cacheRatingGroup(ll.iterator());
			if (givenNparam > 0) {
				dataMgr.cacheRatingGroup(givenNlist.iterator());
			}
		} catch (Exception e) {
			System.out.println("Error: " + e);
		}

		//print movies previously in cache
		//System.out.println("\n\n Movies rated, but removed from cache for user:"+userID);
		Iterator iter = ll.iterator();
		while (iter.hasNext()) {
			ItemRating tr = (ItemRating) iter.next();
			//System.out.println("Movie:"+tr.getItemID()+"  Rating:"+tr.getRating());

			t.totalRemoved++;
			t.iNCRC[(new Float(tr.getRating()).intValue()) - 1]++;
		}

		if (givenNparam > 0) {
			Iterator itt = givenNlist.iterator();
			while (itt.hasNext()) {
				ItemRating tr = (ItemRating) itt.next();
				//System.out.println("Movie:"+tr.getItemID()+"  Rating:"+tr.getRating());

				t.totalRemoved++;
				t.iNCRC[(new Float(tr.getRating()).intValue()) - 1]++;
			}
		}

		int numRecs = ratings.length < NUM_RECS ? ratings.length : NUM_RECS;

		//print reccomendations
		/*System.out.println("\n\n Movies recommended to user:"+userID);
		for(int i = 0; i < numRecs; i++){
			System.out.println((i+1)+".  Movie:"+ratings[i].getItemID()+"  Rating:"+ratings[i].getPrediction());
		}
		*/

		//System.out.println("\n\n******Results of movies recommended to user********");
		float numRated = 0;
		float[] results = new float[10];
		Arrays.fill(results, 0);
		for (int currRating = 0; currRating < numRecs; currRating++) {
			int itemID = ratings[currRating].getItemID();
			float predict = ratings[currRating].getPrediction();
			int realVal = (int) dataMgr.getRatingCached(userID, itemID);
			//System.out.println("Movie:"+itemID+"   Predicted:"+predict+"    Actual:"+realVal);
			t.totalTried++;
			if (realVal != dataMgr.getErrRating()) {
				numRated++;
				//Precision
				if (numRated == 1 && realVal > 3)
					t.topNPrecision[0]++;
				if (numRated <= 5 && realVal > 3)
					t.topNPrecision[1]++;
				if (numRated <= 10 && realVal > 3)
					t.topNPrecision[2]++;
				if (numRated <= 15 && realVal > 3)
					t.topNPrecision[3]++;
				if (numRated <= 20 && realVal > 3)
					t.topNPrecision[4]++;

				//Test hit amount
				if (t.totalTried == 1)
					t.topNHitRate[0]++;
				if (t.totalTried <= 5)
					t.topNHitRate[1]++;
				if (t.totalTried <= 10)
					t.topNHitRate[2]++;
				if (t.totalTried <= 15)
					t.topNHitRate[3]++;
				if (t.totalTried <= 20)
					t.topNHitRate[4]++;
			}
			if (numRated >= 20)
				break;
		}

		//fill percent correct for return
		if ((t.iNCRC[3] + t.iNCRC[4]) >= 1)
			t.topNPrecision[0] = t.topNPrecision[0];
		else
			t.topNPrecision[0] = -999;
		if ((t.iNCRC[3] + t.iNCRC[4]) >= 5)
			t.topNPrecision[1] = t.topNPrecision[1] / 5;
		else
			t.topNPrecision[1] = -999;
		if ((t.iNCRC[3] + t.iNCRC[4]) >= 10)
			t.topNPrecision[2] = t.topNPrecision[2] / 10;
		else
			t.topNPrecision[2] = -999;
		if ((t.iNCRC[3] + t.iNCRC[4]) >= 15)
			t.topNPrecision[3] = t.topNPrecision[3] / 15;
		else
			t.topNPrecision[3] = -999;
		if ((t.iNCRC[3] + t.iNCRC[4]) >= 20)
			t.topNPrecision[4] = t.topNPrecision[4] / 20;
		else
			t.topNPrecision[4] = -999;

		//Fill hit rate for return
		if (t.totalRemoved >= 1)
			t.topNHitRate[0] = t.topNHitRate[0];
		else
			t.topNHitRate[0] = -999;
		if (t.totalRemoved >= 5)
			t.topNHitRate[1] = t.topNHitRate[1] / 5;
		else
			t.topNHitRate[1] = -999;
		if (t.totalRemoved >= 10)
			t.topNHitRate[2] = t.topNHitRate[2] / 10;
		else
			t.topNHitRate[2] = -999;
		if (t.totalRemoved >= 15)
			t.topNHitRate[3] = t.topNHitRate[3] / 15;
		else
			t.topNHitRate[3] = -999;
		if (t.totalRemoved >= 20)
			t.topNHitRate[4] = t.topNHitRate[4] / 20;
		else
			t.topNHitRate[4] = -999;

		dataMgr.removeCachedRatingGroup(ll.iterator());

		//System.out.println(t);
		//	System.out.println("UserID="+userID+" totalTried="+totalTried+" numRated="+numRated);
		return t;
	}

	private void testCoherence() {
		Iterator iter = dataMgr.getAllUsers();
		int report = 5;
		float totalerror = 0;
		int totalPreds = 0;
		while (iter.hasNext()) {
			try {
				int currentUser = ((Integer) iter.next()).intValue();
				ItemPrediction[] recs =
					algorithm.getRecommendations(currentUser, 50);
				for (int i = 0; i < recs.length; i++) {
					ItemPrediction rec = recs[i];
					int item = rec.getItemID();
					float predRating =
						algorithm
							.predictRating(currentUser, item)
							.getPrediction();
					totalPreds++;
					totalerror += Math.abs(rec.getPrediction() - predRating);
					if (report > 0 && i > 40 && i < 50) {
						System.out.println(
							"Rec:"
								+ rec.getPrediction()
								+ "\tPred:"
								+ predRating);
					}
				}
				report--;
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		System.out.println("Avg error:" + (totalerror / totalPreds));
	}

	private class TestResults {
		public float[] topNPrecision;
		public float[] topNHitRate;
		public int[] iNCRC;
		//individual non cached rating count: keep tract of our possible positive recommendations
		public int totalCached;
		public int totalRemoved;
		public int userID;
		public int totalTried;

		public TestResults(
			float[] topNPrecision,
			float[] topNHitRate,
			int totalRatings,
			int[] iNCRC,
			int totalCached,
			int totalRemoved,
			int userID,
			int totalTried) {

			this.topNPrecision = topNPrecision;
			this.topNHitRate = topNHitRate;
			this.iNCRC = iNCRC;
			this.totalCached = totalCached;
			this.totalRemoved = totalRemoved;
			this.userID = userID;
			this.totalTried = totalTried;
		}

		public TestResults() {
			this.topNPrecision = new float[5];
			this.topNHitRate = new float[5];
			this.iNCRC = new int[5];
			for (int i = 0; i < 5; i++) {
				topNPrecision[i] = 0;
				topNHitRate[i] = 0;
				iNCRC[i] = 0;
			}

			this.totalCached = 0;
			this.totalRemoved = 0;
			this.userID = 0;
			this.totalTried = 0;
		}

		public String toString() {
			String printS =
				"\n*********Test Results for User ID:"
					+ userID
					+ "**********\n";
			printS = printS + "Total Cached Ratings: " + totalCached + "\n";
			printS =
				printS + "Total Non Cached Ratings: " + totalRemoved + "\n";
			printS =
				printS
					+ "Total Tried Recs to reach top 20: "
					+ totalTried
					+ "\n\n";
			printS = printS + "Top N Precision:\n";
			printS = printS + "1:" + topNPrecision[0] + "\n";
			printS = printS + "5:" + topNPrecision[1] + "\n";
			printS = printS + "10:" + topNPrecision[2] + "\n";
			printS = printS + "15:" + topNPrecision[3] + "\n";
			printS = printS + "20:" + topNPrecision[4] + "\n\n";
			printS = printS + "Top N Hit Rate:\n";
			printS = printS + "1:" + topNHitRate[0] + "\n";
			printS = printS + "5:" + topNHitRate[1] + "\n";
			printS = printS + "10:" + topNHitRate[2] + "\n";
			printS = printS + "15:" + topNHitRate[3] + "\n";
			printS = printS + "20:" + topNHitRate[4] + "\n";

			return printS;
		}
	}

	//////////////////////////////////////////////////////////////////////////

	/** Private method: Added for finding the corresponding methods
	 */
	private static Method findpredictRating(Class cl) throws Exception {
		Method[] methods = cl.getMethods();
		for (int i = 0; i < methods.length; i++) {
			if (methods[i].getName().equals("predictRating")) {
				return methods[i];
			}
		}

		return null;
	}

	/** Private method: Added for finding the corresponding methods
	 */
	private static Method findgetUpdateUser(Class cl) throws Exception {
		Method[] methods = cl.getMethods();
		for (int i = 0; i < methods.length; i++) {
			if (methods[i].getName().equals("updateUser")) {
				return methods[i];
			}
		}

		return null;
	}

	//////////////////////////////////////////////////////////////////////////////
}
