/* 
 * BayesNet.java
 * 
 * Version 1.2
 * 
 * 13/02/2002
 *
 * Copyright (c) 2003 Oregon State University. All rights reserved.
 */
package org.recommender.algorithms.Experimental;

import org.recommender.server.*;
import java.net.*;
import java.util.*;
import java.io.*;

// XML parsing libraries, freely available from http://jclark.com/xml/xp/
import com.jclark.xml.parse.*;
import com.jclark.xml.parse.base.*;

/**
 * In-memory representation of a Bayesian network with decision trees
 * at each node.  Includes functionality to load a Bayesian network from
 * a file, and for this reason implements the Application interface from
 * the XP XML parser.  It also depends on the xp.jar library.
 *
 * Methods that lack comments are the implementation of the Application
 * interface.  These are effectively callbacks, made public so that
 * they're accessible to the XP parser.
 *
 * @author Daniel Lowd &lt;dlowd@cs.orst.edu&gt;
 */
public class BayesNet implements Application
{
    DocumentParser parser;
    PrintStream out;

    /**
     * Keep track of our state in the XML file
     */
    boolean inDecisionTree = false;
    
    /**
     * Keep track of our state in the XML file
     */
    boolean inCdata = false;

    /**
     * Character data found in the contents of an element.
     */ 
    StringBuffer elementCdata;
    
    /**
     * Character data found in the contents of a CDATA section.
     */  
    StringBuffer cdataCdata;

    /**
     * Decision trees
     */
    DecisionTree[] decisionTrees;
    DecisionTree dtree;

    Integer currVariable;

    /**
     * Child lists (linked lists of Integers)
     */
    List[] children;

    /** 
     * Construct a new Bayesian network.
     * @param maxVar highest variable index used by the network.  This
     *     could be inferred from the file, but it's easier if it's known
     *     ahead of time.
     * @param filename path and filename of the Bayesian network model.
     *     The expected file format is .xmod, as output by the WinMine
     *     toolkit, with decision trees at each node.
     */
    public BayesNet(int maxVar, String filename)
    {
        decisionTrees = new DecisionTree[maxVar+1];
        children = new List[maxVar+1];

        for (int i = 0; i <= maxVar; i++)
        {
            decisionTrees[i] = null;
            children[i] = new LinkedList();
        }

        cdataCdata   = null;
        elementCdata = new StringBuffer();

        parser = new DocumentParser();

        try {
            InputStream in = new FileInputStream(filename);
            load(in);
        } catch (IOException e) {
            System.err.println("Error: could not open file \"" + filename
                + "\"");
        }
    }

    /**
     * Get an iterator of all children of the specified variable. 
     *
     * @param var index of the variable whose children we want.
     */
    public Iterator getChildren(int var)
    {
        return children[var].iterator();
    }

    /**
     * Get the probability that a specified variable is in a specified
     * state given the state of every other variable.
     * @param var index of the variable in question
     * @param state variable state of which to calculate the probability
     * @param allStates array containing the state of each variable
     *     in this network.  Simply returns the approviate value from
     *     the variable's probability distribution in its decision tree.
     *     This method does not do inference, not even Markov-blanket
     *     inference.  The caller is expected to implement that.
     */
    public float getProb(int var, int state, int[] allStates)
    {
        if (decisionTrees[var] == null)
            return 0;

        return decisionTrees[var].getProb(state, allStates);
    }

    /** 
     * Begin parsing an xmod file from the specified InputStream.
     * @param in input xmod file
     */
    public void load(InputStream in)
    {
        // Run the parser (and catch exceptions)
        try {
            OpenEntity oe = new OpenEntity(in, "", 
                    new URL("http://cs.orst.edu/"));
            parser.parse(oe, new EntityManagerImpl(), this, Locale.US);
        } catch (IOException e) {
            System.err.println("IOException: " + e);
        } catch (ApplicationException f) {
            System.err.println("ApplicationException: " + f);
        }
    }

    public void startDocument() { }

    public void endProlog(EndPrologEvent e) { }

    public void startElement(StartElementEvent startElement)
    {
        String name = startElement.getName();

        if (name.equals("LocalModel"))
        {
            // Switch variables
            String varname = startElement.getAttributeValue("variable");
            currVariable = new Integer(varname);
        }
        else if (name.equals("Input"))
        {
            // Add appropriate dependency
            String varname = startElement.getAttributeValue("variable");
            try {
                int var = Integer.parseInt(varname);
                children[var].add(currVariable);
            } catch (NumberFormatException e) {}
        }
        // Special handling for decision trees
        else if (name.equals("DecisionTree"))
        {
            inDecisionTree = true;
            dtree = new DecisionTree();
            decisionTrees[currVariable.intValue()] = dtree;
        }
        else if (name.equals("Vertex") && inDecisionTree)
            dtree.beginVertex(startElement);
        else if (name.equals("Branch") && inDecisionTree)
            dtree.beginBranch();
        else if (name.equals("Multinomial") && inDecisionTree)
            dtree.beginMultinomial();

        elementCdata = new StringBuffer();
    }

    public void characterData(CharacterDataEvent characterData)
    {
        // Keep track of text in the contents of the current element, or
        // a CDATA section.
        char[] pcdata = new char[characterData.getLengthMax()];
        characterData.copyChars(pcdata, 0);
        if (inCdata)
            cdataCdata.append(pcdata);
        else
            elementCdata.append(pcdata);
    }

    public void endElement(EndElementEvent endElement)
    {
        String name = endElement.getName();

        // DecisionTree parsing
        if (inDecisionTree)
        {
            if (name.equals("DecisionTree"))
                inDecisionTree = false;
            else if (name.equals("Vertex"))
                dtree.endVertex();
            else if (name.equals("Branch"))
                dtree.endBranch();
            else if (name.equals("Values"))
                dtree.endValues(elementCdata.toString());
            else if (name.equals("Probs"))
                dtree.endProbs(elementCdata.toString());
        }

        elementCdata = new StringBuffer();
    }

    public void endDocument() { }

    public void startCdataSection(StartCdataSectionEvent e)
    {
        cdataCdata = new StringBuffer();
        inCdata = true;
    }

    public void endCdataSection(EndCdataSectionEvent e) 
    {
        elementCdata.append(cdataCdata);
        inCdata = false;
    }

    // Lots of junk we don't need...
    public void processingInstruction(ProcessingInstructionEvent e) {}
    public void comment(CommentEvent e) {}
    public void startEntityReference(StartEntityReferenceEvent e) {}
    public void endEntityReference(EndEntityReferenceEvent e) {}
    public void startDocumentTypeDeclaration(
            StartDocumentTypeDeclarationEvent e) {}
    public void endDocumentTypeDeclaration(EndDocumentTypeDeclarationEvent e) {}
    public void markupDeclaration(MarkupDeclarationEvent e) {}
}
