/*
 * BayesNetAlgorithm.java
 *
 * Version 1.1
 *
 * 23/08/2002
 *
 * Copyright (c) 2003 Oregon State University. All rights reserved.
 */
package org.recommender.algorithms.Experimental;

import org.recommender.algorithms.*;
import org.recommender.algorithms.Experimental.*;
import org.recommender.server.*;
import java.io.*;
import java.util.*;
import java.sql.*;

/**
 * Describes an implementation of the CFAlgorithm interface
 * to make CF predictions by using Markov-blanket inference
 * on a Bayesian network.  The network must be created by
 * the WinMine Toolkit, freely available from Microsoft Research.
 *
 * @author Daniel Lowd
 */
public class BayesNetAlgorithm implements CFAlgorithm
{
    private DataManager dataMgr;
    BayesNet model;

    /**
     * Variables to keep track of the average time for a call to predictRating
     */
    private long Time = 0;
    private long NumPredictions = 0;

    private int maxItem;

    /**
     * Variables for supporting the discretization of a continuous data set.
     * Rather than having discrete values representing the actual ratings,
     * we have several "bins," each of which contains a sub-range.  It is
     * assumed that the Bayesian network was trained with an equivalent
     * discretization, some number of equally sized ranges with the given min
     * and max.  For predicting, actual ratings are converted to their bin
     * indices, then we use the model to compute the probability of the answer
     * being in each bin, and finally we convert that distribution to an
     * expected value in the original range.
     *
     * NOTE: The code that does this has not been well-tested!!!
     */ 
    private int NUM_BINS;
    private int BIN_MIN;
    private int BIN_MAX;
    private float[] binMeans;

    /**
     * True - each prediction is the expected value
     * False - each prediction is the one with maximum probability
     */
    private boolean USE_EXPECTED_VALUE;

	/**
	 * Get algorithm properties and initialize a BayesNet object.
	 * @param mgr the DataManager of data
	 */
    public BayesNetAlgorithm(DataManager mgr) throws CFNotImplementedException
    {
		dataMgr = mgr;

        maxItem = dataMgr.getMaxItem();

        String prefix = "CFServer.alg.bn.";
        USE_EXPECTED_VALUE = SystemProperties.getBoolean(
                prefix + "useExpectedValue", true);
        String filename = SystemProperties.getString(
                prefix + "modelfile");

        // For discretized models, the number of bins used.
        NUM_BINS = SystemProperties.getInt(prefix + "numBins", -1);

        // Lower range of the first bin
        // (all lower values also go in the first bin)
        BIN_MIN = SystemProperties.getInt(prefix + "binMin", -1);

        // Upper range of the last bin
        // (all higher values also go in the last bin)
        BIN_MAX = SystemProperties.getInt(prefix + "binMax", -1);

        if (NUM_BINS > 0)
        {
            binMeans = new float[NUM_BINS];
            for (int i = 0; i < NUM_BINS; i++)
            {
                binMeans[i] = BIN_MIN + 
                        (BIN_MAX - BIN_MIN)*(i+(float)0.5)/NUM_BINS;
                // DEBUG
                System.out.println("binMeans[" + i + "] = " + binMeans[i]);
            }
        }

        model = new BayesNet(maxItem, filename);
    }

	/**
	 * Returns top n recommendations
	 * Not yet complete
	 */
	public ItemPrediction[] getRecommendations(int activeUser, int n)
            throws SQLException, CFNotImplementedException
    {
    	throw new CFNotImplementedException("getRecommendations not implemented yet");
	}

	// Returns top n recommendations
	public ItemPrediction[] getRecommendationsByType(int activeUser, int n, int type)
            throws SQLException, CFNotImplementedException
    {
    	throw new CFNotImplementedException("getRecommendationsByType not implemented yet");
	}

    /**
     * predicts the rating for the given user and item.
     * @param activeUser
     * @param item
     */
    public ItemPrediction predictRating (int activeUser, int item) throws CFNotImplementedException
    {
        long startTime = System.currentTimeMillis();

        int numDiscreteRatings;

        if (NUM_BINS <= 0)
            numDiscreteRatings = (int)(dataMgr.getMaxRating()
                                       - dataMgr.getMinRating() + 1);
        else
            numDiscreteRatings = NUM_BINS + 1;

        // Construct array with all the user's ratings in a
        // revised scale (0 = missing, 1-N = min, min+1, ..., max)
        int[] userRatings = new int[maxItem+1];
        Arrays.fill(userRatings, 0);

        ItemRating[] ratings = dataMgr.getUserRatingList(activeUser);
        for (int i = 0; i < ratings.length; i++)
        {
            ItemRating ir = ratings[i];
            if (NUM_BINS <= 0)
            {
                // Simple transform from raw rating to rating index.
                userRatings[ir.getItemID()] = (int)(ir.getRating()
                                              - dataMgr.getMinRating() + 1);
            }
            else
            {
                // Convert a raw rating into a bin number.
                int binNum = (int)Math.round((ir.getRating() - BIN_MIN)
                    * NUM_BINS / (BIN_MAX - BIN_MIN));

                // Edge bins get anything that doesn't fit elsewhere.
                if (binNum < 0)
                    binNum = 0;
                else if (binNum >= NUM_BINS)
                    binNum = NUM_BINS - 1;

                // Add 1, so that 0 is reserved for "Missing"
                userRatings[ir.getItemID()] = binNum + 1;
            }
        }


        // Calculate the probability distribution using Markov-blanket inference
        float[] ratingProbs = new float[numDiscreteRatings+1];

        for (int i = 1; i <= numDiscreteRatings; i++)
        {
            userRatings[item] = i;

            float p = model.getProb(item, i, userRatings);

            // DEBUG
            if (Float.isNaN(p)) {
                System.out.println("user: " + activeUser + ", item: " + item
                        + ", rating: " + i);
            }

            Iterator iter = model.getChildren(item);
            while (iter.hasNext())
            {
                int childID = ((Integer)iter.next()).intValue();
                p *= model.getProb(childID, (int)userRatings[childID],
                        userRatings);
            }

            ratingProbs[i] = p;
        }

        float prediction;

        if (USE_EXPECTED_VALUE)
        {
            // Get expected value
            float num = 0;
            float denom = 0;
            boolean errRating = true;
            for (int i = 1; i <= numDiscreteRatings; i++)
            {
                if (NUM_BINS > 0)
                    num += binMeans[i-1]*ratingProbs[i];
                else
                    num += (i-1)*ratingProbs[i];

                denom += ratingProbs[i];

                /*
                if (Math.abs(ratingProbs[i] - denom/numDiscreteRatings)
                        > 0.00001)
                    errRating = false;
                    */
            }

            if (/*errRating ||*/ denom <= 0)
                prediction = dataMgr.getErrRating();
            else
            {
                if (NUM_BINS > 0)
                    prediction = num/denom;
                else
                    prediction = num/denom + dataMgr.getMinRating();
            }
        }
        else
        {
            // Get maximum probability
            int bestRating = 1;
            for (int i = 1; i <= numDiscreteRatings; i++)
            {
                if (ratingProbs[i] > ratingProbs[bestRating])
                    bestRating = i;
            }

            if (NUM_BINS > 0)
                prediction = binMeans[bestRating-1];
            else
                prediction = bestRating - 1 + dataMgr.getMinRating();
        }

        long stopTime = System.currentTimeMillis();
        Time += stopTime - startTime;
        NumPredictions++;

        return new ItemPrediction(activeUser,item,prediction);
    }

    /**
     * Update user information, not yet complete.
     */
    public void updateUser(int userID) throws CFNotImplementedException
    {
        throw new CFNotImplementedException("update User not implemented yet");
    }

    /**
     * Return average prediction time
     * @return average prediction time
     */
    public long getAveragePredictionTime() throws CFNotImplementedException
    {
        return (Time/NumPredictions);
    }

    /**
     * Return the name of this algorithm
     * @return String the name of this algorithm
     */
    public String toString()
    {
        return "BayesNetAlgorithm.java";
    }
}
