/*
 * ClusteringAlgorithm.java
 *
 * Version 1.1
 *
 * 15/08/2002
 *
 * Copyright (c) 2003 Oregon State Universtity. All rights reserved.
 */
package org.recommender.algorithms.Experimental;

import org.recommender.algorithms.*;
import org.recommender.server.*;
import java.sql.*;
import java.util.*;

/**
 * Collaborative filtering algorithm using Bayesian clustering, to divide
 * users into different "classes," and recommend items based on membership
 * in these classes.
 *
 * @author Daniel Lowd <dlowd@cs.orst.edu>
 */
public class ClusteringAlgorithm implements CFAlgorithm
{
    DataManager dataMgr;
    float[][][] wArray;
    float[] classPriors;
    int maxClass;
    int maxItem;
    int maxRating;

    final static float GRADIENT_RATIO = (float) 0.05;
    boolean USE_EXPECTED_VALUE;

    /**
     * Variables to keep track of the average time for a call to predictRating
     */
    private long Time = 0;
    private long NumPredictions = 0;

    /**
     * Constructor, initialize DataManager, parameters and wArray.
     * @param dataMgr
     */
    public ClusteringAlgorithm(DataManager dataMgr)
    {
        this.dataMgr = dataMgr;
        String prefix = "CFServer.alg.cluster.";
        int numClasses = SystemProperties.getInt(prefix + "numClasses", 20);
        int numSteps = SystemProperties.getInt(prefix + "numTrainingSteps", 50);
        USE_EXPECTED_VALUE = SystemProperties.getBoolean(
                prefix + "useExpectedValue", true);

        maxItem = dataMgr.getMaxItem();
        maxRating = (int)dataMgr.getMaxRating();

        initRandomArray(numClasses);

        for (int i = 0; i < numSteps; i++)
            gradientAscentStep();
    }

    /**
     * Create the wArray and fill it with random values.
     * @param numClasses
     */
    public void initRandomArray(int numClasses)
    {
        maxClass = numClasses;
        wArray = new float[maxItem+1][maxRating+1][maxClass];

        long begin = System.currentTimeMillis();

        // Start out with a random array.
        for (int i = 0; i <= maxItem; i++)
            for (int j = 0; j <= maxRating; j++)
                for (int k = 0; k < maxClass; k++)
                    wArray[i][j][k] = (float) Math.random();

        normalizeArray();

        // Assume each class is equally likely to begin with
        classPriors = new float[maxClass];
        for (int k = 0; k < maxClass; k++)
           classPriors[k] = (float)1.0/maxClass;

        long end = System.currentTimeMillis();
        System.out.println("Init time: " + (end - begin)/1000 + "s");
    }

    /**
     * Do one step of gradient ascent to refine the values in wArray
     */
    float gradientAscentStep()
    {
        long begin = System.currentTimeMillis();

        // Create the replacement array and initialize it to 0
        float[][][] wNew = new float[maxItem+1][maxRating+1][maxClass];
        for (int i = 0; i <= maxItem; i++)
            for (int j = 0; j <= maxRating; j++)
                for (int k = 0; k < maxClass; k++)
                    wNew[i][j][k] = 0;

        float[] newPriors = new float[maxClass];
        for (int k = 0; k < maxClass; k++)
            newPriors[k] = 0;

        // Iterate over all data...
        int maxUser = dataMgr.getMaxUser();
        for (int uid = 1; uid <= maxUser; uid++)
        {
            if (dataMgr.getNumRatings(uid) == 0)
                continue;

            // Given this data, determine the probabilities of various
            // class memberships.
            float[] classProb = calcAllProbs(dataMgr.getUserRatingList(uid));
            for (int k = 0; k < maxClass; k++)
                newPriors[k] += classProb[k];

            // Add in the effect of each rating
            ItemRating[] userRatings = dataMgr.getUserRatingList(uid);
            for (int i = 0; i < userRatings.length; i++)
            {
                ItemRating ir = userRatings[i];
                int id = ir.getItemID();
                int rating = (int)ir.getRating();

                for (int k = 0; k < maxClass; k++)
                    wNew[id][rating][k] += classProb[k]/wArray[id][rating][k];
            }
        }

        // Back-up wArray
        float[][][] wBackup = new float[maxItem+1][maxRating+1][maxClass];
        for (int i = 0; i <= maxItem; i++)
            for (int j = 0; j <= maxRating; j++)
                for (int k = 0; k < maxClass; k++)
                    wBackup[i][j][k] = wArray[i][j][k];

        // Update wArray
        for (int i = 0; i <= maxItem; i++)
            for (int j = 0; j <= maxRating; j++)
                for (int k = 0; k < maxClass; k++)
                    wArray[i][j][k] += GRADIENT_RATIO * wNew[i][j][k];

        // Renormalize, so the probabilities are valid again
        normalizeArray();

        float delta = 0;
        for (int i = 0; i <= maxItem; i++)
            for (int j = 0; j <= maxRating; j++)
                for (int k = 0; k < maxClass; k++)
                    delta += Math.abs(wArray[i][j][k] - wBackup[i][j][k]);


        // Normalize new class priors
        classPriors = newPriors;
        float denominator = 0;
        for (int k = 0; k < maxClass; k++)
            denominator += classPriors[k];
        for (int k = 0; k < maxClass; k++)
            classPriors[k] /= denominator;

        long end = System.currentTimeMillis();
        System.out.println("Step time: " + (end - begin) + "ms");
        System.out.println(delta);

        return delta;
    }


    /**
     * Normalize wArray so that the probabilities for each
     * rating on an item sum to 1.  This condition is required because
     * ratings are mutually exclusive (no item can have a rating of
     * both 1 and 2 from the same user).
     */
    void normalizeArray()
    {
        for (int i = 0; i <= maxItem; i++)
            for (int k = 0; k < maxClass; k++)
            {
                float denom = 0;
                for (int j = 0; j <= maxRating; j++)
                {
                    // HACK -- avoid divide by zero errors
                    if (wArray[i][j][k] < 0.001)
                        wArray[i][j][k] = (float) 0.001;

                    denom += wArray[i][j][k];
                }

                for (int j = 0; j <= maxRating; j++)
                    wArray[i][j][k] /= denom;
            }
    }

    /**
     * Calculate all probabilities.
     * @param evidence - Array containing evidence
     * @return Array of probabilities
     */
    float[] calcAllProbs(ItemRating[] evidence)
    {
        float[] probs = new float[maxClass];
        int state;

        // Set an initial probability (assumption: identical posteriors)
        for (state = 0; state < maxClass; state++)
            probs[state] = classPriors[state];

        for (int i = 0; i < evidence.length; i++)
        {
            ItemRating ir = evidence[i];
            int rating = (int)ir.getRating();
            int itemID = ir.getItemID();

            for (state = 0; state < maxClass; state++)
                probs[state] *= wArray[itemID][rating][state];
        }

        // Normalize
        float denom = 0;
        for (state = 0; state < maxClass; state++)
        {
            denom += probs[state];
        }

        if (denom > 0)
        {
            for (state = 0; state < maxClass; state++)
                probs[state] /= denom;
        }

        return probs;
    }

    /**
     * Asks the prediction engine to return a prediction for how a
     * particular user will rate a particular item.
     */
    public ItemPrediction predictRating(int userID, int itemID) throws CFNotImplementedException
    {
        long startTime = System.currentTimeMillis();
        float ratings[] = new float [maxRating+1];

        // Calculate the expected value of the rating, given the
        // data.
        float[] probs = calcAllProbs(dataMgr.getUserRatingList(userID));
        for (int j = 0; j <= maxRating; j++)
        {
            ratings[j] = 0;
            for (int k = 0; k < maxClass; k++)
                ratings[j] += probs[k] * wArray[itemID][j][k];
        }

        float retval;
        if (USE_EXPECTED_VALUE)
        {
            // Expected value
            float num = 0;
            float denom = 0;
            for (int j = 0; j <= maxRating; j++)
            {
                num   += j*ratings[j];
                denom += ratings[j];
            }

            if (denom > 0)
                retval = num/denom;
            else
                retval = dataMgr.getErrRating();
        }
        else
        {
            // BUG: if all probs are 0, or are equal, this
            // will just return the lowest rating.

            // Maximum expectation
            int bestIndex = (int)dataMgr.getErrRating();
            float bestTotal = 0;
            for (int j = 0; j <= maxRating; j++)
            {
                if (ratings[j] > bestTotal)
                {
                    bestIndex = j;
                    bestTotal = ratings[j];
                }
            }
            retval = bestIndex;
        }

        long stopTime = System.currentTimeMillis();
        Time += stopTime - startTime;
        NumPredictions++;

        return new ItemPrediction(userID,itemID,retval);
    }

   /**
    * Return average prediction time
    * @return average prediction time
    */
    public long getAveragePredictionTime() throws CFNotImplementedException
    {
        return (Time/NumPredictions);
    }

    /**
     * Not yet implemented.
     */
    public ItemPrediction[] getRecommendations(int activeUser, int n)
            throws SQLException, CFNotImplementedException
    {
        throw new CFNotImplementedException("getReccomendations has not been implemented yet");
        //return null;
    }

    /**
     * Not yet implemented.
     */
    public void updateUser(int userID) throws CFNotImplementedException
    {
    	throw new CFNotImplementedException("update User not implemented yet");
    }

    /**
     * Return the name of this algorithm
     * @return String
     */
    public String toString()
    {
        return "ClusteringAlgorithm.java\n" + maxClass + " classes";
    }

     /**
     * Haven't been implemented yet. Once getRecommendations() implemented, implement this.
     */
     public ItemPrediction[] getRecommendationsByType(int activeUser, int n, int type)
               throws SQLException, CFNotImplementedException
    {
        throw new CFNotImplementedException("getReccomendationsByType has not been implemented yet");
        //return null;
     }
}
