//==========================================================
//============ ContinuousBayesNetAlgorithm.java ============
//==========================================================
package org.recommender.algorithms.Experimental;

import org.recommender.algorithms.*;
import org.recommender.server.*;
import java.io.*;
import java.util.*;
import java.sql.*; //IY

/**
 * Describes an implementation of the CFAlgorithm interface
 * to make CF predictions using a Bayesian network.
 */
public class ContinuousBayesNetAlgorithm implements CFAlgorithm
{
    private DataManager dataMgr;
    ContinuousBayesNet model;

    // Variables to keep track of the average time for a call to predictRating
    private long Time = 0;
    private long NumPredictions = 0;

    private int maxItem;

    /* Constructs an algorithm to make predictions based on a Bayesian
     * network with continuous variables.  Reads in the network from
     * a preconstructed model file.
     */
    public ContinuousBayesNetAlgorithm(DataManager mgr)
    {
		dataMgr = mgr;

        maxItem = dataMgr.getMaxItem();

        String prefix = "CFServer.alg.cbn.";
        String filename = SystemProperties.getString(prefix + "modelfile");

        model = new ContinuousBayesNet(maxItem, filename);
    }

	// Returns top n recommendations
	public ItemPrediction[] getRecommendations(int activeUser, int n)
            throws SQLException, CFNotImplementedException
    {
    	throw new CFNotImplementedException("getRecommendations not implemented yet");
	}

	// Returns top n recommendations
	public ItemPrediction[] getRecommendationsByType(int activeUser, int n, int type)
            throws SQLException, CFNotImplementedException
    {
    	throw new CFNotImplementedException("getRecommendationsByType not implemented yet");
	}

    /** predicts the rating for the given user and item. */
    public ItemPrediction predictRating(int activeUser, int item) 
            throws CFNotImplementedException
    {
        long startTime = System.currentTimeMillis();

        // Construct array with all the user's ratings
        float userMean = dataMgr.getMeanRatingCached(activeUser);
        float[] userRatings = new float[maxItem+1];

        /* Not-a-Number is how we encode "Missing".  It's an abuse of
         * a float, but it works since NaN is never a legal rating.
         */
        Arrays.fill(userRatings, Float.NaN);

        /* Replace "Missing" ratings with the actual ratings, as recorded
         * in the database.
         */
        ItemRating[] rawRatings = dataMgr.getUserRatingList(activeUser);
        for (int i = 0; i < rawRatings.length; i++)
        {
            ItemRating ir = rawRatings[i];
            userRatings[ir.getItemID()] = ir.getRating() - userMean;
        }

        /* Calculate the expected value of this variable using
         * Markov-blanket inference.
         */
        float num = 0;
        float denom = 0;
        float prediction;

        // Numerically integrate the x*PDF over a reasonable range of values.
        for (float x = -dataMgr.getMaxRating();
                x <= dataMgr.getMaxRating(); x += 0.1)
        {
            userRatings[item] = x;

            /* The Markov-blanket probability density is the probability
             * density for this variable at x times the probabilities that our
             * children have their observed values, given this assignment of x.
             */
            float p = model.getProb(item, x, userRatings);

            Iterator iter = model.getChildren(item);
            while (iter.hasNext())
            {
                int childID = ((Integer)iter.next()).intValue();
                p *= model.getProb(childID, userRatings[childID],
                        userRatings);
            }

            /* Since we're taking the expected value, add up the probabilities
             * for each x value, weighted by x.
             */

            /* EXPERIMENT -- tried using Simpson's rule.  Saw no improvement
             * in results.  Code may be buggy, due to floating point
             * error, which would explain it.
            int mult;
            if (Math.abs(x) == dataMgr.getMaxRating())
                mult = 1;
            else if ((x*10.0)%2 == 0)
                mult = 4;
            else
                mult = 2;
            num += mult*x*p;
             */

            num += x*p;
            denom += p;
        }
        //num /= 3.0;

        long stopTime = System.currentTimeMillis();
        Time += stopTime - startTime;
        NumPredictions++;

        if (denom <= 0)
            prediction = dataMgr.getErrRating();
        else
        {
            prediction = num/denom + userMean;
            if (prediction > dataMgr.getMaxRating())
                prediction = dataMgr.getMaxRating();
            else if (prediction < dataMgr.getMinRating())
                prediction = dataMgr.getMinRating();
        }

        return new ItemPrediction(activeUser,item,prediction);
    }

    public void updateUser(int userID) throws CFNotImplementedException
    {
        throw new CFNotImplementedException("updateUser not yet implemented");
    }

    public long getAveragePredictionTime() throws CFNotImplementedException
    {
        return (Time/NumPredictions);
    }

    public String toString()
    {
        return "ContinuousBayesNetAlgorithm.java";
    }
}
