package org.recommender.algorithms.Experimental;

import org.recommender.server.*;
import java.util.*;
import java.io.*;

// XML parsing libraries
import com.jclark.xml.parse.*;
import com.jclark.xml.parse.base.*;

/**
 * Helps parse a decision tree by processing begin/end element events.  Works
 * with continuous probability distributions, represented as binary Gaussians,
 * rather than multinomials.  Stores a representation of the decision tree in
 * memory.
 *
 * @author Daniel Lowd &lt;dlowd@cs.orst.edu&gt;
 */
public class ContinuousDecisionTree
{
    // ArrayList of input variable names
    ArrayList inputVars;
    // HashMap of LinkedLists of state indices for the input variables
    HashMap inputVarStates;

    // Top level vertex, before any branches.  (Note: some decision trees
    // have no vertices at all; in such cases, this is null.)
    Vertex headVertex;

    // Keep track of current position in the tree, as it is built.
    Vertex currVertex;
    Branch currBranch;
    BinGaussian currBinGaussian;

    /* Keep track of the last read mean, standard deviation, and
     * probability of a missing value, so we have the appropriate
     * arguments for building the current BinGaussian.
     */
    float currMean;
    float currSD;
    float currProbMissing;


    int printIndent;

    /**
     * Construct a ContinuousDecisionTree with the given input variables.
     */
    public ContinuousDecisionTree()
    {
        headVertex = null;
        currVertex = null;
        currBranch = null;
        currBinGaussian = null;

        // Mark these values as invalid, to start with
        currMean = currSD = currProbMissing = Float.NaN;
    }

    /** 
     * Returns the relative probability that this variable is in a specified
     * state, given the states of all other variables in the network.  If
     * the state is NaN, this is the probability that the variable is missing.
     * Otherwise, it is the probability density function of the distribution
     * evaluated at the target value.
     *
     * prob state the potential state of this decision tree's variable.
     * prob allStates array of integers specifying the state of each
     *     other variable.  allStates[i] = the state of variable i.
     */
    public float getProb(float state, float[] allStates)
    {
        if (headVertex == null)
        {
            if (currBinGaussian == null)
                return 0;
            else
                return currBinGaussian.getProb(state);
        }

        return headVertex.getProb(state, allStates);
    }

    /** 
     * Method called when a &lt;Vertex&gt; tag is encountered.
     * Creates a new vertex as the child of the current branch.
     */
    public void beginVertex(StartElementEvent vertex)
    {
        String splitVal = vertex.getAttributeValue("split");
        try {
            int var = Integer.parseInt(splitVal);
            currVertex = new Vertex(var, currBranch);
        } catch(NumberFormatException e) {
            System.out.println(
                    "ERROR: invalid split value in decision tree.\n");
        }

        /* If there's no root vertex, set it as such.
         * Otherwise, add it as the child of a branch.
         */
        if (headVertex == null)
            headVertex = currVertex;
        else
            currBranch.setVertex(currVertex);
    }

    /**
     * Method called when a &lt;/Vertex&gt; tag is encountered.
     * Ends the current vertex and moves up in the tree.
     */
    public void endVertex()
    {
        // Move up in the tree
        if (currVertex != headVertex)
            currVertex = currBranch.getParent();
    }

    /**
     * Method called when a &lt;Branch&gt; tag is encountered.
     * Creates a new branch as the child of the current vertex.
     */
    public void beginBranch()
    {
        currBranch = new Branch(currVertex);
        currVertex.addChild(currBranch);
    }

    /**
     * Method called when a &lt;/Values&gt; tag is encountered.
     * Sets the values of the current branch to those provided.
     *
     * @param values contents of the &lt;Values&gt; tag
     */
    public void endValues(String values)
    {
        currBranch.setValues(values);
    }

    /**
     * Method called when a &lt;/Branch&gt; tag is encountered.
     */
    public void endBranch()
    {
        // Move up in the tree
        currBranch = currVertex.getParent();
    }

    /**
     * Method called when a &lt;/Mean&gt; tag is encountered.
     */
    public void endMean(String contents)
    {
        try {
            currMean = Float.parseFloat(contents);
        } catch (NumberFormatException e) {
            currMean = Float.NaN;
        }
    }

    /**
     * Method called when a &lt;/SD&gt; tag is encountered.
     */
    public void endSD(String contents)
    {
        try {
            currSD = Float.parseFloat(contents);
        } catch (NumberFormatException e) {
            currSD = Float.NaN;
        }
    }

    /**
     * Method called when a &lt;/ProbMissing&gt; tag is encountered.
     */
    public void endProbMissing(String contents)
    {
        try {
            currProbMissing = Float.parseFloat(contents);
        } catch (NumberFormatException e) {
            currProbMissing = Float.NaN;
        }
    }

    /** 
     * Method called when a &lt;BinGaussian&gt; tag is encountered.
     */
    public void endBinGaussian()
    {
        /* Detect and report missing/invalid data.  The mean, standard
         * deviation, and missing probability are all required data
         * for creating a BinGaussian.  If these were not present in the
         * XML file, or could not be read, then we have a problem.  Things
         * may break later on, too, since there's no error recovery coded
         * here.
         */
        if (Float.isNaN(currSD) || Float.isNaN(currProbMissing)
                || Float.isNaN(currMean))
        {
            System.out.println(
                "ERROR: contents of BinGaussian tag are invalid or missing.");
        }

        currBinGaussian = new BinGaussian(currMean, currSD, currProbMissing);
        if (currBranch != null)
            currBranch.setLeaf(currBinGaussian);

        // Reset these values, for error detection.
        currMean = currSD = currProbMissing = Float.NaN;
    }

    /**
     * Convert this object to a String representation, in XML.  Contains
     * all information necessary to reconstruct a complete decision tree.
     */
    public String toString()
    {
        String ret = "<DecisionTree>\n";
        printIndent = 1;

        // Recursively call the toString methods of any children.
        if (headVertex == null)
            ret += currBinGaussian.toString(printIndent);
        else
            ret += headVertex;

        printIndent = 0;
        ret += "</DecisionTree>\n";

        return ret;
    }

    /* Generate appropriate number of tabs for printing, based on
     * the current printIndent.
     */
    protected String tabs()
    {
        String ret = "";
        for (int i = 0; i < printIndent; i++)
            ret += '\t';

        return ret;
    }

    /* A vertex in a decision tree.  Splits on some variable, has a branch
     * as its parent, and has branches as children.
     */
    private class Vertex
    {
        /* Index to variable on which we're splitting.
         * i.e. if split = 1 and the second input variable is 130, then
         * this Vertex is really splitting on variable 130.
         */
        int split;
        LinkedList children;
        Branch parent;

        /* Construct a Vertex that splits on the variable with the given index
         * and has the given Branch as its parent.  parent should only be null
         * if this is the head vertex.
         */
        public Vertex(int split, Branch parent)
        {
            this.split = split;
            this.parent = parent;
            children = new LinkedList();
        }

        /* Add another branch as a child of this Vertex
         */
        public void addChild(Branch child)
        {
            children.add(child);
        }

        /* Return the parent branch.  For the head Vertex, this is null.
         */
        public Branch getParent()
        {
            return parent;
        }

        /* Return the probability associated with this set of states.
         * Recursively searches the decision tree.
         */
        public float getProb(float state, float[] allStates)
        {
            float ret = -1;

            // Iterate through child branches until we find one leading to
            // a leaf matching our input states.
            Iterator iter = children.iterator();
            while (iter.hasNext() && (ret < 0))
            {
                Branch child = (Branch)iter.next();
                ret = child.getProb(state, allStates);
            }

            return ret;
        }

        /* Convert this object to a String representation.  Recursively
         * prints out all children.
         */
        public String toString()
        {
            String ret = tabs() + "<Vertex split=\"" + split + "\">\n";
            Iterator iter = children.iterator();
            printIndent++;
            while (iter.hasNext())
            {
                Branch child = (Branch)iter.next();
                ret += child;
            }
            printIndent--;
            ret += tabs() + "</Vertex>\n";

            return ret;
        }
    }


    /* An arc in a decision tree.  May lead to a vertex (which splits on a
     * variable into two or more branches) or a leaf node (a gaussian
     * probability distribution).
     */
    private class Branch
    {
        // Values of our parent vertex's variable
        LinkedList values;
        Vertex parent;

        // A Branch may have either a vertex or a leaf (multinomial) as its
        // child.
        Vertex vChild;
        BinGaussian bgChild;

        // Construct a branch with the given vertex as its parent.
        // Parent should not be null.
        public Branch(Vertex parent)
        {
            this.parent = parent;
            values = new LinkedList();
        }

        // Set the list of values for this branch.
        // The values should be a string of space-delimited integers, such
        // as is found in the contents of a <Values> tag.
        // TODO
        public void setValues(String strValues)
        {
            RangeTokenizer rt = new RangeTokenizer(strValues);
            while (rt.hasMoreTokens())
            {
                /* DEBUG
                Range r = (Range)rt.nextToken();
                if (r.inRange(Float.NaN))
                    System.out.println("0");
                else
                    System.out.println("1");
                 */

                values.add(rt.nextToken());
            }
        }

        // Set the child of this branch to be the given binGaussian
        public void setLeaf(BinGaussian leaf)
        {
            bgChild = leaf;
        }

        // Set the child of this branch to be the given vertex
        public void setVertex(Vertex vertex)
        {
            vChild = vertex;
        }

        // Get the parent Vertex.  Should never return null.
        public Vertex getParent()
        {
            return parent;
        }

        public boolean inRange(float val)
        {
            Iterator iter = values.iterator();
            while (iter.hasNext())
            {
                Range r = (Range)iter.next();
                if (r.inRange(val))
                    return true;
            }

            return false;
        }

        // Return the probability associated with this set of states.
        // Recursively searches the decision tree.
        public float getProb(float state, float[] allStates)
        {
            // First, check to see if the state for our parent vertex's
            // variable is listed in this branch.
            if (!inRange(allStates[parent.split]))
                return -1;
            else if (bgChild != null)
            {
                if (Float.isNaN(bgChild.getProb(state)))
                    System.out.println("ERROR: yup, NaN");

                return bgChild.getProb(state);
            }
            else if (vChild != null)
                return vChild.getProb(state, allStates);
            else
            {
                System.out.print(
            "ERROR: missing leaf or branch in ContinuousDecisionTree.java.");
                return 0;
            }
        }

        /* Generate a String representation of this object in XML.
         */
        public String toString()
        {
            String ret = tabs() + "<Branch>\n" + tabs() + "<Values>";
            Iterator iter = values.iterator();
            while (iter.hasNext())
            {
                Range r = (Range)iter.next();
                ret += r + " ";
            }
            ret += "</Values>\n";

            printIndent++;
            if (bgChild != null)
                ret += bgChild.toString(printIndent);
            else
                ret += vChild;
            printIndent--;

            ret += tabs() + "</Branch>\n";
            return ret;
        }
    }
}
