/* 
 * DecisionTree.java
 * 
 * Version 1.1
 * 
 * 23/08/2002
 *
 * Copyright (c) 2003 Oregon State University. All rights reserved.
 */
package org.recommender.algorithms.Experimental;

import org.recommender.server.*;
import java.util.*;
import java.io.*;

// XML parsing libraries
import com.jclark.xml.parse.*;
import com.jclark.xml.parse.base.*;

/** 
 * Helps parse a decision tree by processing begin/end element events.
 * Stores a representation of the decision tree in memory.
 *
 * @author Daniel Lowd &lt;dlowd@cs.orst.edu&gt;
 */
public class DecisionTree
{
    /**
     * ArrayList of input variable names
     */
    ArrayList inputVars;
    
    /**
     * HashMap of LinkedLists of state indices for the input variables
     */
    HashMap inputVarStates;

    /**
     * Top level vertex, before any branches.  (Note: some decision trees
     * have no vertices at all; in such cases, this is null.)
     */
    Vertex headVertex;

    /**
     * Keep track of current position in the tree, as it is built.
     */
    Vertex currVertex;
    Branch currBranch;
    Multinomial currMultinomial;

    /**
     * Construct a DecisionTree with the given input variables.
     */
    public DecisionTree()
    {
        headVertex = null;
        currVertex = null;
        currBranch = null;
        currMultinomial = null;
    }

    /**
     * Returns the probability that this variable is in a specified
     * state, given the states of all other variables in the network.
     * @param state the potential state of this decision tree's variable.
     * @param allStates array of integers specifying the state of each
     *     other variable.  allStates[i] = the state of variable i.
     * @return prob
     */
    public float getProb(int state, int[] allStates)
    {
        if (headVertex == null)
        {
            if (currMultinomial == null)
                return 0;
            else
                return currMultinomial.getProb(state);
        }

        return headVertex.getProb(state, allStates);
    }

    /** 
     * Method called when a &lt;Vertex&gt; tag is encountered.
     * Creates a new vertex as the child of the current branch.
     * @param vertex
     */
    public void beginVertex(StartElementEvent vertex)
    {
        String splitVal = vertex.getAttributeValue("split");
        try {
            int var = Integer.parseInt(splitVal);
            currVertex = new Vertex(var, currBranch);
        } catch(NumberFormatException e) {
            // TODO: better error handling?
        }

        // If there's no root vertex, set it as such
        if (headVertex == null)
            headVertex = currVertex;

        // Otherwise, add it as the child of a branch
        else
            currBranch.setVertex(currVertex);
    }

    /**
     * Method called when a &lt;/Vertex&gt; tag is encountered.
     * Ends the current vertex and moves up in the tree.
     */
    public void endVertex()
    {
        // Move up in the tree
        if (currVertex != headVertex)
            currVertex = currBranch.getParent();
    }

    /**
     * Method called when a &lt;Branch&gt; tag is encountered.
     * Creates a new branch as the child of the current vertex.
     */
    public void beginBranch()
    {
        currBranch = new Branch(currVertex);
        currVertex.addChild(currBranch);
    }

    /**
     * Method called when a &lt;/Values&gt; tag is encountered.
     * Sets the values of the current branch to those provided.   
     * @param values
     */
    public void endValues(String values)
    {
        currBranch.setValues(values);
    }

    /**
     * Method called when a &lt;/Branch&gt; tag is encountered.
     */
    public void endBranch()
    {
        // Move up in the tree
        currBranch = currVertex.getParent();
    }

    /** 
     * Method called when a &lt;Multinomial&gt; tag is encountered.
     */
    public void beginMultinomial()
    {
        currMultinomial = new Multinomial();
        if (currBranch != null)
            currBranch.setLeaf(currMultinomial);
    }

    /** 
     * Method called when a &lt;/Probs&gt; tag is encountered. 
     * @param probs pcdata contents of the &lt;Probs&gt; tag.
     */
    public void endProbs(String probs)
    {
        currMultinomial.addProbs(probs);
    }

    /**
     * A vertex in a decision tree.  Splits on some variable, has a branch
     * as its parent, and has branches as children.
     */
    private class Vertex
    {
        /**
         * Index to variable on which we're splitting.
         * i.e. if split = 1 and the second input variable is 130, then
         * this Vertex is really splitting on variable 130.
         */
        int split;
        LinkedList children;
        Branch parent;

        /**
         * Construct a Vertex that splits on the variable with the given index
         * and has the given Branch as its parent.  parent should only be null
         * if this is the head vertex.
         */
        public Vertex(int split, Branch parent)
        {
            this.split = split;
            this.parent = parent;
            children = new LinkedList();
        }

        /**
         * Add another branch as a child of this Vertex
         */
        public void addChild(Branch child)
        {
            children.add(child);
        }

        /**
         * Return the parent branch.  For the head Vertex, this is null.
         */
        public Branch getParent()
        {
            return parent;
        }

        /**
         * Return the probability associated with this set of states.
         * Recursively searches the decision tree.
         */
        public float getProb(int state, int[] allStates)
        {
            float ret = -1;

            // Iterate through child branches until we find one leading to
            // a leaf matching our input states.
            Iterator iter = children.iterator();
            while (iter.hasNext() && (ret < 0))
            {
                Branch child = (Branch)iter.next();
                ret = child.getProb(state, allStates);
            }

            return ret;
        }
    }


    /**
     * An arc in a decision tree.  May lead to a vertex (which splits on a
     * variable into two or more branches) or a leaf node (a multinomial
     * probability distribution).
     */
    private class Branch
    {
        /**
         * Values of our parent vertex's variable
         */
        LinkedList values;
        Vertex parent;

        /**
         * A Branch may have either a vertex or a leaf (multinomial) as its
         * child.
         */
        Vertex vChild;
        Multinomial mChild;

        /**
         * Construct a branch with the given vertex as its parent.
         * Parent should not be null.
         */
        public Branch(Vertex parent)
        {
            this.parent = parent;
            values = new LinkedList();
        }

        /**
         * Set the list of values for this branch.
         * The values should be a string of space-delimited integers, such
         * as is found in the contents of a <Values> tag.
         */
        public void setValues(String strValues)
        {
            StringTokenizer st = new StringTokenizer(strValues);
            while (st.hasMoreTokens())
                values.add(new Integer(st.nextToken()));
        }

        /**
         * Set the child of this branch to be the given multinomial
         */
        public void setLeaf(Multinomial leaf)
        {
            mChild = leaf;
        }

        /**
         * Set the child of this branch to be the given vertex
         */
        public void setVertex(Vertex vertex)
        {
            vChild = vertex;
        }

        /**
         * Get the parent Vertex.  Should never return null.
         */
        public Vertex getParent()
        {
            return parent;
        }

        /**
         * Return the probability associated with this set of states.
         * Recursively searches the decision tree.
         */
        public float getProb(int state, int[] allStates)
        {
            // First, check to see if the state for our parent vertex's
            // variable is listed in this branch.
            if (!values.contains(new Integer(allStates[parent.split])))
                return -1;
            else if (mChild != null)
                return mChild.getProb(state);
            else
                return vChild.getProb(state, allStates);
        }
    }
}
