/*
 * DumbPopAlgorithm.java
 *
 * Copyright (c) 2003 Oregon State University. All rights reserved.
 */
package org.recommender.algorithms.Experimental;
import org.recommender.algorithms.*;
import org.recommender.server.*;
import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.sql.*;

/**
 * Describes an implementation of the CFAlgorithm interface to make
 * CF predictions using a poor non-personalized prediction algorithm.
 * The ratings it predicts are simply the average ratings for each movie.
 * This generally doesn't work as well as PopAlgorithm, which takes into
 * account different user means.
 *
 * @author Daniel Lowd &lt;dlowd@cs.orst.edu&gt;
 */
public class DumbPopAlgorithm implements CFAlgorithm
{
    /**
     * DataManager of this algorithm
     */
    private DataManager dataMgr;

    /**
     * Variables to keep track of the average time for a call to predictRating
     */
    private long Time = 0;
    private long NumPredictions = 0;

	/**
	 * Constructor, simply assigns DataManager.
	 * @param mgr
	 */
    public DumbPopAlgorithm(DataManager mgr)
    {
		dataMgr = mgr;
    }

    /**
     * Yun Wang
     *
     * Return top n recommendations of specific type.
     * @param activeUser
     * @param n
     * @param type
     * @return itemPredList
     * @throws SQLException
     */
    public ItemPrediction[] getRecommendationsByType(int activeUser, int n, int type)
               throws SQLException, CFNotImplementedException
   {
   		throw new CFNotImplementedException("getRecommendationsByType not implemented yet");
         //return null;

   }

    /**
     * Returns top n recommendations.  Uses the same algorithm as
     * predictRating, but predicts all items the user has not rated, rather
     * than just one.  Sorts the items in reverse order of rating and returns
     * the list.
     *
	 * @param activeUser user to predict for
	 * @param n number of recommendations to acquire
	 * @return itemPredList
	 */
	public ItemPrediction[] getRecommendations(int activeUser, int n)
            throws SQLException, CFNotImplementedException
    {
		ItemPrediction[] itemPredList;
		int index = 0;
        int maxItem = dataMgr.getMaxItem() + 1;
		float[] num = new float[maxItem];
		float[] den = new float[maxItem];

		for (int u = 0; u < maxItem; u++) {
			num[u] = 0;
			den[u] = 0;
		}

		itemPredList = new ItemPrediction[maxItem];

        int maxUser = dataMgr.getMaxUser() + 1;
        for (int i = 0; i < maxUser; i++)
        {
            if (dataMgr.getNumRatings(i) == 0)
                continue;

            Iterator iter = dataMgr.getItemList(i);
            while (iter.hasNext())
            {
                ItemRating ir = (ItemRating)iter.next();
                num[ir.getItemID()] += ir.getRating();
                den[ir.getItemID()]++;
            }
        }

		itemPredList[0] =  new ItemPrediction(0, 0, 0);
		for (int i = 1; i < num.length; i++)
        {
			if (num[i] > 0 && den[i] > 0 && dataMgr.getRatingCached(
                        activeUser, i) == dataMgr.getErrRating())
            {
				itemPredList[i] = new ItemPrediction(0, i, num[i]/den[i]);
            }
			else
				itemPredList[i] = new ItemPrediction(0, i, 0);
		}

		Arrays.sort(itemPredList, new CompPrediction(dataMgr));
		return itemPredList;
	}


    /* predicts the rating for the given user and item
     * Returns a simple average of all user ratings for
     * the given item.
     */
    public ItemPrediction predictRating (int activeUser, int item) throws CFNotImplementedException
    {
        long startTime = System.currentTimeMillis();

        // Return the average rating
        float num = 0;
        int den = 0;

        Iterator ratings = dataMgr.getNeighborhood(item);
        while (ratings.hasNext())
        {
            ItemRating ir = (ItemRating)ratings.next();
            num += ir.getRating();
            den++;
        }

        long stopTime = System.currentTimeMillis();
        Time += stopTime - startTime;
        NumPredictions++;

        if (den == 0)
            return new ItemPrediction(activeUser,item,dataMgr.getErrRating());
        else
            return new ItemPrediction(activeUser, item, num/den);
    }


    /**
     * Requires no action.
     */
    public void updateUser(int userID) throws CFNotImplementedException
    {
        throw new CFNotImplementedException("updateUser not implemented yet");
    }


    /**
     * Return average prediction time.
     * @return long
     */
    public long getAveragePredictionTime() throws CFNotImplementedException
    {
        return (Time/NumPredictions);
    }

    /**
     * Return the name of this algorithm.
     * @return String
     */
    public String toString()
    {
        return "DumbPopAlgorithm.java";
    }
}
