/*
 * HortingAlgorithm.java
 *
 * Version 1.1
 *
 * 15/08/2002
 *
 * Copyright (c) 2003 Oregon State Universtity. All rights reserved.
 */
package org.recommender.algorithms.Experimental;
import org.recommender.algorithms.*;
import org.recommender.server.*;
import java.io.*;
import java.util.*;
import java.sql.*;

/**
 * Describes an implementation of the CFAlgorithm interface
 * to make CF predictions using the graph-theoretic algorithm
 * described by Aggarwal, et al based on notions of horting
 * and predictability.
 *
 * @author  Daniel Lowd &lt;dlowd@cs.orst.edu&gt;
 */
public class HortingAlgorithm implements CFAlgorithm
{
    private DataManager dataMgr;

    /**
     * Keep track of all adjacencies
     */
    private ArrayList neighborLists;

    /*
     * Algorithm Parameters -- changing these can modify the speed and accuracy
     * of the algorithm.  In general, looser requirements allow more links in
     * the graph and thus more accuracy.  More stringent requirements result
     * in fewer links in the graph and thus more speed.  Of course, looser
     * requirements sometimes result in using worse data... so it's not a
     * simple tradeoff.
     */

    /**
     * We hort any user that:
     *   Has rated MAX_MATCH items that we've rated OR
     *     Has rated ratio HORT_RATIO of the items we've rated
     *     (at least MIN_MATCH)
     */
    private float HORT_RATIO;
    private float MAX_MATCH;
    private float MIN_MATCH;

    /**
     * We predict any user that:
     *   We hort (see above) AND
     *     Has an average rating difference of less than or equal to
     *     MAX_DISTANCE (after normalizing transformation)
     */
    private float MAX_DISTANCE;

    private float MAX_PATH_LENGTH;

    /**
     * If we find MIN_NEIGHBORS neighbors at distance &lt;= d from us, then
     *    we will not search at level d+1.
     * If we have found MAX_NEIGHBORS neighbors in our search thus far, we
     *    will stop immediately.  Never will more than MAX_NEIGHBORS neighbors
     *    be used for a prediction.  (-1 implies no limit)
     */
    private int MIN_NEIGHBORS;
    private int MAX_NEIGHBORS;

    /**
     * Extension: enable or disable negative coefficients
     */
    private boolean USE_NEGATIVE_COEFFS;

    /**
     * Limit the number of users that may predict a given user.
     */
    private int MAX_PREDICTORS;

    /**
     * Set the effect of the number of items in common on the
     * calculated distance, when calculating predictors.
     */
    private int DISTANCE_WEIGHTING;

    private final static int LINEAR_WEIGHT = 0;
    private final static int LOG_WEIGHT    = 1;
    private final static int NO_WEIGHT     = 2;
    private final static int SQRT_WEIGHT   = 3;

    /**
     * Minimal and maximal rating for this dataset
     */
    private int MIN_RATING;
    private int MAX_RATING;

    /**
     * Member vars for bucket-sort to keep track of which ratings are from
     * whom.  Since only the getHorts() method uses them, they would be static
     * variables there if this were C++.
     */
    private int maxUser;
    private LinkedList[] buckets;
    private int[] bucketSize;
    private boolean initialized;

    private int[][] st_transform_base;

    private boolean[] visited;


    private long NumPredictions = 0;

    /**
     * Variable to keep track of the average time for a call to predictRating
     */
    private long Time = 0;

    /**
     * Constructor, when neighborLists unavailable
     * @param dataMgr
     */
    public HortingAlgorithm(DataManager dataMgr)
    {
        this.dataMgr  = dataMgr;
        loadProperties();

        maxUser = dataMgr.getMaxUser();
        neighborLists = new ArrayList(maxUser+1);
        for (int i = 0; i <= maxUser; i++)
            neighborLists.add(new LinkedList());

        initPredictArrays();
        initHortingArrays();

        // Try to load in pre-computed graph data
        //int start = graphMgr.readFromFile(GRAPH_FILENAME);
        int start = 0;

        // Build the rest
        long begin = System.currentTimeMillis();
        if (start*1000 < maxUser)
            buildGraph(start*1000);
        long end = System.currentTimeMillis();

        System.out.println("Total graph-building time: " + (end - begin)/1000
                + "s.");
    }

    /**
     * Constructor with neighborLists available
     * @param dataMgr
     * @param neighborLists
     */
 /*   public HortingAlgorithm(DataManager dataMgr, ArrayList neighborLists)
    {
        this.dataMgr  = dataMgr;
        loadProperties();

        maxUser = neighborLists.size() - 1;

        // Copy lists
        this.neighborLists = new ArrayList(maxUser+1);
        for (int i = 0; i <= maxUser; i++)
        {
            List neighbor = (List)neighborLists.get(i);
            Vector currUser = new Vector(neighbor.size());
            int numRated = dataMgr.getNumRatings(i);

            Iterator transforms = neighbor.iterator();
            while (transforms.hasNext())
            {
                Transform t = (Transform)transforms.next();
                boolean horts = ((t.commonItems > HORT_RATIO*numRated &&
                        t.commonItems > MIN_MATCH) ||
                        t.commonItems > MAX_MATCH);
                boolean predicts = horts && (t.distance < MAX_DISTANCE);

                if (predicts)
                    currUser.add(t);
            }
            currUser.trimToSize();
            this.neighborLists.add(i, currUser);
        }

        initPredictArrays();
        initHortingArrays();
    }
  */  

    /**
     * Initialize algorithm parameters.
     */
    private void loadProperties()
    {
        MIN_RATING = (int)dataMgr.getMinRating();
        MAX_RATING = (int)dataMgr.getMaxRating();

        String prefix = "CFServer.alg.horting.";
        HORT_RATIO = SystemProperties.getFloat(prefix + "hortRatio", (float) 0.1);
        MIN_MATCH  = SystemProperties.getInt(prefix + "minMatch", 2);
        MAX_MATCH  = SystemProperties.getInt(prefix + "maxMatch", 5);

        MAX_DISTANCE = SystemProperties.getFloat(prefix + "maxDistance", 1);
        MAX_PATH_LENGTH = SystemProperties.getInt(prefix + "maxPathLength", 4);

        // Modifications to the algorithm
        MIN_NEIGHBORS = SystemProperties.getInt(prefix + "minNeighbors", 1);
        MAX_NEIGHBORS = SystemProperties.getInt(prefix + "maxNeighbors", 60);
        USE_NEGATIVE_COEFFS = SystemProperties.getBoolean(prefix
                + "useNegativeCoeffs", false);
        MAX_PREDICTORS = SystemProperties.getInt(prefix + "maxPredictors",
                1000);
        String distanceWeighting = SystemProperties.getString(
                prefix + "distanceWeighting", "none");

        // Set the effect of differing numbers of common items.
        // Default to linear.
        if (distanceWeighting.equalsIgnoreCase("log"))
            DISTANCE_WEIGHTING = LOG_WEIGHT;
        else if (distanceWeighting.equalsIgnoreCase("none"))
            DISTANCE_WEIGHTING = NO_WEIGHT;
        else if (distanceWeighting.equalsIgnoreCase("linear"))
            DISTANCE_WEIGHTING = LINEAR_WEIGHT;
        else if (distanceWeighting.equalsIgnoreCase("sqrt"))
            DISTANCE_WEIGHTING = SQRT_WEIGHT;
        else
            DISTANCE_WEIGHTING = LINEAR_WEIGHT;
    }

    /**
     * Initialize transform matrix
     */
    private void initPredictArrays()
    {
        // Bounds on offsets, given coeff of -1
        int minTnegS;
        int maxTnegS;
        if (USE_NEGATIVE_COEFFS)
        {
            minTnegS = 2*MIN_RATING;
            maxTnegS = 2*MAX_RATING;
        }
        else
        {
            minTnegS = 0;
            maxTnegS = -1;
        }

        // Bounds on offsets, given coeff of +1
        int minTposS = MIN_RATING - MAX_RATING;
        int maxTposS = MAX_RATING - MIN_RATING;

        // Total number of transforms = # for coeff +1 pus # for coeff -1
        int numTransforms = (maxTnegS - minTnegS + 1)
            + (maxTposS - minTposS + 1);

        st_transform_base = new int[numTransforms][MAX_RATING + 3];

        int i = 0;

        // Fill in all coefficient and offset values
        for (int t = minTnegS; t <= maxTnegS; t++, i++)
        {
            st_transform_base[i][MAX_RATING + 1] = -1;
            st_transform_base[i][MAX_RATING + 2] = t;
        }

        for (int t = minTposS; t <= maxTposS; t++, i++)
        {
            st_transform_base[i][MAX_RATING + 1] = 1;
            st_transform_base[i][MAX_RATING + 2] = t;
        }

        // Calculate all the transforms
        for (i = 0; i < st_transform_base.length; i++)
            for (int j = MIN_RATING; j <= MAX_RATING; j++)
            {
                st_transform_base[i][j] = st_transform_base[i][MAX_RATING+1]*j
                    + st_transform_base[i][MAX_RATING+2];
            }
    }

    // TODO -- fix this...
    /**
     * Initialize horting arrays
     */
    private void initHortingArrays()
    {
        // Initialize bucket-sort member vars.
        buckets = new LinkedList[maxUser+1];
        bucketSize = new int[maxUser+1];

        // HACK DEBUG -- temporarily putting this here...
        visited = new boolean[maxUser+1];
    }

    /**
     * Return neighborLists
     * @return neighborLists
     */
    public ArrayList getNeighborLists()
    {
        return neighborLists;
    }

    /**
     * Get predicted rating for a user's item
     * @param userID
     * @param itemID
     * @return predicted rating
     */
    public ItemPrediction predictRating(int userID, int itemID) throws CFNotImplementedException
    {
        long startTime = System.currentTimeMillis();

        LinkedList currQueue = new LinkedList();
        LinkedList nextQueue = new LinkedList();
        for (int i = 0; i <= maxUser; i++)
            visited[i] = false;

        float numerator = 0;
        float denominator = 0;

        int distance = 0;
        currQueue.add(new Transform(userID, 1, 0, 0, 0));

        int numNeighbors = 0;
        while (distance <= MAX_PATH_LENGTH && numNeighbors < MIN_NEIGHBORS)
        {
            Iterator i = currQueue.iterator();
            while (i.hasNext())
            {
                Transform curr = (Transform)i.next();

                // Get the user's rating for this item, if it exists.
                float r = dataMgr.getRatingCached(curr.targetID, itemID);

                if (r >= MIN_RATING)
                {
                    //float weight = 1/(MAX_DISTANCE + curr.distance);
                    numerator   += curr.translate(r); //*weight;
                    denominator++; // += weight;
                    numNeighbors++;
                    if (numNeighbors >= MAX_NEIGHBORS)
                        break;
                }
            }

            // Stop once we've found enough ratings
            if ( numNeighbors >= MIN_NEIGHBORS )
                break;

            i = currQueue.iterator();
            while (i.hasNext())
            {
                Transform curr = (Transform)i.next();

                // Add all outward arcs to neighbors we have not yet visited.
                List neighborList = (List)neighborLists.get(curr.targetID);
                Iterator neighborIter = neighborList.iterator();
                while (neighborIter.hasNext())
                {
                    Transform nextNeighbor = (Transform)neighborIter.next();
                    if ( !visited[nextNeighbor.targetID] )
                    {
                        visited[nextNeighbor.targetID] = true;
                        nextQueue.add(new Transform(curr, nextNeighbor));
                    }
                }
            } // while

            // Search on this next level of distance
            currQueue = nextQueue;
            nextQueue = new LinkedList();
            distance++;
        }

        if (numNeighbors < MIN_NEIGHBORS || denominator == 0)
            return new ItemPrediction(userID, itemID,dataMgr.getErrRating());

        // System.out.println("Users rated: " + usersRated
        //        + " (" + distance + ")");

        long stopTime = System.currentTimeMillis();
        NumPredictions++;
        Time += (stopTime - startTime);

        return new ItemPrediction(userID, itemID,((float)numerator)/denominator);
    }

    /**
     * Get average prediction time.
     * @return average prediction time
     */
    public long getAveragePredictionTime() throws CFNotImplementedException
    {
        return (Time/NumPredictions);
    }


     /**
     *  Yun Wang
     *
     * Return top n recommendations of specific type.
     * @param activeUser
     * @param n
     * @param type
     * @return itemPredList
     * @throws SQLException
     */
     public ItemPrediction[] getRecommendationsByType(int activeUser, int n, int type)
               throws SQLException, CFNotImplementedException
     {
        long beginTime = System.currentTimeMillis();

        int userID = activeUser;

        List neighborList = (List)neighborLists.get(userID);
        Iterator neighbors = neighborList.iterator();
        int numNeighbors = 0;

        int maxItem = dataMgr.getMaxItem();

        float[] num = new float[maxItem+1];
        float[] den = new float[maxItem+1];

     //  int[] typeArr = new int[maxItem+1];     // for storing type

		for (int u = 0; u <= maxItem; u++)
        {
			num[u] = 0;
			den[u] = 0;
		//	typeArr[u] = 0;
		}

     //  long start= System.currentTimeMillis();

   //    Iterator typedIt = dataMgr.getCachedTypedItemList(type);
    // System.out.println("Get the Item List");
       int size = dataMgr.getSizeOfTypedItemList(type);
    // System.out.println("The size of typed item list is " + size);

     /*  while (typedIt.hasNext())
        {
            int i = ((Integer) typedIt.next()).intValue();
            typeArr[i] = 1;
        }*/

     //   long stop= System.currentTimeMillis();
    //  System.out.println("Time to get typed item list = " + (stop - start));

        while (neighbors.hasNext())
        {
            Transform t = (Transform)neighbors.next();

            // Transform and add this user's ratings
            ItemRating[] targetRatings = dataMgr.getUserRatingList(t.targetID);
            for (int i = 0; i < targetRatings.length; i++)
            {
                ItemRating rating = targetRatings[i];
                num[rating.getItemID()] += t.translate(rating.getRating());
                den[rating.getItemID()]++;
            }

            numNeighbors++;
        }

        long postTransformTime = System.currentTimeMillis();
        System.out.println("Time to transform: " + (postTransformTime
                    - beginTime));

        ItemPrediction[] ratingsArray= new ItemPrediction[size+1];

        long start = System.currentTimeMillis();
        int arrayIdx = 0;
        float MIN_GOOD_RATING = dataMgr.getMaxRating();
        while (arrayIdx < n && MIN_GOOD_RATING > dataMgr.getMinRating())
        {
           /* for (int i = 0; i <= maxItem; i++)
            {
                if (den[i] != 0 && typeArr[i] > 0)   //typed item list filter
                {
                    float r = num[i]/den[i];

                    // Only bother sorting the really good scores.
                    if (r > MIN_GOOD_RATING)
                        ratingsArray[arrayIdx++] = new ItemPrediction(0, i, r);
                }
            }*/

           Object[] typedIt = ((Vector) dataMgr.getCachedTypedItemList(type)).toArray();

          int v;
          for(int i = 0; i < typedIt.length; i++)
          {
              v = ((Integer) typedIt[i]).intValue();
              if (den[v] != 0)
              {
                    float r = num[v]/den[v];

                    // Only bother sorting the really good scores.
                    if (r > MIN_GOOD_RATING)
                        ratingsArray[arrayIdx++] = new ItemPrediction(0, v, r);
                }

            }

          /* Iterator typedIt = dataMgr.getCachedTypedItemList(type);

            int i;
            while (typedIt.hasNext())
            {
                i = ((Integer) typedIt.next()).intValue();

                if (den[i] != 0)
                {
                    float r = num[i]/den[i];

                    // Only bother sorting the really good scores.
                    if (r > MIN_GOOD_RATING)
                        ratingsArray[arrayIdx++] = new ItemPrediction(0, i, r);
                }

            }*/

            MIN_GOOD_RATING -= 0.2;
        }
        long stop = System.currentTimeMillis();
        System.out.println("Time to fill the predArray: " + (stop - start));

        long preSortTime = System.currentTimeMillis();

		Arrays.sort(ratingsArray, 0, arrayIdx, new CompPrediction(dataMgr));
        long postSortTime = System.currentTimeMillis();
      System.out.println("Time to sort array: " + (postSortTime
                   - preSortTime));

       System.out.println("Total time: " + (System.currentTimeMillis()
                  - beginTime) + "ms.");

        return ratingsArray;


     }

    /**
     * Returns top n recommendations.
     * @param activeUser
     * @param n
     * @return ratingsArray
     */
    public ItemPrediction[] getRecommendations(int activeUser, int n)
            throws SQLException, CFNotImplementedException
    {
        long beginTime = System.currentTimeMillis();

        // TODO: find the best way to implement this algorithm...
        // Ideas:
        // - Incorporate indirect neighbors as well
        // - Penalize ratings of users that are more distant?
        int userID = activeUser;

        List neighborList = (List)neighborLists.get(userID);
        Iterator neighbors = neighborList.iterator();
        int numNeighbors = 0;

        int maxItem = dataMgr.getMaxItem();
        float[] num = new float[maxItem+1];
        float[] den = new float[maxItem+1];

		for (int u = 0; u <= maxItem; u++)
        {
			num[u] = 0;
			den[u] = 0;
		}

        while (neighbors.hasNext())
        {
            Transform t = (Transform)neighbors.next();

            // Transform and add this user's ratings
            ItemRating[] targetRatings = dataMgr.getUserRatingList(t.targetID);
            for (int i = 0; i < targetRatings.length; i++)
            {
                ItemRating rating = targetRatings[i];
                num[rating.getItemID()] += t.translate(rating.getRating());
                den[rating.getItemID()]++;
            }
            // OLD iterVector.add(transformed.iterator());

            numNeighbors++;
        }

        long postTransformTime = System.currentTimeMillis();
        System.out.println("Time to transform: " + (postTransformTime
                    - beginTime));

        //ItemRating[] ratingsArray = new ItemRating[maxItem+1];
        ItemPrediction[] ratingsArray = new ItemPrediction[maxItem+1];

        long start = System.currentTimeMillis();
        int arrayIdx = 0;
        float MIN_GOOD_RATING = dataMgr.getMaxRating();
        while (arrayIdx < n && MIN_GOOD_RATING > dataMgr.getMinRating())
        {
            for (int i = 0; i <= maxItem; i++)
            {
                if (den[i] != 0)
                {
                    float r = num[i]/den[i];

                    // Only bother sorting the really good scores.
                    if (r > MIN_GOOD_RATING)
                        ratingsArray[arrayIdx++] = new ItemPrediction(0, i, r);
                }
            }
            MIN_GOOD_RATING -= 0.2;
        }
        long stop = System.currentTimeMillis();
        System.out.println("Time to fill the predArray: " + (stop - start));

        long preSortTime = System.currentTimeMillis();
		//Arrays.sort(ratingsArray, 0, arrayIdx, new Comp());
		Arrays.sort(ratingsArray, 0, arrayIdx, new CompPrediction(dataMgr));
        long postSortTime = System.currentTimeMillis();
        System.out.println("Time to sort array: " + (postSortTime
                    - preSortTime));

        System.out.println("Total time: " + (System.currentTimeMillis()
                  - beginTime) + "ms.");

        return ratingsArray;
    }

    /**
     * Return the name of this algorithm
     * @return String
     */
    public String toString()
    {
        return "HortingAlgorithm.java";
    }



    /**
     * Build graph of predictors.
     * @param start
     */
    private void buildGraph(int start)
    {
        try {
        SystemUtils.printDebug("Building graph of predictors...");

        Iterator allUsers = dataMgr.getAllUsers();

        int last_save = start;
        int userID = 0;
        while (allUsers.hasNext())
        {
            Integer intUserID = (Integer)allUsers.next();
            userID = intUserID.intValue();

            if (userID < start)
                continue;

            linkToPredictors(userID);

            /*
            if (userID - last_save >= 1000)
            {
                graphMgr.saveUsers(last_save, last_save + 1000 - 1,
                        GRAPH_FILENAME + "." + last_save/1000);
                last_save += 1000;
            }
             */
        }

        /*
        graphMgr.saveUsers(last_save, userID, GRAPH_FILENAME + "."
                + last_save/1000);
         */

        System.out.println("Total hort time: " + hortTime/1000 + "s.");
        System.out.println("Total predict time: " + predictTime/1000 + "s.");

        } catch (RuntimeException e) {
            System.out.println("Error: " + e);
            e.printStackTrace();
            throw e;
        }
    }

    /**
     * When new user comes in, update the neighborLists and maxUser and HortingArray.
     * @param userID
     */
    public void updateUser(int userID) throws CFNotImplementedException
    {
        // Handle new users
        if (userID > maxUser)
        {
            for (int i = maxUser + 1; i <= userID; i++)
                neighborLists.add(i, new LinkedList());

            maxUser = userID;
            initHortingArrays();
        }

        linkToPredictors(userID);
    }


    /**
     * Keep track of total time spent horting and predicting
     */
    long hortTime = 0;
    long predictTime = 0;

    /**
     * Find out the top predictors
     * @param userID
     */
    private void linkToPredictors(int userID)
    {
        // Create a new linked list to hold our new neighbors.
        List neighborList = new LinkedList();

        long start = System.currentTimeMillis();
        Iterator horts = getHorts(userID).iterator();
        long stop = System.currentTimeMillis();
        hortTime += stop - start;

        start = System.currentTimeMillis();
        while (horts.hasNext())
        {
            // Initialize vars to keep track of which transforms are good
            int st_size = st_transform_base.length;
            int st_transform[][] = new int[st_size][];
            int distance[] = new int[st_size];
            for (int i = 0; i < st_size; i++)
            {
                st_transform[i] = st_transform_base[i];
                distance[i] = 0;
            }

            LinkedList hortRatings = (LinkedList)horts.next();
            int hortLen = hortRatings.size();
            int maxDist = (int)(hortLen * MAX_DISTANCE);
            Iterator hortIter = hortRatings.iterator();

            // NOTE: we're guaranteed that we have rated every item listed
            // in hortRatings.
            while (hortIter.hasNext())
            {
                // Get the ratings, both for the other user and for us
                ItemRating rating = (ItemRating)hortIter.next();
                int ourRating = (int)dataMgr.getRatingCached(userID,
                            rating.getItemID());
                int theirRating = (int)rating.getRating();

                // We should have rated this...
                if (ourRating < 0)
                {
                    System.out.println("Missing rating!");
                    System.out.println("User: " + userID + "; Item: "
                            + rating.getItemID() + "; Neighbor: "
                            + rating.getUserID());
                    continue;
                }

                // Try all linear transform values
                for (int i = 0; i < st_size; i++)
                {
                    distance[i] += Math.abs(st_transform[i][theirRating]
                            - ourRating);
                    if (distance[i] > maxDist)
                    {
                        st_size--;
                        distance[i] = distance[st_size];
                        st_transform[i] = st_transform[st_size];

                        // Try this one over, since we just replaced it...
                        i--;
                    }
                }

                if (st_size == 0)
                    break;
            }

            // Add an arc if there's a good tranform
            float d = MAX_DISTANCE + (float) 0.01;
            int s = 1;
            int t = 0;
            for (int j = 0; j < st_size; j++)
            {
                // Select the best s,t pair
                if (((float)distance[j])/hortLen < d)
                {
                    d = ((float)distance[j])/hortLen;
                    s = st_transform[j][MAX_RATING + 1];
                    t = st_transform[j][MAX_RATING + 2];
                }
            }
            if (d <= MAX_DISTANCE)
            {
                int targetID = ((ItemRating)hortRatings.getFirst()).getUserID();
                float dist = d;

                if (DISTANCE_WEIGHTING == LOG_WEIGHT)
                    dist = d/ (float) Math.log(hortLen);
                else if (DISTANCE_WEIGHTING == LINEAR_WEIGHT)
                    dist = d/hortLen;
                else if (DISTANCE_WEIGHTING == SQRT_WEIGHT)
                    dist = d/(float) Math.sqrt(hortLen);
                else
                    dist = d;

                neighborList.add(new Transform(targetID, s, t, dist, hortLen));
            }
        }

        // Sort the transforms so that we consider the best matches first.
        Collections.sort(neighborList);

        // Select only the top predictors
        if (neighborList.size() <= MAX_PREDICTORS)
            neighborLists.set(userID, new Vector(neighborList));
        else
            neighborLists.set(userID, new Vector(neighborList.subList(0,
                        MAX_PREDICTORS)));

        stop = System.currentTimeMillis();
        predictTime += stop - start;
    }


    /*  Failed experiment -- find a user's predictors not by applying a
     *  discrete transform, but by subtracting each user's mean rating from
     *  their ratings, then comparing.

    private void linkToPredictors(int userID)
    {
        // Create a new linked list to hold our new neighbors.
        List neighborList = new LinkedList();

        long start = System.currentTimeMillis();
        Iterator horts = getHorts(userID).iterator();
        long stop = System.currentTimeMillis();
        hortTime += stop - start;

        start = System.currentTimeMillis();
        while (horts.hasNext())
        {
            LinkedList hortRatings = (LinkedList)horts.next();
            int neighborID = ((ItemRating)hortRatings.getFirst()).getUserID();
            float meanDiff = dataMgr.getMeanRatingCached(neighborID) -
                dataMgr.getMeanRatingCached(userID);

            float distance = 0;
            Iterator hortIter = hortRatings.iterator();
            // NOTE: we're guaranteed that we have rated every item listed
            // in hortRatings.
            while (hortIter.hasNext())
            {
                // Get the ratings, both for the other user and for us
                ItemRating rating = (ItemRating)hortIter.next();
                float ourRating = dataMgr.getRatingCached(userID,
                            rating.getItemID());
                float theirRating = rating.getRating();

                // We should have rated this...
                if (ourRating < 0)
                {
                    System.out.println("Missing rating!");
                    System.out.println("User: " + userID + "; Item: "
                            + rating.getItemID() + "; Neighbor: "
                            + rating.getUserID());
                    continue;
                }

                distance += Math.abs(theirRating - ourRating + meanDiff);
            }

            // Add an arc if there's a good tranform
            distance /= hortRatings.size();
            if (distance <= MAX_DISTANCE)
            {
                int targetID = ((ItemRating)hortRatings.getFirst()).getUserID();
                neighborList.add(new Transform(targetID, 1, meanDiff,
                            distance));
            }
        }

        // Sort the transforms so that we consider the best matches first.
        Collections.sort(neighborList);

        // Select only the top predictors
        if (neighborList.size() <= MAX_PREDICTORS)
            neighborLists[userID] = neighborList;
        else
            neighborLists[userID] = neighborList.subList(0, MAX_PREDICTORS);

        stop = System.currentTimeMillis();
        predictTime += stop - start;
    }
    */


    /**
     * Get a list of all the users we hort using bucket sort.
     * @param userID
     * @return a LiskedList of horts
     */
    private LinkedList getHorts(int userID)
    {
        ItemRating[] userRatings = dataMgr.getUserRatingList(userID);

        if (userRatings.length == 0)
            return new LinkedList();

        // Initialize array for horting calcs.
        for (int i = 0; i <= maxUser; i++)
            bucketSize[i] = 0;

        // Loop through the neighborhoods of each item we've rated
        int numRated = 0;
        for (int i = 0; i < userRatings.length; i++)
        {
            Iterator itemRatings = dataMgr.getNeighborhood(
                    userRatings[i].getItemID());
            while (itemRatings.hasNext())
            {
                ItemRating ir = (ItemRating)itemRatings.next();

                /* Add the rating to the appropriate user's bucket,
                 * creating the bucket if necessary.
                 */
                int uid = ir.getUserID();

                if (bucketSize[uid] == 0)
                    buckets[uid] = new LinkedList();

                buckets[uid].add(ir);
                bucketSize[uid]++;
            }

            numRated++;
        }

        // DEBUG
        int numHorts = 0;
        LinkedList horts = new LinkedList();

        // Skip users who haven't rated anything.
        if ( numRated == 0 )
            return horts;

        for (int i = 0; i <= maxUser; i++)
        {
            if (((bucketSize[i] > HORT_RATIO*numRated
                        && bucketSize[i] > MIN_MATCH)
                        || bucketSize[i] > MAX_MATCH)
                        && i != userID) // We don't hort ourselves
            {
                    horts.add(buckets[i]);
                    // DEBUG
                    numHorts++;
            }
        }

        // DEBUG
        //System.out.println("H: " + numHorts + " R: " + numRated);

        return horts;
    }


    /**
     * A discrete linear transformation from one user's ratings to another's.
     */
    private class Transform implements Comparable
    {
        /**
         * UserID of our target, the user whose ratings predict another's.
         */
        int targetID;

        /**
         * Parameters for the linear transformation associating the target
         * user's ratings with the current one's.
         */
        byte coeff;
        byte offset;
        short commonItems;
        float distance;

        /**
         * Constructor, create a transform to the target user.  The target's ratings
         * will be used to predict the source user's ratings after
         * translating them by multiplying by coeff and adding offset.
         * @param targetID
         * @param coeff
         * @param offset
         * @param distance
         * @param commonItems
         */
        public Transform(int targetID, int coeff, int offset, float distance,
                int commonItems)
        {
            this.targetID    = targetID;
            this.coeff       = (byte)coeff;
            this.offset      = (byte)offset;
            this.distance    = (float)distance;
            this.commonItems = (short)commonItems;
        }

        /**
         * Constructor, create a Transform by composing two other transforms
         * @param Transform 1
         * @param Transform 2
         */
        public Transform(Transform t1, Transform t2)
        {
            targetID = t2.targetID;
            coeff    = (byte)(t1.coeff * t2.coeff);
            offset   = (byte)(t1.coeff * t2.offset + t1.offset);
            distance = t1.distance + t2.distance;
            // We don't care about # of common items when composing transforms...
        }

        /**
         * Translate a rating from the target user to one for the current user
         * by applyings the linear transformation associated with this arc.
         * @param original rating
         * @return translated rating
         */
        public float translate(float rating)
        {
            float retVal = coeff*rating + offset;

            if (retVal < MIN_RATING)
                retVal = MIN_RATING;
            else if (retVal > MAX_RATING)
                retVal = MAX_RATING;
            return retVal;
        }

        /**
         * Return the translated ItemRating.
         * @param original ItemRating
         * @return translated ItemRating
         */
        public ItemRating translate(ItemRating rating)
        {
            return new ItemRating(rating.getUserID(), rating.getItemID(),
                    translate(rating.getRating()));
        }

        /**
         * Implemented to satisfy Comparable.  Allows Transform
         * to be sorted by distance.
         * @param Object
         * @return 0, -1 or 1
         */
        public int compareTo(Object o)
        {
            Transform t = (Transform) o;
            if (distance == t.distance)
                return 0;
            else if (distance < t.distance)
                return -1;
            else
                return 1;
        }
    }
}
