/*
 * ItemItemAlgorithm.java
 *
 * Version 1.1
 *
 * 23/08/2002
 *
 * Copyright (c) 2003 Oregon State University. All rights reserved.
 */
package org.recommender.algorithms.Experimental;
import org.recommender.algorithms.*;
import java.io.*;
import java.util.*;
import java.sql.*;

import org.recommender.server.*;

/**
 * Describes an implementation of the CFAlgorithm interface
 * to make CF predictions using the Item-Item adjusted cosine
 * algorithm described in Sarwar, et al, 2001.
 *
 * @author Daniel Lowd &lt;dlowd@cs.orst.edu&gt;
 * @author  <A HREF="mailto:mclaughm@cs.orst.edu">Matt McLaughlin</A>
 */
public class ItemItemAlgorithm implements CFAlgorithm
{
    private DataManager dataMgr;

    private long NumPredictions  = 0;
    private long Time  = 0;

    private int MAX_NEIGHBORHOOD_SIZE;
    private int MODEL_SIZE;
    private int MIN_COMMON_USERS;
    private int MAX_COMMON_USERS;
    private boolean DISCOUNT_ITEM_AVERAGE;

    private int maxItem;
    List[] neighborLists;
    float[] itemAverages;

    /**
     * Constructor, initialize algorithm parameters and DataManager.
     * @param dataMgr
     */
    public ItemItemAlgorithm(DataManager dataMgr)
    {
        this.dataMgr = dataMgr;

        // Load parameters
        String prefix = "CFServer.alg.item.";
        MAX_NEIGHBORHOOD_SIZE = SystemProperties.getInt(
                prefix + "maxNeighborhoodSize", 30);
        MODEL_SIZE = SystemProperties.getInt(prefix+"modelSize", -1);
        MIN_COMMON_USERS = SystemProperties.getInt(prefix+"minCommonUsers", 2);
        MAX_COMMON_USERS = SystemProperties.getInt(prefix+"maxCommonUsers", 1);
        DISCOUNT_ITEM_AVERAGE = SystemProperties.getBoolean(
                prefix + "discountItemAverage", true);

        setUpNeighborhoods();
        calcItemAverages();
    }

    /**
     * Precompute most similar neighbors for each item.  By doing this
     * ahead of time, we can compute predictions for each user very
     * quickly.
     */
    private void setUpNeighborhoods()
    {
        System.out.println("Setting up neighborhoods...");
        maxItem = dataMgr.getMaxItem();
        neighborLists = new List[maxItem + 1];
        for (int item = 0; item <= maxItem; item++)
        {
            List neighbors = new LinkedList();

            if (dataMgr.getNumRaters(item) == 0)
            {
                neighborLists[item] = neighbors;
                continue;
            }

            for (int neighbor = 0; neighbor <= maxItem; neighbor++)
            {
                SimilarityWeight w = getWeight(item, neighbor);
                if (w.getWeight() > 0)
                    neighbors.add(w);
            }

            // Sort neighbors by similarity weights, so we process the best
            // ones first.
            Collections.sort(neighbors, Collections.reverseOrder());
            if (MODEL_SIZE > 0 && neighbors.size() > MODEL_SIZE)
                neighbors = neighbors.subList(0, MODEL_SIZE);
            neighborLists[item] = new Vector(neighbors);
        }
    }

    /**
     * Calculate item averages.
     */
    private void calcItemAverages()
    {
        // Calculate item averages
        itemAverages = new float[maxItem+1];
        for (int i = 0; i < maxItem; i++)
        {
            float accum = 0;
            int denom = 0;
            Iterator itemRatings = dataMgr.getNeighborhood(i);
            while (itemRatings.hasNext())
            {
                ItemRating ir = (ItemRating)itemRatings.next();
                accum += ir.getRating();
                denom++;
            }

            if (denom > 0)
                itemAverages[i] = accum/denom;
            else
                itemAverages[i] = 0;
        }
    }

    /**
     * Return Iterator of neighborhood for a particular item.
     * @param itemID
     * @return Iterator of neighborhood
     */
    private Iterator getNeighborhood(int itemID)
    {
        return neighborLists[itemID].iterator();
    }

     /**
     * Yun Wang
     *
     * Return top n recommendations of specific type.
     * @param activeUser
     * @param n
     * @param type
     * @return itemPredList
     * @throws SQLException
     */
    public ItemPrediction[] getRecommendationsByType(int activeUser, int n, int type)
               throws SQLException, CFNotImplementedException
   {
        int maxItem = neighborLists.length;
        float[] numerators = new float[maxItem+1];
        float[] denominators = new float[maxItem+1];
        Arrays.fill(numerators, 0);
        Arrays.fill(denominators, 0);

     //   int[] typeArr = new int[maxItem+1];     // for storing type
    //    Arrays.fill(typeArr, 0);

       long start= System.currentTimeMillis();

    //   Iterator typedIt = dataMgr.getCachedTypedItemList(type);
    // System.out.println("Get the Item List");
       int size = dataMgr.getSizeOfTypedItemList(type);
    // System.out.println("The size of typed item list is " + size);

     /*   while (typedIt.hasNext())
        {
            int i = ((Integer) typedIt.next()).intValue();
            typeArr[i] = 1;
        }*/
        long stop= System.currentTimeMillis();
    //  System.out.println("Time to get typed item list = " + (stop - start));

        // Iterate through each item that we've rated
        ItemRating[] userItems = dataMgr.getUserRatingList(activeUser);
        for (int i = 0; i < userItems.length; i++)
        {
            // Add in the effect of each item we've rated on its
            // neighbors.  This approach is actually backwards --
            // the neighbor lists select the neighbors that predict
            // us, not that we predict, but it should work both
            // ways reasonably well.
            ItemRating ir = userItems[i];
            int itemID = ir.getItemID();

            // Avoid exceptions... ideally, this never happens.
            if (itemID > maxItem || neighborLists[itemID] == null)
            {
                SystemUtils.printDebug("Unknown itemID " + itemID
                        + " in ItemItem.getRecommendations()");
                continue;
            }

            float offset = ir.getRating() - itemAverages[itemID];
            Iterator itemNeighbors = getNeighborhood(itemID);
            while(itemNeighbors.hasNext())
            {
                SimilarityWeight sw = (SimilarityWeight)itemNeighbors.next();
                float weight = sw.getWeight();
                int neighborID = sw.getNeighborID();
                numerators[neighborID] += weight*offset;
                denominators[neighborID] += weight;
            }
        }

        int index = 0;
       // ItemPrediction[] predList = new ItemPrediction[maxItem+1];
       ItemPrediction[] predList = new ItemPrediction[size+1];
      /*  for (int i = 0; i <= maxItem; i++)
        {
            // Don't recommend items we've already rated
            if (dataMgr.getRatingCached(activeUser, i) !=
                    dataMgr.getErrRating())
            {
                continue;
            }

            if (typeArr[i] == 0) continue;

            if (denominators[i] != 0)
            {
                predList[index++] = new ItemPrediction(activeUser, i,
                        itemAverages[i] + numerators[i]/denominators[i]);
            }
        }*/

      /*  int i;
        while (typedIt.hasNext())
        {
           i = ((Integer) typedIt.next()).intValue();

           if (dataMgr.getRatingCached(activeUser, i) !=
                    dataMgr.getErrRating())
            {
                continue;
            }

            if (denominators[i] != 0)
            {
                predList[index++] = new ItemPrediction(activeUser, i,
                        itemAverages[i] + numerators[i]/denominators[i]);
            }
        } */

         Object[] typedIt = ((Vector) dataMgr.getCachedTypedItemList(type)).toArray();

          int v;
          for(int i = 0; i < typedIt.length; i++)
          {
              v = ((Integer) typedIt[i]).intValue();

              if (dataMgr.getRatingCached(activeUser, v) !=
                    dataMgr.getErrRating())
              {
                continue;
              }

              if (denominators[v] != 0)
             {
                predList[index++] = new ItemPrediction(activeUser, v,
                        itemAverages[v] + numerators[v]/denominators[v]);
             }
          }


        // Sort recommendations
        //Arrays.sort(predList, 0, index, new Comp());
        Arrays.sort(predList, 0, index, new CompPrediction(dataMgr));

        if(index > n) index = n;

        ItemPrediction[] returnList = new ItemPrediction[index];

        for(int i = 0; i<returnList.length; i++){
			if(predList[i].getPrediction() > dataMgr.getMaxRating()) predList[i] = new ItemPrediction(predList[i].getUserID(), predList[i].getItemID(), dataMgr.getMaxRating());
			if(predList[i].getPrediction() < dataMgr.getMinRating()) predList[i] = new ItemPrediction(predList[i].getUserID(), predList[i].getItemID(), dataMgr.getMinRating());


			returnList[i] = predList[i];
		}

        return returnList;


   }

    /**
     * Get top n item recommendations for the activeUser.
     */
    public ItemPrediction[] getRecommendations(int activeUser, int n)
            throws SQLException, CFNotImplementedException
    {
        int maxItem = neighborLists.length;
        float[] numerators = new float[maxItem+1];
        float[] denominators = new float[maxItem+1];
        Arrays.fill(numerators, 0);
        Arrays.fill(denominators, 0);

        // Iterate through each item that we've rated
        ItemRating[] userRatings = dataMgr.getUserRatingList(activeUser);
        for (int i = 0; i < userRatings.length; i++)
        {
            // Add in the effect of each item we've rated on its
            // neighbors.  This approach is actually backwards --
            // the neighbor lists select the neighbors that predict
            // us, not that we predict, but it should work both
            // ways reasonably well.
            ItemRating ir = userRatings[i];
            int itemID = ir.getItemID();

            // Avoid exceptions... ideally, this never happens.
            if (itemID > maxItem || neighborLists[itemID] == null)
            {
                SystemUtils.printDebug("Unknown itemID " + itemID
                        + " in ItemItem.getRecommendations()");
                continue;
            }

            float offset = ir.getRating() - itemAverages[itemID];
            Iterator itemNeighbors = getNeighborhood(itemID);
            while(itemNeighbors.hasNext())
            {
                SimilarityWeight sw = (SimilarityWeight)itemNeighbors.next();
                float weight = sw.getWeight();
                int neighborID = sw.getNeighborID();
                numerators[neighborID] += weight*offset;
                denominators[neighborID] += weight;
            }
        }

        int index = 0;
        ItemPrediction[] predList = new ItemPrediction[maxItem+1];
        for (int i = 0; i <= maxItem; i++)
        {
            // Don't recommend items we've already rated
            if (dataMgr.getRatingCached(activeUser, i) !=
                    dataMgr.getErrRating())
            {
                continue;
            }

            if (denominators[i] != 0)
            {
                predList[index++] = new ItemPrediction(activeUser, i,
                        itemAverages[i] + numerators[i]/denominators[i]);
            }
        }

        // Sort recommendations
        //Arrays.sort(predList, 0, index, new Comp());
        Arrays.sort(predList, 0, index, new CompPrediction(dataMgr));

        if(index > n) index = n;

        ItemPrediction[] returnList = new ItemPrediction[index];

        for(int i = 0; i<returnList.length; i++){
			if(predList[i].getPrediction() > dataMgr.getMaxRating()) predList[i] = new ItemPrediction(predList[i].getUserID(), predList[i].getItemID(), dataMgr.getMaxRating());
			if(predList[i].getPrediction() < dataMgr.getMinRating()) predList[i] = new ItemPrediction(predList[i].getUserID(), predList[i].getItemID(), dataMgr.getMinRating());


			returnList[i] = predList[i];
		}

        return returnList;
    }


    /*
     * Alternate implementation of getRecommendations.  Actually
     * calls predictRating for each item in the database that we
     * haven't rated.  Probably significantly slower, but it could
     * be more accurate.

    public ItemRating[] getRecommendations(int activeUser, int n)
            throws SQLException
    {
        int maxItem = neighborLists.length;
        ItemRating[] predList = new ItemRating[maxItem+1];

        int index = 0;
        for (int i = 0; i <= maxItem; i++)
        {
            if (dataMgr.getNumRaters(i) < 5)
                continue;

            if (dataMgr.getRatingCached(activeUser, i) !=
                    dataMgr.getErrRating())
            {
                continue;
            }

            // TODO: try supplementing with PopAlgorithm...
            float rating = predictRating(activeUser, i);
            if (rating == dataMgr.getErrRating())
                continue;

            predList[index++] = new ItemRating(0, i, rating);
        }

        // Sort recommendations
        Arrays.sort(predList, 0, index, new Comp());
        return predList;
    }
    */



    /**
     * predicts the rating for the given user and item
     * @param activeUser
     * @param item
     * @return predicted rating
     */
    public ItemPrediction predictRating (int activeUser, int item) throws CFNotImplementedException
    {
        float numerator = 0;
        float denominator = 0;

        long startTime = System.currentTimeMillis();

        // Calculate the prediction based on each neighbor
        int neighborsUsed = 0;
        Iterator iter = getNeighborhood(item);
        while (iter.hasNext())
        {
            SimilarityWeight sw = (SimilarityWeight) iter.next();
            float weight = sw.getWeight();
            float rating = dataMgr.getRatingCached(activeUser,
                    sw.getNeighborID());

            if (rating != dataMgr.getErrRating())
            {
                if (DISCOUNT_ITEM_AVERAGE)
                    numerator += weight *
                        (rating - itemAverages[sw.getNeighborID()]);
                else
                    numerator += weight * rating;

                denominator += weight;
                neighborsUsed++;

                if (neighborsUsed >= MAX_NEIGHBORHOOD_SIZE)
                    break;
            }
        }

        float retval;
        if (denominator != 0)
        {
            if (DISCOUNT_ITEM_AVERAGE)
                retval = itemAverages[item] + numerator/denominator;
            else
                retval = numerator/denominator;

            if (retval > dataMgr.getMaxRating())
                retval = dataMgr.getMaxRating();
            else if (retval < dataMgr.getMinRating())
                retval = dataMgr.getMinRating();
        }
        else
        	retval = dataMgr.getErrRating();

        long stopTime = System.currentTimeMillis();
        NumPredictions++;
        Time += (stopTime-startTime);

        return new ItemPrediction(activeUser,item,retval); // return predicted rating
    }

    /**
     * Currently does not need updateUser
     */
    public void updateUser(int userID) throws CFNotImplementedException
    {
        //throw new CFNotImplementedException("updateUser not implemented yet");
    }

    /**
     * Return average prediction time.
     * @return average prediction time
     */
	public long getAveragePredictionTime() throws CFNotImplementedException
    {
		return (Time/NumPredictions);
	}

	/**
	 * Calculate similarity weight between two items.
     * Uses the adjusted cosine similarity measure.
	 * @param activeitem
	 * @param neighbor
	 * @return SimilarityWeight
	 */
	public SimilarityWeight getWeight(int activeitem, int neighbor)
    {
        float accum = 0;
        float den1 = 0;
        float den2 = 0;

        int numUsers = 0;

        // Consider all the users who have rated this item
        Iterator itemRatings = dataMgr.getNeighborhood(activeitem);
        while (itemRatings.hasNext())
        {
            ItemRating ir = (ItemRating)itemRatings.next();
            float ourRating = ir.getRating();
            float neighborRating = dataMgr.getRatingCached(ir.getUserID(),
                    neighbor);

            // Only consider users who have rated the neighbor item as well
            if (neighborRating == dataMgr.getErrRating())
                continue;

            float userMean = dataMgr.getMeanRatingCached(ir.getUserID());
            float ourOffset = ourRating - userMean;
            float theirOffset = neighborRating - userMean;
            accum += ourOffset * theirOffset;
            den1 += (ourOffset*ourOffset);
            den2 += (theirOffset*theirOffset);
            numUsers++;
        }

        float weight;
        if (den1 > 0 && den2 > 0)
            weight = accum/((float) Math.sqrt(den1) * (float) Math.sqrt(den2));
        else
            weight = 0;

        // Additional modifications to the similarity weight.
        if (numUsers < MIN_COMMON_USERS)
            weight = 0;

        if (numUsers < MAX_COMMON_USERS)
            weight *= ((float)numUsers)/MAX_COMMON_USERS;

        return new SimilarityWeight(neighbor, weight);
    }

    /**
     * Return the name of this algorithm.
     * @return String
     */
    public String toString()
    {
        return "ItemItemAlgorithm.java\n";
    }
}
