/*
 * PersonalityAlgorithm.java
 *
 * Version 1.1
 *
 * 15/08/2002
 *
 * Copyright (c) 2003 Oregon State Universtity. All rights reserved.
 */

package org.recommender.algorithms.Experimental;
import org.recommender.algorithms.*;
import org.recommender.server.*;
import java.io.*;
import java.util.*;
import java.sql.*;

/**
 * Describes an implementation of the CFAlgorithm interface
 * to make CF predictions using the Personality Diagnosis
 * algorithm described in Pennock et al, 1999.
 *
 * @author  Daniel Lowd &lt;dlowd@cs.orst.edu&gt;
 */
public class PersonalityAlgorithm implements CFAlgorithm
{
    private DataManager dataMgr;

    private long NumPredictions  = 0;
    private long Time  = 0;

    private float SIGMA;

    private int MIN_RATING;
    private int MAX_RATING;
    private int ERR_RATING;

    // Modifications to algorithm...
    private int MAX_NEIGHBORHOOD_SIZE;

    // Speed hack -- don't recalculate similarities if we
    // don't have to.  Mostly useful for accuracy experiments.
    private boolean CACHE_LAST_NEIGHBORS;

    float[][] probTable;

    /**
     * Constructor, initialize algorithm parameters and DataManager.
     * @param dataMgr
     */
    public PersonalityAlgorithm(DataManager dataMgr)
    {
        this.dataMgr = dataMgr;

        MIN_RATING = (int)dataMgr.getMinRating();
        MAX_RATING = (int)dataMgr.getMaxRating();
        ERR_RATING = (int)dataMgr.getErrRating();

        String prefix = "CFServer.alg.pd.";
        SIGMA = SystemProperties.getFloat(prefix + "variance", (float) 2.5);
        CACHE_LAST_NEIGHBORS = SystemProperties.getBoolean(
                prefix + "cacheLastNeighbors", false);

        // DEBUG
        System.out.println("Variance: " + SIGMA);

        MAX_NEIGHBORHOOD_SIZE = SystemProperties.getInt(
                prefix + "maxNeighborhoodSize", -1);

        // Generate probability table
        probTable = new float[MAX_RATING + 1][MAX_RATING + 1];
        for (int i = MIN_RATING; i <= MAX_RATING; i++)
        {
            float denom = 0;
            for (int j = MIN_RATING; j <= MAX_RATING; j++)
            {
                int diff = Math.abs(i - j);
                probTable[j][i] = (float) Math.exp(-diff*diff/(2*SIGMA*SIGMA));
                denom += probTable[j][i];
            }

            // Normalize
            for (int j = 0; j <= MAX_RATING; j++)
            {
                probTable[j][i] /= denom;
            }
        }
    }

    /**
     * Not yet complete
     */
    public ItemPrediction[] getRecommendationsByType(int activeUser, int n, int type)
               throws SQLException, CFNotImplementedException
   {
   		throw new CFNotImplementedException("getRecommendationsByType not implemented yet");
   }

    /**
     * Not yet implemented.
     */
    public ItemPrediction[] getRecommendations(int activeUser, int n) throws SQLException, CFNotImplementedException {
        throw new CFNotImplementedException("getRecommendations not implemented yet");
    }


    /**
     * predicts the rating for the given user and item
     */
    public ItemPrediction predictRating(int activeUser, int item) throws CFNotImplementedException
    {
        float numerator = 0;
        float denominator = 0;

        long startTime = System.currentTimeMillis();

        float ratingTotals[] = new float[MAX_RATING+1];
        for (int i = MIN_RATING; i <= MAX_RATING; i++)
            ratingTotals[i] = 0;


        // Calculate similarity to each other user
        List neighborWeights;
        if (CACHE_LAST_NEIGHBORS)
        {
            neighborWeights = getSimilarityWeightsCached(activeUser);
        }
        else
        {
            Iterator neighborRatings = dataMgr.getNeighborhood(item);
            neighborWeights = getSimilarityWeights(activeUser, neighborRatings);

            if (MAX_NEIGHBORHOOD_SIZE != -1)
                Collections.sort(neighborWeights, Collections.reverseOrder());
        }


        // Calculate the prediction based on each neighbor
        int neighborsUsed = 0;
        Iterator iter = neighborWeights.iterator();
        while (iter.hasNext() && neighborsUsed != MAX_NEIGHBORHOOD_SIZE)
        {
            SimilarityWeight w = (SimilarityWeight)iter.next();
            float rating = dataMgr.getRatingCached(w.getNeighborID(), item);
            float weight = w.getWeight();

            if (rating == dataMgr.getErrRating()
                    // HACK
                    || rating < MIN_RATING || rating > MAX_RATING)
                continue;

            for (int i = MIN_RATING; i <= MAX_RATING; i++)
            {
                ratingTotals[i] += noiseProb(i, rating) * weight;
            }

            neighborsUsed++;
        }

        // Determine the most probable rating (not the expected value,
        // as in other algorithms).
        float denom = 0;
        for (int r = MIN_RATING; r <= MAX_RATING; r++)
            denom += ratingTotals[r];

        int bestR = ERR_RATING;
        float bestTotal = denom/(MAX_RATING - MIN_RATING + 1);
        for (int r = MIN_RATING; r <= MAX_RATING; r++)
        {
            if (ratingTotals[r] > bestTotal)
            {
                bestR = r;
                bestTotal = ratingTotals[r];
            }
        }

        long stopTime = System.currentTimeMillis();
        NumPredictions++;
        Time += (stopTime-startTime);

        return new ItemPrediction(activeUser,item,bestR);
    }

    /**
     * Return average prediction time.
     * @return average prediction time
     */
	public long getAveragePredictionTime()
    {
		return (Time/NumPredictions);
	}

    /**
     * Find the probability that an observed rating of r1 corresponds
     * to the "true" rating of r2, assuming that observed ratings
     * differ from "true" ratings by adding Gaussian noise.
     * @param r1 observed rating
     * @param r2 hypothesized "true" rating
     * @return probability
     */
    protected float noiseProb(float r1, float r2)
    {
        return probTable[(int)r1][(int)r2];
    }

	/**
	 * Calculate relative probability that the activeuser is of the
     * same personality type as the neighbor.  Note that the probabilities
     * returned by this method will not sum to one; they must be normalized.
     * This probability is returned as a SimilarityWeight, for convenience
     * in sorting.
	 * @param activeuser
	 * @param neighbor
	 * @return SimilarityWeight
	 */
	public SimilarityWeight getWeight(int activeuser, int neighbor)
    {
        float prob = 1;
        ItemRating[] userRatings = dataMgr.getUserRatingList(activeuser);
        for (int i = 0; i < userRatings.length; i++)
        {
            ItemRating ir = userRatings[i];
            float ourRating = ir.getRating();
            float theirRating = dataMgr.getRatingCached(neighbor,
                    ir.getItemID());

            if (theirRating != ERR_RATING)
                prob *= (MAX_RATING - MIN_RATING)
                        * noiseProb(ourRating, theirRating);
        }

        return new SimilarityWeight(neighbor, prob);
   	}

	/**
	 * Return a List of SimilarityWeights representing the relative
     * probabilities that the active user is the same personality type
     * as each neighbor.
	 * @param activeUser
	 * @param neighborRatings
	 * @return List of SimilarityWeights
	 */
    private List getSimilarityWeights(int activeUser, 
            Iterator neighborRatings)
    {
        LinkedList vals = new LinkedList();

        while (neighborRatings.hasNext())
        {
            ItemRating ir = (ItemRating)neighborRatings.next();
            vals.add(getWeight(activeUser, ir.getUserID()));
        }

        return vals;
	}

    /* Keep track of the last user we calculated weights for.
     * If the same user is requested twice in a row, then we can
     * save time with this simple caching.
     */
    int lastUser = -1;
    List similarityWeightList = null;

    private List getSimilarityWeightsCached(int activeUser)
    {
        if (activeUser == lastUser)
            return similarityWeightList;

        // Calculate similarity with each other user.
        similarityWeightList = new LinkedList();
        Iterator neighbors = dataMgr.getAllUsers();
        while(neighbors.hasNext())
        {
            Integer n = (Integer)neighbors.next();
            similarityWeightList.add(getWeight(activeUser, n.intValue()));
        }

        // Sort in reverse order, so the best similarities are used first.
        if (MAX_NEIGHBORHOOD_SIZE != -1)
            Collections.sort(similarityWeightList, Collections.reverseOrder());

        lastUser = activeUser;
        return similarityWeightList;
    }

    /**
     * Not yet implemented.
     */
    public void updateUser(int userID) throws CFNotImplementedException
    {
		throw new CFNotImplementedException("updateUser not implemented yet");
    }

    /**
     * Return the name of this algorithm.
     * @return String
     */
    public String toString()
    {
        return "PersonalityAlgorithm.java\n";
    }
}
