/*
 * PopAlgorithm.java
 *
 * Version 1.1
 *
 * 15/08/2002
 *
 * Copyright (c) 2003 Oregon State University. All rights reserved.
 */
package org.recommender.algorithms.Experimental;
import org.recommender.algorithms.*;
import org.recommender.server.*;
import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.sql.*;

/**
 * Describes an implementation of the CFAlgorithm interface to make
 * CF predictions using the non-personalized prediction algorithm
 * described by Herlocker, et al 1999.
 *
 * Summary of algorithm: Compares all users rating for the item with
 * their average ratings, and computes and average offset for this
 * item.  Returns the active user's mean rating plus this average
 * offset.
 *
 * This turns out to be much more effective than a simple average over
 * all user ratings for the item, since it takes into account
 * different rating scales for different users.
 *
 * @author Daniel Lowd &lt;dlowd@cs.orst.edu&gt;
 */
public class PopAlgorithm implements CFAlgorithm
{
    /**
     * DataManager of this algorithm
     */
    private DataManager dataMgr;

    /**
     * Variables to keep track of the average time for a call to predictRating
     */
    private long Time = 0;
    private long NumPredictions = 0;

	/**
	 * Constructor, simply assigns DataManager.
	 * @param mgr
	 */
    public PopAlgorithm(DataManager mgr)
    {
		dataMgr = mgr;
    }


     /**
     * Yun Wang
     *
     * Return top n recommendations of specific type.
     * @param activeUser
     * @param n
     * @param type
     * @return itemPredList
     * @throws SQLException
     */
    public ItemPrediction[] getRecommendationsByType(int activeUser, int n, int type) throws CFNotImplementedException, SQLException
   {
         throw new CFNotImplementedException("getRecommendationsByType not implemented yet");
         //return null;

   }



    /**
     * Returns top n recommendations.  Uses the same algorithm as
     * predictRating, but predicts all items the user has not rated, rather
     * than just one.  Sorts the items in reverse order of rating and returns
     * the list.
     *
	 * @param activeUser user to predict for
	 * @param n number of recommendations to acquire
	 * @return itemPredList
	 */
	public ItemPrediction[] getRecommendations(int activeUser, int n)
            throws SQLException, CFNotImplementedException
    {
		//ItemRating[] itemPredList;
		ItemPrediction[] itemPredList;

		int index = 0;
        int maxItem = dataMgr.getMaxItem() + 1;
		float[] num = new float[maxItem];
		float[] den = new float[maxItem];

		for (int u = 0; u < maxItem; u++) {
			num[u] = 0;
			den[u] = 0;
		}

		itemPredList = new ItemPrediction[maxItem];

        int maxUser = dataMgr.getMaxUser() + 1;
        for (int i = 0; i < maxUser; i++)
        {
            if (dataMgr.getNumRatings(i) == 0)
                continue;

            float userAverage = dataMgr.getMeanRatingCached(i);
            ItemRating[] userRatings = dataMgr.getUserRatingList(i);
            for (int j = 0; j < userRatings.length; j++)
            {
                ItemRating ir = userRatings[j];
                num[ir.getItemID()] += ir.getRating() - userAverage;
                den[ir.getItemID()]++;
            }
        }

		//itemPredList[0] =  new ItemRating(0, 0, 0);
		itemPredList[0] =  new ItemPrediction(0, 0, 0);

		for (int i = 1; i < num.length; i++) {
			if (num[i] > 0 && den[i] > 0 && dataMgr.getRatingCached(
                        activeUser, i) == dataMgr.getErrRating())
            {
                float activeUserAverage
                    = dataMgr.getMeanRatingCached(activeUser);
				//itemPredList[i] = new ItemRating(0, i, ItemPrediction
				itemPredList[i] = new ItemPrediction(0, i,
                        activeUserAverage + num[i]/den[i]);
            }
			else
				//itemPredList[i] = new ItemRating(0, i, 0);
				itemPredList[i] = new ItemPrediction(0, i, 0);
		}

		//Arrays.sort(itemPredList, new Comp());
		Arrays.sort(itemPredList, new CompPrediction(dataMgr));
		return itemPredList;
	}


    /**
     * Predicts the rating for the given user and item.
     * @param activeUser
     * @param item
     * @return predicted rating
     */
    public ItemPrediction predictRating (int activeUser, int item) throws CFNotImplementedException
    {
        long startTime = System.currentTimeMillis();

        // Return the user's mean plus the average offset for this movie
        float num = 0;
        int den = 0;

        Iterator itemRatings = dataMgr.getNeighborhood(item);
        while(itemRatings.hasNext())
        {
            ItemRating ir = (ItemRating)itemRatings.next();
            float rating = ir.getRating();
            float mean = dataMgr.getMeanRatingCached(ir.getUserID());

            if (!Float.isNaN(mean))
            {
                num += rating - mean;
                den++;
            }
        }

        float prediction;
        float activeUserMean = dataMgr.getMeanRatingCached(activeUser);
        if (den != 0 && !Float.isNaN(activeUserMean))
        {
            prediction = activeUserMean + num/den;
            if (prediction > dataMgr.getMaxRating())
                prediction = dataMgr.getMaxRating();
            else if (prediction < dataMgr.getMinRating())
                prediction = dataMgr.getMinRating();
        }
        else
            prediction = dataMgr.getErrRating();

        long stopTime = System.currentTimeMillis();
        Time += stopTime - startTime;
        NumPredictions++;

        return new ItemPrediction(activeUser,item,prediction);
    }

    /**
     * Not yet implemented.
     */
    public void updateUser(int userID) throws CFNotImplementedException
    {
        throw new CFNotImplementedException("updateUser not implemented yet");
    }


    /**
     * Return average prediction time.
     * @return long
     */
    public long getAveragePredictionTime() throws CFNotImplementedException
    {
        return (Time/NumPredictions);
    }

    /**
     * Return the name of this algorithm.
     * @return String
     */
    public String toString()
    {
        return "PopAlgorithm.java";
    }
}
