/*
 * SimpleDistributionAlgorithm.java
 *
 * Version 1.1
 *
 * 15/08/2002
 *
 * Copyright (c) 2003 Oregon State Universtity. All rights reserved.
 */
package org.recommender.algorithms.Experimental;

import org.recommender.algorithms.*;
import org.recommender.server.*;

import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.sql.*;
import java.lang.Math.*;

/**
 * Describes an implementation of the CFAlgorithm interface
 * to make CF predictions using the Simple Pearson Correlation
 * combined with distribtion vectors.
 *
 * @author  <A HREF="mailto:sameerk@cs.orst.edu">Sameer Kadam</A>
 * @author  <A HREF="mailto:iyoon@cs.orst.edu">Irwin Yoon</A>
 * @author  <A HREF="mailto:godde@cs.orst.edu">Olivier Godde</A>
 * @author  <A HREF="mailto:dlowd@cs.orst.edu">Daniel Lowd</A>
 * @author  <A HREF="mailto:mclaughm@cs.orst.edu">Matt McLaughlin</A>
 */
public class SimpleDistributionAlgorithm implements CFAlgorithm {

    /**
     * DataManager for this algorithm
     */
    private DataManager dataMgr;

    ///////////// Parameters /////////////////
    // DL Some explanations are ambiguous or incomplete.  Some may change...

    /**
     * Minimum number of neighbors required.  If we have fewer than
     * this number of neighbors, return ERR_RATING rather than a
     * prediction.
     */
    private int MIN_NEIGHBORHOOD_SIZE;

    /**
     * Maximum number of neighbors to use for generating a prediction.
     */
    private int MAX_NEIGHBORHOOD_SIZE;

    /**
     * Maximum ratio of possible neighbors to use.  i.e., if this is set
     * to 0.1, then we'll use no more than 10% of the users with which
     * we can compute a correlation.
     */
    private float NEIGHBORHOOD_SIZE_RATIO;

    /**
     * Minimum number of items in common for a correlation to be valid.
     */
    private int MIN_COMMON_ITEMS;

    /**
     * Ratio of common items required for a correlation to be valid.
     * Superceded by MIN_COMMON_ITEMS.
     */
    private float HORT_RATIO;

    /**
     * Maximum number of items to consider for weighting a correlation
     * based on the number of items.  Correlations benefit from being
     * based on a larger number of items up to this number.
     */
    private int MAX_COMMON_ITEMS;

    /**
     * Neighbor weight threshold.  Neighbors with weights below this value
     * will not be used.
     */
    private float MIN_CORRELATION;

    private long NumPredictions = 0;
    private long Time = 0;
    private boolean CACHE_LAST_NEIGHBORS;

    /**
     * Array Containing the selected distributions
     */
    private float[][] relativeDifDist;
    private float[] normDist;

    /**
     * Variables declaring how the algorithm will run
     */
    //int compPredNumber; // select which prediction comparator to use
    //int compOptionNumber; //select the option to use in the comparator
    int formula; // 0 = Standard Pearson style,  1 = Min's
    //boolean activeMeanRec; //for Reccomendations: true=Active User mean, false=Item Mean
    //boolean activeMeanPred;//for Predictions: true=Active User mean, false=Item Mean
    int predictionStyle;//0=Expected Value. 1=Value with highest Probability
    //float cWeight;//weight to be placed on ActiveMean



    /**
     * Constructor, pass in dataMgr and initialize algorithm parameters.
     */
    public SimpleDistributionAlgorithm(DataManager dataMgr) {
        this.dataMgr = dataMgr;

        formula = 0;
        predictionStyle = 0;

        // Get parameters
        String prefix = "CFServer.alg.distribution.";
        MIN_NEIGHBORHOOD_SIZE = SystemProperties.getInt(
                prefix + "minNeighborhoodSize", 1);
        MAX_NEIGHBORHOOD_SIZE = SystemProperties.getInt(
                prefix + "maxNeighborhoodSize", 50);
        NEIGHBORHOOD_SIZE_RATIO = SystemProperties.getFloat(
                prefix + "neighborhoodSizeRatio", 1);
        HORT_RATIO = SystemProperties.getFloat(
                prefix + "hortRatio", 1);
        MIN_COMMON_ITEMS = SystemProperties.getInt(
                prefix + "minCommonItems", 2);
        MAX_COMMON_ITEMS = SystemProperties.getInt(
                prefix + "maxCommonItems", 60);
        MIN_CORRELATION = SystemProperties.getFloat(
                prefix + "minCorrelation", 0);
        CACHE_LAST_NEIGHBORS = SystemProperties.getBoolean(
                prefix + "cacheLastNeighbors", false);

        formula = SystemProperties.getInt(
                prefix + "formula", 0);

        if (formula == 1)
            System.out.println("Using Min's formula");
        else
            System.out.println("Using Standard formula");

        // fill relativeDif and norm arrays with starting distrubitions
        normDist = new float[9];
        for (int i = 0; i < 9; i++) {
            normDist[i] = (1 / 9);
        }

        relativeDifDist = new float[9][9];

        relativeDifDist[0][0] = (float) 0.01; // difference is 0 at -4 possibility
        relativeDifDist[0][1] = (float) 0.01; // difference is 0 at -3 possibility
        relativeDifDist[0][2] = (float) 0.02; // difference is 0 at -2 possibility
        relativeDifDist[0][3] = (float) 0.06; // difference is 0 at -1 possibility
        relativeDifDist[0][4] = (float) 0.80; // difference is 0 at 0  possibility
        relativeDifDist[0][5] = (float) 0.06; // difference is 0 at 1  possibility
        relativeDifDist[0][6] = (float) 0.02; // difference is 0 at 2  possibility
        relativeDifDist[0][7] = (float) 0.01; // difference is 0 at 3  possibility
        relativeDifDist[0][8] = (float) 0.01; // difference is 0 at 4  possibility

        relativeDifDist[1][0] = (float) 0.01; // difference is 1 at -4 possibility
        relativeDifDist[1][1] = (float) 0.01; // difference is 1 at -3 possibility
        relativeDifDist[1][2] = (float) 0.01; // difference is 1 at -2 possibility
        relativeDifDist[1][3] = (float) 0.02; // difference is 1 at -1 possibility
        relativeDifDist[1][4] = (float) 0.06; // difference is 1 at 0  possibility
        relativeDifDist[1][5] = (float) 0.80; // difference is 1 at 1  possibility
        relativeDifDist[1][6] = (float) 0.06; // difference is 1 at 2  possibility
        relativeDifDist[1][7] = (float) 0.02; // difference is 1 at 3  possibility
        relativeDifDist[1][8] = (float) 0.01; // difference is 1 at 4  possibility

        relativeDifDist[2][0] = (float) 0.01; // difference is 2 at -4 possibility
        relativeDifDist[2][1] = (float) 0.01; // difference is 2 at -3 possibility
        relativeDifDist[2][2] = (float) 0.01; // difference is 2 at -2 possibility
        relativeDifDist[2][3] = (float) 0.01; // difference is 2 at -1 possibility
        relativeDifDist[2][4] = (float) 0.02; // difference is 2 at 0  possibility
        relativeDifDist[2][5] = (float) 0.06; // difference is 2 at 1  possibility
        relativeDifDist[2][6] = (float) 0.80; // difference is 2 at 2  possibility
        relativeDifDist[2][7] = (float) 0.06; // difference is 2 at 3  possibility
        relativeDifDist[2][8] = (float) 0.02; // difference is 2 at 4  possibility

        relativeDifDist[3][0] = (float) 0.01; // difference is 3 at -4 possibility
        relativeDifDist[3][1] = (float) 0.01; // difference is 3 at -3 possibility
        relativeDifDist[3][2] = (float) 0.01; // difference is 3 at -2 possibility
        relativeDifDist[3][3] = (float) 0.01; // difference is 3 at -1 possibility
        relativeDifDist[3][4] = (float) 0.02; // difference is 3 at 0  possibility
        relativeDifDist[3][5] = (float) 0.02; // difference is 3 at 1  possibility
        relativeDifDist[3][6] = (float) 0.06; // difference is 3 at 2  possibility
        relativeDifDist[3][7] = (float) 0.80; // difference is 3 at 3  possibility
        relativeDifDist[3][8] = (float) 0.06; // difference is 3 at 4  possibility

        relativeDifDist[4][0] = (float) 0.005;// difference is 4 at -4 possibility
        relativeDifDist[4][1] = (float) 0.005;// difference is 4 at -3 possibility
        relativeDifDist[4][2] = (float) 0.01; // difference is 4 at -2 possibility
        relativeDifDist[4][3] = (float) 0.01; // difference is 4 at -1 possibility
        relativeDifDist[4][4] = (float) 0.02; // difference is 4 at 0  possibility
        relativeDifDist[4][5] = (float) 0.03; // difference is 4 at 1  possibility
        relativeDifDist[4][6] = (float) 0.04; // difference is 4 at 2  possibility
        relativeDifDist[4][7] = (float) 0.08; // difference is 4 at 3  possibility
        relativeDifDist[4][8] = (float) 0.80; // difference is 4 at 4  possibility

        relativeDifDist[5][0] = (float) 0.01; // difference is -1 at -4 possibility
        relativeDifDist[5][1] = (float) 0.02; // difference is -1 at -3 possibility
        relativeDifDist[5][2] = (float) 0.06; // difference is -1 at -2 possibility
        relativeDifDist[5][3] = (float) 0.80; // difference is -1 at -1 possibility
        relativeDifDist[5][4] = (float) 0.06; // difference is -1 at 0  possibility
        relativeDifDist[5][5] = (float) 0.02; // difference is -1 at 1  possibility
        relativeDifDist[5][6] = (float) 0.01; // difference is -1 at 2  possibility
        relativeDifDist[5][7] = (float) 0.01; // difference is -1 at 3  possibility
        relativeDifDist[5][8] = (float) 0.01; // difference is -1 at 4  possibility

        relativeDifDist[6][0] = (float) 0.02; // difference is -2 at -4 possibility
        relativeDifDist[6][1] = (float) 0.06; // difference is -2 at -3 possibility
        relativeDifDist[6][2] = (float) 0.80; // difference is -2 at -2 possibility
        relativeDifDist[6][3] = (float) 0.06; // difference is -2 at -1 possibility
        relativeDifDist[6][4] = (float) 0.02; // difference is -2 at 0  possibility
        relativeDifDist[6][5] = (float) 0.01; // difference is -2 at 1  possibility
        relativeDifDist[6][6] = (float) 0.01; // difference is -2 at 2  possibility
        relativeDifDist[6][7] = (float) 0.01; // difference is -2 at 3  possibility
        relativeDifDist[6][8] = (float) 0.01; // difference is -2 at 4  possibility

        relativeDifDist[7][0] = (float) 0.06; // difference is -3 at -4 possibility
        relativeDifDist[7][1] = (float) 0.80; // difference is -3 at -3 possibility
        relativeDifDist[7][2] = (float) 0.06; // difference is -3 at -2 possibility
        relativeDifDist[7][3] = (float) 0.02; // difference is -3 at -1 possibility
        relativeDifDist[7][4] = (float) 0.02; // difference is -3 at 0  possibility
        relativeDifDist[7][5] = (float) 0.01; // difference is -3 at 1  possibility
        relativeDifDist[7][6] = (float) 0.01; // difference is -3 at 2  possibility
        relativeDifDist[7][7] = (float) 0.01; // difference is -3 at 3  possibility
        relativeDifDist[7][8] = (float) 0.01; // difference is -3 at 4  possibility

        relativeDifDist[8][0] = (float) 0.80; // difference is -4 at -4 possibility
        relativeDifDist[8][1] = (float) 0.08; // difference is -4 at -3 possibility
        relativeDifDist[8][2] = (float) 0.04; // difference is -4 at -2 possibility
        relativeDifDist[8][3] = (float) 0.03; // difference is -4 at -1 possibility
        relativeDifDist[8][4] = (float) 0.02; // difference is -4 at 0  possibility
        relativeDifDist[8][5] = (float) 0.01; // difference is -4 at 1  possibility
        relativeDifDist[8][6] = (float) 0.01; // difference is -4 at 2  possibility
        relativeDifDist[8][7] = (float) 0.005; // difference is -4 at 3  possibility
        relativeDifDist[8][8] = (float) 0.005; // difference is -4 at 4  possibility
    }

    /* Keep track of the last user whose neighbors were requested.
     * If the same user is requested twice in a row, we can save time
     * with this computation.
     */
    private static int lastActiveUser = -1;
    private static List lastNeighbors = null;

    private List getNeighborsCached(UserInfo activeUserInfo) {

        if (activeUserInfo.userID == lastActiveUser) {
            return lastNeighbors;
        }

        // Compute similarity with all available users
//        lastNeighbors = getSimilarityWeightsTopN(activeUser,  dataMgr.getAllUsers());
        lastNeighbors = getSimilarityWeightsTopN(activeUserInfo, dataMgr.getAllUsers());

        // Sort similarities in reverse order, so we use the best ones first.
        Collections.sort(lastNeighbors, Collections.reverseOrder());

        lastActiveUser = activeUserInfo.userID;
        return lastNeighbors;
    }

    /**
     * Return top n recommendations of specific type.
     * @param activeUser
     * @param n
     * @param type
     * @return itemPredList
     * @throws SQLException
     */
    public ItemPrediction[] getRecommendationsByType(int activeUser, int n, int type)
            throws SQLException, CFNotImplementedException {
        //ItemPrediction[] itemPredList;
        LinkedList topNCoeff = new LinkedList();
        int index = 0;
        float item_prediction = 0;
        int maxItem = dataMgr.getMaxItem();
        float[][] num = new float[maxItem + 1][9];
        float[] den = new float[maxItem + 1];
        ItemPrediction[] predictions = new ItemPrediction[maxItem + 1];

        int[] typeArr = new int[maxItem + 1];     // for storing type

        UserInfo activeUserInfo = dataMgr.getUserInfo(activeUser);

        for (int u = 0; u <= maxItem; u++) {
            for (int i = 0; i < 9; i++) num[u][i] = 0;
            den[u] = 0;
            typeArr[u] = 0;
        }


        Iterator typedIt = dataMgr.getCachedTypedItemLists(type);
        int size = dataMgr.getSizeOfTypedItemList(type);

        while (typedIt.hasNext()) {
            int i = ((Integer) typedIt.next()).intValue();
            typeArr[i] = 1;
        }
        // Get correlation of db_users based on activeUser
        long start = System.currentTimeMillis();
        List neighborCoeffs;

        //get the lock here, so you don't have to worry about waiting in between
        //your get weight calls and you insure your average matches what you weights
        // are calculated at.
        if (!activeUserInfo.getReadersLock(true)) return new ItemPrediction[0];


        //long start = System.currentTimeMillis();
        if (CACHE_LAST_NEIGHBORS) {
            neighborCoeffs = getNeighborsCached(activeUserInfo);
            activeUserInfo.releaseReadersLock();
        } else {
            /* get list of neighbors, calculate weights, and sort in
             * reverse order, so we use the best ones first.
             */
            //Iterator neighbors = getNeighborhood(item);
            //neighborCoeffs = getSimilarityWeights(activeUser, neighbors, item);


            neighborCoeffs = getSimilarityWeightsTopN(
                    activeUserInfo, dataMgr.getAllUsers());
            activeUserInfo.releaseReadersLock();

            Collections.sort(neighborCoeffs, Collections.reverseOrder());
        }


        int numNeighbors = (int) (neighborCoeffs.size()
                * NEIGHBORHOOD_SIZE_RATIO);

        if (numNeighbors < MIN_NEIGHBORHOOD_SIZE) {
            return new ItemPrediction[0];
        }

        //long stop = System.currentTimeMillis();
        //    System.out.println("Time to select best neighbors= " + (stop - start));

        Iterator iter = neighborCoeffs.iterator();
        //start = System.currentTimeMillis();

        ItemRating r = null;
        float rating = 0;
        float mean = 0;
        // While there are neighbors in topNCoeff
        while (iter.hasNext()) {
            //Get new neighbor
            SimilarityWeight w = (SimilarityWeight) iter.next();

            if (w.getNeighborID() == activeUser) {
                continue;
            }

            UserInfo neighborInfo = dataMgr.getUserInfo(w.getNeighborID());
            if (neighborInfo == null) continue;
            if (!neighborInfo.getReadersLock(false)) continue;


            // mean = dataMgr.getMeanRatingCached(w.getNeighborID());
            mean = neighborInfo.getUserMean();

            /*// Get neighbor's item list
            Iterator i_index = (Iterator) dataMgr.getItemList(w.getNeighborID());

            while (i_index.hasNext())   //While there are items in list
            {
                r = (ItemRating) i_index.next();
                rating = r.getRating();
                int indexT = 0;
                if (rating > 0) {
                    // Get neighbor's weight
                    float weight = w.getWeight();
                    indexT = transfer_to_index(rating - mean);
                    try {
                        if (formula == 1) {
                            // For Min's style
                            for (int i = 0; i < 9; i++) num[r.getItemID()][i] = num[r.getItemID()][i] + (((relativeDifDist[indexT][i]) * weight) + ((1 - weight) / 9));
                            den[r.getItemID()]++;
                        } else {
                            //Standard Pearson Style
                            for (int i = 0; i < 9; i++) {
                                num[r.getItemID()][i] = num[r.getItemID()][i] + ((relativeDifDist[indexT][i]) * weight);
                            }
                            den[r.getItemID()] += weight;
                        }
                    } catch (Exception ex) {
                    }
                }
            }
        }     */
            float weight = w.getWeight();
            int[] itemIDs = neighborInfo.getRatings().ids;
            float[] ratings = neighborInfo.getRatings().ratings;
            int indexT = 0;
            for (int i = 0; i < neighborInfo.getRatings().getLength(); i++) {
                int ind = itemIDs[i];
                rating=ratings[i];
                if (typeArr[ind] > 0) {
                    indexT = transfer_to_index(rating - mean);
                    for (int k = 0; k < 9; k++) num[ind][k] = num[ind][k] + (((relativeDifDist[indexT][k]) * weight) + ((1 - weight) / 9));
                    //den[ind] += weight;
                    den[ind] +=1;
                }
            }
            neighborInfo.releaseReadersLock();
        }
        //For standard way of measuring
        for (int i = 0; i < num.length; i++) {
            int test = new Float(dataMgr.getRatingCached(activeUser, i)).intValue();
            if (typeArr[i] > 0 && test == dataMgr.getErrRating()) {


                //for using ItemDist
                //float[] tActiveDist = getItemDist(i);

                float[] tActiveDist = new float[5];
                for (int z = 0; z < tActiveDist.length; z++) tActiveDist[z] = (float)1 / (float)tActiveDist.length;

                float itemMean = dataMgr.getMeanItemRatingCached(i);
                //for(int j = 0; j<5;j++) itemMean += (tActiveDist[j] * (j+1));

                for (int j = 0; j < 5; j++) tActiveDist[j] = (tActiveDist[j]) / (den[i] + 1);
                for (int j = 0; j < 9; j++) num[i][j] = num[i][j] / (den[i] + 1);


                predictions[i] = computeRating(tActiveDist, num[i], itemMean, i);
            } else
                predictions[i] = new ItemPrediction(0, i, dataMgr.getErrRating());
        }

        double stop = System.currentTimeMillis();
        //System.out.println("Time to compute predictions=" + (stop - start));
        start = System.currentTimeMillis();
        Arrays.sort(predictions, 0, num.length - 1, new CompPrediction(dataMgr));
        stop = System.currentTimeMillis();
        //System.out.println("Time to sort the array= " + (stop - start));

        int predIndex = num.length - 1;
        while (predIndex < n) {
            predictions[predIndex++] = new ItemPrediction(0, 0, dataMgr.getErrRating());
        }
        return (ItemPrediction[]) predictions;
    }


    /**
     * Return top n recommendations.
     * @param activeUser
     * @param n
     * @return itemPredList
     * @throws SQLException
     */
    public ItemPrediction[] getRecommendations(int activeUser, int n) throws SQLException, CFNotImplementedException {
        //ItemPrediction[] itemPredList;
        LinkedList topNCoeff = new LinkedList();
        int index = 0;
        float item_prediction = 0;
        int maxItem = dataMgr.getMaxItem();
        float[][] num = new float[maxItem + 1][9];
        float[] den = new float[maxItem + 1];
        ItemPrediction[] predictions = new ItemPrediction[maxItem + 1];


        UserInfo activeUserInfo = dataMgr.getUserInfo(activeUser);

        for (int u = 0; u <= maxItem; u++) {
            for (int i = 0; i < 9; i++) num[u][i] = 0;
            den[u] = 0;
        }

        // Get correlation of db_users based on activeUser
        long start = System.currentTimeMillis();
        List neighborCoeffs;

        //get the lock here, so you don't have to worry about waiting in between
        //your get weight calls and you insure your average matches what you weights
        // are calculated at.
        if (!activeUserInfo.getReadersLock(true)) return new ItemPrediction[0];


        //long start = System.currentTimeMillis();
        if (CACHE_LAST_NEIGHBORS) {
            neighborCoeffs = getNeighborsCached(activeUserInfo);
            activeUserInfo.releaseReadersLock();
        } else {
            /* get list of neighbors, calculate weights, and sort in
             * reverse order, so we use the best ones first.
             */
            //Iterator neighbors = getNeighborhood(item);
            //neighborCoeffs = getSimilarityWeights(activeUser, neighbors, item);


            neighborCoeffs = getSimilarityWeightsTopN(
                    activeUserInfo, dataMgr.getAllUsers());
            activeUserInfo.releaseReadersLock();

            Collections.sort(neighborCoeffs, Collections.reverseOrder());
        }


        int numNeighbors = (int) (neighborCoeffs.size()
                * NEIGHBORHOOD_SIZE_RATIO);

        if (numNeighbors < MIN_NEIGHBORHOOD_SIZE) {
            return new ItemPrediction[0];
        }

        //long stop = System.currentTimeMillis();
        //    System.out.println("Time to select best neighbors= " + (stop - start));

        Iterator iter = neighborCoeffs.iterator();
        //start = System.currentTimeMillis();

        ItemRating r = null;
        float rating = 0;
        float mean = 0;
        // While there are neighbors in topNCoeff
        while (iter.hasNext()) {
            //Get new neighbor
            SimilarityWeight w = (SimilarityWeight) iter.next();

            if (w.getNeighborID() == activeUser) {
                continue;
            }

            UserInfo neighborInfo = dataMgr.getUserInfo(w.getNeighborID());
            if (neighborInfo == null) continue;
            if (!neighborInfo.getReadersLock(false)) continue;


            // mean = dataMgr.getMeanRatingCached(w.getNeighborID());
            mean = neighborInfo.getUserMean();

            float weight = w.getWeight();
            int[] itemIDs = neighborInfo.getRatings().ids;
            float[] ratings = neighborInfo.getRatings().ratings;
            int indexT = 0;
            for (int i = 0; i < neighborInfo.getRatings().getLength(); i++) {

                int ind = itemIDs[i];
                rating=ratings[i];
                if (rating > 0) {
                indexT = transfer_to_index(rating - mean);
                for (int k = 0; k < 9; k++) num[ind][k] = num[ind][k] + (((relativeDifDist[indexT][k]) * weight) + ((1 - weight) / 9));
                //den[ind] += weight;
                den[ind] += 1;
                }
            }
            neighborInfo.releaseReadersLock();
        }
        //For standard way of measuring
        for (int i = 0; i < num.length; i++) {
            int test = new Float(dataMgr.getRatingCached(activeUser, i)).intValue();
            if (test == dataMgr.getErrRating()) {

                float[] tActiveDist = new float[5];
                for (int z = 0; z < tActiveDist.length; z++) tActiveDist[z] = (float)1 / (float)tActiveDist.length;
                float itemMean = dataMgr.getMeanItemRatingCached(i);


                for (int j = 0; j < 5; j++) tActiveDist[j] = (tActiveDist[j]) / (den[i] + 1);
                for (int j = 0; j < 9; j++) num[i][j] = num[i][j] / (den[i] + 1);
                //for (int j = 0; j < 9; j++) num[i][j] = num[i][j] / (den[i]);

                predictions[i] = computeRating(tActiveDist, num[i], itemMean, i);
            } else
                predictions[i] = new ItemPrediction(0, i, dataMgr.getErrRating());
        }

        double stop = System.currentTimeMillis();
        //System.out.println("Time to compute predictions=" + (stop - start));
        start = System.currentTimeMillis();
        Arrays.sort(predictions, 0, num.length - 1, new CompPrediction(dataMgr));
        stop = System.currentTimeMillis();
        //System.out.println("Time to sort the array= " + (stop - start));

        int predIndex = num.length - 1;
        while (predIndex < n) {
            predictions[predIndex++] = new ItemPrediction(0, 0, dataMgr.getErrRating());
        }
        return (ItemPrediction[]) predictions;
    }



    /**
     * Predicts the rating for the given user and item.
     * @param activeUser
     * @param item
     * @return predicted rating
     */
    public ItemPrediction predictRating(int activeUser, int item) throws CFNotImplementedException {
        float[] num = new float[9];
        float den = 0;
        float acumWeight = 0;
        //float[] activeUserDist = getActiveUserDist(activeUser);
        float retval = 0;
        long startTime = System.currentTimeMillis();
        ItemPrediction prediction;
        List neighborCoeffs;

        UserInfo activeUserInfo = dataMgr.getUserInfo(activeUser);
        //get the lock here, so you don't have to worry about waiting inbetween
        //your get weight calls and you insure your average matches what you weights
        // are calculated at.
        if (!activeUserInfo.getReadersLock(true)) return new ItemPrediction(activeUser, item, dataMgr.getErrRating());

        float activeUserMean = dataMgr.getMeanRatingCached(activeUser);
        if (CACHE_LAST_NEIGHBORS) {
            neighborCoeffs = getNeighborsCached(activeUserInfo);
            activeUserInfo.releaseReadersLock();
        } else {
            /* get list of neighbors, calculate weights, and sort in
             * reverse order, so we use the best ones first.
             */
            //Iterator neighbors = getNeighborhood(item);
            //neighborCoeffs = getSimilarityWeights(activeUser, neighbors, item);

            neighborCoeffs = getSimilarityWeightsTopN(
                    activeUserInfo, dataMgr.getAllUsers());
            activeUserInfo.releaseReadersLock();
            Collections.sort(neighborCoeffs, Collections.reverseOrder());
        }

        int numNeighbors = (int) (neighborCoeffs.size()
                * NEIGHBORHOOD_SIZE_RATIO);

        if (numNeighbors < MIN_NEIGHBORHOOD_SIZE) {
            return new ItemPrediction(activeUser, item, dataMgr.getErrRating());
        } else if (MAX_NEIGHBORHOOD_SIZE > 0 &&
                numNeighbors > MAX_NEIGHBORHOOD_SIZE) {
            numNeighbors = MAX_NEIGHBORHOOD_SIZE;

            // now, calculate the prediction based on each neighbor
        }
        Iterator iter = neighborCoeffs.iterator();
        int neighborsUsed = 0;

        while (iter.hasNext() && neighborsUsed < numNeighbors) {
            SimilarityWeight w = (SimilarityWeight) iter.next();
            int neighbor = w.getNeighborID();
            // HACK -- Avoid correlations with ourself. (DLowd)
            if (neighbor == activeUser) {
                continue;
            }

            float rating = dataMgr.getRatingCached(neighbor, item);
            float neighborMean = dataMgr.getMeanRatingCached(neighbor);
            float weight = w.getWeight();

            if (rating == dataMgr.getErrRating()) {
                continue;
            }
            if (rating > 0) {
                //System.out.println("Weight = "+weight);
                // Get neighbor's weight
                int index = transfer_to_index(rating - neighborMean);
                try {
                    for (int i = 0; i < 9; i++) {
                        //Min form
                        if (formula == 1)
                            num[i] = num[i] + ((relativeDifDist[index][i] * weight) + ((1 - weight) / 9));
                        //Standard form
                        else
                            num[i] = num[i] + (relativeDifDist[index][i] * weight);
                    }
                    den++;
                    acumWeight = acumWeight + Math.abs(weight);

                } catch (Exception ex) {
                }
            }
            // Select only the best neighbors.
            neighborsUsed++;
            if (neighborsUsed >= numNeighbors) break;
        }

        // start the final calc!
        float[] activeDist = new float[5];
        //Min's form
        if (formula == 0) {
            for (int i = 0; i < 9; i++) {
                num[i] = num[i] / (den + 1);
            }
            for (int i = 0; i < 5; i++) {
                activeDist[i] = (float) .2 / (den + 1);
            }
        }

        //Standard form
        else {
            for (int i = 0; i < 9; i++) {
                num[i] = num[i] / (acumWeight + 1);
            }
            for (int i = 0; i < 5; i++) {
                activeDist[i] = (float) .2 / (acumWeight + 1);
            }
        }

        if (den != 0 && neighborsUsed >= MIN_NEIGHBORHOOD_SIZE) {
            prediction = computeRating(activeDist, num, activeUserMean, 0);
            retval = prediction.getPrediction();
            if (retval > dataMgr.getMaxRating())
                retval = dataMgr.getMaxRating();
            else if (retval < dataMgr.getMinRating())
                retval = dataMgr.getMinRating();
        } else
            retval = dataMgr.getErrRating();


        long stopTime = System.currentTimeMillis();

        /* SystemUtils.printDebug ("TIMER - TOTAL TIME TO PREDICT RATING: "
                                 + (stopTime-startTime)
                                 + " milliseconds.");*/
        //float[] pd = prediction.getDistribution();

        NumPredictions++;
        Time += (stopTime - startTime);

        System.out.println("Retval:"+retval);
        return new ItemPrediction(activeUser, item, retval); // return predicted rating
    }

    /**
     * Return average prediction time.
     * @return average prediction time
     */
    public long getAveragePredictionTime() throws CFNotImplementedException {
        if (NumPredictions > 0)
            return (Time / NumPredictions);
        else
            return (long) 0;
    }

    /**
     * Calculate CorrelationCoeff for an item between the activeuser and one neighbor.
     * This method is the bottleneck of this algorithm - do not add anything unnecessary -
     * Remove all logging or debugging statements after you are done. Even a harmless
     * if statement can slow things down.
     *
     * @param neighbor
     * @param itemId
     * @return SimilarityWeight
     */
    public SimilarityWeight getWeight(RatingArrays activeUserRatings,
                                      int neighbor, int itemId) {

        //****Declarations*****************************************
        float weight = 0;

        float combinedSum = 0;
        float activeSum = 0;
        float activeSqrSum = 0;
        float neighborSum = 0;
        float neighborSqrSum = 0;
        int numCommonItems = 0;

        //     RatingArrays activeUserRatings;

        // Get active user and neighbor average rating and standard deviation
        //if(!activeUserInfo.getReadersLock(true)) return new SimilarityWeight(neighbor,0);

        /*
        synchronized (activeUserInfo) {
            // Synchronize to ensure that ratings are consistent
            // This synchronized block does decrease the performance of the algorithm
            // by a noticable amount 160 topns/second -> 150 topns/second on a
            // 2.2ghz P4.
            activeUserRatings = activeUserInfo.getRatings();
        }
        */

        UserInfo neighborUserInfo = dataMgr.getUserInfo(neighbor);
        RatingArrays neighborRatings;

        if (neighborUserInfo == null) {
            return new SimilarityWeight(neighbor, 0);
        }

        if (!neighborUserInfo.getReadersLock(false)) return new SimilarityWeight(neighbor, 0);

        //synchronized (neighborUserInfo) {
        neighborRatings = neighborUserInfo.getRatings();
        //}

        int activeUserNumRatings = activeUserRatings.getLength();
        int nborNumRatings = neighborRatings.getLength();

        int p = 0;
        int q = 0;

        /* Iterate through the ratings of the active and neighbor users,
         * considering each item that has been rated by both.
         * Because the ratings are ordered by item ID, we can do this search
         * in O(n) rather than O(n^2).
         */
        while (p < activeUserNumRatings && q < nborNumRatings) {
            int activeUserItemID = activeUserRatings.ids[p];
            int nborItemID = neighborRatings.ids[q];

            // Don't use ratings for the actual item we're predicting!
            if (activeUserItemID == itemId) {
                p++;
            } else if (activeUserItemID < nborItemID) {
                p++;
            } else if (activeUserItemID > nborItemID) {
                q++;
            } else { // equal

                float activeRating = activeUserRatings.ratings[p];
                float neighborRating = neighborRatings.ratings[q];

                //Sum components
                activeSum += activeRating;
                neighborSum += neighborRating;
                activeSqrSum += (activeRating * activeRating);
                neighborSqrSum += (neighborRating * neighborRating);
                combinedSum += (activeRating * neighborRating);

                p++;
                q++;
                numCommonItems++;
            }
        }
        //activeUserInfo.releaseReadersLock();
        neighborUserInfo.releaseReadersLock();

        // 0 or 1 item is not enough to compute a correlation.
        if (numCommonItems < 2) {
            return new SimilarityWeight(neighbor, 0);
        }

        float activeUserMean = activeSum / numCommonItems;
        float neighborMean = neighborSum / numCommonItems;

        //Matt McLaughlin
        //Compute weight (Pearson correlation coefficient) as found in
        //"Applied Linear Regression, Second Edition", Sanford Weisberg, p.9
        weight = (combinedSum - numCommonItems * activeUserMean * neighborMean)
                / (float) Math.sqrt((activeSqrSum - numCommonItems * activeUserMean * activeUserMean)
                * (neighborSqrSum - numCommonItems * neighborMean * neighborMean));

        // Theoretically, the weight should never be NaN, nor more than 1;
        // however, floating point error sometimes leads to strange situations.
        if (Float.isNaN(weight)) {
            weight = 0;
        } else if (weight > 1) {
            weight = 1;
        }

        // Apply correlation threshold
        if (weight < MIN_CORRELATION) {
            weight = 0;
        }

        // If too few ratings then the correlation is irrelevant => set to zero
        int activeNumRatings = activeUserRatings.getLength();
        if (numCommonItems < MIN_COMMON_ITEMS
                && numCommonItems < HORT_RATIO * activeNumRatings) {
            weight = 0;
        }

        // Penalize correlations with fewer ratings if below the cutoff
        else if (numCommonItems < MAX_COMMON_ITEMS) {
            weight *= ((float) numCommonItems) / MAX_COMMON_ITEMS;
        }

        return (new SimilarityWeight(neighbor, weight));
    }


    /**
     * Return a List of SimilarityWeights for an item between the activeUser and each of
     * his/her neighbors.
     * @param activeUser
     * @param neighbors
     * @param itemId
     * @return List of SimilarityWeights
     */
    private List getSimilarityWeights(UserInfo activeUser, Iterator neighbors, int itemId) {

        LinkedList vals = new LinkedList();

        while (neighbors.hasNext()) {
            ItemRating neighborInitial = (ItemRating) neighbors.next();
            vals.add(getWeight(activeUser.getRatings(), neighborInitial.getUserID(), itemId));
        }

        return vals;
    }


    /**
     * Method to find neighborhood of users for input user.
     * @param itemID
     * @return Iterator of neighborhood
     */
    private Iterator getNeighborhood(int itemID) {
        // Neighborhood includes all other users who have rated this item
        Iterator iter = (Iterator) dataMgr.getNeighborhood(itemID);

        return iter;
    }

    /**
     * Return a List of SimilarityWeight between activeUser and neighbors.
     * @param activeUserInfo
     * @param neighbors
     * @return List of SimilarityWeight
     */
    private List getSimilarityWeightsTopN(UserInfo activeUserInfo, Iterator neighbors) {
        LinkedList vals = new LinkedList();
        Integer n;
//UserInfo activeUserInfo = dataMgr.getUserInfo(activeUser);
        RatingArrays activeUserRatings = activeUserInfo.getRatings();


        while (neighbors.hasNext()) {
            n = (Integer) neighbors.next();
            if (n.intValue() == activeUserInfo.userID) continue;
// SimilarityWeight w = getWeight(activeUserInfo, n.intValue(), -1);
            SimilarityWeight w = getWeight(activeUserRatings, n.intValue(), -1);
            if (w.getWeight() > 0) {
                vals.add(w);
            }
        }
        return vals;
    }

    /**
     * Haven't been implemented.
     */
    public void updateUser(int userID) throws CFNotImplementedException {
        throw new CFNotImplementedException("update User not implemented yet");
    }

    /**
     * Return the name of this algorithm.
     * @return String
     */
    public String toString() {
        return "SimpleDistributionAlgorithm.java\n";
    }


    /**
     * Return the return the sum/Mapping of the ActiveDistribution and the DifDistribution
     * @param ActiveUserDist
     * @param DifDist
     * @return Sum/Mapping of Distributions
     */
    public float[] mapDist(float[] ActiveUserDist, float[] DifDist, float ActiveAverage) {
        if (ActiveUserDist.length != 5 || DifDist.length != 9) return null;

        int active_average_i;
        active_average_i = (int) Math.round(ActiveAverage);

        for (int i = 0; i <= 8; i++) {
            if ((i - 4 + active_average_i) > 0 && (i - 4 + active_average_i) < 5) {
                ActiveUserDist[i - 4 + active_average_i] += DifDist[i];
            } else {
                if ((i - 4 + active_average_i) <= 0) {
                    ActiveUserDist[0] += DifDist[i];
                } else {
                    ActiveUserDist[4] += DifDist[i];
                }
            }
        }
        return ActiveUserDist;
    }

    /**
     * Return the return the mapped index of the relativeDifDist
     * @param difference
     * @return Mapped Index
     */

    public int transfer_to_index(float difference) {
        int tempi;
        tempi = (int) Math.round(difference);
        if (tempi >= 0) {
            if (tempi > 4) {
                return (int) 4;
            } else {
                return (int) tempi;
            }
        } else {
            if (tempi == -1) {
                return (int) 5;
            }
            if (tempi == -2) {
                return (int) 6;
            }
            if (tempi == -3) {
                return (int) 7;
            }
            if (tempi <= -4) {
                return (int) 8;
            }
        }
        return tempi;
    }


    /**
     * Return the DistributionArray of the active user
     * @param activeuser
     * @return ActiveUserDistArray
     */
    public float[] getActiveUserDist(int activeuser) {

        float[] activeUserDist = new float[5];
        //initialize array
        for (int i = 0; i < 5; i++) activeUserDist[i] = 0;
        Iterator iter = (Iterator) dataMgr.getItemList(activeuser);
        int count = 0;

        while (iter.hasNext()) {
            ItemRating activeUserRating = (ItemRating) iter.next();
            float rating;
            rating = activeUserRating.getRating();
            if (rating < 1 || rating > 5) continue;
            activeUserDist[(int) rating - 1] += 1;
            count++;
        }

        if (count <= 5) {
            for (int i = 0; i < 5; i++) activeUserDist[i] = (float) 0.2;
            return activeUserDist;
        }
        for (int i = 0; i < 5; i++) {
            activeUserDist[i] = activeUserDist[i] / count;
        }
        return activeUserDist;
    }

    /**
     * Return the DistributionArray of the active user
     * @param item
     * @return ActiveUserDistArray
     */
    public float[] getItemDist(int item) {

        float[] itemDist = new float[5];
        int count = 0;
        //initialize array
        for (int i = 0; i < 5; i++) itemDist[i] = 0;
        ItemRating[] userArray = dataMgr.getItemRatingList(item);
        for (int i = 0; i < userArray.length; i++) {
            float rating = dataMgr.getRatingCached(userArray[i].getUserID(), item);
            if (rating < 1 || rating > 5) continue;
            itemDist[(int) rating - 1] += 1;
            count++;
        }

        if (count <= 5) {
            for (int i = 0; i < 5; i++) itemDist[i] = (float) 0.2;
            return itemDist;
        }
        for (int i = 0; i < 5; i++) {
            itemDist[i] = itemDist[i] / count;
        }
        return itemDist;
    }


    /**
     * Return the computed prediction
     * @param activeDist
     * @param difDist
     * @param activeMean
     * @return predicted rating
     */
    public ItemPrediction computeRating(float[] activeDist, float[] difDist, float activeMean, int itemNumber) {

        float predictedRating = 0;
        float[] predictedDist = mapDist(activeDist, difDist, activeMean);

        //Predicted Rating equals the rating with the highest probability
        /*if (predictionStyle == 1) {
            float highestProb = predictedDist[0];
            for (int i = 1; i < predictedDist.length; i++) {
                if (highestProb < predictedDist[i]) {
                    highestProb = predictedDist[i];
                    predictedRating = (float) i + 1;
                }
            }
        }

        //Predicted rating equals the expected value of the distribution

        else {*/
        //    for (int i = 0; i < 5; i++) {
          //      predictedRating += (predictedDist[i] * (i + 1));
                //System.out.print(""+predictedDist[i]+"\t");
           //}
        //System.out.println();
        //} */

        //For testing purposes only (just adding 4 + 5)
        predictedRating = predictedDist[3] + predictedDist[4];

        //Testing for 5 only
        //predictedRating = predictedDist[4];

        //Normal
        ItemPrediction retval = new ItemPrediction(0, itemNumber, predictedRating);
        return retval;
    }



    /**
     * Return the computed prediction
     * @param activeUserDist
     * @param difDist
     * @param activeUserMean
     * @return predicted rating
     *
     public ItemPredictionExtended computeRating (float[] activeDist, float[] difDist, float activeMean, int itemNumber, float combinedWeight){

     float[] predictedDist = mapDist(activeDist, difDist, activeMean);
     float predictedRating = 0;
     //Predicted Rating equals the rating with the highest probability
     if(predictionStyle == 1){
     float highestProb = predictedDist[0];
     for (int i = 1; i < predictedDist.length; i++){
     if(highestProb < predictedDist[i]){
     highestProb =  predictedDist[i];
     predictedRating = (float)i + 1;
     }
     }
     }

     //Predicted rating equals the expected value of the distribution
     else{
     for (int i = 0; i < 5; i++) predictedRating += (predictedDist[i] * (i+1));
     }

     //Normal
     ItemPredictionExtended retval = new ItemPredictionExtended(0, itemNumber, predictedRating, predictedDist, combinedWeight, difDist, activeMean);
     return retval;
     }*/
}
