/*
 * TransNeighborAlgorithm.java
 *
 * Version 1.1
 *
 * 15/08/2002
 *
 * Copyright (c) 2003 Oregon State Universtity. All rights reserved.
 */
package org.recommender.algorithms.Experimental;
import org.recommender.algorithms.*;
import org.recommender.server.*;
import java.io.*;
import java.util.*;
import java.sql.*;

/**
 * Implementation of a transitive nearest-neighbor CF algorithm.
 * Find neighbors using Pearson correlation.  If none of our neighbors
 * have rated a given item, try our neighbors' neighbors.
 *
 * @author  Daniel Lowd &lt;dlowd@cs.orst.edu&gt;
 *
 */
public class TransNeighborAlgorithm implements CFAlgorithm
{
    private DataManager dataMgr;
    private List[] neighborLists;

    // Parameters from SimplePearsonAlgorithm

    // Minimum number of neighbors required.  If we have fewer than
    // this number of neighbors, return ERR_RATING rather than a
    // prediction.
    private int MIN_NEIGHBORHOOD_SIZE;

    // Maximum number of neighbors to use for generating a prediction.
    private int MAX_NEIGHBORHOOD_SIZE;

    // Maximum ratio of possible neighbors to use.  i.e., if this is set
    // to 0.1, then we'll use no more than 10% of the users with which
    // we can compute a correlation.
    private float NEIGHBORHOOD_SIZE_RATIO;

    // Minimum number of items in common for a correlation to be valid.
    private int MIN_COMMON_ITEMS;

    // Maximum number of items to consider for weighting a correlation
    // based on the number of items.  Correlations benefit from being
    // based on a larger number of items up to this number.
    private int MAX_COMMON_ITEMS;

    // Neighbor weight threshold.  Neighbors with weights below this value
    // will not be used.
    private float MIN_CORRELATION;


    private int maxUser;

    // Parameters from HortingAlgorithm
    private int MAX_PATH_LENGTH;

    // New parameters
    private int PRECOMPUTED_NEIGHBORS;

    /* Member vars for bucket-sort to keep track of which ratings are from
     * whom.  Since only the getHorts() method uses them, they would be static
     * variables there if this were C++.
     */
    private LinkedList[] buckets;
    private int[] bucketSize;
    private boolean initialized;

    private boolean[] visited;

    // Variables to keep track of the average time for a call to predictRating
    private long Time = 0;
    private long NumPredictions = 0;

    /**
     * Constructor, initialize algorithm parameters.
     * @param dataMgr
     */
    public TransNeighborAlgorithm(DataManager dataMgr)
    {
        this.dataMgr = dataMgr;
        maxUser = dataMgr.getMaxUser();

        // Get parameters
        String prefix = "CFServer.alg.trans.";
        MIN_NEIGHBORHOOD_SIZE = SystemProperties.getInt(
                prefix + "minNeighborhoodSize", 5);
        MAX_NEIGHBORHOOD_SIZE = SystemProperties.getInt(
                prefix + "maxNeighborhoodSize", 30);
        NEIGHBORHOOD_SIZE_RATIO = SystemProperties.getFloat(
                prefix + "neighborhoodSizeRatio", 0.1);
        MIN_COMMON_ITEMS  = SystemProperties.getInt(
                prefix + "minCommonItems", 1);
        MAX_COMMON_ITEMS  = SystemProperties.getInt(
                prefix + "maxCommonItems", 60);
        MIN_CORRELATION   = SystemProperties.getFloat(
                prefix + "minCorrelation", 0);

        PRECOMPUTED_NEIGHBORS = SystemProperties.getInt(
                prefix + "precomputedNeighbors", 20);
        MAX_PATH_LENGTH   = SystemProperties.getInt(
                prefix + "maxPathLength", 2);

        initArrays();
        //computeNeighborhoods();
    }

    private void initArrays()
    {
        // Initialize bucket-sort member vars.
        buckets = new LinkedList[maxUser+1];
        bucketSize = new int[maxUser+1];

        // Initialize array for keeping track of which neighbors
        // we've visited
        visited = new boolean[maxUser+1];

        // Initialize array of neighbors for each user
        neighborLists = new List[maxUser+1];
        for (int i = 0; i < maxUser; i++)
            neighborLists[i] = new LinkedList();
    }

    private void computeNeighborhoods()
    {
        Iterator iter = dataMgr.getAllUsers();
        while (iter.hasNext())
        {
            int userID = ((Integer)iter.next()).intValue();
            linkToPredictors(userID);
        }
    }
    /*
    {
        // Array to keep track of the correlation between this user and
        // each other user.
        SimilarityWeight[] weights = new SimilarityWeight[maxUser+1];

        // Iterate through all users
        Iterator userIter = dataMgr.getAllUsers();
        while (userIter.hasNext())
        {
            int activeUser = ((Integer)userIter.next()).intValue();

            // NOTE: if dataMgr.getAllUsers().hasNext() == false,
            // this code will fail!
            Iterator neighborIter = dataMgr.getAllUsers();
            int i = 0;

            // Calculate our similarity with each possible neighbor
            while (neighborIter.hasNext())
            {
                int neighbor = ((Integer)neighborIter.next()).intValue();
                weights[i++] = getWeight(activeUser, neighbor, 4.0);
            }

            Arrays.sort(weights, 0, i-1, Collections.reverseOrder());

            // Take the first items from the array (ie the largest weights)
            // and store them in the user's list
            neighborLists[activeUser] = new LinkedList();
            for (int j = 0; j < PRECOMPUTED_NEIGHBORS && j < i; j++)
            {
                neighborLists[activeUser].add(weights[j]);
            }
        }
    }
    */
    /**
     * Not yet complete
     */
    public ItemPrediction[] getRecommendationsByType(int activeUser, int n, int type)
               throws SQLException, CFNotImplementedException
   {
   		throw new CFNotImplementedException("getRecommendationsByType not implemented yet");
        // return null;

   }

	/**
	 * Returns top n recommendations.
	 * @param activeUser
	 * @param number of recommendation
	 * @return ItemPredList
	 */
	public ItemPrediction[] getRecommendations(int activeUser, int n)
            throws SQLException, CFNotImplementedException
    {
        int maxItem = dataMgr.getMaxItem();
		float[] num = new float[maxItem+1];
		float[] den = new float[maxItem+1];

		for (int u = 0; u <= maxItem; u++) {
			num[u] = 0;
			den[u] = 0;
		}

		Iterator iter = neighborLists[activeUser].iterator();
		while (iter.hasNext())
        {
            // Get new neighbor
			SimilarityWeight w = (SimilarityWeight)iter.next();
            int neighbor = w.getNeighborID();
            float weight = w.getWeight();
			float mean = dataMgr.getMeanRatingCached(neighbor);

			// Include all the neighbor's ratings
			Iterator i_index = (Iterator) dataMgr.getItemList(neighbor);
			while (i_index.hasNext())
            {
				ItemRating r = (ItemRating)i_index.next();
                num[r.getItemID()] += (r.getRating() - mean) * weight;
                den[r.getItemID()] += weight;
			}
		}

        // Predict all items that neighbors have rated
		float activeUserMean = dataMgr.getMeanRatingCached(activeUser);
		ItemPrediction[] itemPredList = new ItemPrediction[maxItem+1];
		for (int i = 0; i <= maxItem; i++)
        {
			if (num[i] > 0 && den[i] > 0)
				itemPredList[i] = new ItemPrediction(0, i, activeUserMean
                        + num[i]/den[i]);
			else
				itemPredList[i] = new ItemPrediction(0, i, 0);
		}

        // Select the highest rated items
		Arrays.sort(itemPredList, new Comp());
		return itemPredList;
	}


    /**
     * predicts the rating for the given user and item.
     * @param activeUser
     * @param item
     * @return predicted rating
     */
    public ItemPrediction predictRating (int activeUser, int item) throws CFNotImplementedException
    {
        long startTime = System.currentTimeMillis();
        float numerator = 0;
        float denominator = 0;

        // Calculate similarity with each neighbor in neighborhood
		Iterator neighbors = dataMgr.getNeighborhood(item);
        List neighborWeights = getSimilarityWeights(activeUser, neighbors);

        int numNeighbors = (int)(neighborWeights.size()
                                * NEIGHBORHOOD_SIZE_RATIO);
        if (numNeighbors < MIN_NEIGHBORHOOD_SIZE)
            return dataMgr.getErrRating();
        else if (MAX_NEIGHBORHOOD_SIZE > 0 &&
                numNeighbors > MAX_NEIGHBORHOOD_SIZE)
            numNeighbors = MAX_NEIGHBORHOOD_SIZE;

        // sort descending, so we can use just the best
        Collections.sort(neighborWeights, Collections.reverseOrder());

        // now, calculate the prediction based on each neighbor
        Iterator iter = neighborWeights.iterator();
        int neighborsUsed = 0;
        while (iter.hasNext())
        {
            SimilarityWeight w = (SimilarityWeight)iter.next();

            float mean = dataMgr.getMeanRatingCached (w.getNeighborID());
			float rating = dataMgr.getRatingCached(w.getNeighborID(), item);
            float weight = w.getWeight();

            if (weight < MIN_CORRELATION)
                break;

            numerator += weight * (rating - mean);
            denominator += weight;

            // Select only the best neighbors.
            neighborsUsed++;
            if (neighborsUsed >= numNeighbors)
                break;
        }

        float retval;
        if (denominator != 0)
        {
            float activeUserMean = dataMgr.getMeanRatingCached(activeUser);
            retval = activeUserMean + (numerator/denominator);

            if (retval > dataMgr.getMaxRating())
                retval = dataMgr.getMaxRating();
            else if (retval < dataMgr.getMinRating())
                retval = dataMgr.getMinRating();
        }
        else
        {
            // retval = transPredictRating(activeUser, item);
            retval = dataMgr.getErrRating();
        }

        long stopTime = System.currentTimeMillis();
        Time += stopTime - startTime;
        NumPredictions++;

        return new ItemPrediction(activeUser,item,retval);
    }

    /**
     * Not yet implemented.
     */
    public void updateUser(int userID) throws CFNotImplementedException
    {
    	throw new CFNotImplementedException("updateUser not yet implemented");
    }

    /**
     * Return average prediction time.
     * @return long
     */
    public long getAveragePredictionTime() throws CFNotImplementedException
    {
        return (Time/NumPredictions);
    }


    /*
    // HACK... not done yet...
    public float transPredictRating(int activeUser, int item)
    {
        float numerator = 0;
        float denominator = 0;

        // Calculate similarity with each neighbor in neighborhood
		List neighborWeights = getNeighborhood(activeUser, item);

        // sort descending, so we can use just the best
        if (MAX_NEIGHBORHOOD_SIZE > 0 && neighborWeights.size() > MAX_NEIGHBORHOOD_SIZE)
            Collections.sort(neighborWeights, Collections.reverseOrder());

        // now, calculate the prediction based on each neighbor
        Iterator iter = neighborWeights.iterator();
        int neighborsUsed = 0;
        while (iter.hasNext())
        {
            SimilarityWeight w = (SimilarityWeight) iter.next();
            float mean = dataMgr.getMeanRatingCached (w.getUserID());
			float rating = w.getNeighborRating();
            float weight = w.getWeight();

            numerator += weight * (rating - mean);
            denominator += weight;

            neighborsUsed++;
            if (neighborsUsed == MAX_NEIGHBORHOOD_SIZE)
                break;
        }

        float retval;
        if (denominator != 0)
        {
            float activeUserMean = dataMgr.getMeanRatingCached(activeUser);
            retval = activeUserMean + (numerator/denominator);

            if (retval > dataMgr.getMaxRating())
                retval = dataMgr.getMaxRating();
            else if (retval < dataMgr.getMinRating())
                retval = dataMgr.getMinRating();
        }
        else
        {
            retval = dataMgr.getErrRating();
        }

        return retval;
    }
    */

	/* Compute a correlation between two users
    public SimilarityWeight getWeight(int activeUser, int neighbor)
    {
        // Get active user and neigbhor average rating and standard deviation
        float activeUserMean = dataMgr.getMeanRatingCached(activeUser);
        float activeUserSD = dataMgr.getUserSD(activeUser);
        float neighborMean = dataMgr.getMeanRatingCached(neighbor);
        float neighborSD = dataMgr.getUserSD(neighbor);

        if (activeUserMean < 0 || activeUserSD < 0 ||
                neighborMean < 0 || neighborSD < 0)
        {
            // DEBUG
            System.out.println("Error: could not access stats for user "
                    + activeUser + " or " + neighbor + ".");
            return new SimilarityWeight(neighbor, 0, rating);
        }

        // If either user has only rated one item, the SD is zero
        // and the correlation is worthless.
        if (activeUserSD == 0 || neighborSD == 0)
        {
            return new SimilarityWeight(neighbor, 0, rating);
        }

        // Iterate through each item rated by the active user
        int numCommonItems = 0;
        float accum = 0;

        Iterator iter = dataMgr.getItemList(activeUser);
        Iterator iter2 = dataMgr.getItemList(neighbor);
        ItemRating activeUserRating = (ItemRating)iter.next();
        ItemRating neighborRating = (ItemRating)iter2.next();
        while (iter.hasNext() || iter2.hasNext())
        {
            if (activeUserRating.getItemID() < neighborRating.getItemID())
                if (iter.hasNext())
                    activeUserRating = (ItemRating)iter.next();
                else
                    break;
            else if (activeUserRating.getItemID() > neighborRating.getItemID())
                if (iter2.hasNext())
                    neighborRating = (ItemRating)iter2.next();
                else
                    break;
            else
            {
                // The item has been rated by both
                numCommonItems++;

                // get the rating values
                float activeUserRatingVal = activeUserRating.getRating();
                float neighborRatingVal = neighborRating.getRating();

                // add the contribution of this common item to the sum
                // (Pearson correlation numerator)
                accum += (activeUserRatingVal - activeUserMean)
                        * (neighborRatingVal - neighborMean);

                // Move on to the next items
                if (iter.hasNext())
                    activeUserRating = (ItemRating)iter.next();
                else
                    break;

                if (iter2.hasNext())
                    neighborRating = (ItemRating)iter2.next();
                else
                    break;
            }
        }

        // Calculate the weight (Pearson correlation)
        float weight = accum/(activeUserSD*neighborSD);

        // If too few ratings then the weight is set to zero
        if (numCommonItems < MIN_COMMON_ITEMS)
            weight = 0;

        // Penalize neighbors with fewer ratings if below the cutoff
        else if (numCommonItems < MAX_COMMON_ITEMS)
            weight *= ((float)numCommonItems)/MAX_COMMON_ITEMS;

        return (new SimilarityWeight(neighbor, weight));
    }
*/

	/**
	 * Compute a correlation between activeUser and one neighbor.
	 * @param activeUser
	 * @param neighbor
	 * @return SimilarityWeight
	 */
    public SimilarityWeight getWeight(int activeUser, int neighbor)
    {
        // Get active user and neigbhor average rating and standard deviation
        float activeUserMean = dataMgr.getMeanRatingCached(activeUser);
        float activeUserSD = dataMgr.getUserSD(activeUser);
        float neighborMean = dataMgr.getMeanRatingCached(neighbor);
        float neighborSD = dataMgr.getUserSD(neighbor);

        if (activeUserMean < 0 || activeUserSD < 0 ||
                neighborMean < 0 || neighborSD < 0)
        {
            // DEBUG
            System.out.println("Error: could not access stats for user "
                    + activeUser + " or " + neighbor + ".");
            return new SimilarityWeight(neighbor, 0);
        }

        // If either user has only rated one item, the SD is zero
        // and the correlation is worthless.
        if (activeUserSD == 0 || neighborSD == 0)
        {
            return new SimilarityWeight(neighbor, 0);
        }

        // Iterate through each item rated by the active user
        int numCommonItems = 0;
        float accum = 0;

        Iterator iter = dataMgr.getItemList(activeUser);
        while (iter.hasNext())
        {
            ItemRating activeUserRating = (ItemRating)iter.next();
            float neighborRatingVal = dataMgr.getRatingCached(neighbor,
                    activeUserRating.getItemID());

            if (neighborRatingVal != dataMgr.getErrRating())
            {
                // The item has been rated by both
                numCommonItems++;

                // get the rating values
                float activeUserRatingVal = activeUserRating.getRating();

                // add the contribution of this common item to the sum
                // (Pearson correlation numerator)
                accum += (activeUserRatingVal - activeUserMean)
                        * (neighborRatingVal - neighborMean);
            }
        }

        // Calculate the similarity weight
        float weight = accum/(activeUserSD*neighborSD);

        // If too few ratings then the correlation is irrelevant => set to zero
        if (numCommonItems < MIN_COMMON_ITEMS)
            weight = 0;

        // Penalize correlations with fewer ratings if below the cutoff
        else if (numCommonItems < MAX_COMMON_ITEMS)
            weight *= ((float)numCommonItems)/MAX_COMMON_ITEMS;

        return (new SimilarityWeight(neighbor, weight));
    }

	/**
	 * private method
	 */
    private List getNeighborhood(int userID, int itemID)
    {
        int distance = 0;
        LinkedList neighbors = new LinkedList();

        for (int i = 0; i <= maxUser; i++)
            visited[i] = false;

        // Start by iterating through all our neighbors
        Iterator iter = neighborLists[userID].iterator();
        while (neighbors.isEmpty() && distance <= MAX_PATH_LENGTH)
        {
            LinkedList neighborQueue = new LinkedList();
            while (iter.hasNext())
            {
                SimilarityWeight w = (SimilarityWeight)iter.next();
                int neighborID = w.getNeighborID();
                float rating = dataMgr.getRatingCached(neighborID, itemID);

                // If this neighbor has rated the item, remember it!
                if (rating != dataMgr.getErrRating())
                {
                    neighbors.add(new SimilarityWeight(neighborID,
                                w.getWeight()));
                }

                // Only add neighbors if there are no good neighbors at this
                // level.
                if (neighbors.isEmpty())
                {
                    Iterator i = neighborLists[neighborID].iterator();
                    while (i.hasNext())
                    {
                        // Get neighbor weight
                        SimilarityWeight nw = (SimilarityWeight)i.next();
                        if ( !visited[nw.getNeighborID()] )
                        {
                            neighborQueue.add(new SimilarityWeight(
                                    nw.getNeighborID(), w.getWeight()
                                    * nw.getWeight()));
                        }
                    }
                }
            }

            // Iterate through indirect neighbors
            iter = neighborQueue.iterator();
            distance++;
        }

        return neighbors;
    }

    /**
     * private method.
     */
    private List getSimilarityWeights(int activeUser, Iterator neighbors)
    {
        LinkedList vals = new LinkedList();

        while (neighbors.hasNext())
        {
            ItemRating ir = (ItemRating) neighbors.next();
            vals.add(getWeight(activeUser, ir.getUserID()));
        }

        return vals;
    }

    /**
     * Link to predictors.
     * @param userID
     */
    public void linkToPredictors(int userID)
    {
        // Create a new linked list to hold our new neighbors.
        List neighborList = new LinkedList();

        Iterator horts = getHorts(userID).iterator();
        while (horts.hasNext())
        {
            float weight = 0;
            float accum = 0;

            LinkedList hortRatings = (LinkedList)horts.next();
            int hortLen = hortRatings.size();
            Iterator hortIter = hortRatings.iterator();

            int neighbor = ((ItemRating)hortRatings.getFirst()).getUserID();
            float ourMean = dataMgr.getMeanRatingCached(userID);
            float theirMean = dataMgr.getMeanRatingCached(neighbor);

            // NOTE: we're guaranteed that we have rated every item listed
            // in hortRatings.
            while (hortIter.hasNext())
            {
                // Get the ratings, both for the other user and for us
                ItemRating rating = (ItemRating)hortIter.next();
                float ourRating = dataMgr.getRatingCached(userID,
                            rating.getItemID());
                float theirRating = rating.getRating();

                // We should have rated this...
                if (ourRating < 0)
                {
                    System.out.println("Missing rating!");
                    System.out.println("User: " + userID + "; Item: "
                            + rating.getItemID() + "; Neighbor: "
                            + rating.getUserID());
                    continue;
                }

                accum += (ourRating - ourMean)*(theirRating - theirMean);
            }

            float ourSD = dataMgr.getUserSD(userID);
            float theirSD = dataMgr.getUserSD(neighbor);
            weight = accum/(ourSD * theirSD);

            // Penalize correlations with fewer ratings if below the cutoff
            if (hortLen < MAX_COMMON_ITEMS)
                weight *= ((float)hortLen)/MAX_COMMON_ITEMS;

            neighborList.add(new SimilarityWeight(neighbor, weight));
        }

        // Sort the transforms so that we consider the best matches first.
        Collections.sort(neighborList);

        // Select only the top predictors
        if (neighborList.size() <= PRECOMPUTED_NEIGHBORS)
            neighborLists[userID] = neighborList;
        else
            neighborLists[userID] = neighborList.subList(0,
                    PRECOMPUTED_NEIGHBORS);
    }

    /**
     * private method
     */
    // Get a list of all the users we hort using bucket sort.
    private LinkedList getHorts(int userID)
    {
        Iterator rated = dataMgr.getItemList(userID);

        if (!rated.hasNext())
            return new LinkedList();

        // Initialize array for horting calcs.
        for (int i = 0; i <= maxUser; i++)
            bucketSize[i] = 0;

        // Loop through the neighborhoods of each item we've rated
        int numRated = 0;
        while (rated.hasNext())
        {
            ItemRating currItem = (ItemRating)rated.next();
            Iterator iter = dataMgr.getNeighborhood(currItem.getItemID());
            while (iter.hasNext())
            {
                ItemRating ir = (ItemRating)iter.next();

                /* Add the rating to the appropriate user's bucket,
                 * creating the bucket if necessary.
                 */
                int uid = ir.getUserID();

                if (bucketSize[uid] == 0)
                    buckets[uid] = new LinkedList();

                buckets[uid].add(ir);
                bucketSize[uid]++;
            }

            numRated++;
        }

        // DEBUG
        int numHorts = 0;
        LinkedList horts = new LinkedList();

        // Skip users who haven't rated anything.
        if ( numRated == 0 )
            return horts;

        for (int i = 0; i <= maxUser; i++)
        {
            // Add any user that's not us and who has rated at least one
            // item in common with us.
            if (bucketSize[i] > 0 && i != userID)
            {
                horts.add(buckets[i]);
            }
        }

        return horts;
    }

    /**
     * Returns a simple text identifier for this algorithm
     * @return String
     */
    public String toString()
    {
        return "TransNeighborAlgorithm.java";
    }
}
