/*
 * UserItem2Algorithm.java
 *
 * Version 1.1
 *
 * 23/08/2002
 *
 * Copyright (c) 2003 Oregon State University. All rights reserved.
 */
package org.recommender.algorithms.Experimental;
import org.recommender.algorithms.*;
import org.recommender.server.*;
import java.io.*;
import java.util.*;
import java.sql.*;

/**
 * Describes an implementation of the CFAlgorithm interface
 * to make CF predictions using a Pearson nearest neighbor algorithm,
 * but limited to examining similar items using the Item-Item adjusted
 * cosine algorithm described in Sarwar, et al, 2001.
 *
 * Item similarities are precomputed.  Then, to make an individual
 * prediction for a user and an item, we select the N items rated by
 * the user most similar to the item we're predicting.  This subset
 * of the user's ratings is used for correlations, which in turn
 * yield a list of neighbors for.
 *
 * @author Daniel Lowd &lt;dlowd@cs.orst.edu&gt;
 */

/*
 * It turned out to perform no better than Pearson nearest neighbor,
 * and may thus be considered a failed experiment.  Code was saved
 * for posterity, just in case it someday proved useful.
 *
 * THINGS NOT YET TRIED: play with neighborhood size more (maybe a
 * larger or smaller # of user neighbors would be useful), or
 * maybe this will work better with a more dense graph.
 *
 * Comments may be incomplete, but many of the workings are identical
 * to those of SimplePearsonAlgorithm and ItemItemAlgorithm.
 */

public class UserItem2Algorithm implements CFAlgorithm
{
    private DataManager dataMgr;

    private long NumPredictions  = 0;
    private long Time  = 0;

    private int MAX_COMMON_ITEMS;
    private int MAX_NEIGHBORHOOD_SIZE;
    private int MAX_RELEVANT_ITEMS;
    private int MODEL_SIZE;
    private int MIN_COMMON_USERS;
    private int MAX_COMMON_USERS;

    private int maxItem;
    List[] neighborLists;

    /**
     * Constructor, load parameters and set up neighborhoods.
     * @param datamgr the DataManager to get data from.
     */
    public UserItem2Algorithm(DataManager dataMgr)
    {
        this.dataMgr = dataMgr;

        // Load parameters
        String prefix = "CFServer.alg.useritem2.";
        MAX_COMMON_ITEMS = SystemProperties.getInt(
                prefix + "maxCommonItems", 60);
        MAX_RELEVANT_ITEMS   = SystemProperties.getInt(
                prefix + "maxRelevantItems", 30);
        MAX_NEIGHBORHOOD_SIZE   = SystemProperties.getInt(
                prefix + "maxNeighborhoodSize", 50);
        MODEL_SIZE              = SystemProperties.getInt(
                prefix + "modelSize", -1);
        MIN_COMMON_USERS        = SystemProperties.getInt(
                prefix + "minCommonUsers", 2);
        MAX_COMMON_USERS        = SystemProperties.getInt(
                prefix + "maxCommonUsers", 1);

        setUpNeighborhoods();
    }

    /**
     * Set up neighborhoods.
     */
    void setUpNeighborhoods()
    {
        // Set up neighborhoods...
        maxItem = dataMgr.getMaxItem();
        neighborLists = new List[maxItem + 1];
        for (int item = 0; item <= maxItem; item++)
        {
            List neighbors = new LinkedList();

            if (dataMgr.getNumRaters(item) == 0)
            {
                neighborLists[item] = neighbors;
                continue;
            }

            for (int neighbor = 0; neighbor <= maxItem; neighbor++)
            {
                SimilarityWeight coeff = getItemCoeff(item, neighbor);
                if (coeff.getWeight() > 0)
                    neighbors.add(coeff);
            }

            // Sort neighbors by similarity weights, so we process the best
            // ones first.
            Collections.sort(neighbors, Collections.reverseOrder());
            if (MODEL_SIZE > 0 && neighbors.size() > MODEL_SIZE)
                neighbors = neighbors.subList(0, MODEL_SIZE);
            neighborLists[item] = new Vector(neighbors);
        }
    }

    /**
     * Return an iterator of neighborhood for one item.
     * @param itemID
     * @return iterator of neighborhood
     */
    private Iterator getNeighborhood(int itemID)
    {
        return neighborLists[itemID].iterator();
    }

    /**
     * Not yet implemented.
     */
    public ItemRating[] getRecommendations(int activeUser, int n)
            throws SQLException, CFNotImplementedException
    {
    	throw new CFNotImplementedException("getRecommendationsByType not implemented yet");
        return null;
    }

    /**
     * Not yet complete
     */
    public ItemPrediction[] getRecommendationsByType(int activeUser, int n, int type)
               throws SQLException, CFNotImplementedException
   {
   		throw new CFNotImplementedException("getRecommendationsByType not implemented yet");
         return null;

   }

    /**
     * predicts the rating for the given user and item
     * @param activeUser
     * @param item
     * @return predicted rating
     */
    public ItemPrediction predictRating (int activeUser, int item) throws CFNotImplementedException
    {
        float numerator = 0;
        float denominator = 0;

        long startTime = System.currentTimeMillis();

        // Select top-N items that the active user has rated, by relevance
        // We will thus calculate our correlations using only the relevant
        // items.
        List userItems = new LinkedList();
        Iterator topItems = neighborLists[item].iterator();
        int itemsSelected = 0;

        while(topItems.hasNext() && itemsSelected < MAX_RELEVANT_ITEMS)
        {
            SimilarityWeight c = (SimilarityWeight)topItems.next();
            int itemID = c.getUserID();
            float rating = dataMgr.getRatingCached(activeUser, itemID);
            if (rating != dataMgr.getErrRating())
            {
                userItems.add(new ItemRating(activeUser, itemID, rating));
                itemsSelected++;
            }
        }

        // get list of neighbors
        Iterator neighbors = dataMgr.getNeighborhood(item);
        List neighborCoeffs = getSimilarityWeights(userItems, neighbors);

        // sort descending, so we can use just the best
        Collections.sort(neighborCoeffs, Collections.reverseOrder());

        // now, calculate the prediction based on each neighbor
        Iterator iter = neighborCoeffs.iterator();
        int neighborsUsed = 0;
        while (iter.hasNext())
        {
            SimilarityWeight w = (SimilarityWeight) iter.next();
            float weight = w.getWeight();
            int neighbor = w.getUserID();
            float rating = dataMgr.getRatingCached(neighbor, item);

            // This should never be the case...
            if (rating == dataMgr.getErrRating())
            {
                System.out.println("Error: rating missing in " +
                        "UserItem2Algorithm.predictRating()!");
            }

            numerator += weight *
                    (rating - dataMgr.getMeanRatingCached(neighbor));

            denominator += weight;
            neighborsUsed++;

            if (neighborsUsed >= MAX_NEIGHBORHOOD_SIZE)
                break;
        }

        float retval;
        if (denominator != 0)
        {
            retval = dataMgr.getMeanRatingCached(activeUser)
                    + numerator/denominator;

            if (retval > dataMgr.getMaxRating())
                retval = dataMgr.getMaxRating();
            else if (retval < dataMgr.getMinRating())
                retval = dataMgr.getMinRating();
        }
        else
        	retval = dataMgr.getErrRating();

        long stopTime = System.currentTimeMillis();
        NumPredictions++;
        Time += (stopTime-startTime);

        return new ItemPrediction(activeUser,item,retval); // return predicted rating
    }

    /**
     * Not yet implemented.
     */
    public void updateUser(int userID) throws CFNotImplementedException
    {
    	throw new CFNotImplementedException("updateUser not yet implemented");
        // TODO: how should we take into account new ratings for this?
    }

	public long getAveragePredictionTime() throws CFNotImplementedException
    {
		return (Time/NumPredictions);
	}

    /**
     * Determine the similarity between two items.  Uses the same
     * similarity metric as ItemItemAlgorithm, namely, the adjusted
     * cosine metric.
     */
	private SimilarityWeight getItemWeight(int activeitem, int neighbor)
    {
        float accum = 0;
        float den1 = 0;
        float den2 = 0;

        int numUsers = 0;

        // Consider all the users who have rated this item
        Iterator iter = dataMgr.getNeighborhood(activeitem);
        while (iter.hasNext())
        {
            ItemRating ir = (ItemRating)iter.next();
            float ourRating = ir.getRating();
            float neighborRating = dataMgr.getRatingCached(ir.getUserID(),
                    neighbor);

            // Only consider users who have rated the neighbor item as well
            if (neighborRating == dataMgr.getErrRating())
                continue;

            float userMean = dataMgr.getMeanRatingCached(ir.getUserID());
            float ourOffset = ourRating - userMean;
            float theirOffset = neighborRating - userMean;
            accum += ourOffset * theirOffset;
            den1 += (ourOffset*ourOffset);
            den2 += (theirOffset*theirOffset);
            numUsers++;
        }

        float weight;
        if (den1 > 0 && den2 > 0)
            weight = accum/(Math.sqrt(den1) * Math.sqrt(den2));
        else
            weight = 0;

        // Additional modifications to the similarity weight.
        if (numUsers < MIN_COMMON_USERS)
            weight = 0;

        if (numUsers < MAX_COMMON_USERS)
            weight *= ((float)numUsers)/MAX_COMMON_USERS;

        return new SimilarityWeight(neighbor, weight);
    }

    /**
     * Calculate the similarity between a subset of a user's ratings
     * and a neighbor's ratings.  Similarity weight is a Pearson
     * correlation.
     * @param userItems
     * @param neighbor
     * @return SimilarityWeight
     */
    private SimilarityWeight getUserWeight(List userItems, int neighbor)
    {
        float accum = 0;
        float weight = 0;

        if (userItems.isEmpty())
            return new SimilarityWeight(neighbor, 0);

        int activeUser = ((ItemRating)userItems.get(0)).getUserID();

        float activeUserSD = dataMgr.getUserSD(activeUser);
        float activeUserMean = dataMgr.getMeanRatingCached(activeUser);
        float neighborSD = dataMgr.getUserSD(neighbor);
        float neighborMean = dataMgr.getMeanRatingCached(neighbor);

        int numCommonItems = 0;
        Iterator iter = userItems.iterator();
        while (iter.hasNext())
        {
            ItemRating activeUserRating = (ItemRating)iter.next();
            float neighborRatingVal = dataMgr.getRatingCached(neighbor,
                    activeUserRating.getItemID());

            if (neighborRatingVal != dataMgr.getErrRating())
            {
                // The item has been rated by both
                numCommonItems++;

                // get the rating values
                float activeUserRatingVal = activeUserRating.getRating();

                // add the contribution of this common item to the sum
                // (Pearson correlation numerator)
                accum += (activeUserRatingVal - activeUserMean)
                    * (neighborRatingVal - neighborMean);
            }
        }


   		// Compute similarity weight
        weight = accum/activeUserSD/neighborSD;

        if (weight  < 0)
            weight  = 0;

        // Penalize correlations with fewer ratings if below the cutoff
        if (numCommonItems < MAX_COMMON_ITEMS)
            weight *= ((float)numCommonItems)/MAX_COMMON_ITEMS;

        return new SimilarityWeight(neighbor,weight);
    }

    /**
     * Generate a list of similarity weights to all neighbors, based
     * on the specified subset of a user's item ratings.
     */
    private List getSimilarityWeights(List userItems, Iterator neighbors)
    {
        LinkedList vals = new LinkedList();

        while (neighbors.hasNext())
        {
            ItemRating ir = (ItemRating)neighbors.next();
            vals.add(getUserCoeff(userItems, ir.getUserID()));
	    }

	    return vals;
	}

    /**
     * Return the name of this algorithm.
     * @return the name of this algorithm.
     */
    public String toString()
    {
        return "UserItem2Algorithm.java\n";
    }
}
