/*
 * VectorSimilarityAlgorithm.java
 *
 * Version 1.1
 *
 * 15/08/2002
 *
 * Copyright (c) 2003 Oregon State Universtity. All rights reserved.
 */

package org.recommender.algorithms.Experimental;
import org.recommender.algorithms.*;
import org.recommender.server.*;
import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.sql.*;
import java.lang.Math.*;

/**
 * Describes an implementation of the CFAlgorithm interface
 * to make CF predictions using the vector similarity method.
 * Includes an implementation of inverse user frequency
 * weighting from Breese et al 1998, but it doesn't help
 * and isn't recommended.  May be buggy.
 *
 * @author  Daniel Lowd &lt;dlowd@cs.orst.edu&gt;
 */
public class VectorSimilarityAlgorithm implements CFAlgorithm
{
    private DataManager dataMgr;

    private long NumPredictions  = 0;
    private long Time  = 0;
    private boolean USE_INVERSE_USER_FREQUENCY;
    private int MAX_NEIGHBORHOOD_SIZE;

    private int totalNumUsers;


    // Cached user denominators, for the vector similarity calculation
    private float[] userDenom;

    // Total sum of item frequencies
    private float iufSum;

    // Sum of item frequency times rating for each user
    private float[] frArray;

    // Sum of item frequency times rating squared for each user
    private float[] frrArray;

    /**
     * Constructor.
     * @param dataMgr
     */
    public VectorSimilarityAlgorithm(DataManager dataMgr)
    {
        this.dataMgr = dataMgr;

        String prefix = "CFServer.alg.vsim.";
        USE_INVERSE_USER_FREQUENCY = SystemProperties.getBoolean(
                prefix + "useIUF", false);

        MAX_NEIGHBORHOOD_SIZE = SystemProperties.getInt(
                prefix + "maxNeighborhoodSize", 50);

        if (USE_INVERSE_USER_FREQUENCY)
            init();

        int maxUser = dataMgr.getMaxUser();
        userDenom = new float[maxUser+1];
        Arrays.fill(userDenom, -1);
    }

    /**
     * Initialize data structures
     */
    private void init()
    {
        // Initialization required for inverse user frequency calculations
        int maxUser = dataMgr.getMaxUser();

        totalNumUsers = 0;
        frArray  = new float[maxUser+1];
        frrArray = new float[maxUser+1];

        for (int i = 0; i <= maxUser; i++)
        {
            if (dataMgr.getNumRatings(i) > 0)
                totalNumUsers++;

            float total = 0;
            float squares = 0;
            float iufSum = 0;

            // Calculate user totals
            Iterator iter = dataMgr.getItemList(i);
            while (iter.hasNext())
            {
                ItemRating ir = (ItemRating)iter.next();
                float r = ir.getRating();
                float iuf = getIUF(ir.getItemID());

                total   += iuf*r;
                squares += iuf*r*r;
            }

            frArray[i]  = total;
            frrArray[i] = squares;
        }

        iufSum = 0;
        for (int i = 0; i <= dataMgr.getMaxItem(); i++)
        {
            iufSum += getIUF(i);
        }
    }

    /**
     * Not yet complete
     */
    public ItemPrediction[] getRecommendationsByType(int activeUser, int n, int type)
               throws SQLException, CFNotImplementedException
   {
   		throw new CFNotImplementedException("getRecommendationsByType not implemented yet");
         //return null;

   }

    /**
     * Not yet complete
     */
    public ItemPrediction[] getRecommendations(int activeUser, int n)
        throws SQLException, CFNotImplementedException
    {
    	throw new CFNotImplementedException("getRecommendations not implemented yet");
        //return null;
    }

    /**
     * Predict rating for activeUser' item.
     * @param activeUser
     * @param item
     */
    public ItemPrediction predictRating (int activeUser, int item) throws CFNotImplementedException
    {
        float numerator = 0;
        float denominator = 0;

        long startTime = System.currentTimeMillis();
        // get list of neighbors' ratings
        Iterator itemRatings = dataMgr.getNeighborhood(item);

        List neighborCoeffs = getSimilarityWeights(activeUser, itemRatings);

        // sort descending, so we can use just the best
        Collections.sort(neighborCoeffs, Collections.reverseOrder());

        // now, calculate the prediction based on each neighbor
        int numNeighbors = 0;
        Iterator iter = neighborCoeffs.iterator();
        while (iter.hasNext())
        {
            SimilarityWeight c = (SimilarityWeight) iter.next();
            float rating = dataMgr.getRatingCached(c.getNeighborID(), item);
            float weight = c.getWeight();

            numerator   += weight * rating;
            denominator += Math.abs(weight);

            numNeighbors++;
            if (numNeighbors >= MAX_NEIGHBORHOOD_SIZE)
                break;
        }

        // final calc!
        float retval;
        if (denominator != 0)
            retval = numerator/denominator;
        else
        	retval= dataMgr.getErrRating();


        long stopTime = System.currentTimeMillis();
        NumPredictions++;
        Time += (stopTime-startTime);

        return new ItemPrediction(activeUser,item,retval); // return predicted rating
    }

    /**
     * Return average prediction time.
     * @return average prediction time
     */
	public long getAveragePredictionTime() throws CFNotImplementedException
    {
		return (Time/NumPredictions);
	}

    /**
     * Return inverse user frequency (IUF) for the given item.  Inverse user
     * frequency is equal to the natural log of the total number of users
     * divided by the number who rated the given item.
     * @param itemID
     * @return 0, if number of raters is 0, IUF, if not
     */
    private float getIUF(int itemID)
    {
        if (dataMgr.getNumRaters(itemID) == 0)
            return 0;

        return (float) Math.log(totalNumUsers/dataMgr.getNumRaters(itemID));
    }

    /**
     * Returns the magnitude of the user ratings vector, when considered
     * as an n-dimensional vector.  This is calculated, like any distance
     * in Cartesian space, as the square root of the sum of the users'
     * ratings squared.
     *
     * @param userID user to consider.
     * @return magnitude or "length" of the user ratings vector.
     */
    private float getUserDenom(int userID)
    {
        if (userDenom[userID] < 0) {
            float accum = 0;

            ItemRating[] userRatings = dataMgr.getUserRatingList(userID);
            for (int i = 0; i < userRatings.length; i++)
            {
                float r = userRatings[i].getRating();
                accum += r*r;
            }

            userDenom[userID] = (float) Math.sqrt(accum);
        }

        return userDenom[userID];
    }

    /**
     * Calculate weight between two users using inverse user frequency.
     * @param activeUser
     * @param neighbor
     * @return SimilarityWeight
     */
    public SimilarityWeight getWeightIUF(int activeUser, int neighbor)
    {
        float U = iufSum*(frrArray[activeUser]
                - frArray[activeUser]*frArray[activeUser]);
        float V = iufSum*(frrArray[neighbor]
                - frArray[neighbor]*frArray[neighbor]);
        float denom = (float) Math.sqrt(U*V);

        float accum = 0;
        Iterator iter = dataMgr.getItemList(activeUser);
        while (iter.hasNext())
        {
            ItemRating ir = (ItemRating)iter.next();
            float ourRating = ir.getRating();
            float theirRating = dataMgr.getRatingCached(neighbor,
                    ir.getItemID());

            if (theirRating != dataMgr.getErrRating())
                accum += getIUF(ir.getItemID()) * ourRating * theirRating;
        }

        float num = (iufSum*accum - frArray[activeUser]*frArray[neighbor]);
        return new SimilarityWeight(neighbor, num/denom);
    }

	/**
	 * Calculate similarity weight between two users.
	 * @param activeUser
	 * @param neighbor
	 * @return SimilarityWeight
	 */
	private SimilarityWeight getWeight(int activeUser, int neighbor)
    {
        float weight = 0;
        float denom = getUserDenom(activeUser) * getUserDenom(neighbor);

        if (denom == 0)
            return new SimilarityWeight(neighbor, 0);

        ItemRating[] userRatings = dataMgr.getUserRatingList(activeUser);
        for (int i = 0; i < userRatings.length; i++)
        {
            ItemRating activeUserRating = userRatings[i];
            float activeUserRatingVal = activeUserRating.getRating();
            float neighborRatingVal = dataMgr.getRatingCached(neighbor,
                    activeUserRating.getItemID());

            if (neighborRatingVal != dataMgr.getErrRating())
                weight += activeUserRatingVal * neighborRatingVal/denom;
        }

        return new SimilarityWeight(neighbor,weight);
   	}


    /**
     * Return a list of weights between the activeUser and each of his/her
     * neighbors.
     * @param activeUser
     * @param neighbors
     * @return List
     */
    private List getSimilarityWeights (int activeUser, Iterator neighbors)
    {
        LinkedList vals = new LinkedList();

        while (neighbors.hasNext())
        {
            ItemRating ir = (ItemRating)neighbors.next();

            if (USE_INVERSE_USER_FREQUENCY)
                vals.add(getWeightIUF(activeUser, ir.getUserID()));
            else
                vals.add(getWeight(activeUser, ir.getUserID()));
        }

        return vals;
	}

    /**
     * Not yet complete.  This algorithm will utterly fail if users with
     * userID's greater than the previously largest userID are added while
     * it is running.
     */
    public void updateUser(int userID) throws CFNotImplementedException
    {
        // Update user denominator.
        getUserDenom(userID);
    }

    /**
     * Return the name of this algorithm.
     * @return String
     */
    public String toString()
    {
        return "VectorSimilarityAlgorithm.java\n";
    }
}
