/* AlgorithmTuner.java
 *
 * Automatically optimize an algorithm's parameters for the loaded dataset.
 *
 * Author: Daniel Lowd <dlowd@cs.orst.edu>
 */
package org.recommender.algorithms.Experimental.tuning;
import org.recommender.server.*;
import java.io.*;
import java.util.*;
import java.sql.*;

public class AlgorithmTuner
{
    int VERSION = 1;
    DataManager dataMgr;

    float bestMAE;
    MutableParameterSet bestParams;
    CFAlgorithm bestAlg;

    ArrayList testRatings;
    ItemRating[] randomTestSet;

    public AlgorithmTuner(DataManager dataMgr, String filename, 
            String testSetWhereClause)
    {
        this.dataMgr = dataMgr;

        try {
            bestParams = new MutableParameterSet(filename);
        } catch (FileNotFoundException e) {
            System.out.println("Error: could not find parameter file + \""
                + filename + "\".");
            bestParams = null;
        }

        testRatings = new ArrayList();
        fillTestSet(testSetWhereClause);

        bestMAE = 1000.0;
    }

    public void fillTestSet(String testSetWhereClause)
    {
        int i = 0;
        Iterator iter = null;
        
        try {
            iter = dataMgr.getRatingList(testSetWhereClause);
        } catch (SQLException e) {
            System.out.println("Error: could not load test set \"" +
                testSetWhereClause + "\".");
            return;
        }

        CFAlgorithm pop = new PopAlgorithm(dataMgr, null);
        while (iter.hasNext())
        {
            ItemRating ir = (ItemRating)iter.next();
            if (dataMgr.getNumRatings(ir.getUserID()) >=5
                    && pop.predictRating(ir.getUserID(), ir.getItemID())
                        != SystemConstants.ERR_RATING)
                testRatings.add(i++, ir);
        }

        System.out.println("Ratings loaded: " + testRatings.size());
    }

    // Simulated annealing
    public MutableParameterSet tune()
    {
        System.out.println("Simulated annealing, v" + VERSION + "...");
        initMAE();

        //FullHortingAlgorithm baseAlg = new FullHortingAlgorithm(dataMgr, 
        //        bestParams);
        //List[] neighborLists = baseAlg.getNeighborLists();
        //bestAlg = baseAlg;
        bestAlg = new BoostedPearsonAlgorithm(dataMgr, bestParams);
        CFAlgorithm currAlg = bestAlg;
        MutableParameterSet currParams = bestParams;

        int numIters = 0;
        for (int i = 0; i < 1000; i++)
        {
            if (i % 50 == 0)
            {
                printMAE(bestAlg);
            }

            MutableParameterSet otherParams = currParams.permute();
            
           // CFAlgorithm otherAlg = new FullHortingAlgorithm(dataMgr, 
           //         otherParams, neighborLists);
           CFAlgorithm otherAlg = new BoostedPearsonAlgorithm(dataMgr, 
                   otherParams);

            float deltaE = compAlgs(currAlg, otherAlg, 100);
            //System.out.println("Delta E = " + deltaE);
            //float T = 0.1 * Math.pow(0.5, i/100.0);
            float K = 1.00;
            float T = 1/Math.log(i);
            if (deltaE < 0 || Math.random() < Math.exp(-deltaE/(K*T)))
            {
                currAlg = otherAlg;
                currParams = otherParams;
                System.out.println("Switching params.");
            }

            // See if we actually have a new best
            if (deltaE < 0 && compAlgs(bestAlg, currAlg, 100+i) < 0)
            {
                System.out.println("********* NEW BEST! ***********");
                bestAlg = currAlg;
                bestParams = currParams;
                System.out.println(currParams);
                numIters = 0;
            }
            // Verify that we're still making progress
            else if (numIters > 100)
            {
                if (compAlgs(bestAlg, currAlg, 100+i) < 0)
                {
                    System.out.println("!!!!!!!!!!!! NEW BEST! !!!!!!!!!!!");
                    bestAlg = currAlg;
                    bestParams = currParams;
                    System.out.println(currParams);
                    numIters = 0;
                }
                else
                    numIters = 0;
                /*
                else
                {
                    System.out.println("100 iters, no change, we're done.");
                    break;
                }
                */
            }

            numIters++;
        }

        if (compAlgs(bestAlg, currAlg, 1000) < 0)
            bestAlg = currAlg;

        System.out.println("Final best parameters: " + bestParams);
        return bestParams;
    }

    public void initMAE()
    {
        randomTestSet = new ItemRating[100];
        for (int i = 0; i < 100; i++)
            randomTestSet[i] = getRandomRating();
    }

    public float compAlgsv2(CFAlgorithm alg1, CFAlgorithm alg2)
    {
        return (getMAE(alg2) - getMAE(alg1));
    }

    public float getMAE(CFAlgorithm alg)
    {
        float num = 0;

        for (int i = 0; i < 100; i++)
        {
            ItemRating ir = randomTestSet[i];
            float rating = alg.predictRating(ir.getUserID(), ir.getItemID());
            // Assume that we can predict this rating (we should have tested
            // this already).
            num += Math.abs(rating - ir.getRating());
        }

        return num/100;
    }

    /*
    // Iteration
    public MutableParameterSet iterTune()
    {
        CFAlgorithm bestAlg = new FullPearsonAlgorithm(dataMgr, bestParams);
        CFAlgorithm currAlg = bestAlg;
        MutableParameterSet currParams = bestParams;
        int numParams = bestParams.size();

        int numIters = 0;
        for (int i = 0; i < 10; i++)
        {
            currParams = new MutableParameterSet(bestParams);
            for (int j = 0; j < numParams; j++)
            {
                Parameter originalParam = currParams.get(j);
                Parameter param = originalParam.iterPermute();
                while (!param.equals(originalParam))
                {
                    currParams.set(j, param);
                    currAlg = new FullPearsonAlgorithm(dataMgr, currParams);
                    param = param.iterPermute();
                    if (compAlgs(bestAlg, currAlg, 100) < 0)
                    {
                        bestParams = new MutableParameterSet(currParams);
                        bestAlg = currAlg;
                    }
                }

                System.out.println(originalParam.getName() + ": old - "
                        + originalParam.getValue() + "; new - "
                        + bestParams.get(j).getValue());
            }
        }

        System.out.println("Final best parameters: " + bestParams);
        return bestParams;
    }
    */

    public float compAlgs(CFAlgorithm alg1, CFAlgorithm alg2, int numTests)
    { 
        if (VERSION == 1)
            return compAlgsv1(alg1, alg2, numTests);
        else
            return compAlgsv2(alg1, alg2);
    }

    public float compAlgsv1(CFAlgorithm alg1, CFAlgorithm alg2, int numTests)
    {
        Stats stats = new Stats();

        for (int i = 0; i < numTests; i++)
        {
            ItemRating testRating = getRandomRating();

            float rating1 = alg1.predictRating(testRating.getUserID(),
                    testRating.getItemID());
            float rating2 = alg2.predictRating(testRating.getUserID(),
                    testRating.getItemID());

            if (rating1 != SystemConstants.ERR_RATING 
                    && rating2 != SystemConstants.ERR_RATING
                    && Math.abs(rating1 - rating2) > 0.0001)
            {
                float actual = testRating.getRating();
                float err1 = (rating1 - actual)*(rating1 - actual);
                float err2 = (rating2 - actual)*(rating2 - actual);
                stats.addNum(err1 - err2);
                //stats.addNum(Math.abs(rating1 - testRating.getRating())
                //        - Math.abs(rating2 - testRating.getRating()));
            }
        }
            
        /*
        // Paired t-test
        float t = stats.calcMean()*Math.sqrt(stats.size())/stats.calcSD();

        // If we're 99% sure this is better, return true.
        if ( t > 2.358 ) // v = 120, t_.010
            return true;
        else if (t < -1.289)
            return false;
         */

        // Calculate the minimum mean difference required in order to
        // have ~95% certainty that this is an improvement.
        //float goalMean = 1.6*stats.calcSD()/Math.sqrt(stats.size());
        float goalMean = 1.2*stats.calcSD()/Math.sqrt(stats.size());

        // Return the difference between that goal and what we actually got.
        return (goalMean - stats.calcMean());
    }


    public void printMAE(CFAlgorithm alg)
    { 
        float totalDist = 0;
        int numRatings = 0;
        for (int i = 0; i < 1000; i++)
        {
            ItemRating testRating = getRandomRating();
            /*
            System.out.println("u: " + testRating.getUserID() + "; i:" +
                    testRating.getItemID() + "; r: " + testRating.getRating());
                    */

            float rating = alg.predictRating(testRating.getUserID(),
                    testRating.getItemID());
            if (rating != SystemConstants.ERR_RATING)
            {
                totalDist += Math.abs(rating - testRating.getRating());
                numRatings++;
            }
        }
            
        System.out.println("Best MAE: " + totalDist/numRatings);
    }

    /*
    // Select one of the ratings in the database at random.
    ItemRating getRandomRating()
    {
        // Select a user
        int userID = 0;
        int numRatings = 0;
        while (numRatings == 0)
        {
            userID = (int)(Math.random()*dataMgr.getMaxUser() + 0.5);
            numRatings = dataMgr.getNumRatings(userID);
        }

        // Select a rating at random
        int rating = (int)(Math.random() * numRatings);
        Iterator userRatings = dataMgr.getItemList(userID);
        while (rating-- > 0)
        {
            userRatings.next();
        }
        return (ItemRating)userRatings.next();
    }
    */

    // Select one of the ratings in the database at random.
    ItemRating getRandomRating()
    {
        int ratingNum = (int)(Math.random() * testRatings.size());
        return (ItemRating)testRatings.get(ratingNum);
    }


    float testOne(CFAlgorithm alg, ItemRatingComparable testRating)
    {
        float actualRating = testRating.getRating();
        int userID = testRating.getUserID();
        int itemID = testRating.getItemID();
        float predictedRating = alg.predictRating(userID, itemID);

        if (predictedRating == SystemConstants.ERR_RATING)
            return SystemConstants.ERR_RATING;
        else
            return Math.abs(actualRating - predictedRating); 
    }
}
