/* FloatParameter.java
 *
 * Author: Daniel Lowd <dlowd@cs.orst.edu>
 *
 * Represents a mutable algorithm parameter with a float value.  Contains
 * methods for parsing itself (given a StringTokenizer) and randomly permuting
 * itself.  Knows its own minimum and maximum values, as well as an average
 * increment.  When randomly permuted, it will increase or decrease by as much
 * as twice this increment, as determined by Math.random().  Used by
 * MutableParameterSet.
 */
package org.recommender.algorithms.Experimental.tuning;

import java.util.*;

public class DoubleParameter extends Parameter
{
    String name;
    float min;
    float max;
    float increment;
    float value;

    public DoubleParameter(String name, StringTokenizer st)
    {
        this.name = name;

        min         = Float.parseFloat(st.nextToken());
        max         = Float.parseFloat(st.nextToken());
        increment   = Float.parseFloat(st.nextToken());
        value       = Float.parseFloat(st.nextToken());
    }

    public DoubleParameter(String name, float min, float max,
            float increment, float value)
    {
        this.name       = name;
        this.min        = min;
        this.max        = max;
        this.increment  = increment;
        this.value      = value;
    }

    public DoubleParameter(DoubleParameter other)
    {
        name        = other.name;
        min         = other.min;
        max         = other.max;
        increment   = other.increment;
        value       = other.value;
    }

    // Generate a new parameter with a value increased or decreased by up 
    // to twice the increment
    public Parameter permute()
    {
        float newValue = value;
        
        while (newValue == value)
        {
            newValue = value + increment*4.0*(Math.random() - 0.5);
            if (newValue > max)
                newValue = max;
            else if (newValue < min)
                newValue = min;
        }

        return new DoubleParameter(name, min, max, increment, newValue);
    }

    public Parameter iterPermute()
    {
        float newValue = value + increment;
        if (value > max)
            value = min;

        return new DoubleParameter(name, min, max, increment, newValue);
    }

    public boolean equals(Parameter other)
    {
        if (other.getType() != TYPE_DOUBLE)
            return false;

        DoubleParameter p = (DoubleParameter)other;

        return (Math.abs(value - p.value) < 0.0001);
    }

    public int getType()
    {
        return TYPE_DOUBLE;
    }

    public String getName()
    {
        return name;
    }

    public String getValue()
    {
        return "" + value;
    }
}
