/* IntParameter.java
 *
 * Author: Daniel Lowd <dlowd@cs.orst.edu>
 *
 * Represents a mutable algorithm parameter with an integer value.  Contains
 * methods for parsing itself (given a StringTokenizer) and randomly permuting
 * itself.  Knows its own minimum and maximum values, as well as an average
 * increment.  When randomly permuted, it will increase or decrease by as much
 * as twice this increment, as determined by Math.random().  Used by
 * MutableParameterSet.
 */
package org.recommender.algorithms.Experimental.tuning;

import java.util.*;

public class IntParameter extends Parameter
{
    String name;
    int min;
    int max;
    int increment;
    int value;

    public IntParameter(String name, StringTokenizer st)
    {
        this.name = name;

        min = Integer.parseInt(st.nextToken());
        max = Integer.parseInt(st.nextToken());
        increment = Integer.parseInt(st.nextToken());
        value = Integer.parseInt(st.nextToken());
    }

    public IntParameter(String name, int min, int max, int increment, int value)
    {
        this.name       = name;
        this.min        = min;
        this.max        = max;
        this.increment  = increment;
        this.value      = value;
    }

    public IntParameter(IntParameter other)
    {
        name        = other.name;
        min         = other.min;
        max         = other.max;
        increment   = other.increment;
        value       = other.value;
    }

    // Generate a new parameter with a value increased or decreased by up 
    // to twice the increment
    public Parameter permute()
    {
        int newValue = value;
        
        while (newValue == value)
        {
            newValue = value + (int)(increment*4.0*(Math.random()-0.5) + 0.5);
            if (newValue > max)
                newValue = max;
            else if (newValue < min)
                newValue = min;
        }

        return new IntParameter(name, min, max, increment, newValue);
    }

    public Parameter iterPermute()
    {
        int newValue = value + increment;
        if (newValue > max)
            newValue = min;

        return new IntParameter(name, min, max, increment, newValue);
    }

    public boolean equals(Parameter other)
    {
        if (other.getType() != TYPE_INT)
            return false;

        IntParameter p = (IntParameter)other;

        return (value == p.value);
    }

    public int getType()
    {
        return TYPE_INT;
    }

    public String getName()
    {
        return name;
    }

    public String getValue()
    {
        return "" + value;
    }
}
