/* MutableParameterSet.java
 *
 * Author: Daniel Lowd <dlowd@cs.orst.edu>
 *
 * A MutableParameterSet is a set of parameters that can be randomly
 * permuted according to provided specifications.
 */
package org.recommender.algorithms.Experimental.tuning;

import java.io.*;
import java.util.*;

public class MutableParameterSet extends ParameterSet
{
    LinkedList params;

    public MutableParameterSet(MutableParameterSet other)
    {
        params = new LinkedList(other.params);
    }

    public MutableParameterSet()
    {
        params = new LinkedList();
    }

    public MutableParameterSet(String filename) throws FileNotFoundException
    {
        params = new LinkedList();

        // Reader for the parameter file
        BufferedReader in;

        // Open the file
        in = new BufferedReader(new FileReader(filename));

        try {
            // Read a line containing a parameter and its name
            String currLine;
            while ((currLine = in.readLine()) != null)
            {
                // Parse the line into a parameter name and a value
                StringTokenizer st = new StringTokenizer(currLine);
                String name = st.nextToken();
                String type = st.nextToken();

                if (type.equalsIgnoreCase("int"))
                    params.add(new IntParameter(name, st));
                else if (type.equalsIgnoreCase("float"))
                    params.add(new FloatParameter(name, st));
                else if (type.equalsIgnoreCase("bool"))
                    params.add(new BoolParameter(name, st));
                else if (type.equalsIgnoreCase("string"))
                    params.add(new StringParameter(name, st));
                else
                    System.out.println("ERROR: parameter type \"" + type 
                        + "\" not implemented!");
            }
        } catch (IOException e) {
            // When we encounter an error, stop and use what we have.
        }
    }

    // Randomly permute one of our parameters
    public MutableParameterSet permute()
    {
        MutableParameterSet retVal = new MutableParameterSet();

        float denom = params.size();
        Iterator iter = params.iterator();
        while (iter.hasNext())
        {
            Parameter curr = (Parameter)iter.next();

            if (curr.getType() != Parameter.TYPE_STRING && denom > 0
                    && Math.random() < 1/denom)
            {
                retVal.params.add(curr.permute());
                denom = 0;
            }
            else
            {
                retVal.params.add(curr);
                denom--;
            }
        }

        return retVal;
    }

    public int size()
    {
        return params.size();
    }

    public Parameter get(int i)
    {
        return (Parameter)params.get(i);
    }

    public void set(int i, Parameter p)
    {
        params.set(i, p);
    }

    public Iterator iterator()
    {
        return params.iterator();
    }

    public String getString(String paramName, String defaultVal)
    {
        String retVal = getString(paramName);
        if (retVal == null)
        {
            setString(paramName, defaultVal);
            retVal = defaultVal;
        }

        return retVal;
    }

    public void setString(String paramName, String value)
    {
        // Remove the parameter from our list, if it exists
        Iterator iter = params.iterator();
        while (iter.hasNext())
        {
            Parameter p = (Parameter)iter.next();
            if (paramName.equals(p.getName()))
                iter.remove();
        }

        // Add it
        params.add(new StringParameter(paramName, value));
    }

    public String getString(String paramName)
    {
        Iterator iter = params.iterator();

        while (iter.hasNext())
        {
            Parameter curr = (Parameter)iter.next();
            if (paramName.equalsIgnoreCase(curr.getName()))
                return curr.getValue();
        }

        return null;
    }

    public String toString()
    {
        String retVal = "";
        Iterator iter = params.iterator();

        while (iter.hasNext())
        {
            Parameter curr = (Parameter)iter.next();
            retVal += curr.getName() + " " + curr.getValue() + "\n";
        }

        return retVal;
    }

    public String getAllParams()
    {
        return toString();
    }
}
