/* 
 * ParameterSet.java
 * 
 * Version 1.1
 * 
 * 15/08/2002
 *
 * Copyright (c) 2003 Oregon State Universtity. All rights reserved.
 */ 
 
package org.recommender.algorithms;

import org.recommender.server.*;
/**
 * A ParameterSet represents all parameters necessary to specify an algorithm.
 * Parameters are algorithm-dependent.  The abstract class ParameterSet
 * provides utility methods for retrieving parameters as integers, floats,
 * or booleans.  Implementing classes must provide a method for getting
 * the original string value of a parameter.
 *
 * @author Daniel Lowd <dlowd@cs.orst.edu>
 */
abstract public class ParameterSet
{
    /**
     * Child classes are responsible for implementing this method,
     * as well as some way to construct a ParameterSet.
     */
    abstract public String getString(String paramName);
    
    /**
     * Child classes are responsible for implementing this method,
     * as well as some way to construct a ParameterSet.
     */
    abstract public String getString(String paramName, String defaultVal);
    
    /**
     * Child classes are responsible for implementing this method,
     * as well as some way to construct a ParameterSet.
     */
    abstract public void setString(String paramName, String value);
    
    /**
     * Child classes are responsible for implementing this method,
     * as well as some way to construct a ParameterSet.
     */
    abstract public String getAllParams();

    /**
     * Get parameter's int value, if not defined, return default value
     * @param paramName
     * @param defaultVal
     * @return int value of the parameter
     */     
    public int getInt(String paramName, int defaultVal)
    {
        int retVal;
        String paramVal = getString(paramName, ""+defaultVal);

        try {
            retVal = Integer.parseInt(paramVal);
        } catch (NumberFormatException e) {
            retVal = defaultVal;
            setString(paramName, ""+defaultVal);
        }

        return retVal;
    }

    /**
     * Get parameter's float value, if not defined, return default value
     * @param paramName
     * @param defaultVal
     * @return float value of the parameter
     */   
    public float getFloat(String paramName, float defaultVal)
    {
        float retVal;
        String paramVal = getString(paramName, ""+defaultVal);

        try {
            retVal = Float.parseFloat(paramVal);
        } catch (NumberFormatException e) {
            retVal = defaultVal;
            setString(paramName, ""+defaultVal);
        }

        return retVal;
    }

    /**
     * Get parameter's boolean value, if not defined, return default value
     * @param paramName
     * @param defaultVal
     * @return boolean value of the parameter
     */   
    public boolean getBoolean(String paramName, boolean defaultVal)
    {
        boolean retVal;
        String paramVal;
        String defaultString;
        
        if (defaultVal)
            defaultString = "true";
        else
            defaultString = "false";

        paramVal = getString(paramName, defaultString);

        if (paramVal.equalsIgnoreCase("true"))
            retVal = true;
        else if (paramVal.equalsIgnoreCase("false"))
            retVal = false;
        else
        {
            setString(paramName, defaultString);
            retVal = defaultVal;
        }
        
        return retVal;
    }
}
