/* 
 * SimpleParameterSet.java
 * 
 * Version 1.1
 * 
 * 15/08/2002
 *
 * Copyright (c) 2003 Oregon State Universtity. All rights reserved.
 */ 
 
package org.recommender.algorithms;

import org.recommender.server.*;
import java.io.*;
import java.util.*;

/**
 * Parse a file containing algorithm parameters, and allow access to
 * those parameters.
 *
 * @author Daniel Lowd <dlowd@cs.orst.edu>
 */
public class SimpleParameterSet extends ParameterSet
{
    /**
     * Keep track of all parameters
     */
    Hashtable params;
    
    /**
     * Keep a string with the file's entire contents.
     */
    String allParams;

    /**
     * Constructor, loads the file with the given filename.
     * @param filename
     * @throws FileNotFoundException
     */
    public SimpleParameterSet(String filename) throws FileNotFoundException
    {
        params = new Hashtable(); 
        allParams = "";
       

        // Reader for the parameter file
        BufferedReader in;

        // Open the file
        in = new BufferedReader(new FileReader(filename));

        try {
            // Read a line containing a parameter and its name
            String currLine;
            while ((currLine = in.readLine()) != null)
            {
                // Parse the line into a parameter name and a value
                StringTokenizer st = new StringTokenizer(currLine);

                String parameter, value;
                try {
                    parameter = st.nextToken();
                    value     = st.nextToken();
                } catch (NoSuchElementException e) {
                    continue;
                }


                // Add it to the hashtable
                params.put(parameter, value);

                // Add this line to our complete string
                allParams += currLine + "\n";
            }
        } catch (IOException e) {
            // When we encounter an error, stop and use what we have.
        }
    }

    /**
     * Constructor, gets params from props directly and changes them into strings
     * @param props
     */
    public SimpleParameterSet(Properties props)
    {
        this.params = props;

        // Get string representation of parameters
        try {
            ByteArrayOutputStream outbuf = new ByteArrayOutputStream();
            props.store(outbuf, "");
            allParams = outbuf.toString();
        } catch (IOException e) {
            allParams = "# ERROR STORING PROPERTIES!\n";
        }
    }

    /**
     * Constructor, creates dummy parameter file
     */
    public SimpleParameterSet()
    {
        params = new Hashtable();
        allParams = "";
    }

    /**
     * Get a parameter's value.  If the parameter has no value, create a
     * parameter with the specified default value and return that value.
     * This way, we have a record of any default values being used.
     *
     * One caveat: if the same parameter is requested twice with two
     * different default values specified, only the first will be used.
     * Once it is requested the first time, the parameter will be set to
     * the first default value provided, just as if it had been specified
     * in a parameter file.
     * @param paramName
     * @param defaultVal
     * @return param's value
     */
    public String getString(String paramName, String defaultVal)
    {
        String retVal = getString(paramName);
        if (retVal == null)
        {
            setString(paramName, defaultVal);
            retVal = defaultVal;
        }

        return retVal;
    }

    /**
     * Get a parameter's value.  Return null if not found.
     * @param paramName
     * @return param's value, null if not found
     */
    public String getString(String paramName)
    {
        return (String)params.get(paramName);
    }

    /**
     * Sets a parameter's value.  This may alter the allParams string
     * so that it contains two copies of a parameter.  This should probably
     * be considered a BUG.  Perhaps we can live with it for now.
     * @param paramName
     * @param value
     */
    public void setString(String paramName, String value)
    {
        params.put(paramName, value);
        allParams += paramName + " " + value + "\n";
    }

    /**
     * Return a string containing all parameters.  This includes
     * default parameters, if default parameters were used with
     * the getString method.
     * @return allParams
     */
    public String getAllParams()
    {
        return allParams;
    }
}
