/*
 * SimplePearsonAlgorithm.java
 *
 * Version 1.2
 *
 * 2/15/2003
 *
 * Copyright (c) 2003 Oregon State University. All rights reserved.
 */
package org.recommender.algorithms;

import org.recommender.server.*;

import java.util.*;
import java.sql.*;

/**
 * Describes an implementation of the CFAlgorithm interface
 * to make CF predictions using the Simple Pearson Correlation
 * method.
 *
 * @author  <A HREF="mailto:sameerk@cs.orst.edu">Sameer Kadam</A>
 * @author  <A HREF="mailto:iyoon@cs.orst.edu">Irwin Yoon</A>
 * @author  <A HREF="mailto:godde@cs.orst.edu">Olivier Godde</A>
 * @author  <A HREF="mailto:dlowd@cs.orst.edu">Daniel Lowd</A>
 * @author  <A HREF="mailto:wangyun@cs.orst.edu">Yun Wang</A>
 * @author  <A HREF="mailto:mclaughm@cs.orst.edu">Matt McLaughlin</A>
 */
public class SimplePearsonAlgorithm
        implements CFAlgorithm {

    /**
     * DataManager for this algorithm
     */
    private DataManager dataMgr;

    ///////////// Parameters /////////////////
    // DL Some explanations are ambiguous or incomplete.  Some may change...

    /**
     * Minimum number of neighbors required.  If we have fewer than
     * this number of neighbors, return ERR_RATING rather than a
     * prediction.
     */
    private int MIN_NEIGHBORHOOD_SIZE;

    /**
     * Maximum number of neighbors to use for generating a prediction.
     */
    private int MAX_NEIGHBORHOOD_SIZE;

    /**
     * Maximum ratio of possible neighbors to use.  i.e., if this is set
     * to 0.1, then we'll use no more than 10% of the users with which
     * we can compute a correlation.
     */
    private float NEIGHBORHOOD_SIZE_RATIO;

    /**
     * Minimum number of items in common for a correlation to be valid.
     *
     * (Generally left at 2.)
     */
    private int MIN_COMMON_ITEMS;

    /**
     * Ratio of common items required for a correlation to be valid.
     * Superceded by MIN_COMMON_ITEMS.
     *
     * (Generally unused; not found to be helpful.)
     */
    private float HORT_RATIO;

    /**
     * Maximum number of items to consider for weighting a correlation
     * based on the number of items.  Correlations benefit from being
     * based on a larger number of items up to this number.
     */
    private int MAX_COMMON_ITEMS;

    /**
     * Neighbor weight threshold.  Neighbors with weights below this value
     * will not be used.
     */
    private float MIN_CORRELATION;

    /**
     * If false, we'll compute the active user's similarity to each neighbor
     * who has rated the target item whenever a rating is requested.  This
     * is the default behavior.  When this option is true, we instead
     * compute the active user's similarity to *all* users.  In most cases,
     * this is slower.  However, it also allows us to reuse the similarity
     * computation if we're requesting a lot of predictions in a row for
     * the same user, as is the case for most experiments.
     */
    private boolean CACHE_LAST_NEIGHBORS;

    //private List[] neighborLists;
    //private int maxUser;

    private long NumPredictions = 0;
    private long Time = 0;
	private Stats timeStats = new Stats();

    /**
     * Constructor, pass in dataMgr and initialize algorithm parameters.
     */
    public SimplePearsonAlgorithm(DataManager dataMgr) {
        this.dataMgr = dataMgr;

        // Get parameters
        String prefix = "CFServer.alg.pearson.";
        MIN_NEIGHBORHOOD_SIZE = SystemProperties.getInt(
                prefix + "minNeighborhoodSize", 1);
        MAX_NEIGHBORHOOD_SIZE = SystemProperties.getInt(
                prefix + "maxNeighborhoodSize", 50);
        NEIGHBORHOOD_SIZE_RATIO = SystemProperties.getFloat(
                prefix + "neighborhoodSizeRatio", 1);
        HORT_RATIO = SystemProperties.getFloat(
                prefix + "hortRatio", 1);
        MIN_COMMON_ITEMS = SystemProperties.getInt(
                prefix + "minCommonItems", 2);
        MAX_COMMON_ITEMS = SystemProperties.getInt(
                prefix + "maxCommonItems", 60);
        MIN_CORRELATION = SystemProperties.getFloat(
                prefix + "minCorrelation", 0);
        CACHE_LAST_NEIGHBORS = SystemProperties.getBoolean(
                prefix + "cacheLastNeighbors", false);
    }

    /* Keep track of the last user whose neighbors were requested.
     * If the same user is requested twice in a row, we can save time
     * with this computation.
     */
    private static int lastActiveUser = -1;
    private static List lastNeighbors = null;

    private List getNeighborsCached(UserInfo activeUserInfo) {

        if (activeUserInfo.userID == lastActiveUser) {
            return lastNeighbors;
        }

        // Compute similarity with all available users
//        lastNeighbors = getSimilarityWeightsTopN(activeUser,  dataMgr.getAllUsers());
		lastNeighbors = getSimilarityWeightsTopN(activeUserInfo,  dataMgr.getAllUsers());

        // Sort similarities in reverse order, so we use the best ones first.
        Collections.sort(lastNeighbors, Collections.reverseOrder());

        lastActiveUser = activeUserInfo.userID;
        return lastNeighbors;
    }

    /**
     * Yun Wang
     *
     * Return top n recommendations of specific type.
     * @param activeUser
     * @param n
     * @param type
     * @return itemPredList
     * @throws SQLException
     */
    public ItemPrediction[] getRecommendationsByType(int activeUser, int n,
                                                     int type) throws SQLException, CFNotImplementedException {

        ItemPrediction[] itemPredList;
        int maxItem = dataMgr.getMaxItem();

        //    ItemPrediction[] tempPredList = new ItemPrediction[maxItem + 1];
        float[] num = new float[maxItem + 1];
        float[] den = new float[maxItem + 1];

        int[] typeArr = new int[maxItem + 1]; // for storing type

        for (int u = 0; u <= maxItem; u++) {
            num[u] = 0;
            den[u] = 0;
            typeArr[u] = 0;
        }

        UserInfo activeUserInfo = dataMgr.getUserInfo(activeUser);

        Iterator typedIt = dataMgr.getCachedTypedItemLists(type);
        int size = dataMgr.getSizeOfTypedItemList(type);

        ItemPrediction[] tempPredList = new ItemPrediction[size + 1];

        while (typedIt.hasNext()) {
            int i = ((Integer) typedIt.next()).intValue();

            typeArr[i] = 1;
        }

        List neighborCoeffs;
        //get the lock here, so you don't have to worry about waiting in between
        //your get weight calls and you insure your average matches what you weights
        // are calculated at.
        if(!activeUserInfo.getReadersLock(true)) return new ItemPrediction[0];


        //long start = System.currentTimeMillis();
        if (CACHE_LAST_NEIGHBORS) {
            neighborCoeffs = getNeighborsCached(activeUserInfo);
            activeUserInfo.releaseReadersLock();
        } else {
            /* get list of neighbors, calculate weights, and sort in
             * reverse order, so we use the best ones first.
             */
            //Iterator neighbors = getNeighborhood(item);
            //neighborCoeffs = getSimilarityWeights(activeUser, neighbors, item);


            neighborCoeffs = getSimilarityWeightsTopN(
                    activeUserInfo, dataMgr.getAllUsers());
            activeUserInfo.releaseReadersLock();

            Collections.sort(neighborCoeffs, Collections.reverseOrder());
        }



        int numNeighbors = (int) (neighborCoeffs.size()
                * NEIGHBORHOOD_SIZE_RATIO);

        if (numNeighbors < MIN_NEIGHBORHOOD_SIZE) {
            return new ItemPrediction[0];
        }

        //long stop = System.currentTimeMillis();
        //    System.out.println("Time to select best neighbors= " + (stop - start));

        Iterator iter = neighborCoeffs.iterator();
        //start = System.currentTimeMillis();

        ItemRating r = null;
        float rating = 0;
        float mean = 0;

        // While there are neighbors less than max number
        while (iter.hasNext()) {
            //Get new neighbor
            SimilarityWeight w = (SimilarityWeight) iter.next();
            // Get new neighbor's average rating

            if (w.getNeighborID() == activeUser) {
                continue;
            }

            UserInfo neighborInfo = dataMgr.getUserInfo(w.getNeighborID());
            if(neighborInfo == null) continue;
            if(!neighborInfo.getReadersLock(false)) continue;


           // mean = dataMgr.getMeanRatingCached(w.getNeighborID());
            mean = neighborInfo.getUserMean();

            // Get neighbor's item list
         /*  Iterator i_index = dataMgr.getItemList(w.getNeighborID());

            // Get typed item list
            //  Iterator typed_index = typedItemList.iterator();

            while (i_index.hasNext()) { //While there are items in list

                r = (ItemRating) i_index.next();

                rating = r.getRating();
                int item = r.getItemID();

                if (rating > 0 && typeArr[item] > 0) {
                    // Get neighbor's weight
                    float weight = w.getWeight();
                    num[item] += ((rating - mean) * weight);
                    den[item] += Math.abs(weight);

                }

            } // while (i_index)*/
            float weight = w.getWeight();
            int[] itemIDs = neighborInfo.getRatings().ids;
            float[] ratings = neighborInfo.getRatings().ratings;
            for(int i = 0; i < neighborInfo.getRatings().getLength(); i++){
                int index = itemIDs[i];
                if (typeArr[index] > 0) {
                	num[index] += ((ratings[i] - mean) * weight);
                	den[index] += weight;
                }
            }
            neighborInfo.releaseReadersLock();

        } // while (iter)


        //stop = System.currentTimeMillis();
        //    System.out.println("Time to compute predictions=" + (stop - start));

        //start = System.currentTimeMillis();
        tempPredList[0] = new ItemPrediction(0, 0, 0);

        int predIndex = 0;
        float activeUserMean = dataMgr.getMeanRatingCached(activeUser);
        for (int i = 1; i < num.length; i++) {
            // int i = ((Integer) typedItemList.get(k)).intValue();

            //check that activeUser has not already rated item
            if (typeArr[i] > 0 && num[i] > 0 && den[i] > 0 &&
                    dataMgr.getRatingCached(activeUser, i) == dataMgr.getErrRating()) {

                /*itemPredList[predIndex++] = new ItemPrediction(0, i,
                        active_user_avr + num[i]/den[i]);*/

                //Updated: now uses item mean for better performance
                //tempPredList[predIndex++] = new ItemPrediction(0, i,
                //        dataMgr.getMeanItemRatingCached(i) + num[i] / den[i]);

                tempPredList[predIndex++] = new ItemPrediction(0, i,
                        activeUserMean + num[i] / den[i]);

            }

        }


        //stop = System.currentTimeMillis();
//    System.out.println("Time to fill pred array=" + (stop - start));

        //start = System.currentTimeMillis();
        Arrays.sort(tempPredList, 0, predIndex, new CompPrediction(dataMgr));
        //stop = System.currentTimeMillis();
//    System.out.println("Time to sort the array= " + (stop - start));

        if (predIndex > n) {
            predIndex = n;
        }
        itemPredList = new ItemPrediction[predIndex];

        //fill array to be returned
        for (int i = 0; i < itemPredList.length; i++) {
            if (tempPredList[i].getPrediction() > dataMgr.getMaxRating()) {
                tempPredList[i] = new ItemPrediction(tempPredList[i].getUserID(),
                        tempPredList[i].getItemID(),
                        dataMgr.getMaxRating());
            }
            if (tempPredList[i].getPrediction() < dataMgr.getMinRating()) {
                tempPredList[i] = new ItemPrediction(tempPredList[i].getUserID(),
                        tempPredList[i].getItemID(),
                        dataMgr.getMinRating());

            }

            itemPredList[i] = tempPredList[i];
        }

        return itemPredList;

    } // getRecommendationsByType()

    /**
     * Return top n recommendations.
     * @param activeUser
     * @param n
     * @return itemPredList
     * @throws SQLException
     */
    public ItemPrediction[] getRecommendations(int activeUser, int n) throws
            SQLException, CFNotImplementedException {
        ItemPrediction[] itemPredList;
        ItemPrediction[] tempPredList;
        int maxItem = dataMgr.getMaxItem();
        tempPredList = new ItemPrediction[maxItem + 1];
        float[] num = new float[maxItem + 1];
        float[] den = new float[maxItem + 1];

        for (int u = 0; u <= maxItem; u++) {
            num[u] = 0;
            den[u] = 0;
        }

        List neighborCoeffs;
        UserInfo activeUserInfo = dataMgr.getUserInfo(activeUser);

        //get the lock here, so you don't have to worry about waiting inbetween
        //your get weight calls and you insure your average matches what you weights
        // are calculated at.
        float activeUserMean = dataMgr.getMeanRatingCached(activeUser);

        if(!activeUserInfo.getReadersLock(true)) return new ItemPrediction[0];


        if (CACHE_LAST_NEIGHBORS) {
            neighborCoeffs = getNeighborsCached(activeUserInfo);
            activeUserInfo.releaseReadersLock();
        } else {
            /* get list of neighbors, calculate weights, and sort in
             * reverse order, so we use the best ones first.
             */
            //Iterator neighbors = getNeighborhood(item);
            //neighborCoeffs = getSimilarityWeights(activeUser, neighbors, item);

        //    neighborCoeffs = getSimilarityWeightsTopN(activeUser, dataMgr.getAllUsers());

        	neighborCoeffs = getSimilarityWeightsTopN(activeUserInfo, dataMgr.getAllUsers());

            activeUserInfo.releaseReadersLock();
            Collections.sort(neighborCoeffs, Collections.reverseOrder());
        }


        int numNeighbors = (int) (neighborCoeffs.size()
                * NEIGHBORHOOD_SIZE_RATIO);

        if (numNeighbors < MIN_NEIGHBORHOOD_SIZE) {
            return new ItemPrediction[0];
        }

        Iterator iter = neighborCoeffs.iterator();
        ItemRating r = null;
        float rating = 0;
        float mean = 0;

        // While there are neighbors in topNCoeff
        while (iter.hasNext()) {
            //Get new neighbor
            SimilarityWeight w = (SimilarityWeight) iter.next();
            float weight = w.getWeight();
            if (w.getNeighborID() == activeUser) {
                continue;
            }

            UserInfo neighborInfo = dataMgr.getUserInfo(w.getNeighborID());
            if(neighborInfo == null) continue;
            if(!neighborInfo.getReadersLock(false)) continue;

            // Get new neighbor's average rating
            mean = neighborInfo.getUserMean();

            int[] itemIDs = neighborInfo.getRatings().ids;
            float[] ratings = neighborInfo.getRatings().ratings;
            for(int i = 0; i < neighborInfo.getRatings().getLength(); i++){
                int index = itemIDs[i];
                num[index] += ((ratings[i] - mean) * weight);
                den[index] += weight;
            }
            neighborInfo.releaseReadersLock();

            // Get neighbor's item list
            /*Iterator i_index = dataMgr.getItemList(w.getNeighborID());

            while (i_index.hasNext()) { //While there are items in list
                r = (ItemRating) i_index.next();
                rating = r.getRating();
                if (rating > 0) {
                    // Get neighbor's weight
                    float weight = w.getWeight();
                    num[r.getItemID()] += ((rating - mean) * weight);
                    den[r.getItemID()] += Math.abs(weight);
                }
            }*/
        }

        tempPredList[0] = new ItemPrediction(0, 0, 0);
        int predIndex = 0;

        for (int i = 1; i < num.length; i++) {
            //Make sure you are not dividing by zero and activeUser has not already rated item
            if (num[i] > 0 && den[i] > 0 &&
                    dataMgr.getRatingCached(activeUser, i) == dataMgr.getErrRating()) {
                // Since the user's mean rating is constant for every prediction, it doesn't help
                // us to differentiate between items. So we use the meanItemRating instead.
                // Unfortunately, this means it could be inconsistent with the normal getPrediction()
                //tempPredList[predIndex++] = new ItemPrediction(0, i,
                //dataMgr.getMeanItemRatingCached(i) + num[i] / den[i]);
                tempPredList[predIndex++] = new ItemPrediction(0, i,
                        activeUserMean + num[i] / den[i]);
            } else {
//            	System.out.println("It is not possible to measure the rate of item "
//            		+ i + " for user " + activeUser);
            }
        }

        Arrays.sort(tempPredList, 0, predIndex, new CompPrediction(dataMgr));

        if (predIndex > n) {
            predIndex = n;
        }
        itemPredList = new ItemPrediction[predIndex];
        for (int i = 0; i < itemPredList.length; i++) {
            if (tempPredList[i].getPrediction() > dataMgr.getMaxRating()) {
                tempPredList[i] = new ItemPrediction(tempPredList[i].getUserID(),
                        tempPredList[i].getItemID(),
                        dataMgr.getMaxRating());
            }
            if (tempPredList[i].getPrediction() < dataMgr.getMinRating()) {
                tempPredList[i] = new ItemPrediction(tempPredList[i].getUserID(),
                        tempPredList[i].getItemID(),
                        dataMgr.getMinRating());

            }
            itemPredList[i] = tempPredList[i];
        }
        return itemPredList;
    }

    /**
     * Predicts the rating for the given user and item.
     * @param activeUser
     * @param item
     * @return predictedRating
     */
    public ItemPrediction predictRating(int activeUser, int item) throws
            CFNotImplementedException {
        float numerator = 0;
        float denominator = 0;

        long startTime = System.currentTimeMillis();

        List neighborCoeffs;

        UserInfo activeUserInfo = dataMgr.getUserInfo(activeUser);
        //get the lock here, so you don't have to worry about waiting inbetween
        //your get weight calls and you insure your average matches what you weights
        // are calculated at.
        if(!activeUserInfo.getReadersLock(true)) return new ItemPrediction(activeUser,item,dataMgr.getErrRating());

        float activeUserMean = dataMgr.getMeanRatingCached(activeUser);
        if (CACHE_LAST_NEIGHBORS) {
            neighborCoeffs = getNeighborsCached(activeUserInfo);
            activeUserInfo.releaseReadersLock();
        } else {
            /* get list of neighbors, calculate weights, and sort in
             * reverse order, so we use the best ones first.
             */
            //Iterator neighbors = getNeighborhood(item);
            //neighborCoeffs = getSimilarityWeights(activeUser, neighbors, item);

            neighborCoeffs = getSimilarityWeightsTopN(
                    activeUserInfo, dataMgr.getAllUsers());
            activeUserInfo.releaseReadersLock();
            Collections.sort(neighborCoeffs, Collections.reverseOrder());
        }

        int numNeighbors = (int) (neighborCoeffs.size()
                * NEIGHBORHOOD_SIZE_RATIO);

        if (numNeighbors < MIN_NEIGHBORHOOD_SIZE) {
            return new ItemPrediction(activeUser, item, dataMgr.getErrRating());
        } else if (MAX_NEIGHBORHOOD_SIZE > 0 &&
                numNeighbors > MAX_NEIGHBORHOOD_SIZE) {
            numNeighbors = MAX_NEIGHBORHOOD_SIZE;

            // now, calculate the prediction based on each neighbor
        }
        Iterator iter = neighborCoeffs.iterator();
        int neighborsUsed = 0;

        while (iter.hasNext() && neighborsUsed <  numNeighbors) {
            SimilarityWeight w = (SimilarityWeight) iter.next();
            int neighbor = w.getNeighborID();
            // HACK -- Avoid correlations with ourself. (DLowd)
            if (neighbor == activeUser) {
                continue;
            }

            float rating = dataMgr.getRatingCached(neighbor, item);
            float neighborMean = dataMgr.getMeanRatingCached(neighbor);
            float weight = w.getWeight();

            if (rating == dataMgr.getErrRating()) {
                continue;
            }

            numerator += (weight * (rating - neighborMean));
            denominator += Math.abs(weight);

            // Select only the best neighbors.
            neighborsUsed++;
        }

        // start the final calc!

        float retval;
        if (denominator != 0 && neighborsUsed >= MIN_NEIGHBORHOOD_SIZE) {

            retval = activeUserMean + (numerator / denominator);

            if (retval > dataMgr.getMaxRating()) {
                retval = dataMgr.getMaxRating();
            } else if (retval < dataMgr.getMinRating()) {
                retval = dataMgr.getMinRating();
            }
        } else {
            retval = dataMgr.getErrRating();

        }

        long stopTime = System.currentTimeMillis();

        NumPredictions++;
        Time += (stopTime - startTime);
        timeStats.addNum(stopTime - startTime);
        
        ItemPrediction predictedRating = new ItemPrediction(activeUser, item, retval);
        return predictedRating; // return predicted rating
    }

    /**
     * Return average prediction time.
     * @return average prediction time
     */
    public long getAveragePredictionTime() throws CFNotImplementedException {
        return (Time / NumPredictions);
    }

    /**
     * Wrapper for getWeight(UserInfo, int, int) that first calls datamanager to lookup the UserInfo structure
     * @param activeUser integer id of active user
     * @param neighbor integer id of neighbor
     * @param itemId integer id of the current item being predicted (so the active user's rating is ignored)
     * @return The SimilarityWeight between the two users.
     */
  /*  public SimilarityWeight getWeight(int activeUser, int neighbor, int itemId) {
        UserInfo userInfo = dataMgr.getUserInfo(activeUser);
        return getWeight(userInfo, neighbor, itemId);
    }    */

    /**
     * Calculate CorrelationCoeff for an item between the activeuser and one neighbor.
     * This method is the bottleneck of this algorithm - do not add anything unnecessary -
     * Remove all logging or debugging statements after you are done. Even a harmless
     * if statement can slow things down.
     *
     * @param neighbor
     * @param itemId
     * @return SimilarityWeight
     */
    public SimilarityWeight getWeight(RatingArrays activeUserRatings,
                                      int neighbor, int itemId) {

        //****Declarations*****************************************
        float weight = 0;

        float combinedSum = 0;
        float activeSum = 0;
        float activeSqrSum = 0;
        float neighborSum = 0;
        float neighborSqrSum = 0;
        int numCommonItems = 0;

        //     RatingArrays activeUserRatings;

        // Get active user and neighbor average rating and standard deviation
        //if(!activeUserInfo.getReadersLock(true)) return new SimilarityWeight(neighbor,0);

        /*
        synchronized (activeUserInfo) {
            // Synchronize to ensure that ratings are consistent
            // This synchronized block does decrease the performance of the algorithm
            // by a noticable amount 160 topns/second -> 150 topns/second on a
            // 2.2ghz P4.
            activeUserRatings = activeUserInfo.getRatings();
        }
        */

        UserInfo neighborUserInfo = dataMgr.getUserInfo(neighbor);
        RatingArrays neighborRatings;

        if (neighborUserInfo == null) {
            return new SimilarityWeight(neighbor, 0);
        }

        if(!neighborUserInfo.getReadersLock(false)) return new SimilarityWeight(neighbor, 0);

        //synchronized (neighborUserInfo) {
            neighborRatings = neighborUserInfo.getRatings();
        //}

        int activeUserNumRatings = activeUserRatings.getLength();
        int nborNumRatings = neighborRatings.getLength();

        int p = 0;
        int q = 0;

        /* Iterate through the ratings of the active and neighbor users,
         * considering each item that has been rated by both.
         * Because the ratings are ordered by item ID, we can do this search
         * in O(n) rather than O(n^2).
         */
        while (p < activeUserNumRatings && q < nborNumRatings) {
            int activeUserItemID = activeUserRatings.ids[p];
            int nborItemID = neighborRatings.ids[q];

            // Don't use ratings for the actual item we're predicting!
            if (activeUserItemID == itemId) {
                p++;
            } else if (activeUserItemID < nborItemID) {
                p++;
            } else if (activeUserItemID > nborItemID) {
                q++;
            } else { // equal

                float activeRating   = activeUserRatings.ratings[p];
                float neighborRating = neighborRatings.ratings[q];

                //Sum components
                activeSum   += activeRating;
                neighborSum += neighborRating;
                activeSqrSum   += (activeRating * activeRating);
                neighborSqrSum += (neighborRating * neighborRating);
                combinedSum += (activeRating * neighborRating);

                p++;
                q++;
                numCommonItems++;
            }
        }
        //activeUserInfo.releaseReadersLock();
        neighborUserInfo.releaseReadersLock();

        // 0 or 1 item is not enough to compute a correlation.
        if (numCommonItems < 2) {
            return new SimilarityWeight(neighbor, 0);
        }

        float activeUserMean = activeSum / numCommonItems;
        float neighborMean   = neighborSum / numCommonItems;

        //Matt McLaughlin
        //Compute weight (Pearson correlation coefficient) as found in
        //"Applied Linear Regression, Second Edition", Sanford Weisberg, p.9
        weight = (combinedSum - numCommonItems * activeUserMean * neighborMean)
                / (float) Math.sqrt((activeSqrSum - numCommonItems * activeUserMean * activeUserMean)
                * (neighborSqrSum - numCommonItems * neighborMean * neighborMean));

        // Theoretically, the weight should never be NaN, nor more than 1;
        // however, floating point error sometimes leads to strange situations.
        if (Float.isNaN(weight)) {
            weight = 0;
        } else if (weight > 1) {
            weight = 1;
        }

        // Apply correlation threshold
        if (weight < MIN_CORRELATION) {
            weight = 0;
        }

        // If too few ratings then the correlation is irrelevant => set to zero
        int activeNumRatings = activeUserRatings.getLength();
        if (numCommonItems < MIN_COMMON_ITEMS
                && numCommonItems < HORT_RATIO * activeNumRatings) {
            weight = 0;
        }

        // Penalize correlations with fewer ratings if below the cutoff
        else if (numCommonItems < MAX_COMMON_ITEMS) {
            weight *= ((float) numCommonItems) / MAX_COMMON_ITEMS;
        }

        return (new SimilarityWeight(neighbor, weight));
    }

    /**
     * Return a List of SimilarityWeights for an item between the activeUser and each of
     * his/her neighbors.
     * @param activeUser
     * @param neighbors
     * @param itemId
     * @return List of SimilarityWeights
     */
 /*   private List getSimilarityWeights(int activeUser, Iterator neighbors,
                                      int itemId) {

        LinkedList vals = new LinkedList();

        while (neighbors.hasNext()) {
            ItemRating neighborInitial = (ItemRating) neighbors.next();
            SimilarityWeight w = getWeight(activeUser, neighborInitial.getUserID(), itemId);
            System.out.println("alg.getSimilarityWeight. after getWeight user = "+activeUser + " neighbor " + neighborInitial.getUserID());
            if (w.getWeight() > 0) {
                vals.add(w);
            }
        }

        return vals;
    }  */

    /**
     * Method to find neighborhood of users for input user.
     * @param itemID
     * @return Iterator of neighborhood
     */
    private Iterator getNeighborhood(int itemID) {
        // Neighborhood includes all other users who have rated this item
        Iterator iter = dataMgr.getNeighborhood(itemID);

        return iter;
    }

    /**
     * Return a List of SimilarityWeight between activeUser and neighbors.
     * @param activeUser
     * @param neighbors
     * @return List of SimilarityWeight
     */
 /*   private List getSimilarityWeightsTopN(int activeUser, Iterator neighbors) {
        LinkedList vals = new LinkedList();
        Integer n;
UserInfo activeUserInfo = dataMgr.getUserInfo(activeUser);
        while (neighbors.hasNext()) {
            n = (Integer) neighbors.next();

            SimilarityWeight w = getWeight(activeUserInfo, n.intValue(), -1);
            if (w.getWeight() > 0) {
                vals.add(w);
            }
        }
        return vals;
    }*/

	private List getSimilarityWeightsTopN(UserInfo activeUserInfo, Iterator neighbors) {
        LinkedList vals = new LinkedList();
        Integer n;
        //UserInfo activeUserInfo = dataMgr.getUserInfo(activeUser);
        RatingArrays activeUserRatings = activeUserInfo.getRatings();

		// long st = System.currentTimeMillis();
		while (neighbors.hasNext()) {
            n = (Integer) neighbors.next();
            if(n.intValue() == activeUserInfo.userID) continue;
            // SimilarityWeight w = getWeight(activeUserInfo, n.intValue(), -1);
            SimilarityWeight w = getWeight(activeUserRatings, n.intValue(), -1);
            if (w.getWeight() > 0) {
                vals.add(w);
            }
        }
//		long et = System.currentTimeMillis();
//		System.out.println("Time to get SimilarityWeightsTopN = " + (et-st));
		return vals;
    }

    /**
     * Doesn't currently need upateUser.
     */
    public void updateUser(int userID) throws CFNotImplementedException {
        //throw new CFNotImplementedException("update User not implemented yet");
    }

    /**
     * Return the name of this algorithm.
     * @return String
     */
    public String toString() {
        return "SimplePearsonAlgorithm.java\n";
    }
	/**
	 * @return
	 */
	public Stats getTimeStats() {
		return timeStats;
	}

}
