/* 
 * TestStats.java
 * 
 * Version 1.1
 * 
 * 23/08/2002
 *
 * Copyright (c) 2003 Oregon State Universtity. All rights reserved.
 */

package org.recommender.algorithms;

import org.recommender.server.*;
/**
 * This class provides methods to store and print out experiment results.
 *
 * @author Daniel Lowd
 * 
 */
public class TestStats
{
    /**
     * Number of ratings predicted.
     */
    int numRatings = 0;

    /**
     * Number of ratings we could not predict
     */
    int numFailed = 0;

    float totalDistance = 0;
    float totalDistSquared = 0;

    /**
     * UserID of the last rating added.
     * (We assume that all ratings from one user are grouped together.)
     */
    int lastUserID = 0;

    /**
     * Statistics for the current user, allowing us to do averages
     * over all users rather than all ratings.
     */
    float userDistance = 0;
    float userDistSquared = 0;
    float totalUserDistance = 0;
    float totalUserDistSquared = 0;

    int numUserRated = 0;
    int numUserFailed = 0;
    float totalUserCoverage = 0;

    /**
     * Total number of users tested.
     */
    int usersTested = 0;
    
    /**
     * Values used in Breese computation.
     */
	private float breeseNumerator, breeseDenominator, usersTestedInBreese;

    /**
     * Keep track of successful and failed predictions.
     * Rows represent actual rating; columns represent
     * predicted rating.  The column MAX_RATING+1 represents
     * ratings that the algorithm could not predict.
     */
    int[][] confusionMatrix;

    DataManager dataMgr;
    int MAX_RATING;

    /**
     * Constructor, initialize a TestStats object.
     * @param dataMgr the DataManager to get data from
     */
    public TestStats(DataManager dataMgr)
    {
        this.dataMgr = dataMgr;
        MAX_RATING = (int)dataMgr.getMaxRating();

        // Initialize confusion matrix
        confusionMatrix = new int[MAX_RATING+1][MAX_RATING+2];
        for (int i = 0; i <= MAX_RATING; i++)
            for (int j = 0; j <= MAX_RATING+1; j++)
                confusionMatrix[i][j] = 0;
    }

    /**
     * When a new experiment result comes, update result stats.
     * @param userID
     * @param itemID
     * @param actualRating
     * @param predictedRating
     */
    public void addRating(int userID, int itemID, float actualRating,
            float predictedRating)
    {
        // When we switch users...
        if (lastUserID != userID)
        {
            /* Add in this user's avg distance and coverage, so we can
             * compute the mean.
             */
            if (numUserRated > 0)
            {
                 totalUserDistance += userDistance/numUserRated;
                 totalDistSquared += userDistSquared/numUserRated;
                 usersTested++;
            }

            if (numUserRated + numUserFailed > 0)
                totalUserCoverage += (float)numUserRated/(numUserRated +
                        numUserFailed);

            lastUserID = userID;
            numUserRated = 0;
            numUserFailed = 0;
            userDistance = 0;
            userDistSquared = 0;
        }

        if (predictedRating == dataMgr.getErrRating())
        {
            // Update confusion matrix
            try {
            	confusionMatrix[(int)actualRating][MAX_RATING+1]++;
            } catch (Exception e) {
            	System.out.println("actualRat = " + actualRating);
				System.out.println("max_rating = " + MAX_RATING);
            }

            numUserFailed++;
            numFailed++;
        }
        else
        {
            float err = Math.abs(actualRating-predictedRating);
            userDistance += err;
            userDistSquared += err*err;
            totalDistance += err;
            totalDistSquared += err*err;
            numUserRated++;
            numRatings++;

            // Update confusion matrix
            try {
            	confusionMatrix[(int)actualRating][(int)(predictedRating + 0.50)]++;
            } catch (ArrayIndexOutOfBoundsException e2) {
            	System.out.println("Erro em TestStats (actualRating, predicted) = " + actualRating + " , " + predictedRating);
            }
        }
    }

    /**
     * Convert the major results into a String.
     * @return retVal String of results
     */
    public String toString()
    {
        String retVal = "";

        retVal += "MAE: " + getMAE() + "\n";
        retVal += "MSE: " + getMSE() + "\n";
        retVal += "Breese: " + getBreese() + "\n";
        retVal += "Coverage: " + getCoverage() + "\n";
        retVal += "Rated: " + numRatings + "; Failed: " + numFailed + "\n";
        retVal += "Number of users: " + usersTested + "\n";
        retVal += "\n";

        for (int i = 0; i <= MAX_RATING; i++)
        {
            for (int j = 0; j <= MAX_RATING + 1; j++)
                retVal += confusionMatrix[i][j] + "\t";
            retVal += "\n";
        }

        int correct = 0;
        for (int i = 0; i <= MAX_RATING; i++)
            correct += confusionMatrix[i][i];

        retVal += "Correct: " + ((float)correct)/numRatings;

        return retVal;
    }

    /**
     * Accessor, return MAE
     * @return MAE
     */
    public float getMAE()
    {
        if (numRatings == 0)
            return -1;

        return totalDistance/numRatings;
    }

    /**
     * Accessor, return MSE
     * @return MSE
     */
    public float getMSE()
    {
        if (numRatings == 0)
            return -1;

        return totalDistSquared/numRatings;
    }

    /**
     * Accessor, return Coverage
     * @return Coverage
     */
    public float getCoverage()
    {
        return (float)numRatings/(numRatings + numFailed);
    }

    /**
     * Accessor, return average MAE
     * @return Average MAE
     */
    public float getAvgMAE()
    {
        if (usersTested == 0)
            return -1;

        return totalUserDistance/usersTested;
    }

    /**
     * Accessor, return average MSE
     * @return Average MSE
     */
    public float getAvgMSE()
    {
        if (usersTested == 0)
            return -1;

        return totalUserDistSquared/usersTested;
    }

    /**
     * Accessor, return average coverage
     * @return Average coverage
     */
    public float getAvgCoverage()
    {
        return totalUserCoverage/usersTested;
    }

	public void addBreeseValues(float rMax, float rVal) {
		// Add up the ratio, so we can calc the mean later
		if (rMax > 0) {
			breeseNumerator += rVal;
			breeseDenominator += rMax;
			usersTestedInBreese++;
		}
	}
	
	public float getBreese() {
		return ((usersTestedInBreese>0) 
					? (100 * breeseNumerator / breeseDenominator) : -1);
	}
}
