/**
 * This class provides methods for the client to access the CFEngine server.
 * All calls to the CFEngine server should come from this class.
 * We are using a singleton class pattern - only one instance of this class
 * is expected to exist at any time. In theory, this is cleaner OO programming.
 * For example, apparently when you subclass, you can't really override static methods,
 * you only hide them.
 *
 * Created by
 * User: herlock
 * Date: Apr 15, 2003
 * Time: 10:02:21 AM
 */
package org.recommender.clients.console;

import org.recommender.server.*;

import java.rmi.Naming;
import java.rmi.RemoteException;

/**
 * A wrapper class for the CFEngine interface. The console client
 * makes requests to this class, which then passes the requests on
 * to the client. All references to the CFEngine are isolated in
 * this class.
 *
 * @author Jon Herlocker
 * @see CFEngine
 */
public class ClientCFManager {
    private static ClientCFManager ourInstance;
    private CFEngine server = null;    // RMI object to access cfengine
    private ClientDBManager dbManager;
    private String cfEngineHost = null;

    /**
     * @return A singleton ClientCFManager instance
     */
    public synchronized static ClientCFManager getInstance(String serverHostName) {
        if (ourInstance == null) {
            ourInstance = new ClientCFManager(serverHostName);
        }
        return ourInstance;
    }

    /**
     * The constructor is private, since this is a singleton
     * class. Only the static getInstance() method should be
     * used to construct this class.
     */
    private ClientCFManager(String host) {
        dbManager = ClientDBManager.getInstance();

        if (host == null) {
            cfEngineHost = "localhost";
        } else {
            cfEngineHost = host;
        }
        connectToServer();

    }

    /**
     * Lookup the cfengine service via RMI on the host identified by the cfEngineHost field,
     * and execute the test() method on that remote object.
     */
    private void connectToServer() {
        // When a server starts that wants to be accessible via RMI,
        // it registers itself as a service with an RMI nameserver. Clients that
        // want to connect to that service contact the nameserver to determine
        // what host and port the desired service is running on.
        //    The cfengine service is known by the name "cfengine". This code
        // connects to the RMI name server to find out where the cfengine
        // service is running.

        String url = "rmi://" + cfEngineHost + "/";
        String bindName = SystemProperties.getString("CFClient.rmi.objectname");
        try {
            //server = (CFEngine)Naming.lookup(url + "cfengine");
			server = (CFEngine)Naming.lookup(url + bindName);
		
            //
            // server is an object that encapsulates access to the remote
            // cfengine server. When we execute a method on engine, RMI
            // arranges for that same method to be executed on the cfengine
            // server.
            //     In this case, engine.test() is a simple method to check
            // that we are successfully able to communicate with the remote
            // server. It will throw an exception if the remote server is not
            // running.
            String resultString = server.test();
            System.err.println(resultString);
        } catch (Exception e) {
            System.err.println("Error: Cannot connect to service \"cfengine\" via RMI registry on "
                               + cfEngineHost + e);
            e.printStackTrace();
            System.exit(1);
        }
    }


    /**
     * Setter.
     * @param engine The CFEngine interface used to communicate with the server
     */
    public void setServer(CFEngine engine) {
        server = engine;
    }

    /**
     * Wrapper for CFEngine.getItemRatingList
     * @param item
     * @return An array of User objects
     * @see CFEngine
     */

    public User [] getItemRatingList (Item item){
         ItemRating idArray [] = null;

        do {
            try {
                idArray = server.getItemRatingList(item.getID());
                break;
            } catch (RemoteException e) {
                connectToServer();
            } catch (CFIllegalParameterException e) {
                e.printStackTrace();
                System.exit(1);
            }
        } while (true);

        User userArray [] = new User[idArray.length];

        for (int i = 0; i < idArray.length; i++) {
            int id = idArray[i].getUserID();
            userArray[i] = dbManager.clientLookupUser(id);
        }
        return userArray;
    }

    /**
     * Wrapper for CFEngine.getUserRatingList()
     * @param user
     * @return An array of Item objects
     * @see CFEngine
     */

    public Rating [] getUserRatingList (User user) {
        ItemRating[] irArray;

        do {
            try {
                irArray = server.getUserRatingList(user.getID());
                break;
            } catch (RemoteException e) {
                connectToServer();
            } catch (CFIllegalParameterException e) {
                e.printStackTrace();
                System.exit(1);
            }
        } while (true);

        Rating[] ratingArray = new Rating[irArray.length];
        for (int i = 0; i < irArray.length; i++) {
            int id = irArray[i].getItemID();
            ratingArray[i] = new Rating(user, dbManager.clientLookupItem(id), irArray[i].getRating());
        }
        return ratingArray;
    }

    /**
     * Wrapper for CFEngine.getRating()
     * @param user
     * @param item
     * @return A Rating object
     * @see CFEngine
     */
    public Rating getRating(User user, Item item) {
        ItemRating rating = null;


        int tryCount = 0;
        do {
            try {
                rating = server.getRating(user.getID(), item.getID());
                break;
            } catch (RemoteException e) {
                connectToServer();
            } catch (CFIllegalParameterException e) {
                e.printStackTrace();
                System.exit(1);
            } catch (CFInternalErrorException e) {
                // In case the internal error is transient, we try several times
                tryCount++;
                if (tryCount >= 3) {
                    e.printStackTrace();
                    System.exit(1);
                }
            }
        } while (true);

        return new Rating(user, item, rating.getRating());
    }

    /**
     * Wrapper for CFEngine.setRating()
     * @param user
     * @param item
     * @param rating
     * @see CFEngine
     */
    public void addRating(User user, Item item, int rating) {
        int tryCount = 0;
        do {
            try {
                server.setRating(user.getID(), item.getID(), rating);
                return;
            } catch (java.rmi.RemoteException e1) {
                // connnectToServer will terminate the application if it cannot connect to the server
                connectToServer();
            } catch (CFIllegalParameterException e1) {
                // Either a negative id was passed to user or item or rating
                // was outside of the server's configured range
                e1.printStackTrace();
                System.exit(1);
            } catch (CFInternalErrorException e) {
                // In case the internal error is transient, we try several times
                tryCount++;
                if (tryCount >= 3) {
                    e.printStackTrace();
                    System.exit(1);
                }
            }
        } while (true);
    }

    /**
     * Wrapper for CFEngine.setRatingList
     * @param ratings
     * @return the number of rating added
     * @see CFEngine
     */
    public int addRatingList(Rating [] ratings) {
        ItemRating[] newRatingList = new ItemRating[ratings.length];
		int number = 0;

        for (int i = 0; i < ratings.length; i++) {
            Rating r = ratings[i];
            newRatingList[i] = new ItemRating(r.getUser().getID(), r.getItem().getID(), r.getRatingValue());
        }

        int tryCount = 0;
        do {
            try {
                number = server.setRatingList(newRatingList);
                return number;
            } catch (java.rmi.RemoteException e1) {
                // connnectToServer will terminate the application if it cannot connect to the server
                connectToServer();
            } catch (CFIllegalListParameterException e1) {
                // Either a negative id was passed to user or item or rating
                // was outside of the server's configured range
                System.err.println("Only " + e1.getNumSuccessfull() + " ratings were sucessfully removed");
                e1.printStackTrace();
                System.exit(1);
            } catch (CFInternalErrorException e) {
                // In case the internal error is transient, we try several times
                tryCount++;
                if (tryCount >= 3) {
                    e.printStackTrace();
                    System.exit(1);
                }
            }
        } while (true);
    }

    /**
     * Removes a single rating from the CFEngine.
     * @param user
     * @param item
     */
    public void removeRating(User user, Item item) {
        int tryCount = 0;
        do {
            try {
                server.removeRating(user.getID(), item.getID());
                return;
            } catch (java.rmi.RemoteException e1) {
                // connnectToServer will terminate the application if it cannot connect to the server
                connectToServer();
            } catch (CFIllegalParameterException e1) {
                // Either a negative id was passed to user or item or rating
                // was outside of the server's configured range
                e1.printStackTrace();
                System.exit(1);
            } catch (CFInternalErrorException e) {
                // In case the internal error is transient, we try several times
                tryCount++;
                if (tryCount >= 3) {
                    e.printStackTrace();
                    System.exit(1);
                }
            }
        } while (true);
    }

    /**
        * Wrapper for CFEngine.removeRatingList
     * @param user
        * @param items - An array of items for which ratings are to be removed
        * @return the number of rating added
        * @see CFEngine
        */
    public int removeRatingList(User user, Item [] items) {
           int[] idList = new int[items.length];

           for (int i = 0; i < items.length; i++) {
               idList[i] = items[i].getID();
           }

           int tryCount = 0;
           do {
               try {
                   server.removeRatingList(user.getID(), idList);
                   return idList.length;
               } catch (java.rmi.RemoteException e1) {
                   // connnectToServer will terminate the application if it cannot connect to the server
                   connectToServer();
               } catch (CFIllegalListParameterException e1) {
                   // Either a negative id was passed to user or item or rating
                   // was outside of the server's configured range
                   System.err.println("Only " + e1.getNumSuccessfull() + " ratings were sucessfully removed");
                   e1.printStackTrace();
                   System.exit(1);
               } catch (CFInternalErrorException e) {
                // In case the internal error is transient, we try several times
                tryCount++;
                if (tryCount >= 3) {
                    e.printStackTrace();
                    System.exit(1);
                }
            }
           } while (true);
       }

    /**
     * Wrapper for CFEngine.getPredictedRating
     * @param user
     * @param item
     * @return A single Prediction object
     * @see CFEngine
     */
    public Prediction getPredictedRating(User user, Item item) {
        ItemPrediction predictedRating = null;

        int tryCount = 0;
        do {
            try {
                predictedRating = server.getPredictedRating(user.getID(), item.getID());
                return new Prediction(item, predictedRating.getPrediction());
            } catch (RemoteException e) {
                connectToServer();
            } catch (CFIllegalParameterException e) {
                e.printStackTrace();
                System.exit(1);
            } catch (CFInternalErrorException e) {
                // In case the internal error is transient, we try several times
                tryCount++;
                if (tryCount >= 3) {
                    e.printStackTrace();
                    System.exit(1);
                }
            }
        } while (true);
    }

    /**
     * Wrapper for CFEngine.getPredictedRatings()
     * @param user
     * @param itemArray
     * @return An array of Prediction objects
     * @see CFEngine
     */
    public Prediction[] getPredictedRatings(User user, Item[] itemArray) {
        int[] itemIDs = new int[itemArray.length];
        ItemPrediction[] predictedRatingValues = new ItemPrediction[itemArray.length];

        for (int i = 0; i < itemArray.length; i++) {
            Item item = itemArray[i];
            itemIDs[i] = item.getID();
        }
        int tryCount = 0;
        do {
            try {
                predictedRatingValues = server.getPredictedRatingList(user.getID(), itemIDs);
                break;
            } catch (RemoteException e) {
                connectToServer();
            } catch (CFIllegalListParameterException e) {
                e.printStackTrace();
                System.exit(1);
            } catch (CFInternalErrorException e) {
                // In case the internal error is transient, we try several times
                tryCount++;
                if (tryCount >= 3) {
                    e.printStackTrace();
                    System.exit(1);
                }
            }
        } while (true);


        Prediction[] predictedRatings = new Prediction[predictedRatingValues.length];
        for (int i = 0; i < predictedRatingValues.length; i++) {
            predictedRatings[i] = new Prediction(itemArray[i], predictedRatingValues[i].getPrediction());
        }
        return predictedRatings;
    }

    /**
     * Wrapper for CFEngine.getRecommendations
     * @param user
     * @param number
     * @param offset
     * @return An array of Prediction objects
     * @see CFEngine
     */
    public Prediction[] getRecommendations(User user, int number, int offset) {
        ItemPrediction[] recommendations = null;

        long start = System.currentTimeMillis();
        int tryCount = 0;
        do {
            try {
                recommendations = server.getRecommendations(user.getID(), number, offset);
                break;
            } catch (RemoteException e) {
                connectToServer();
            } catch (CFIllegalParameterException e) {
                e.printStackTrace();
                System.exit(1);
            } catch (CFInternalErrorException e) {
                // In case the internal error is transient, we try several times
                tryCount++;
                if (tryCount >= 3) {
                    e.printStackTrace();
                    System.exit(1);
                }
            }
        } while (true);

        long stop = System.currentTimeMillis();
        System.err.println("Clock time for top N recommendation= " + (stop - start));

        Prediction[] returnedRecs = new Prediction[recommendations.length];
        for (int i = 0; i < recommendations.length; i++) {
            Item item = dbManager.clientLookupItem(recommendations[i].getItemID());
            returnedRecs[i] = new Prediction(item, recommendations[i].getPrediction());
        }
        return returnedRecs;
    }

    /**
     * Wrapper for CFEngine.getRecommendationsByType
     * @param user
     * @param type
     * @param number
     * @param offset
     * @return An array of Prediction objects
     */
    Prediction[] getRecommendationsByType(User user, Type type, int number, int offset) {
        ItemPrediction[] recommendations = null;

        int tryCount = 0;
        do {
            try {
                recommendations = server.getRecommendationsByType(
                        user.getID(), number, offset, type.getTypeId());
                break;
            } catch (RemoteException e) {
                e.printStackTrace();
                System.exit(1);
            } catch (CFIllegalParameterException e) {
                e.printStackTrace();
                System.exit(1);
            } catch (CFInternalErrorException e) {
                // In case the internal error is transient, we try several times
                tryCount++;
                if (tryCount >= 3) {
                    e.printStackTrace();
                    System.exit(1);
                }
            }
        } while (true);


        Prediction[] predictions = new Prediction[recommendations.length];
        for (int i = 0; i < recommendations.length; i++) {
            ItemPrediction recommendation = recommendations[i];
            Item item = dbManager.clientLookupItem(recommendation.getItemID());
            predictions[i] = new Prediction(item, recommendation.getPrediction());
        }
        return predictions;
    }

    /**
     * Query the server to determine the minimum rating value supported
     * for the current data.
     * @return return minimum rating value from the user
     */
    public float getMinRating() {
        float rating = -1;

        do {
            try {
                rating = server.getMinRating();
                return rating;
            } catch (java.rmi.RemoteException e) {
                connectToServer();
            }
        } while (true);
    }

    /**
     * Query the server to determine the maximum rating value supported
     * for the current data.
     * @return    return maximum rating value from the user
     */
    public float getMaxRating() {
        do {
            try {
                return server.getMaxRating();
            } catch (java.rmi.RemoteException e) {
                connectToServer();
            }
        } while (true);
    }

    /**
     * Query the server to determine the rating value that is returned
     * to indicate that the system has no rating for an item or
     * cannot produce a prediction for an item .
     * @return    return the default value indicating no rating or no prediction
     */
    public float getErrRating() {
        do {
            try {
                return server.getErrRating();
            } catch (java.rmi.RemoteException e) {
                connectToServer();
            }
        } while (true);
    }
}

