package org.recommender.clients.console;

import org.recommender.server.SystemProperties;

import java.sql.*;
import java.util.Vector;

/**
 * Created by
 * User: herlock
 * Date: Apr 16, 2003
 * Time: 8:49:26 AM
 */
public class ClientDBManager {
    private static ClientDBManager ourInstance;

    private Connection conn = null; // JDBC connetion class.

    // Maps text user names to cfengine userids
    private static String userTableName = ClientConstants.USER_TABLE;

        // Name of username field in user table
    private static String userTableNameField = ClientConstants.USER_TABLE_FIELD_NAME;

           // Name of id field in user table
    private static String userTableIdField = ClientConstants.USER_TABLE_FIELD_ID;

    // Maps text user names to cfengine userids
    private static String itemTableName = ClientConstants.ITEM_TABLE;

            // Name of username field in user table
    private static String itemTableNameField = ClientConstants.ITEM_TABLE_FIELD_NAME;

           // Name of id field in user table
    private static String itemTableIdField = ClientConstants.ITEM_TABLE_FIELD_ID;

    // Used to store the SQL query that is sent to the DB in many places.
    private Statement stmt = null;

    // Name of table mapping type ids to names
    private static String typeTableName = ClientConstants.TYPE_TABLE;

    // Field of above table for the type id
    private static String typeTableIdField = ClientConstants.TYPE_TABLE_FIELD_ID;

    // Field of above table for the type name
    private static String typeTableNameField = ClientConstants.TYPE_TABLE_FIELD_NAME;

    public synchronized static ClientDBManager getInstance() {
        if (ourInstance == null) {
            ourInstance = new ClientDBManager();
        }
        return ourInstance;
    }

    private ClientDBManager() {
    }

     /**
         Get connection with database and cache it in a private instance
         variable. Upon return from this function, there should be a 99%
         guarantee that the connection is still valid.

         // @todo - should also check to see if cached connection still works

         @throws java.sql.SQLException
         */
    public void getConnection() throws SQLException {
            if (conn != null) {
                return;
            } else {
                // @todo We need to create a separate parameters file for the client
                String drivers = SystemProperties.getString("CFClient.jdbc.drivers");
                if (drivers != null)
                    System.setProperty("jdbc.drivers", drivers);

                // the jdbc url should encode the type of database software, the hostname,
                // and the name of the actual database name in one string.
                String url = SystemProperties.getString("CFClient.jdbc.url");
                String username = SystemProperties.getString("CFClient.jdbc.username");
                String password = SystemProperties.getString("CFClient.jdbc.password");

                conn = DriverManager.getConnection(url, username, password);

                return;
            }

        }

    /**
     * Gets a list of all the types that the client is currently aware of
     * @return - An array of Types.
     */
    public Vector retriveAllTypes() {
        String query = "SELECT " + typeTableIdField + ", " + typeTableNameField
            + " FROM " + typeTableName;
        ResultSet rs = null;

        // Establish connection with databases
        try {
            getConnection();
            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);
        } catch (SQLException e) {
            System.out.println("SQLException:");
            while (e != null) {
                System.out.println("SQLState: " + e.getSQLState());
                System.out.println("Message: " + e.getMessage());
                System.out.println("Vendor: " + e.getErrorCode());
                e = e.getNextException();
                System.out.println("");
            }
            System.exit(1);
        }

        Vector typeVector = new Vector();
        try {
            while (rs != null && rs.next()) {
                int id = rs.getInt(1);
                String name = rs.getString(2);
                typeVector.add(new Type(id, name));
            }
        } catch (SQLException e) {
            System.out.println("SQLException:");
            System.exit(1);
        }
     return typeVector;
    }

    /**
       * Check whether there is a client-side record for an item with this server id.
       * @param itemId - cfengine server itemid for the item
       * @return a populated Item object
       */
      public Item clientLookupItem(int itemId) {
          Item matchedItem = null;
          int id = 0;
          String name;

          String statement = "SELECT * FROM " + itemTableName
                  + " WHERE " + itemTableIdField + " = ?";
          PreparedStatement ps;                  // JDBC statment
          ResultSet rs;                          // JDBC results
          try {
              getConnection();
              ps = conn.prepareStatement(statement);
              ps.setInt(1, itemId);
              rs = ps.executeQuery();
              boolean itemExists = rs.next();

              if (itemExists) {
                  id = rs.getInt(itemTableIdField);
                  name = rs.getString(itemTableNameField);
                  matchedItem = new Item(id, name);
              }
          } catch (Exception ex) {
              System.err.println("Exception while accessing database: " + ex.toString());
              System.exit(1);
          }

          if (matchedItem == null) {
              matchedItem = new Item(id);
          }

          return matchedItem;
      }

      /**
       * Check whether there is a Item with this name.
       * @param itemName
       * @return a populated Item object
       * @throws ClientNoSuchItemException
       */
      public Item clientLookupItem(String itemName) throws ClientNoSuchItemException {
          Item matchedItem = null;

          int id;
          String name;

          String statement = "SELECT * FROM " + itemTableName
                  + " WHERE " + itemTableNameField + " = ?";

          PreparedStatement ps;                  // JDBC statment
          ResultSet rs;                          // JDBC results

          try {
              getConnection();
              ps = conn.prepareStatement(statement);
              ps.setString(1, itemName);
              rs = ps.executeQuery();
              boolean itemExists = rs.next();

              if (itemExists) {
                  id = rs.getInt(itemTableIdField);
                  name = rs.getString(itemTableNameField);
                  matchedItem = new Item(id, name);
              }
          } catch (Exception ex) {
              System.err.println("Exception while accessing database: " + ex.toString());
              System.exit(1);
          }

          if (matchedItem == null) {
              throw new ClientNoSuchItemException();
          }

          return matchedItem;
      }


    /**
     * Check whether there is a Item with this name. Exactly one of the parameters is used.
     *
     * @param typeName - name of the type. Null indicates id should be used instead.
     * @param id - id of the type. Ignored if typename is non-null.
     * @return a populated Type object
     */
    public Type clientLookupType(String typeName, int id) {
        Type matchedType;

        String statement = "SELECT * FROM " + typeTableName
                + " WHERE ";
        if (typeName == null) {
            statement = statement + typeTableIdField;
        } else {
           statement = statement + typeTableNameField;
        }
        statement = statement + " = ?";

        PreparedStatement ps;                  // JDBC statment
        ResultSet rs;                          // JDBC results

        matchedType = null;
        try {
            getConnection();
            ps = conn.prepareStatement(statement);
            if (typeName == null) {
                ps.setInt(1, id);
            } else {
                ps.setString(1, typeName);
            }
            rs = ps.executeQuery();

            if (rs.next()) {
                int foundId = rs.getInt(typeTableIdField);
                String foundName = rs.getString(typeTableNameField);
                matchedType = new Type(foundId, foundName);
            }
        } catch (Exception ex) {
            System.err.println("Exception while accessing database: " + ex.toString());
            System.exit(1);
        }

        return matchedType;
    }

    public Type clientLookupType(String typeName) {
        if (typeName == null) {
            return null;
        }
        // Second parameter will be ignored.
        return clientLookupType(typeName, 0);
    }

     public Type clientLookupType(int id) {
        return clientLookupType(null, id);
    }

        /**
     * Fill an array with all the current usernames from the client
     * database.
     */
        public Vector retrieveUserVector() {
        Vector userVector = new Vector();  // vector to hold the usernames
        PreparedStatement ps;                  // JDBC statment
        ResultSet rs;                          // JDBC results
        String statement = "SELECT " + userTableIdField + ", " + userTableNameField + " FROM " + userTableName;
        try {
            getConnection();
            ps = conn.prepareStatement(statement);
            rs = ps.executeQuery();
            while (rs.next()) {
                int id = rs.getInt(1);
                String name = rs.getString(2);
                User user = new User(id, name);
                userVector.add(user);
            }
        } catch (Exception ex) {
            System.err.println("Exception while accessing database: " + ex.toString());
            System.exit(1);
        }

        return userVector;
    }

    /**
       * Returns a User object with all User information
       *
       * @param userName Client-side username.
       * @return new User - null if userName does not exist
       */
      public User clientLookupUser(String userName) {
          // second parameter is ignored.
          if (userName == null) {
              return null;
          }
          return clientLookupUser(userName, 0);
      }

      /**
       * Returns a User object with all User information
       *
       * @param userId server-side userId.
       * @return new User - null if userId does not exist in client database
       */
      public User clientLookupUser(int userId) {
          return clientLookupUser(null, userId);
      }

      /**
       * Returns a User object with all User information
       * returns NULL if invalid user or password is incorrect. If
       * userName is null, the lookup is done by userId, otherwise
       * userId is ignored.
       * @param userName - null if this is a lookup by userID
       * @param userId - ignored if userName is non-null
       * @return new User.
       */
      public User clientLookupUser(String userName, int userId) {
          PreparedStatement ps;                  // JDBC statment
          ResultSet rs = null;                          // JDBC result

          String statement = "SELECT * " + " FROM " + userTableName + " WHERE ";

          // @todo broken prepared statement - not sure why. Should be ?
          if (userName != null) {
              statement = statement + userTableNameField + " = '" + userName + "'";
          } else {
              statement = statement + userTableIdField + " = " + userId;
          }
          //print(statement);

          try {
              getConnection();
              ps = conn.prepareStatement(statement);
              if (userName != null) {
                  // @todo broken prepared statement and I don't know why.
                  //ps.setString(1, userName);
              } else {
                  //ps.setInt(1, userId);
              }
              rs = ps.executeQuery(statement);
          } catch (Exception e) {
              System.out.println("Fatal DB error " + e);
              e.printStackTrace();  //To change body of catch statement use Options | File Templates.
              System.exit(1);
          }

          try {
              while (rs != null && rs.next()) {
                  return new User(rs.getInt("UserID"),
                          rs.getString("UserName"));
              }
          } catch (SQLException e) {
              System.out.println("Fatal DB error " + e);
              e.printStackTrace();  //To change body of catch statement use Options | File Templates.
              System.exit(1);
          }

          return null;
      }

      /**
     * Adds the user to the database and returns the assigned userid for
     * this user.
     * @param user
     * @return userID - returns -1 if unable to add the user user.
     */
    public int clientAddNewUser(User user) {
        String command;
        ResultSet rs;
        String value = "'" + user.getUserName() + "'";

          // The following is mysql dependent - we have an autoincrement
          // column.
        command = "INSERT INTO " + userTableName
                + " VALUES (NULL , " + value + ")";

        try {
            getConnection();
            stmt = conn.createStatement();
            stmt.executeUpdate(command);
        } catch (SQLException e) {
            try {
                getConnection();
                stmt = conn.createStatement();
                stmt.executeUpdate(command);
            } catch (Exception ex) {
                System.err.println("Fatal DB error: " + ex);
                System.exit(1);
            }
        }

        rs = null;
        command = "SELECT " + userTableIdField + " FROM "
                + userTableName
                + " WHERE " + userTableNameField + " = '" + user.getUserName() + "'";

        try {
            getConnection();
            stmt = conn.createStatement();
            rs = stmt.executeQuery(command);
        } catch (SQLException e) {
            try {
                getConnection();
                stmt = conn.createStatement();
                rs = stmt.executeQuery(command);
            } catch (Exception ex) {
                System.err.println("Fatal DB error: " + ex);
            }
        }

        try {
            if (rs != null && rs.next()) {
                return rs.getInt(userTableIdField);
            }
        } catch (SQLException e) {
            System.err.println("Fatal db exception: " + e);
            System.exit(1);
        }

        return -1;
    }

      /**
     * Returns a vector of names of all items that the client knows about.
     * @return A vector of Strings - each the name of one item
     */
    public Vector clientRetrieveAllItems() {
        Vector items = new Vector();

        PreparedStatement ps;                  // JDBC statment
        ResultSet rs;                          // JDBC results
        String statement = "SELECT " + itemTableNameField + ", " + itemTableIdField
                + " FROM " + itemTableName;
        try {
            getConnection();
            ps = conn.prepareStatement(statement);
            rs = ps.executeQuery();
            while (rs.next()) {
                items.add(new Item(rs.getInt(2), rs.getString(1)));
            }
        } catch (Exception ex) {
            System.err.println("Exception while accessing database: " + ex);
            System.exit(1);
        }

        return items;
    }
}
