/*
* ClientTextUI.java
*
* Version 1.1
*
* 15/08/2002
*
* Copyright (c) 2002 Oregon State Universtity. All rights reserved.
 */


package org.recommender.clients.console;

import org.recommender.server.*;

import java.util.Vector;

/**
 * This class contains the text based Client session.
 *
 * @version 1.14 07/08/2002
 * @author Irwin Yoon
 * @author Jon Herlocker
 */
public class ClientTextUI {
    private User curUser = null;

    // The ClientDBManager handles all connections to the DB.
    private ClientDBManager dbManager;

     // These hold range of allowable ratings for items. Retrieved from the server
    private float minRating;
    private float maxRating;

    // @todo errRating not used - we should probably check the results retuned by predictions for errRating
    private float errRating = 0;

    private ClientCFManager cfManager;

    /**
     * Constructor, Create access to database from the client side.
     */
    public ClientTextUI(ClientCFManager cfmanager) {
        cfManager = cfmanager;
        minRating = cfManager.getMinRating();
        maxRating = cfManager.getMaxRating();
        errRating = cfManager.getErrRating();

        // ClientDBManager is a singleton
        dbManager = ClientDBManager.getInstance();
    }

    public void setCFManager(ClientCFManager cfmanager) {
        this.cfManager = cfmanager;
    }

       /**
     * Simple methods to add a level of indirection to output
     * Also reduces the numbers of characters to type.
     * (compare to System.out.println())
     */
    public static void print(String s) {
        System.out.println(s);
    }

    private static void print() {
        System.out.println();
    }

    /**
     * Start the text based Client session.
     */
    public void menuStartSession() {
        // This will loop until endSession() is called, which
        // calls System.Exit();
        while (true) {
            print();
            print("Welcome to the CF Prediction Engine");
            print("===================================");
            print("Please choose one option:");
            print("1. Create a New User");
            print("2. Existing User: View Ratings/Recommendations or Send Ratings");
            print("3. Exit");

            // One constant for each menu choice
            final int NEW_USER = 1;
            final int EXISTING_USER = 2;
            final int EXIT = 3;

            int choice; // Which menu item the user selects

            do {
                choice =
                        Console.readInt("Enter choice: ");
            } while (choice != NEW_USER &&
                    choice != EXISTING_USER &&
                    choice != EXIT);

            if (choice == NEW_USER)
                getUserInfo();
            else if (choice == EXISTING_USER)
                menuSelectUser();
            else if (choice == EXIT)
                endSession();
        }
    }

    /**
     * Exit the text based Client session.
     */
    private void endSession() {
        print();
        print(
                "Thanks for using the CF Prediction Engine...");
        print();
        System.exit(0);
    }

    /**
     * Get existing user input - check if the user exists.
     */
    private void getExistingUserInput() {
        print();
        int numOfAttempts = 1;
        String userName;

        do {
            if (numOfAttempts > 1)
                print(
                        "Error: User name not found. Please try again\n");
            userName = Console.readLine("User Name:");
            if (userName.compareTo("") == 0) {
                return;
            }
            numOfAttempts++;
        } while (!checkUser(userName));

        curUser = dbManager.clientLookupUser(userName);
        menuGetMainActions();
    }

    /**
     * menuSelectUser
     */
    private void menuSelectUser() {
        while (true) {
            print();
            print("================================");
            print("1. See a list of existing users");
            print("2. Select a user by name");
            print("3. Return");

            final int VIEW_EXISTING = 1;
            final int ENTER_NAME = 2;
            final int RETURN = 3;
            int choice; // Which menu item the user selects
            do {
                choice =
                        Console.readInt("Enter choice: ");
            } while (choice != VIEW_EXISTING
                    && choice != ENTER_NAME
                    && choice != RETURN);
            if (choice == VIEW_EXISTING) {
                viewExistingUsers();
            } else if (choice == ENTER_NAME) {
                getExistingUserInput();
            } else if (choice == RETURN) {
                return;
            }
        }
    }

    /**
     * Print a list of all the existing users, according the client's
     * DB (NOT THE SERVER)
     */
    private void viewExistingUsers() {
        Vector userVector = dbManager.retrieveUserVector();
        print("Existing Usernames:");
        int count = 0;
        for (int i = 0; i < userVector.size(); i++) {
            print(userVector.get(i).toString());
            count++;
        }
        print(count + " users");
    }

    /**
     * Query the user for all information needed for a new user
     */
    private void getUserInfo() {
        print();
        String userName = Console.readLine("Enter User Name:");

        while (checkUser(userName)) {
            print(
                    "Error: This login (" + userName + ") is already in use");
            print("       Please pick another login.");
            userName = Console.readLine("Login Name:");
        }
        curUser = new User(userName);
        curUser.setID(addNewUser(curUser));
    }

    /**
     * Get rating from console and check if it is valid.
     */
    private int getValidRating(String name) {
        int rating = Console.readInt("Rating for " + name + " (" + minRating + "-" + maxRating + "): ");
        while (rating < minRating || rating > maxRating) {
            print(
                    "Only values " + minRating + " through " + maxRating + " are valid. Please try again!");
            rating = Console.readInt(name + " (" + minRating + "-" + maxRating + "): ");
        }
        return rating;
    }

    /**
     * Get choice of main actions from users and call
     * the corresponding methods.
     */
    private void menuGetMainActions() {
        for (; ;) {
            print("===================================");
            print("Client User Name: " + curUser.getUserName());
            print("Server User ID: " + curUser.getID());
            print("===================================");
            String curUserString = " (" + curUser.getUserName() + ")";
            print(
                    "Please choose one of the following options");
            int i = 0;
            print(++i + ". Set a rating for" + curUserString);
            print(++i + ". Set a list of new ratings for" + curUserString);
            print(++i + ". Retrieve a rating for" + curUserString);
            print(++i + ". Delete a rating by " + curUserString);
            print(++i + ". List all items rated by " + curUserString);
            print(++i + ". List all users who have rated a given item");
            print(++i + ". Predict a rating for" + curUserString);
            print(++i + ". Predict a list of ratings for" + curUserString);
            print(++i + ". Get top n recommendations for" + curUserString);
            print(++i + ". Get top n recommendations of specific type for" + curUserString);
            print(++i + ". Return to user selection menu");
            print(++i + ". Exit the System");

            print("===================================");

            int choice = 0;

            /**
             * Actions that an existing user can take.
             */
            // The following weirdness is an experiement to get around the
            // requirement that case labels be constants.
            // With this setup, it initially takes a bit of writing, but is
            // easier to reorder. However, it looks terrible.
            final int RATE_ITEM = 1;
            final int RATE_ITEM_LIST = RATE_ITEM + 1;
            final int GET_RATING = RATE_ITEM_LIST + 1;
            final int REMOVE_RATING = GET_RATING + 1;
            final int GET_ITEMRATED_LIST = REMOVE_RATING + 1;
            final int GET_USERRATED_LIST = GET_ITEMRATED_LIST + 1;
            final int GET_PREDICTEDRATING = GET_USERRATED_LIST + 1;
            final int GET_PREDICTEDRATING_LIST = GET_PREDICTEDRATING + 1;
            final int GET_RECOMMENDATIONS = GET_PREDICTEDRATING_LIST + 1;
            final int GET_RECOMMENDATION_BY_TYPE = GET_RECOMMENDATIONS + 1;
            final int RETURN = GET_RECOMMENDATION_BY_TYPE + 1;
            final int EXIT = RETURN + 1;

            while (choice > EXIT || choice < 1) {
                choice = Console.readInt("Enter choice:");
            }
            switch (choice) {
                case RATE_ITEM:
                    addRating();
                    break;
                case GET_RATING:
                    printRating();
                    break;
                case GET_RECOMMENDATIONS:
                    printRecommendations();
                    break;
                case GET_PREDICTEDRATING:
                    printPredictedRating();
                    break;
                case GET_PREDICTEDRATING_LIST:
                    printPredictedRatingList();
                    break;
                case RATE_ITEM_LIST:
                    addRatingList();
                    break;
                case GET_ITEMRATED_LIST:
                    printItemArray();
                    break;
                case GET_USERRATED_LIST:
                    printUserArray();
                    break;
                case GET_RECOMMENDATION_BY_TYPE:
                    printRecommendationsByType();
                    break;
                case REMOVE_RATING:
                    removeRating();
                    break;
                case RETURN:
                    return;
                case EXIT:
                    endSession();
            }//end of switch
        }//end of for
    }

    private void removeRating() {
        Item itemChoice = menuPromptItem();
        cfManager.removeRating(curUser, itemChoice);
        print("Item deleted");
        print();
    }

    /**
     * Check if the user exists
     * @param userName
     * @return boolean
     */
    private boolean checkUser(String userName) {
       User user = dbManager.clientLookupUser(userName);

        if (user == null) {
            //SystemUtils.printDebug("Username does not exist...");
            return false;
        } else {
            //SystemUtils.printDebug("Username exists...");
            //SystemUtils.printDebug("User ID = " + curUser.getID());
            return true;
        }
    }

    /**
     * Prompt the user to enter an item name or item id,
     * then return an Item object.
     *
     * @return null if the user doesn't select an item, otherwise an Item object
     */
    private Item menuPromptItem() {
        int choice;          // The menu selection the user will make
        int id = -1;             // The server-side id of an item to be rated
        String itemName = null;  // the client-side name of an item to be rated.

        final int BY_CLIENT_NAME = 1;
        final int BY_SERVER_ID = 2;
        final int LIST_ITEMS = 3;
        final int RETURN = 4;
        Item itemChoice = null;  // Item that is chosen by the user

        do { // Until the user select a valid item or returns to previous menu
            do {
                int i = 0;
                print();
                print(++i + ". Select by item name");
                print(++i + ". Select by item id");
                print(++i + ". See a list of all possible items (names and ids)");
                print(++i + ". Return to previous menu");

                choice = 0;
                while (choice > LIST_ITEMS || choice < 1) {
                    choice = Console.readInt("Enter choice:");
                }
                switch (choice) {
                    case BY_CLIENT_NAME:
                        itemName = Console.readLine("Enter item name:");
                        break;
                    case BY_SERVER_ID:
                        id = Console.readInt("Enter item id:");
                        break;
                    case LIST_ITEMS:
                        print();
                        print("All known item names");
                        Vector items = dbManager.clientRetrieveAllItems();
                        for (int j = 0; j < items.size(); j++) {
                            print((items.get(j)).toString());
                        }
                        break;
                    case RETURN:
                        return null;
                }
            } while (choice == LIST_ITEMS);

            if (itemName != null || id != -1) {
                try {
                    if (itemName != null) {
                        itemChoice = dbManager.clientLookupItem(itemName);
                    } else {
                        itemChoice = dbManager.clientLookupItem(id);
                    }
                } catch (ClientNoSuchItemException e) {
                    print(
                            "Error: No such item exists in our database.");
                    print();
                    continue;
                }
            } else {
                System.out.println("This shouldn't happen");
                System.exit(1);
            }
        } while (itemChoice == null);
        return itemChoice;
    }

    /**
     * Print a list of all items that the current user has already rated.
     */
    private void printItemArray() {
        //@todo!!! We need to change the api so that Ratings are returned, not items
        Rating [] rateArray = cfManager.getUserRatingList(curUser);

        print(curUser.getUserName() + " has rated the following items:");
        for (int i = 0; i < rateArray.length; i++) {
            print(rateArray[i].toString());
        }
    }

    /**
     * Get an array of userIDs who have rated a given item.
     */
    private void printUserArray() {
        Item item = menuPromptItem();

        User userArray [] = cfManager.getItemRatingList(item);

        if (userArray.length == 0) {
            print(" Nobody has rated " + item);
        } else {
            print(" The following users have rated " + item + " : ");
        }

        for (int i = 0; i < userArray.length; i++) {
            print(userArray[i].toString());
        }
        print(userArray.length + " users have rated " + item);
    }

    /**
     * Add new user to database.
     * @param user
     * @return userid
     */
    private int addNewUser(User user) {
        print();
        SystemUtils.printDebug("Adding new user to DB...");
        int userid;
        userid = dbManager.clientAddNewUser(user);

        print(" New user's ID " + userid);
        return userid;
    }

    /**
     * Return the rating for a Item which has been rated.
     */
    private void printRating() {
        print();

        Item itemChoice = menuPromptItem();
        Rating rating = cfManager.getRating(curUser, itemChoice);

        // @todo instead of printing the actual value of the error rating, we should show the symbolic constant
        print(ratingToString(rating));
        print();
    }

      /**
     * Set rating for a item
     *
     */
    private void addRating() {
        // @todo explain why we are using integers when system expects floats
        Item itemChoice = menuPromptItem();
        int rating = getValidRating(itemChoice.getName());
        cfManager.addRating(curUser, itemChoice, rating);
        print();
    }

         /**
     * Add a list of new ratings.
     */
    private void addRatingList() {
        print();

        int itemID;
        float rating;

        final int length = Console.readInt(
                "number of items wanted to add new ratings for  ");

        Rating[] newRatingList = new Rating[length];
        for (int i = 0; i < length; i++) {
            itemID = Console.readInt(
                    "item ID wanted to add rating for:  ");
            rating = Console.readFloat(
                    "rating wanted to add: ");

            while (rating < minRating || rating > maxRating) {
                print(
                        "Only values between" + minRating + " and " + maxRating + " are valid. Please try again!");
                rating = Console.readFloat(
                        "rating wanted to add: ");
            }
            newRatingList[i] = new Rating(curUser, new Item(itemID), rating);
        }
        int number = cfManager.addRatingList(newRatingList);

        print(" Done! " + number + " rating added.");
    }

    /**
     * Get the predicted rating for a Item.
     */
    private void printPredictedRating() {
        Item theItem = menuPromptItem();
        Prediction predictedRating = cfManager.getPredictedRating(curUser, theItem);

        print();
        print("==========================");
        print(predictionToString(predictedRating));
        print("==========================");
        print();
    }

    /**
     * Get prediction for a list of items.
     */
    private void printPredictedRatingList() {
        print();

        int length = Console.readInt(
                "number of items wanted to get predictions for  ");

        int[] itemID = new int[length];

        for (int i = 0; i < length; i++) {
            itemID[i] = Console.readInt(
                    "item ID wanted to get prediction for:  ");
        }

        Item[] items = new Item[itemID.length];
        for (int i = 0; i < itemID.length; i++) {
            int id = itemID[i];
            items[i] = dbManager.clientLookupItem(id);
        }
        Prediction[] predictedRatings = cfManager.getPredictedRatings(curUser, items);

        for (int i = 0; i < length; i++) {
            print(predictionToString(predictedRatings[i]));
        }
    }

     /**
     * Get top n recommendations
     */
    private void printRecommendations() {
        print();
        int number = Console.readInt("Number of recommendations wanted:");
        int offset = Console.readInt("Starting from offset  (first rec. is offset 0):");
        if (number <= 0) {
            print();
            return;
        }
        print();
        print("Up to " + number + " recommendations starting from offset " + offset);
        print("===============================");

        Prediction[] recommendations = cfManager.getRecommendations(curUser, number, offset);
        if (recommendations.length > 1) {
            for (int i = 0; i < recommendations.length; i++) {
                print((i + 1) + ". " + recommendations[i]);
            }
        } else {
            print("Server returned no recommendations");
            print("    perhaps the user profile is not large enough");
        }

        print("===============================");
        print();
    }

        /**
     * top n recommendations by type
     */
    private void printRecommendationsByType() {
        Type type = menuPromptType();
        if (type == null) {
            return;
        }
        print();
        int number = Console.readInt(
                "Number of recommendations wanted:");
        int offset = Console.readInt(
                "Starting from offset (first rec. is offset 0):");
        if (number <= 0) {
            print();
            return;
        }

        Prediction recommendations [] = cfManager.getRecommendationsByType(curUser, type, number, offset);
        print();

        if (recommendations.length == 0)
            print("No more recommendations starting from offset " + offset);
        else if (recommendations.length < number)
            print("No more than " + recommendations.length +
                    " recommendations starting from offset " + offset);
        else
            print(number + " recommendations starting from offset " + offset);
        print("===============================");

        if (recommendations.length > 0) {
            for (int i = 0; i < recommendations.length; i++) {
                print((i + 1) + ". "
                        + recommendations[i]);
            }
        } else
            print("No recommendations were returned.");
        print("===============================");
        print();

    }

    /**
     * Prompt the user to enter a type.
     */
    private Type menuPromptType() {
        final int BY_NAME = 1;
        final int BY_ID = BY_NAME + 1;
        final int LIST_TYPES = BY_ID + 1;
        int choice;
        Type type = null;
        do {
            int i = 0;
            print();
            print(++i + ". Enter type by name");
            print(++i + ". Enter type by id");
            print(++i + ". See a list of all types");
            print("");

            do {
                choice = Console.readInt("Enter choice:");
            } while (choice < 0 || choice > LIST_TYPES);

            switch (choice) {
                case BY_NAME:
                    do {
                        String name = Console.readLine("Enter Name: ");
                        if (name.compareTo("") == 0) {
                            break;
                        }
                        type = dbManager.clientLookupType(name);
                    } while (type == null);
                    break;
                case BY_ID:
                    do {
                        int id = Console.readInt("Enter Id: ");
                        type = dbManager.clientLookupType(id);
                    } while (type == null);
                    break;
                case LIST_TYPES:
                    printTypes();
                    break;
            }
        } while (choice == LIST_TYPES);
        return type;
    }

    /**
     * List out all the available types.
     */
    private void printTypes() {
        print();
        print("All available types:");
        Vector allTypes = dbManager.retriveAllTypes();
        for (int i = 0; i < allTypes.size(); i++) {
            Type type = (Type) allTypes.elementAt(i);
            print(type.toString());
        }
    }

    private String getRatingValueString (float ratingValue) {
        String ratingString;
        if (ratingValue == errRating) {
            ratingString = "<No Rating>";
        } else {
            ratingString = "" + ratingValue;
        }
        return ratingString;
    }

    /**
     * Prints a rating object. This is done here and not in a Rating.toString()
     * method because the Rating objects aren't aware of the errRating value
     * @param rating - a Rating object.
     */
    private String ratingToString(Rating rating) {

        return rating.user + ", " + rating.item + ", " + getRatingValueString(rating.ratingValue);
    }

    /**
     * Prints a rating object. This is done here and not in a Rating.toString()
     * method because the Rating objects aren't aware of the errRating value
     * @param prediction - a Rating object.
     */
    private String predictionToString(Prediction prediction) {
       return "[" + getRatingValueString(prediction.getPredictionValue()) + "] " + prediction.getItem();
    }

 }//end of class