/*
 * CFEngine.java
 *
 * Copyright (c) 2003 Oregon State Universtity. All rights reserved.
 */

package org.recommender.server;

import java.rmi.*;

/**
 * Describes the interface for the CF Engine.
 * Contains methods that will be invoked by remote
 * clients of the CF Engine (CFEngineClient) and
 * will be implemented by the implementation
 * class (CFEngineImpl) and created by a CF Engine
 * server (CFEngineServer).
 *
 * @since   JDK 1.2
 *
 */

public interface CFEngine
        extends Remote {

    /**
     * Test whether the CFEngine is alive on the host. Useful to test if an
     * RMI connection was successful. If this method doesn't
     * throw an exception, then the server was successfully contacted.
     *
     * @return String
     * @throws RemoteException
     */
    String test() throws RemoteException;

    /**
     * Add new rating to both database and the cache. The rating is added immediately
     * to the memory cache, and there may be a short delay before it reaches the
     * database. If the server crashes before the rating is written to the database
     * that rating may be lost.
     *
     * Note: If rating for the same (user, item) already exists, it replaces it.
     *
     * @param user
     * @param item
     * @param rating
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalParameterException Throws exception for out of range user,item, or rating
     */
    public void setRating(int user, int item, float rating) throws java.rmi.
            RemoteException, CFIllegalParameterException, CFInternalErrorException;

    /**
     * Add a list of new ItemRatings to both database and the cache.
     * Like setRating(), if a rating already exists for a given (user, item),
     * it will be replaced if you specify a new rating for the same (user, item) pair.
     * While ratings are written as fast as possible to the database, they are not
     * guranteed, and if the server crashes, the most recently sent ratings may be
     * lost.
     *
     * @param newRatings a list of new ratings wanted to add
     * @return numSuccess Returns the number of new ratings that were successfully added to the database
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalListParameterException Throws exception if ItemRating[] has invalid contents
     */
    public int setRatingList(ItemRating[] newRatings)
            throws java.rmi.RemoteException, CFIllegalListParameterException, CFInternalErrorException;

    /**
     * Removes rating from both database and cache.  If rating doesn't exist, it
     * just returns.
     *
     * @param user
     * @param item
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalParameterException Throws exception for out of range user or item
     */
    public void removeRating(int user, int item)
            throws java.rmi.RemoteException, CFIllegalParameterException, CFInternalErrorException;

    /**
     * Removes list of ratings from users profile. Ratings that do not
     * exist are just ignored.
     *
     * @param user
     * @param itemIDs
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalListParameterException Throws exception if userID or itemID are invalid
     */
    public void removeRatingList(int user, int[] itemIDs)
            throws java.rmi.RemoteException, CFIllegalListParameterException, CFInternalErrorException;

    /**
     * Retrieves one user's rating for a given item.
     *
     * If the user has not rated the item, returns an ItemRating with
     * an ErrRating. @see getErrRating()
     *
     * @param userID
     * @param itemID
     * @return rating
     *
     * @throws RemoteException Throws exception when connection to server is lost
     * @throws CFIllegalParameterException Throws exception if userID or Item ID are out of range
     */
    ItemRating getRating(int userID, int itemID)
            throws RemoteException, CFIllegalParameterException, CFInternalErrorException;

    /**
     * Retrieves the ratings by the user for a list of given items.
     *
     * If the user has not rated the item, returns an ItemRating with
     * an ErrRating. @see getErrRating()
     *
     * @param userID
     * @param itemIDs
     * @return
     * @throws RemoteException
     * @throws CFIllegalListParameterException Throws exception if userID or itemID[] are invalid
     */
    public ItemRating[] getRatingList(int userID, int[] itemIDs)
            throws RemoteException, CFIllegalListParameterException, CFInternalErrorException;

    /**
     * Get a user's predicted rating for an item.
     *
     * If the CFEngine algorithm cannot predict a rating for any reason, returns an ItemPrediction with
     * an ErrRating. @see getErrRating().
     * The most common reason for not being able to predict a rating is not having enough information
     *
     * @param userID
     * @param itemID
     * @return predictedRating
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalParameterException Thrown if userID or itemID are out of range
     */
    public ItemPrediction getPredictedRating(int userID, int itemID)
            throws RemoteException, CFIllegalParameterException, CFInternalErrorException;

    /**
     * Get ten items with high predicted ratings for current user. This
     * method is not guaranteed to be optimal. It may not return the
     * items with the absolutely highest predicted ratings. However, the results
     * it returns should be pretty close.
     *
     * Returns array of size equal to the number of valid predictions in the
     * range requested.  If there are no valid recommendations in the specified
     * range, returns 0 length array.
     *
     * @param curUser User to get recommendations for.
     * @param number Number of recommendations to be retrieved.
     * @param offset Returns number of recommendations, starting at the offset
     * @return ItemPrediction[]
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalParameterException Throws exception for invalid curUser, number or offset
     */
    public ItemPrediction[] getRecommendations(int curUser, int number,
                                               int offset)
            throws RemoteException, CFIllegalParameterException, CFInternalErrorException;

    /**
     * Gets predicted ratings for a list of items. If the server is unable to predict
     * a rating for any reason for a given item, the predicted value will be set to the
     * ErrRating. @see getErrRating()
     *
     * @param userID
     * @param itemID a list of items
     * @return predictions
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalListParameterException Throws exception if userID or itemID[] are invalid
     */
    public ItemPrediction[] getPredictedRatingList(int userID, int[] itemID)
            throws java.rmi.RemoteException, CFIllegalListParameterException, CFInternalErrorException;

    /**
     * Return an array of all known ratings for the given user.
     *
     * Note: returns 0 length array if there are no ratings.
     *
     * @param userID
     * @return an array of itemIDs
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalParameterException Throws exception if userID is invalid
     */
    public ItemRating[] getUserRatingList(int userID)
            throws java.rmi.RemoteException, CFIllegalParameterException;

    /**
     * Return an array of all known ratings for the given item.
     *
     * Note: returns 0 length array if there are no ratings.
     *
     * @param itemID
     * @return an array of userID
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalParameterException Throws exception if itemID is out of range
     */
    public ItemRating[] getItemRatingList(int itemID)
            throws java.rmi.RemoteException, CFIllegalParameterException;

    /**
     * Shuts the server down.
     * @throws RemoteException
     */
    public void shutdown() throws java.rmi.RemoteException;

    /**
     * Get top n recommendations for current user of specific type.
     *
     * Returns array of size equal to the number of valid predictions in the
     * range requested.  If there are no valid recommendations in the specified
     * range, returns 0 length array.
     *
     * @param curUser User to get recommendations for.
     * @param number Number of recommendations to be retrieved.
     * @param offset Returns number of recommendations, starting at the offset
     * @param type Catagory to gat recommendations from.
     * @return ItemPrediction[]
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalParameterException Throws exception if curUser,number,offset, or type are out of range
     */
    public ItemPrediction[] getRecommendationsByType(int curUser, int number, int offset, int type)
            throws RemoteException, CFIllegalParameterException, CFInternalErrorException;

    /**
     *  Get the maximum rating allowed for the current server configuration. This
     * may be useful if you have clients that can adapt to multiple rating ranges.
     *
     * @return max
     *
     * @throws RemoteException Thrown upon client/server connection error
     */
    public float getMaxRating() throws java.rmi.RemoteException;

    /**
     *  Get the minimum rating allowed for the current server configuration.
     * This may be useful if you have clients that can adapt to multiple rating ranges
     *
     * @return min
     *
     * @throws RemoteException Thrown upon client/server connection error
     */
    public float getMinRating() throws java.rmi.RemoteException;

    /**
     *  Get the error rating for the current server configuration.
     * This is the rating that will be returned when a rating or
     * prediction is not available.
     *
     * @return error rating
     *
     * @throws RemoteException Thrown upon client/server connection error
     */
    public float getErrRating() throws java.rmi.RemoteException;

    /**
     * @return An unused integer userid that is guaranteed to be unique now and in the future
     */
    public int getNextUserId() throws java.rmi.RemoteException, CFInternalErrorException;

    /**
     * @return An unused integer itemid that is guaranteed to be unique now and in the future
     */
    public int getNextItemId() throws java.rmi.RemoteException, CFInternalErrorException;

    /**
     *
     * @return A string describing key runtime statistics of the CFEngine server, such as cache hit rate
     * @throws java.rmi.RemoteException
     */
    public String getStatistics() throws java.rmi.RemoteException;

}
