/*
 * CFEngineImpl.java
 *
 * Version 1.1
 *
 * 15/08/2002
 *
 * Copyright (c) 2003 Oregon State Universtity. All rights reserved.
 */
package org.recommender.server;

import java.rmi.*;
import java.rmi.server.*;
import java.lang.reflect.*;

import org.recommender.algorithms.*;

//import javax.rmi.*;

import java.sql.*;
import java.util.Iterator;

/**
 * Implementation of the CFEngine interface.
 *
 * @author  <A HREF="mailto:sameerk@cs.orst.edu">Sameer Kadam</A>
 * @author  <A HREF="mailto:iyoon@cs.orst.edu">Irwin Yoon</A>
 * @author  <A HREF="mailto:godde@cs.orst.edu">Olivier Godde</A>
 * @author  <A HREF="mailto:dlowd@cs.orst.edu">Daniel Lowd</A>
 * @author  <A HREF="mailto:nong@cs.orst.edu">Shuzhen Nong</A>
 * @author  <A HREF="mailto:mclaughm@cs.orst.edu">Matt McLaughlin</A>
 * @since   JDK 1.2
 *
 */
public class CFEngineImpl
        extends UnicastRemoteObject
        implements CFEngine {
    //extends PortableRemoteObject implements CFEngine {

    private String host;
    private DataManager dataMgr = null;
    //private CFAlgorithm algorithm = null;

    public Class algorithm = null;
    public Object newAlgObj = null;

    /**
     * Constructor, do nothing.
     * @throws RemoteException
     */
    public CFEngineImpl() throws RemoteException {

    }

    /**
     * Constructor, create a DataManager with a mode: experiment,
     * autotest, or server, instantiate an algorithm and start
     * a DBManager which is reponsible for dynamic database access.
     * @param propertiesFilename path and filename to the properties
     *     file that defines the configuration of this server.
     * @throws RemoteException
     */
    public CFEngineImpl(String propertiesFilename) throws RemoteException, CFInternalErrorException {

        // Load properties
        SystemProperties.loadProperties(propertiesFilename);

        host = SystemProperties.getString("CFServer.rmi.nameserver",
                "localhost");
        String testWhere = SystemProperties.getString("CFServer.db.testSet",
                "0");
        String mode = SystemProperties.getString("CFServer.mode", "server");

        if (mode.equals("experiment")) {
            dataMgr = new DataManager(false, false);
        } else if (mode.equals("autotest")) {
            dataMgr = new DataManager(true, false);
        } else {
            dataMgr = new DataManager(true, true);

        }

        // Instantiate algorithm
        // algorithm = CFAlgorithmFactory.constructAlgorithm(dataMgr);
        String algName = SystemProperties.getString("CFServer.alg.name");

        try {
            algorithm = Class.forName(algName);
            SystemUtils.printDebug("Using algorithm: " + algorithm.getName());

            Constructor[] cons = algorithm.getConstructors();

            SystemUtils.printDebug("Constructing a new algorithm object, please wait...");

            long start = System.currentTimeMillis();

            newAlgObj = cons[0].newInstance(new Object[]{dataMgr});

            long stop = System.currentTimeMillis();

            SystemUtils.printDebug("Finish constructing a new algorithm object, time = " +
                    (stop - start) + " ms");

        } catch (Exception e) {
            // Various exceptions can be thrown by the reflect API, so we catch them all here
            //System.out.println(" class not found Exception" + e);
            throw new CFInternalErrorException(" class not found Exception", e);
        }

        if (algorithm == null) {
            // Unknown algorithm name.  Print an error and quit.
            System.out.println("ERROR: Could not construct algorithm.");
            System.exit(0);
        }

        // Run some automatic tests on our algorithm.
        if (mode.equals("autotest")) {
            AutoTester.runTests(dataMgr, (CFAlgorithm)newAlgObj, testWhere);
        }


        if (mode.equals("autotest")) {
            SystemUtils.printDebug("Tests done, killing server.");
            System.exit(0);
        }

        // Start a new Thread here, DBManager which is responsible for any dynamic database access
        DBManager db = new DBManager();
        db.start(); //Start DBManager thread

        // The following Thread is responsible for refreshing the cache whenever it fills up
        CacheUpdate cacheThread = new CacheUpdate(dataMgr);
        cacheThread.start();

        // The following Thread is responsible for resampling the data
        Resample samplingThread = new Resample(dataMgr);
        samplingThread.start();
    }

    /**
     * Test wether the CFEngine is alive on the host.
     * @return String
     * @throws RemoteException
     */
    public String test() throws RemoteException {
        return "CFEngine is alive on " + host;
    }

    /**
     * Retrieves a rating for user' item from cache.
     *
     * If the user has not rated the item, returns an ItemRating with
     * an ErrRating. @see getErrRating()
     *
     * @param userID
     * @param itemID
     * @return rating
     *
     * @throws RemoteException Throws exception when connection to server is lost
     * @throws CFIllegalParameterException Throws exception if userID or Item ID are out of range
     */
    public ItemRating getRating(int userID, int itemID) throws RemoteException,
            CFIllegalParameterException, CFInternalErrorException {

        if (userID < 0) {
            throw new CFIllegalParameterException("User ID " + userID +
                    " is out of bounds");
        }
        if (itemID < 0) {
            throw new CFIllegalParameterException("Item ID " + itemID +
                    " is out of bounds");
        }

        ItemRating ir = null;
        try {
            ItemInfo it = dataMgr.getItemInfo(itemID);
            if(it == null) ir = new ItemRating(userID, itemID, dataMgr.getErrRating());
            else{
                if (!it.getReadersLock(true)) {
                    throw new CFInternalErrorException("Unable to obtain lock");
                }
                ir = (ItemRating) it.getItemRatingsHash().get(userID);
                it.releaseReadersLock();
                if (ir == null) {
                    ir = new ItemRating(userID, itemID, dataMgr.getErrRating());
                }
            }
        } catch (Exception e) {
            throw new CFInternalErrorException("Error in getRating", e);
        }
        return ir;
    }

    /**
     * Retrieves ratings for user's items from cache.
     *
     * If the user has not rated the item, returns an ItemRating with
     * an ErrRating. @see getErrRating()
     *
     * @param userID
     * @param itemIDs
     * @return
     * @throws RemoteException
     * @throws CFIllegalListParameterException Throws exception if userID or itemID[] are invalid
     */
    public ItemRating[] getRatingList(int userID, int[] itemIDs)
            throws RemoteException, CFIllegalListParameterException, CFInternalErrorException {

        if (userID < 0) {
            throw new CFIllegalListParameterException("User ID " + userID +
                    " is out of bounds", 0);
        }
        if (itemIDs == null) {
            throw new CFIllegalListParameterException("itemIDs cannot be null", 0);
        }

        ItemRating[] returnList = new ItemRating[itemIDs.length];
        ItemInfo ii = null;
        for (int i = 0; i < itemIDs.length; i++) {
            if (itemIDs[i] < 0) {
                throw new CFIllegalListParameterException("Item ID " + itemIDs[i] +
                        " is out of bounds", i);
            }
            try {
                ii = dataMgr.getItemInfo(itemIDs[i]);
                if (ii == null)
                    returnList[i] = new ItemRating(userID, itemIDs[i], dataMgr.getErrRating());
                else {
                    ii.getReadersLock(true);
                    returnList[i] = (ItemRating) ii.getItemRatingsHash().get(userID);
                    ii.releaseReadersLock();
                    if (returnList[i] == null) returnList[i] = new ItemRating(userID, itemIDs[i], dataMgr.getErrRating());
                }
            } catch (Exception e) {
                throw new CFInternalErrorException("Error in getRatingList", e);
            }
        }
        return returnList;
    }

    /**
     * Get top n recommendation for user.
     *
     * Returns array of size equal to the number of valid predictions in the
     * range requested.  If there are no valid recommendations in the specified
     * range, returns 0 length array.
     *
     * @param userID User to get recommendations for.
     * @param number Number of recommendations to be retrieved.
     * @param offset Returns number of recommendations, starting at the offset
     * @return ItemPrediction[]
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalParameterException Throws exception for invalid curUser, number or offset
     */

    public ItemPrediction[] getRecommendations(int userID, int number, int offset) throws
            RemoteException, CFIllegalParameterException, CFInternalErrorException {

        // First load the active user into the cache if nessary.
        if (userID < 0) {
            throw new CFIllegalParameterException("User ID " + userID +
                    " is out of bounds");
        }
        if (number < 0) {
            throw new CFIllegalParameterException(
                    "param number must be greater than 0.  number = " + number);
        }
        if (offset < 0) {
            throw new CFIllegalParameterException(
                    "param offset must be greater or equal to 0.  offset = " + offset);
        }
        //if((number+offset) > dataMgr.getNumTopNRecords()) throw new CFIllegalParameterException("Max of 300 recomendations cached.  number + offset cannot exceed this amount");

        if (userID > dataMgr.getMaxUser()) {
            return new ItemPrediction[0];
        }

        //This function will load the active user into the cache if necessary
        //Perform this check here since we don't have login sessions
        try {
            dataMgr.checkCachedUser(userID);
            // Also sets the inUse field, which ensures that userID's record is not
            // removed from the cache until we clear it.
        } catch (Exception e) {
            throw new CFInternalErrorException("Error in getRecommendations", e);
        }

        ItemPrediction ip[];
        // We used the cached topN unless the user has added any new ratings.
        if (!dataMgr.checkCachedTopN(userID) ||
                ((offset == 0) && dataMgr.getAddedRatings(userID) > 0)) {

            Object retVal = null;
            Object[] argObj = {
                new Integer(userID), new Integer(dataMgr.getNumTopNRecords())};

            try {
                Method met = findgetRecommendations(algorithm);
                retVal = met.invoke(newAlgObj, argObj);

            } catch (Exception e) {
                dataMgr.decUseCount(userID);
                throw new CFInternalErrorException(" exception in invoking getRecommendations method ", e);
            }

            dataMgr.cacheUserTopN(userID, (ItemPrediction[]) retVal);
            dataMgr.resetUserCounter(userID);

            ip = dataMgr.getCachedTopN(userID, number, offset);
        } else {
            ip = dataMgr.getCachedTopN(userID, number, offset);
        }
        dataMgr.decUseCount(userID);
        return ip;
    }

    /**
     * Get top n recommendations for current user of specific type.
     *
     * Returns array of size equal to the number of valid predictions in the
     * range requested.  If there are no valid recommendations in the specified
     * range, returns 0 length array.
     *
     * @param userID User to get recommendations for.
     * @param number Number of recommendations to be retrieved.
     * @param offset Returns number of recommendations, starting at the offset
     * @param type Catagory to gat recommendations from.
     * @return ItemPrediction[]
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalParameterException Throws exception if curUser,number,offset, or type are out of range
     */

    public ItemPrediction[] getRecommendationsByType(int userID, int number,
                                                     int offset, int type) throws RemoteException,
            CFIllegalParameterException, CFInternalErrorException {

        if (userID < 0) {
            throw new CFIllegalParameterException("User ID " + userID +
                    " is out of bounds");
        }
        if (number < 0) {
            throw new CFIllegalParameterException(
                    "param number must be greater than 0.  number = " + number);
        }
        if (offset < 0) {
            throw new CFIllegalParameterException(
                    "param offset must be greater or equal to 0.  offset = " + offset);
        }
        if (type < 0 || type > dataMgr.getMaxType()) {
            throw new CFIllegalParameterException(
                    "param type must be valid.  type = " + type);
        }

        if (userID > dataMgr.getMaxUser()) {
            return new ItemPrediction[0];
        }

        try {
            dataMgr.checkCachedUser(userID);
            // Also sets the inUse field, which ensures that userID's record is not
            // removed from the cache until we clear it.
        } catch (Exception e) {
            throw new CFInternalErrorException("Error caching user", e);
        }

        Object retVal = null;
        Object[] argObj = {
            new Integer(userID), new Integer(dataMgr.getNumTopNRecords()),
            new Integer(type)};

        try {
            Method met = findgetRecommendationsByType(algorithm);
            retVal = met.invoke(newAlgObj, argObj);

        } catch (Exception e) {
            dataMgr.decUseCount(userID);
            throw new CFInternalErrorException(" exception in invoking getRecommendationsByType method ", e);
        }

        ItemPrediction ip[] = dataMgr.getTopNByType(userID, number, offset, type,
                (ItemPrediction[]) retVal);
        dataMgr.decUseCount(userID);
        return ip;
    }

    /**
     * Get predicted rating for user's item.
     *
     * If there is an error predicting rating, returns an ItemPrediction with
     * an ErrRating. @see getErrRating()
     *
     * @param userID
     * @param itemID
     * @return predictedRating
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalParameterException Thrown if userID or itemID are out of range
     */

    public ItemPrediction getPredictedRating(int userID, int itemID) throws
            RemoteException, CFIllegalParameterException, CFInternalErrorException {

        // First load the active user into the cache if nessary -
        // those who are not in cache or are new users, then call
        // algorithm.getPredictedRating(userID, itemID) to get predicted
        // rating for the user's item.

        if (userID < 0) {
            throw new CFIllegalParameterException("User ID " + userID +
                    " is out of bounds");
        }
        if (itemID < 0) {
            throw new CFIllegalParameterException("Item ID " + itemID +
                    " is out of bounds");
        }
        if (userID > dataMgr.getMaxUser() || itemID > dataMgr.getMaxItem()) {
            return new ItemPrediction(userID, itemID, dataMgr.getErrRating());
        }

        // This function will load the active user into the cache if necessary
        // Perform this check here since we don't have login sessions
        try {
            dataMgr.checkCachedUser(userID);
            // Also sets the inUse field, which ensures that userID's record is not
            // removed from the cache until we clear it.
        } catch (Exception e) {
            throw new CFInternalErrorException("Error caching user", e);
        }

        Object predictedRating = null;
        Object[] argObj = {
            new Integer(userID), new Integer(itemID)};
        try {
            Method predictRating = findpredictRating(algorithm);
            predictedRating = predictRating.invoke(newAlgObj, argObj);

        } catch (Exception e) {
            dataMgr.decUseCount(userID);
            throw new CFInternalErrorException(" exception in invoking predictRating method ", e);
        }
        ItemPrediction ip = (ItemPrediction) predictedRating;
        dataMgr.decUseCount(userID);
        return ip;
    }

    /**
     * Gets predicted ratings for a list of items.
     *
     * For each invalid item in itemID[], returns an ItemPrediction with a
     * predicted value of ErrRating. @see getErrRating()
     *
     * @param userID
     * @param itemID a list of items
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalListParameterException Throws exception if userID or itemID[] are invalid
     */

    public ItemPrediction[] getPredictedRatingList(int userID, int[] itemID) throws
            RemoteException, CFIllegalListParameterException, CFInternalErrorException {

        if (itemID == null) {
            throw new CFIllegalListParameterException("Item ID array is null", 0);
        }

        if (userID < 0) {
            throw new CFIllegalListParameterException("User ID " + userID +
                    " is out of bounds", 0);
        }

        ItemPrediction[] predictions = new ItemPrediction[itemID.length];

        int i = 0;
        try {
            for (i = 0; i < itemID.length; i++) {

                predictions[i] = getPredictedRating(userID, itemID[i]);
            }

        } catch (CFIllegalParameterException e) {
            throw new CFIllegalListParameterException("Element in list failed.", i, e);
        }

        return predictions;

    }

    /**
     * Return an array of itemIDs which the given user has rated.
     *
     * Note: returns 0 length array if there are no ratings.
     *
     * @param userID
     * @return an array of itemIDs
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalParameterException Throws exception if userID is invalid
     */
    public ItemRating[] getUserRatingList(int userID) throws java.rmi.RemoteException,
            CFIllegalParameterException {

        if (userID < 0) {
            throw new CFIllegalParameterException("User ID " + userID +
                    " is out of bounds");
        }
        if (userID > dataMgr.getMaxUser()) {
            return new ItemRating[0];
        }

        UserInfo userInfo = dataMgr.getUserInfo(userID);
        if (userInfo == null) {
            // Return a zero-length array
            return new ItemRating[0];
        }

        if (!userInfo.getReadersLock(true)) return new ItemRating[0];;
        RatingArrays ratings = userInfo.getRatings();
        ItemRating[] ret = ratings.getItemRatingsArray();
        userInfo.releaseReadersLock();
        return ret;
    }

    /**
     * Return an array of userIDs who have rated the given item.
     *
     * Note: returns 0 length array if there are no ratings.
     *
     * @param itemID
     * @return an array of userID
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalParameterException Throws exception if itemID is out of range
     */
    public ItemRating[] getItemRatingList(int itemID) throws java.rmi.RemoteException,
            CFIllegalParameterException {

        if (itemID < 0) {
            throw new CFIllegalParameterException("Item ID " + itemID +
                    " is out of bounds");
        }

        ItemInfo itemInfo = dataMgr.getItemInfo(itemID);
        if(itemInfo == null) return new ItemRating[0];
        if(!itemInfo.getReadersLock(true)) return new ItemRating[0];

        IntHashMap itemHash = itemInfo.getItemRatingsHash();
        if(itemHash == null){
            itemInfo.releaseReadersLock();
            return new ItemRating[0];
        }

        ItemRating[] retArray = new ItemRating[itemHash.size()];
        Iterator iter = itemHash.iterator();
        int i = 0;
        iter = itemHash.iterator();
        while (iter.hasNext()) {
            retArray[i] = (ItemRating) iter.next();
            i++;
        }
        itemInfo.releaseReadersLock();
        return retArray;
    }

    /**
     * Return item's name.
     * @param itemId
     * @return itemName
     * @throws RemoteException
     * @throws SQLException
     */
    /*	public String getItemName(int itemID) throws RemoteException, SQLException {
    return dataMgr.getItemName(itemID);
     } */

    /**
     * Send a new rating the CFEngine server.
     *
     * Note: If rating already exists, it replaces it.
     *
     * @param user
     * @param item
     * @param ratingValue
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalParameterException Throws exception for out of range user,item, or rating
     */
    public void setRating(int user, int item, float ratingValue) throws java.rmi.
            RemoteException, CFIllegalParameterException, CFInternalErrorException {

// First load the active user into the cache if nessary -
// those who are not in cache or are new users, then add new rating
// to both database and the cache and update statistical values.

        if (user < 0) {
            throw new CFIllegalParameterException("User ID " + user +
                    " is out of bounds");
        }
        if (item < 0) {
            throw new CFIllegalParameterException("Item ID " + item +
                    " is out of bounds");
        }
        if (ratingValue < dataMgr.getMinRating() ||
                ratingValue > dataMgr.getMaxRating()) {
            throw new CFIllegalParameterException("Rating Value " + ratingValue +
                    " is out of bounds.");
        }

        ItemRating rating = new ItemRating(user, item, ratingValue);
        try {
            dataMgr.addRating(rating);
        } catch (Exception e) {
            throw new CFInternalErrorException("Error caching user", e);
        }

        // Update any algorithm models that may exist
        //algorithm.updateUser(user); 	Until not sure
    }

    /**
     * Add a list of new ItemRatings to both database and the cache.
     * If a rating already exists for a (user, item), then the old rating
     * will be overwritten.
     *
     * @param newRatings a list of new ratings wanted to add
     * @return numSuccess Returns the number of new ratings that were successfully added to the database
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalListParameterException Throws exception if ItemRating[] has invalid contents
     */
    public int setRatingList(ItemRating[] newRatings) throws java.rmi.
            RemoteException, CFIllegalListParameterException, CFInternalErrorException {

        if (newRatings == null) {
            throw new CFIllegalListParameterException("newRatings can not be null", 0);
        }

        float minRating = dataMgr.getMinRating();
        float maxRating = dataMgr.getMaxRating();

        for (int i = 0; i < newRatings.length; i++) {
            if (newRatings[i].getUserID() < 0 || newRatings[i].getItemID() < 0 ||
                    newRatings[i].getRating() < minRating ||
                    newRatings[i].getRating() > maxRating) {
                throw new CFIllegalListParameterException("Elements parameter was out of range", i);
            }
            try {
                dataMgr.addRating(newRatings[i]);
            } catch (Exception e) {
                throw new CFInternalErrorException("Error caching user", e);
            }
        }
        return newRatings.length;
    }

    /**
     * Remove the rating from both database and the cache.
     *
     * Removes rating from both database and cache.  If rating doesn't exist, it
     * just returns.
     *
     * @param user
     * @param item
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalParameterException Throws exception for out of range user or item
     */

    public void removeRating(int user, int item) throws java.rmi.RemoteException,
            CFIllegalParameterException, CFInternalErrorException {

        //First load the active user into the cache if nessary -
        //those who are not in cache or are new users, then remove rating
        //from both database and cache.

        if (user < 0) {
            throw new CFIllegalParameterException("User ID " + user +
                    " is out of bounds");
        }
        if (item < 0) {
            throw new CFIllegalParameterException("Item ID " + item +
                    " is out of bounds");
        }

        if (user > dataMgr.getMaxUser() || item > dataMgr.getMaxItem()) {
            return;
        }

        try {
            dataMgr.removeRating(user, item);
        } catch (Exception e) {
            throw new CFInternalErrorException("Error caching user", e);
        }

        // Update any algorithm models that may exist
        //algorithm.updateUser(user); 	Until not sure
    }

    /**
     * Removes list of ratings from users profile
     *
     * @param user
     * @param itemIDs
     *
     * @throws RemoteException Thrown upon client/server connection error
     * @throws CFIllegalListParameterException Throws exception if userID or itemID[] are invalid
     */
    public void removeRatingList(int user, int[] itemIDs) throws RemoteException,
            CFIllegalListParameterException, CFInternalErrorException {

        if (user < 0) {
            throw new CFIllegalListParameterException("User ID " + user +
                    " is out of bounds", 0);
        }
        if (itemIDs == null) {
            throw new CFIllegalListParameterException("itemIDs cannot be null", 0);
        }

        for (int i = 0; i < itemIDs.length; i++) {
            if (itemIDs[i] < 0) {
                throw new CFIllegalListParameterException("Item ID " + itemIDs[i] +
                        " is out of bounds", i);
            }
            try {
                dataMgr.removeRating(user, itemIDs[i]);
            } catch (Exception e) {
                throw new CFInternalErrorException("Error caching user", e);
            }
        }
    }


    /**
     *  Get the maximum rating allowed for the current server configuration.
     *
     * @return max
     *
     * @throws RemoteException Thrown upon client/server connection error
     */
    public float getMaxRating() throws java.rmi.RemoteException {
        return dataMgr.getMaxRating();
    }

    /**
     *  Get the minimum rating allowed for the current server configuration.
     *
     * @return min
     *
     * @throws RemoteException Thrown upon client/server connection error
     */
    public float getMinRating() throws java.rmi.RemoteException {
        return dataMgr.getMinRating();
    }

    /**
     *  Get the error rating for the current server configuration.
     *
     * @return error rating
     *
     * @throws RemoteException Thrown upon client/server connection error
     */
    public float getErrRating() throws java.rmi.RemoteException {
        return dataMgr.getErrRating();
    }

    /**
     * @return An unused integer userid that is guaranteed to be unique
     */
    public int getNextUserId() throws java.rmi.RemoteException, CFInternalErrorException {
        int id = -1;
        try {
            id = dataMgr.getNextUserId();
        } catch (Exception e) {
            throw new CFInternalErrorException("Error getting user id", e);
        }
        return id;
    }

    /**
     * @return An unused integer itemid that is guaranteed to be unique
     */
    public int getNextItemId() throws java.rmi.RemoteException, CFInternalErrorException {
        int id = -1;
        try {
            id = dataMgr.getNextItemId();
        } catch (Exception e) {
            throw new CFInternalErrorException("Error getting item id", e);
        }
        return id;
    }

    public String getStatistics() throws RemoteException {
        return dataMgr.getStatistics();
    }

    /**
     * Hides the rating value for the item we are trying to predict for
     * then re-compute the user SD and user mean - allows external application
     * for dynamic cache loading and system coverage computation
     * @param user
     * @param item
     * @throws RemoteException
     * @throws CFIllegalParameterException
     */
    /*    public void hideRating(int user, int item)
         throws java.rmi.RemoteException, CFIllegalParameterException{
     if(user <= 0 || user > dataMgr.getMaxUser()) throw new CFIllegalParameterException("User ID "+user+" is out of bounds");
     if(item < 0 || item > dataMgr.getMaxItem()) throw new CFIllegalParameterException("Item ID "+item+" is out of bounds");
         dataMgr.hideRatingFromRecords(user, item);
        }*/

    /**
     * Load users from an external file.
     * Use param ratingSetOut to determine whether to load this ItemRating.
     * @param fielName
     * @param numUsers
     * @param ratingSetOut
     * @throws RemoteException
     * @throws SQLException
     */
    /*    public void loadRatingsFile(String fileName, int numUsers, int ratingSetOut) //OG
         throws java.sql.SQLException, java.rmi.RemoteException{
         dataMgr.loadRatingsFile(fileName, numUsers, ratingSetOut);
         }*/

    /**
     * Initialize the data structures with the sizes of the max user and item ID.
     * @throws RemoteException
     * @throws SQLException
     */
    /*   public void initializeData()
     throws java.sql.SQLException, java.rmi.RemoteException{
     dataMgr.initializeData();
     }*/

    /**
     * Get the percentage of item entries that are not null
     * i.e., items we can compute a prediction for.
     * @return the percentage
     * @throws RemoteException
     * @throws SQLException
     */
    /*	public float getCoverage()
         throws java.sql.SQLException, java.rmi.RemoteException{
         return dataMgr.getCoverage();
         }*/

    /**
     * Graciously shut the server down. Execute all pending DB requests before exiting.
     * @throws RemoteException
     */
    public void shutdown() throws java.rmi.RemoteException {
        dataMgr.shutdown();
    }

    //////////////////////////////////////////////////////////////////////////

    /** Private method: Added for finding the corresponding methods
     */
    private static Method findpredictRating(Class cl) throws Exception {
        Method[] methods = cl.getMethods();
        for (int i = 0; i < methods.length; i++) {
            if (methods[i].getName().equals("predictRating")) {
                return methods[i];
            }
        }

        return null;
    }

    /** Private method: Added for finding the corresponding methods
     */
    private static Method findgetRecommendations(Class cl) throws Exception {
        Method[] methods = cl.getMethods();
        for (int i = 0; i < methods.length; i++) {
            if (methods[i].getName().equals("getRecommendations")) {
                return methods[i];
            }
        }

        return null;
    }

    /** Private method: Added for finding the corresponding methods
     */
    private static Method findgetRecommendationsByType(Class cl) throws
            Exception {
        Method[] methods = cl.getMethods();
        for (int i = 0; i < methods.length; i++) {
            if (methods[i].getName().equals("getRecommendationsByType")) {
                return methods[i];
            }
        }

        return null;
    }

    //////////////////////////////////////////////////////////////////////////////

}
