/*
 * CFEngineServer.java
 *
 * Version 1.1
 *
 * 15/08/2002
 *
 * Copyright (c) 2003 Oregon State Universtity. All rights reserved.
 */

package org.recommender.server;

import java.rmi.*;
import java.rmi.registry.*;

/**
 * The CF Engine server implementation.
 * Creates a CFEngineImpl server object and registers it with
 * the RMI registry. Remote clients send lookup this object
 * from the registry and send there CF requests to this server.
 *
 * @since   JDK 1.2
 */
public class CFEngineServer {

    private static CFEngine engine;

    /**
     * Create a CFEngineImpl server object, start API, and
     * create a DBMananger which is responsible for dynamic database access.
     */
    public static void main(String[] args) {
        try {
            long serverBootStart = System.currentTimeMillis();
            System.out.println("Constructing CF Engine implementation...");

            try {
                String installdir = System.getProperty("cf.path", ".");
                String filename = installdir + "/lib/"
                        + "CFServer.properties";
                engine = new CFEngineImpl(filename);

            } catch (java.rmi.RemoteException re) {
                System.out.println("WARNING:Cannot instanciate CFEngineImpl object");
                System.out.println(re.toString());
            }

            System.out.println("Binding CF Engine implementation to registry...");
            Registry reg;
            boolean bound = false;
            String bindName; 

            // Nested catches are ugly, but I don't immediately see a better
            // way to handle this. There are two steps, and RemoteException has
            // to be handled different in each step. AccessException is handled the same
            // whether it occurs in either step.
            try {
                try {
                    // First try and connect to a local registry.
                    reg = LocateRegistry.getRegistry(SystemProperties.getString("CFServer.rmi.nameserver"));                    
                    //reg.rebind("cfengine", engine);
                    
                    bindName = SystemProperties.getString("CFServer.rmi.objectname");
                    reg.rebind(bindName, engine);
                    bound = true;
                } catch (RemoteException re) {
                    // The docs say that getRegistry will succeed even if a registry
                    // is not yet running. So if we reach this point,
                    // rebind() has probably thrown an exception because
                    // it cannot connect to the named registry.
                    System.out.println("No RMI registry running on \""
                            + SystemProperties.getString("CFServer.rmi.nameserver") +
                            "\" or the rmiregistry that you are trying to connect to " +
                            "does not have the CF classes in its path. This is generally " +
                            "not a problem. The server will start its own registry " +
                            "locally.");
                }

                if (!bound) {
                    // Unable to bind to the named registry, so instead we create a
                    // local registry and bind there instead. If a RemoteException
                    // occurs here, we just punt.

                    System.out.println("Creating a new registry on \"localhost\"");
                    reg = LocateRegistry.createRegistry(Registry.REGISTRY_PORT);
                    //reg.rebind("cfengine", engine);
                    
                    bindName = SystemProperties.getString("CFServer.rmi.objectname");
                    reg.rebind(bindName, engine);
                    bound = true;
                }

            } catch (AccessException ae) {
                System.out.println("Received an AccessException when trying to connect to registry on "
                        + SystemProperties.getString("CFServer.rmi.nameserver") +
                        ". Most likely you are not specifying the appropriate Java Security Policy file. ");
                throw ae;
            } catch (RemoteException re) {
                System.out.println("If the message below indicates that a port is already in use, then there is already an RMI server running that does not have the CFEngine classes in its CLASSPATH."
                        + " Either fix that, or set your codebase appropriately.");
                System.out.println(re.toString());
                throw re;
            }

            long serverBootStop = System.currentTimeMillis();
            System.out.println("Server Boot Time = "
                    + (serverBootStop - serverBootStart) + "ms");

            System.out.println("Waiting for requests from clients...");
        } catch (Exception e) {
            System.out.println("exception in CFengineserver");
            e.printStackTrace();
        }
    }
}
