/*
* DBList.java
*
* Version 1.1
*
* 15/08/2002
*
* Copyright (c) 2003 Oregon State Universtity. All rights reserved.
 */

package org.recommender.server;

import java.io.*;
import java.sql.*;
import java.util.*;
import java.net.*;

/**
   Implements a queue of ratings to be written to the database.
   1 Accessor method to modify the queue (add ratings) shall be called by client threads,
   2 Accessor method to get the content of the queue shall be called by DBManager thread,

   Thread safe property is guaranted by using Object monitors to provide synchronization.

   @author  <A HREF="mailto:godde@cs.orst.edu">Olivier Godde</A>
   */
public class DBList{

    private static LinkedList dbQueries;

    private static String ratingTableName;
    private static String userFieldName;
    private static String itemFieldName;
    private static String ratingFieldName;
    private static boolean flushFlag;
    private static Object flushMonitor;
    private static int listSize = 20;

    /**
     * Static initializer of dbQueries.
     */
    static {
        dbQueries = new LinkedList();  //static initializer
    }

    /**
     * Initialize the table this DBList will contact with.
     * Called by DBManager thread.
     * @param ratingTable
     * @param userField
     * @param itemField
     * @param ratingField
     */
    public static void init(String ratingTable, String userField, String itemField, String ratingField)
    {
        ratingTableName= ratingTable;
        userFieldName  = userField;
        itemFieldName  = itemField;
        ratingFieldName= ratingField;
        flushFlag      = false;
        flushMonitor   = new Object();
    }

    /**
     * Insert clients' ratings into the table.
     * Synchronize so that two clients do not access the list at the same time.
     * When the size of the list increases to 50, notify the waiting DBManager thread
     * to add rating into database.
     * @param ItemRating to be inserted
     */
    public static void insertOrReplaceRating(ItemRating rating){
        // mysql supports the "replace into" as a enhanced SELECT that will
        // delete a row in a unique key if it already exists.
        String query = "REPLACE INTO " + getRatingTableName() + "(" + getUserFieldName() + ", " + getItemFieldName() + ", " + getRatingFieldName() + ")" +
                       " VALUES ('" + rating.getUserID() + "', '" + rating.getItemID() + "', '" + rating.getRating() + "')";

        synchronized (dbQueries) { //Enter this object's monitor
            //if (dbQueries.lastIndexOf(query) == -1)
            dbQueries.addLast(query);

            if (dbQueries.size() >= listSize) //Notify the waiting DBManager thread in that case
                dbQueries.notifyAll();
        }
    }

    /**
     * Remove clients' ratings from the table.
     * Synchronize so that two clients do not access the list at the same time.
     * @param ItemRating to be removed
     */
    public static void removeRating(int userID, int itemID){

        String query = "DELETE FROM " + getRatingTableName() + " WHERE " + getUserFieldName() + "= " +  userID + " AND " +
                       getItemFieldName() + "= " + itemID;

        synchronized(dbQueries){ //Enter this object's monitor
            //if (dbQueries.lastIndexOf(query) == -1)
            dbQueries.addLast(query);

            if (dbQueries.size() >= listSize) //Notify the waiting DBManager thread in that case
                dbQueries.notifyAll();
        }
    }

    /**
     * Update clients' ratings in the table.
     * Synchronize so that two clients do not access the list at the same time.
     * @param ItemRating to be updated
     */
    public static void updateRating(ItemRating rating){

        String query = "UPDATE " + getRatingTableName() + " SET " + getRatingFieldName()   + "= '" + rating.getRating() + "' " +
                       " WHERE " + getUserFieldName() + "= '" + rating.getUserID() + "' AND " +
                       getItemFieldName() + "= '" + rating.getItemID() + "'";

        synchronized(dbQueries){ //Enter this object's monitor
            //if (dbQueries.lastIndexOf(query) == -1)
            dbQueries.addLast(query);

            if (dbQueries.size() >= listSize) //Notify the waiting DBManager thread in that case
                dbQueries.notifyAll();
        }
    }

    /**
     * Flushes the list of queries, forcing execution of all the pending database requests.
     */
    public static void flush(){

        synchronized(dbQueries){  // Notify the DBManager thread, so that it executes the pending DB accesses
            flushFlag = true;
            dbQueries.notifyAll();
        }

        try{
            synchronized(flushMonitor){ // Wait in this object monitor. Will be awakened by DBManager when it is done with the DB queries.
                flushMonitor.wait();    // Useful for quitting the application (wait before exit)
            }
        } catch(Exception ex){
            System.out.println("Cannot execute wait call");
        }
    }


    /**
     * Wakes up a Thread that executed a "flush" operation.
     */
    public static void wakeUpFlushingThread(){
        synchronized(flushMonitor){
            flushMonitor.notifyAll();
        }
    }

    /**
     * Accessor function, to be called by DBManager Thread. The Thread enters the
     * monitor for the DBList and calls wait whenever there is less than 50
     * ratings waiting and it's been less than 30 seconds since the last time it
     * ran. Otherwise, it enters the monitor again to get a copy of the List and
     * returns this copy.
     * @return returnList
     */
    public static LinkedList getQueries(){

        synchronized (dbQueries) {
            if (dbQueries.size() < listSize && !flushFlag) {
                try{
                    dbQueries.wait(DBManager.minimumRunIntervalMs);
                } catch(java.lang.InterruptedException ex){
                    System.out.println("DBList WARNING: Cannot execute wait call, thread still running.");
                }
            }
        }

        LinkedList returnedList= null;

        synchronized (dbQueries) {
            returnedList= (LinkedList) dbQueries.clone();

            //re-initialze in that case, since the ratings will be written by DBManager
            if (dbQueries.size() >= listSize  || flushFlag)
                dbQueries.clear();

            if (flushFlag)
                flushFlag = false;
        }

        return returnedList;
    }

    /**
     * Accessor, return ratingTableName.
     * @return ratingTableName
     */
    public static String getRatingTableName() {
        return ratingTableName;
    }

    /**
     * Accessor, return userFieldName.
     * @return userFiledName
     */
    public static String getUserFieldName() {
        return userFieldName;
    }

    /**
     * Accessor, return itemFieldName.
     * @return itemFieldName
     */
    public static String getItemFieldName() {
        return itemFieldName;
    }

    /**
     * Accessor, return ratingFieldName.
     * @return ratingFieldName
     */
    public static String getRatingFieldName() {
        return ratingFieldName;
    }

}