/*
* DBManager.java
*
* Version 1.1
*
* 15/08/2002
*
* Copyright (c) 2003 Oregon State Universtity. All rights reserved.
 */

package org.recommender.server;

import java.io.IOException;
import java.sql.*;
import java.util.LinkedList;
import java.util.Iterator;

/**
 * Implementation of the DBManager.
 * This class handles the database accesses necessary to the CF Engine
 * Application. This includes reading and writing out User Info,
 * Item Info, Ratings Info.
 * The interface should be generic enough.
 *
 * @author  <A HREF="mailto:godde@cs.orst.edu">Olivier Godde</A>
 */
public class DBManager extends Thread
{

    // The minimal interval at which the DBManager runs in miliseconds
    public final static long minimumRunIntervalMs = 30000;

    private static String ratingTableName;
    private static String userFieldName;
    private static String itemFieldName;
    private static String ratingFieldName;

    private int maxUsers = 0;
    private int maxItems = 0;

    private static Statement stmt = null;

    /**
     * Constructor, get the database to contact with and
     * establish connection with it.
     */
    public DBManager(){

        SystemUtils.printDebug("Starting DB Manager...");

        // Init names...
        ratingTableName = SystemProperties.getString(
                "CFServer.db.ratingsTable", "");
        userFieldName   = SystemProperties.getString(
                "CFServer.db.userField", "");
        itemFieldName   = SystemProperties.getString(
                "CFServer.db.itemField", "");
        ratingFieldName = SystemProperties.getString(
                "CFServer.db.ratingField", "");

        DBList.init(ratingTableName, userFieldName, itemFieldName,
                    ratingFieldName);

        // Establish connection with databases
        Connection conn = null;
        try {
            conn = getConnection();
            stmt = conn.createStatement();
        } catch (SQLException ex) {
            try {
                conn = getConnection();
                stmt = conn.createStatement();
            } catch (SQLException e) {
                System.out.println("SQLException:");
                while (e != null) {
                    System.out.println("SQLState: " + e.getSQLState());
                    System.out.println("Message: " + e.getMessage());
                    System.out.println("Vendor: " + e.getErrorCode());
                    e = e.getNextException();
                    System.out.println("");
                }
            }
        }
        try {
        	conn.close();
        } catch (SQLException e) {
            System.out.println("SQLException:" + e);
        }
    }

    /**
     * An infinite loop, whenever the size of DBList is greater than 50,
     * get a copy of the dblist and insert all ratings in the list into
     * database.
     */
    public void run()
    {
        while(true) {  // While there are ratings to insert, perform this operation
            //SystemUtils.printDebug("DBManager: Entering Loop...");

            // The DBManager works with a copy of the original list
            LinkedList queryList = DBList.getQueries();
            //SystemUtils.printDebug("List size= " + queryList.size());

            try{
                executeQueries(queryList);
          /*  } catch (SQLException e) {
            	System.out.println("SQLException:" + e);
            	e.printStackTrace();
            } catch (IOException e) {
            	System.out.println("IOException:" + e);
            	e.printStackTrace();*/
            } catch(Exception ex){
                System.out.println("DBManager WARNING: Failure proceeding queries." );
            	//ex.printStackTrace();
            }
        }
    }

    /**
     * When the size of the DBList reaches its limit or needs to be flusehed, insert all ratings
     * waiting in this list into the database.
     * @param  queryList
     * @throws SQLException
     * @throws IOException
     */
    private void executeQueries(LinkedList queryList)  throws SQLException, IOException {
        Iterator iter= queryList.iterator();
        String query;
		Connection conn = null;
		
        // Check for connection persistance
        /*if (stmt == null){
            conn = getConnection();
            stmt = conn.createStatement();
        }*/
        try {
        	conn = getConnection();
            stmt = conn.createStatement();
        } catch (SQLException ex) {
        	//try again
            try {
        	    conn = getConnection();
                stmt = conn.createStatement();
            } catch (SQLException e) {
            	try {	
         			conn = getConnection();
         			stmt = conn.createStatement();
         		}  catch (SQLException ae) {         		
        	   		System.out.println("Fatal Database Error: " + ae);
                	ae.printStackTrace();
                }
            }
        }
        ResultSet rs= null;

        // Insert new ratings
        while(iter.hasNext()){
            query= (String) iter.next();

            try {
                rs = stmt.executeQuery(query);
                rs.close();
            } catch (SQLException e) {
                try{
                    conn = getConnection();
                    stmt = conn.createStatement();
                    rs = stmt.executeQuery(query);
                    rs.close();
                } catch(Exception ex){
                	// we can't risk loosing one rating, give a time limit
                	long start = System.currentTimeMillis();
                	while (conn == null && System.currentTimeMillis() < (start + 100)) 
                	{
                		conn = getConnection();
                    	stmt = conn.createStatement();
                    	rs = stmt.executeQuery(query);
                    	rs.close();
                    }                		
                  //  System.out.println("DBManager WARNING: DB access Failed");
                  //  System.out.println(query);
               }
            }
            yield();
        }
        
        conn.close();

        DBList.wakeUpFlushingThread();
    }

    /**
     * Get connection with database
     * @return Connection
     * @throws SQLException
     */
    private Connection getConnection() throws SQLException {
       /* String drivers = SystemProperties.getString("CFServer.jdbc.drivers");
        if (drivers != null)
            System.setProperty("jdbc.drivers", drivers);

        String url = SystemProperties.getString("CFServer.jdbc.url");
        String username = SystemProperties.getString("CFServer.jdbc.username");
        String password = SystemProperties.getString("CFServer.jdbc.password");

        return DriverManager.getConnection(url, username, password); */
        
        // We now use a connection pooling approach. The real SQL driver
        // is set in the static initializer of this class.
        return DriverManager.getConnection("jdbc:jdc:jdcpool");
        
    }

    /**
     * Accessor, return ratingTableName
     * @return ratingTableName
     */
    public String getRatingTableName() {
        return ratingTableName;
    }

    /**
     * Accessor, return userFieldName
     * @return userFieldName
     */
    public String getUserFieldName() {
        return userFieldName;
    }

    /**
     * Accessor, return itemFieldName
     * @return itemFieldName
     */
    public String getItemFieldName() {
        return itemFieldName;
    }

    /**
     * Accessor, return ratingFiledName
     * @return ratingFieldName
     */
    public String getRatingFieldName() {
        return ratingFieldName;
    }

    /**
     * Accessor, return maxUsers
     * @return maxUsers
     */
    public int getMaxUser() {
        return maxUsers;
    }

    /**
     * Accessor, return maxItems
     * @return maxItems
     */
    public int getMaxItem() {
        return maxItems;
    }
}