/*
 * DataManager.java *
 * Version 1.1
 *
 * 15/08/2002
 *
 * Copyright (c) 2003 Oregon State Universtity. All rights reserved.
 */

package org.recommender.server;

import java.io.*;
import java.sql.*;
import java.util.*;

/**
 Implementation of the DataManager.
 This class handles all the cached data in a CF Engine
 Application. This includes reading and writing out User Info,
 Item Info, Ratings Info. It will also be responsible for
 deciding when to cache in data and when to write it out to
 disk.
 The interface should be generic enough so that the transition
 is painless.
 @author  <A HREF="mailto:sameerk@cs.orst.edu">Sameer Kadam</A>
 @author  <A HREF="mailto:iyoon@cs.orst.edu">Irwin Yoon</A>
 @author  <A HREF="mailto:godde@cs.orst.edu">Olivier Godde</A>
 @author  <A HREF="mailto:dlowd@cs.orst.edu">Daniel Lowd</A>
 @author  <A HREF="mailto:wangyun@cs.orst.edu">Yun Wang</A>
 @author  <A HREF="mailto:mclaughm@cs.orst.edu">Matt McLaughlin</A>
 */
public class DataManager {
    /*
     *  The following are write-once fields - only modified at startup
     *  Access to them does not need to be synchronized
     */
private int resample = 0;
    // These fields hold the names of relational database table columns
    private static String ratingTableName;
    private static String userFieldName;
    private static String itemFieldName;
    private static String ratingFieldName;
    private static String userInfoTableName;
    private static String userInfoUserFieldName;
    private static String userInfoSamplingFieldName;
    private static String userInfoNumRatingsFieldName;

    // These fields holds the names of relational database tables
    private static String idTableName;
    private static String idHolder;
    private static String idName;
    private static String itemTypeTableName;
    private static String typeFieldName;

    private static int maxSampledUsers;    // Number of users to sample from the database
    private static int maxCachedUsers;     // Number of additional users to sample
    private static int numTopNRecords;     // Number of topNs to cache per user

    private boolean sampling_flag = false;   // Should we sample at all?

    private float minRating;
    private float maxRating;
    private float errRating;

    private List[] TypedItemList;
    private int maxType = 0;


    /*
     * The following fields may be modified more than once
     */

    //
    // These methods are protected with synchronized keywords
    //
    private int maxUsers = 0; // Max user ID
    private int numUsers = 0; // Total # of users in the database
    private int numItems = 0; // Total # of items in the database
    private int maxItems = 0; // Max item ID
    private int numUsersCached = 0; // Total # of cached users

    private final Object cacheMonitor = new Object();

    // Number of times we needed a user and found them already cached
    private int userCacheHits = 0;

    // Number of times we needed a user and had to load them from the database
    private int userCacheMisses = 0;

    //
    // todo These fields are not protected entirely!
    //

    // The IntHashMaps are now synchronized, but the entries inside them are not.
    // So we need to explore synchronizing the UserInfo and ItemInfo objects and
    // their subobjects
    //

    // IntHashMap - most methods are synchronized. IntHashMap allows modification while
    // you are iterating (unlike LinkedList). This seems safe to me for the most part.
    // However, if you don't synchronize, an entry may go away
    // between iter.next() and inthashmap.get(), so make sure you use a try/catch to
    // catch the NoSuchElementException.
    //
    //  The three datastructures below are all just views into the same data
    // If you are going to modify more than one within a method, you should consider
    // if you need to lock all the ones you are modifying for the entire time that
    // any of them are being modified.
    //
    // The UserInfo object has a field "useCount", which should be incremented whenever that
    // user is active. userUnload will never unload a user if useCount is positive.
    //
    // UserInfo objects should now be synchronized in most of the important places.
    //
    // RatingArray objects should only be modified within removeRating() and addRating
    // unless you reconsider the synchronization impacts of your changes. I believe they
    // relatively well protected know within those methods.
    //
    // ItemIndex seems reasonably well synchronized.
    //
    // Next question: What synchronization do we have to perform on the itemInfo?
    private SynIntHashMap ratingsMatrix = null; // Key: userid, Entries: UserInfo objects
    private SynIntHashMap itemIndex = null; // Key: userid, entries: ItemInfo objects
    LinkedList dbUsers;   // Unordered list of UserInfo structures for all users cached

    private PreparedStatement ps1;
    /**
     * Static initialization code.
     */
    static {
        String drivers = SystemProperties.getString("CFServer.jdbc.drivers");
        if (drivers != null) {
            System.setProperty("jdbc.drivers", drivers);

        }
        int test = SystemProperties.getInt("CFServer.Test", 0);
        String url = SystemProperties.getString("CFServer.jdbc.url")+ "?autoReconnect=true";
        if (test == 1)
        	url = SystemProperties.getString("CFServer.test.jdbc.url")+ "?autoReconnect=true";
        else
        	url = SystemProperties.getString("CFServer.jdbc.url")+ "?autoReconnect=true";
        	
        String username = SystemProperties.getString("CFServer.jdbc.username");
        String password = SystemProperties.getString("CFServer.jdbc.password");
        try {
            new org.recommender.util.pool.JDCConnectionDriver(url, username, password);
        } catch (Exception e) {
            // If the above code fails, the server will not be able to start, so
            // exiting is appropriate.
            System.out.println("Exception: " + e);
            System.out.println("Fatal error: Couldn't initialize sql connection pool");
            System.exit(1);
        }
    }

    /**
     Create a new DataManager object, establish connection with database, and cache ratings into memory.
     @param loadRatings
     @param doSampling
     */
    public DataManager(boolean loadRatings, boolean doSampling) {
        SystemUtils.printDebug("Starting Data Manager...");

        sampling_flag = doSampling;

        minRating = SystemProperties.getFloat("CFServer.db.minRating", 0);
        maxRating = SystemProperties.getFloat("CFServer.db.maxRating", 0);
        errRating = minRating - 1000;

        // temporaty
        // typeFieldName = "TypeID";
        itemTypeTableName = SystemProperties.getString("CFServer.db.itemTypeTable",
                "");
        typeFieldName = SystemProperties.getString("CFServer.db.typeField", "");
        idTableName = SystemProperties.getString("CFServer.db.idTable", "");
        idHolder = SystemProperties.getString("CFServer.db.idHolder", "");
        idName = SystemProperties.getString("CFServer.db.idName", "");

        // Load table and field names
        ratingTableName = SystemProperties.getString(
                "CFServer.db.ratingsTable", "");
        userFieldName = SystemProperties.getString("CFServer.db.userField", "");
        itemFieldName = SystemProperties.getString("CFServer.db.itemField", "");
        ratingFieldName = SystemProperties.getString("CFServer.db.ratingField",
                "");

        userInfoTableName = SystemProperties.getString(
                "CFServer.db.userInfo.table", "");
        userInfoUserFieldName = SystemProperties.getString(
                "CFServer.db.userInfo.userField", "");
        userInfoSamplingFieldName = SystemProperties.getString(
                "CFServer.db.userInfo.samplingField", "");
        userInfoNumRatingsFieldName = SystemProperties.getString(
                "CFServer.db.userInfo.numRatingsField", "");

        maxSampledUsers = SystemProperties.getInt("CFServer.mem.maxSampledUsers",
                1200);
        maxCachedUsers = SystemProperties.getInt("CFServer.mem.maxCachedUsers", 500);
        numTopNRecords = SystemProperties.getInt("CFServer.mem.numTopNRecords", 300);

        // Establish connection with databases
        Connection conn = null;
        Statement stmt;
        try {
           conn = getConnection();
           stmt = conn.createStatement();
           precompileSQLStatements(conn);
        } catch (SQLException e) {
            try {
                conn = getConnection();
                stmt = conn.createStatement();
                precompileSQLStatements(conn);
            }  catch (SQLException ex) {
               System.out.println("Fatal DB error: " + ex);
               System.exit(1);
            }
        }
        try {
            if (loadRatings) {
                SystemUtils.printDebug("Caching ratings matrix into memory...");
                String whereClause = SystemProperties.getString(
                        "CFServer.db.trainingSet", "1");
                long start = System.currentTimeMillis();
                cacheRatingsMatrix(whereClause);
                long stop = System.currentTimeMillis();
                SystemUtils.printDebug("Time to cache ratings:  "
                        + (stop - start) + " ms");

                // added by Yun Wang
                // cache typed item list
                start = System.currentTimeMillis();
                cacheTypedItemList();
                stop = System.currentTimeMillis();
                SystemUtils.printDebug("Time to cache typed item list:  "
                        + (stop - start) + " ms");

            } else {
                /*ratingsMatrix = new IntHashMap();
                  itemIndex     = new IntHashMap();
                  dbUsers      = new LinkedList();*/
                initializeData();
            }
        } catch (SQLException e) {
            System.out.println("SQLException:");
            while (e != null) {
                    System.out.println("SQLState: " + e.getSQLState());
                    System.out.println("Message: " + e.getMessage());
                    System.out.println("Vendor: " + e.getErrorCode());
                    e = e.getNextException();
                    System.out.println("");
           }
           System.exit(1);
        }

      /*  } catch (Exception e) {
            System.out.println("Exception: " + e);
            e.printStackTrace();
            System.exit(1);
        }*/
    }

    /**
     Precompile SQLStatements.
     @param conn
     @throws SQLException
     */
   private void precompileSQLStatements(Connection conn) throws SQLException {
        ps1 = conn.prepareStatement("SELECT * FROM " + getRatingDatabaseName()
                                    + " WHERE " + userFieldName + " = ?" );
        //      // IY precompile sql statements
//        // get neighbors
//        /*
//             ps1 = conn.prepareStatement ("SELECT UserID, Rating FROM " +
//             getRatingDatabaseName() +
//             " WHERE MovieID=?");
//                 //getUserObj
//             ps2 = conn.prepareStatement ("SELECT * FROM " +
//             getUserDatabaseName() +
//             " WHERE UserID=?");
//                 // hard code temp table name for now
//             ps3 = conn.prepareStatement ("CREATE TEMPORARY TABLE RatingTmp1 " +
//             "TYPE=HEAP " +
//             "SELECT * FROM " +
//             getRatingDatabaseName() +
//             " WHERE UserID=?");
//             ps4 = conn.prepareStatement ("CREATE TEMPORARY TABLE RatingTmp2 " +
//             "TYPE=HEAP " +
//             "SELECT * FROM " +
//             getRatingDatabaseName() +
//             " WHERE UserID=?");
//             ps5 = conn.prepareStatement ("INSERT INTO RatingTmp2" +
//             " SELECT * FROM " +
//             getRatingDatabaseName() +
//             " WHERE UserID=?");
//             ps6 = conn.prepareStatement ("DELETE FROM RatingTmp2");
//             ps7 = conn.prepareStatement ("SELECT AVG(Rating) FROM " +
//             getRatingDatabaseName() +
//             " WHERE UserID=?");
//             ps8 = conn.prepareStatement ("SELECT * FROM" +
//             " RatingTmp1, RatingTmp2" +
//             " WHERE RatingTmp1.MovieID = RatingTmp2.MovieID");
//         */
//
////        ps9 = conn.prepareStatement("SELECT Rating FROM " +
////                getRatingDatabaseName() +
////                " WHERE UserID =?" +
////                " AND MovieID =?");
//
//        /*  ps10 = conn.prepareStatement ("SELECT MovieID FROM " +
//            getUserDatabaseName() +
//            " WHERE MovieID=?");
//         */
//
    }

    /**
     Get connection with database.
     Get database properties and establish connection to database.
     @return Connection
     @throws SQLException
     */
    private Connection getConnection() throws SQLException {
        /*String drivers = SystemProperties.getString("CFServer.jdbc.drivers");
                 if (drivers != null)
            System.setProperty("jdbc.drivers", drivers);
                 String url = SystemProperties.getString("CFServer.jdbc.url");
             String username = SystemProperties.getString("CFServer.jdbc.username");
             String password = SystemProperties.getString("CFServer.jdbc.password");
                 return DriverManager.getConnection(url, username, password);*/
        // We now use a connection pooling approach. The real SQL driver
        // is set in the static initializer of this class.
        return DriverManager.getConnection("jdbc:jdc:jdcpool");
    }

    /**
     Get rating from the cache.
     @param userID
     @param itemID
     @return rating
     */
    public float getRatingCached(int userID, int itemID) {
        ItemInfo itemInfo = (ItemInfo) itemIndex.get(itemID);
        if (itemInfo == null) {
            return getErrRating();
        }

       	/* Yun: IntHashMap is no longer synchronized, have to add read lock.
       			I have scaned code. The usage in SimplePearson and DataManager
       			doesn't add lock on it at all, so I add lock here.
       	*/
        itemInfo.getReadersLock(true);

        IntHashMap hash = itemInfo.getItemRatingsHash();
        if (hash == null) {
        	itemInfo.releaseReadersLock();
            return getErrRating();
        }
        ItemRating itemRating = (ItemRating) hash.get(userID);

        if (itemRating != null) {
            itemInfo.releaseReadersLock();
            return itemRating.getRating();
        } else {
            itemInfo.releaseReadersLock();
            return getErrRating();
        }
    }

    /**
     Get rating from the cache.
     @param userID
     @param itemID
     @return rating
     */
    public ItemRating getItemRatingCached(int userID, int itemID) throws SQLException {
        // Also sets the inUse field, which ensures that userID's record is not
        // removed from the cache until we clear it.

        try {
            checkCachedUser(userID);
        } catch (Exception e) {
            System.out.print("checkCachedUser() error");
            e.printStackTrace();
        }

        ItemInfo itemInfo = (ItemInfo) itemIndex.get(itemID);

        if (itemInfo == null) {
            return new ItemRating(userID, itemID, getErrRating());
        }

        // Yun: IntHashMap is no longer synchronized, we have to add read lock
        itemInfo.getReadersLock(true);
        IntHashMap hash = itemInfo.getItemRatingsHash();
        if (hash == null) {
            itemInfo.releaseReadersLock();
            return new ItemRating(userID, itemID, getErrRating());
        }

        ItemRating ir;
        ir = (ItemRating) hash.get(userID);
        itemInfo.releaseReadersLock();

        if (ir == null) {
             ir = new ItemRating(userID, itemID, getErrRating());
        }

        decUseCount(userID);
        return ir;
    }

    /**
     Method to print contents of all tables.
     */
    public void printAllTables() {
        // printUserTable();
        // printItemTable();
        printRatingTable();
        System.out.println();
    }

    /**
     Print ratingTable
     */
    public void printRatingTable() {
        System.out.println();
        System.out.println("Rating Table");
        System.out.println("======================================");
        try {
            showTable(getRatingDatabaseName());
        } catch (SQLException e) {
            System.out.println("Error: Error accesing table "
                    + getRatingDatabaseName());
            System.out.println(e);
        }
    }

    /**
     Print a specified table.
     @param tableName
     @throws SQLException
     */
    private void showTable(String tableName) throws SQLException {
        String query = "SELECT * FROM " + tableName;
        ResultSet rs = null;
        Statement stmt;
        Connection conn = null;
        try {
            conn = getConnection();
            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);
        } catch (Exception e) {
            try {
                conn = getConnection();
                stmt = conn.createStatement();
                rs = stmt.executeQuery(query);
            } catch (Exception ex) {
                System.err.println("Fatal DB error: " + ex);
                System.exit(1);
            }
        }

        ResultSetMetaData rsmd = rs.getMetaData();
        int colCount = rsmd.getColumnCount();
        while (rs != null && rs.next()) {
            for (int i = 1; i <= colCount; i++) {
                if (i > 1) {
                    System.out.print(", ");
                }
                System.out.print(rs.getString(i));
            }
            System.out.println();
        }
        rs.close();
        conn.close();
    }

    /************ Helperfunctions for the CF algorithms *****************/
    /*                              OG
     * This interface must be general enough to be completely independent
     * from the Algorithm part.
     */

    /**
     Returns a list of ItemRatings representing ratings of users.
     A neighbor means someone who has rated itemID.
     @param itemID
     @return Iterator of neighbors
     */
    public Iterator getNeighborhood(int itemID) {
        Iterator iter;
        ItemInfo itemInfo = (ItemInfo) itemIndex.get(itemID);
        if (itemInfo == null) {
            // Return an interator to an empty list
            return (new LinkedList()).iterator();
        }

       	// Yun: IntHashMap is no longer synchronized, have to add read lock
       	itemInfo.getReadersLock(true);
        IntHashMap itemRatingsHash = itemInfo.getItemRatingsHash();
        iter = itemRatingsHash.iterator();
        itemInfo.releaseReadersLock();
        return iter;
    }
    public ItemInfo getItemInfo(int itemID){
        return (ItemInfo)itemIndex.get(itemID);
    }
    /**
     Returns a list of item ratings for a given user using the cached data
     Load the user into the cache if not done yet (because of data sampling).
     @param userID
     @return Iterator of items.
     */
    public Iterator getItemList(int userID) {
        Iterator iter;

        UserInfo userInfo = (UserInfo) ratingsMatrix.get(userID);
        if (userInfo == null) {
            // Return empty list
            return (new LinkedList()).iterator();
        }

   //     userInfo.getReadersLock(true);
        RatingArrays ratings = userInfo.getRatings();
        LinkedList ratingsList = ratings.getRatingsList();
        iter = ratingsList.iterator();
   //     userInfo.releaseReadersLock();
        return iter;
    }

    /**
     * Return an array of userIDs who have rated the given item.
     * @param itemID
     * @return retArray an array of ItemRatings
     */
    public ItemRating[] getItemRatingList(int itemID) {
        IntHashMap itemHash;
        Iterator iter;

        if (itemID > getMaxItem()) {
            return new ItemRating[0];
        }

        ItemInfo itemInfo = (ItemInfo) itemIndex.get(itemID);
        if (itemInfo == null) {
            itemHash = new IntHashMap();
        } else {
            itemHash = itemInfo.getItemRatingsHash();
        }
        ItemRating[] retArray = new ItemRating[itemHash.size()];
        int i = 0;
        iter = itemHash.iterator();
        while (iter.hasNext()) {
            retArray[i] = (ItemRating) iter.next();
            i++;
        }
        return retArray;
    }

    /**
     Return an array of itemIDs which the given user has rated.
     @param userID
     @return retArray an array of itemRatings
     */
    public ItemRating[] getUserRatingList(int userID) {
        if (userID > getMaxUser()) {
            return new ItemRating[0];
        }

        UserInfo userInfo = (UserInfo) ratingsMatrix.get(userID);
        if (userInfo == null) {
            // Return a zero-length array
            return new ItemRating[0];
        }

        if (!userInfo.getReadersLock(true)) return new ItemRating[0];
        RatingArrays ratings = userInfo.getRatings();
        ItemRating[] ret = ratings.getItemRatingsArray();
        userInfo.releaseReadersLock();
        return ret;
    }

    /**
     Return an iterator through all user ID's, each one an Integer object.
     @return Iterator of userIDs
     */
    public Iterator getAllUsers() {
        return dbUsers.iterator();
    }

    /**
     Return the mean of user's ratings from the cache.
     @param userID
     @return mean rating
     */
    public float getMeanRatingCached(int userID) {
        //value = ((Float) userMeans.get(userID));
        //mean = value.floatValue();
        UserInfo userInfo = (UserInfo) ratingsMatrix.get(userID);
        if (userInfo == null) {
            return getErrRating();
        }
        return userInfo.getUserMean();
    }

    /**
     Return the mean of item's ratings from the cache.
     @param itemID
     @return mean rating
     */
    public float getMeanItemRatingCached(int itemID) {
        float mean;
        ItemInfo itemInfo = (ItemInfo) itemIndex.get(itemID);
        if (itemInfo == null) {
            mean = getErrRating();
        } else {
            mean = itemInfo.getItemMean();
        }
        return mean;
    }

    /**
     Recalculate mean and standard deviation for a given user.
     Used when ratings in the database change for testing and
     and a new rating is added.
     @param userID
     */
    public void updateStats(int userID) {

        // Yun: do not add lock here, take care in the higher level
        Stats userStats = new Stats();

        Iterator iter = getItemList(userID);

        while (iter.hasNext()) {
            ItemRating r = (ItemRating) iter.next();
            userStats.addNum(r.getRating());
        }

        float mean = userStats.calcMean();
        float sd = userStats.calcSD();

        if (userID <= getMaxUser() + 1) { //In case the function is called out of the context of addCachedRatingSafe
            UserInfo userInfo = (UserInfo) ratingsMatrix.get(userID);
            userInfo.setUserMean(mean);
            userInfo.setUserSD(sd);
        }

        //mean = ((UserInfo) ratingsMatrix.get(userID)).userMean;
    }

    /**
     Returns the standard deviation of the user's ratings from the cache.
     @param userID
     @return userSD
     */
    public float getUserSD(int userID) {
        //get all items the user has rated
        //SD = ((Float)userSD.get(userID)).floatValue();
        UserInfo userInfo = (UserInfo) ratingsMatrix.get(userID);
        if (userInfo == null) {
            return -1;
        }

        return userInfo.getUserSD();
    }

    /**
     Returns the standard deviation of the item's ratings from the cache.
     @param itemID
     @return itemSD
     */
    public float getItemSD(int itemID) {
        float SD;
        ItemInfo itemInfo = (ItemInfo) itemIndex.get(itemID);
        if (itemInfo == null) {
            SD = -1;
        } else {
            SD = itemInfo.getItemSD();
        }
        return SD;
    }

    /**
     Cache ratings from database into memory - avoid too many disk IO access,
     If sampling_flag is false, then cache all ratings in whereClause,
     otherwise only cache sampled ratings.
     @param whereClause
     @throws SQLException
     */
    private void cacheRatingsMatrix(String whereClause) throws SQLException {

        String query = "SELECT MAX(" + itemFieldName + ") FROM " +
                getRatingDatabaseName();

        int myMaxItems = 0;

        ResultSet rs = null;
        Statement stmt;
        Connection conn = null;
        try {
            conn = getConnection();
            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);
        } catch (Exception e) {
            try {
                conn = getConnection();
                stmt = conn.createStatement();
                rs = stmt.executeQuery(query);
            } catch (Exception ex) {
                System.err.println("Fatal DB error: " + ex);
                System.exit(1);
            }
        }

        if (rs != null && rs.next()) {
            myMaxItems = rs.getInt(1);
        }
        setMaxItems(myMaxItems);

        rs.close();

        //System.out.println(maxItems);
        //initialize item/users vector
        //itemIndex = new ArrayList(maxItems+1);
        itemIndex = new SynIntHashMap(new IntHashMap());

        /*for (int j=0; j <= maxItems; j++)
            itemIndex.add(j, new IntHashMap());*/

        query = "SELECT MAX(" + userFieldName + ") FROM " + getRatingDatabaseName();
        dbUsers = new LinkedList();

        try {
            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);
        } catch (Exception e) {
            try {
                conn = getConnection();
                stmt = conn.createStatement();
                rs = stmt.executeQuery(query);
            } catch (Exception ex) {
                System.err.println("Fatal DB error: " + ex);
                System.exit(1);
            }
        }

        int myMaxUsers = 0;
        if (rs != null && rs.next()) {
            myMaxUsers = rs.getInt(1);
        }
        setMaxUsers(myMaxUsers);
        rs.close();
        conn.close();
        SystemUtils.printDebug("Max users = " + myMaxUsers);


        ratingsMatrix = new SynIntHashMap(new IntHashMap());
        /*ratingsMatrix.add(0, null);
            userMeans = new ArrayList(maxUsers+1);
            userMeans.add(0, null);
            userSD = new ArrayList(maxUsers+1);
            userSD.add(0, null);*/

        if (!sampling_flag) {
            cacheRatings(whereClause);
        } else {
            cacheSampledRatings();
        }
    }

    /**
     Cache all ratings specified in whereClause into memory to avoid going to
     disk everytime ratings are needed,also responsible for update user stats.
     @param whereClause
     @return the last user's ID
     @throws SQLException
     */
    private int cacheRatings(String whereClause) throws SQLException {
        IntHashMap currentUserRatings = new IntHashMap();
        Stats userStats = new Stats();

        int currUserID = 0;
        int prevUserID = 0;
        int myNumUsersCached;

//        synchronized (dbUsers) {
//            synchronized (itemIndex) {
//                synchronized (ratingsMatrix) {
                    SystemUtils.printDebug("Caching Ratings for Users in "
                            + getRatingDatabaseName());
                    String query = "SELECT * FROM " + getRatingDatabaseName()
                            + " WHERE (" + whereClause + ") "
                            + " ORDER BY " + userFieldName + ", " + itemFieldName;

                    System.out.println(query);
                    ResultSet rs = null;
                    Statement stmt;
                    Connection conn = null;
                    try {
                        conn = getConnection();
                        stmt = conn.createStatement();
                        rs = stmt.executeQuery(query);
                    } catch (SQLException e) {
                        try {
                            conn = getConnection();
                            stmt = conn.createStatement();
                            rs = stmt.executeQuery(query);
                        } catch (SQLException ex) {
                            System.err.println("Fatal DB error: " + ex);
                            System.exit(1);
                        }

                    }

                    System.out.println();

                    /*for (int j=0; j <= maxItems; j++)
                    itemIndex.add(j, new IntHashMap());*/

                    //int index= 0;
                    myNumUsersCached = 0;
                    prevUserID = 0;
                    UserInfo userInfo;
                    while (rs != null && rs.next()) {
                        currUserID = rs.getInt("UserID");

                        if (currUserID != prevUserID) { // Got data for new user
                            // Add current user to dbUsers list FHH 3/17/02
                            dbUsers.add(new Integer(currUserID));
                            myNumUsersCached++;

                            //OG add
                            /*userMeans.add(prevUserID, new Float(userStats.calcMean()));
                            userSD.add(prevUserID, new Float(userStats.calcSD()));*/
                            float[] ratings = new float[currentUserRatings.size()];
                            int[] itemIds = new int[currentUserRatings.size()];
                            Iterator iter = currentUserRatings.iterator();
                            int i = 0;
                            while (iter.hasNext()) {
                                ItemRating ir = (ItemRating) iter.next();
                                ratings[i] = ir.getRating();
                                itemIds[i] = ir.getItemID();
                                i++;
                            }
                            userInfo = new UserInfo(prevUserID, new RatingArrays(itemIds, ratings),
                                    userStats.calcMean(), userStats.calcSD(), false);
                            ratingsMatrix.put(prevUserID, userInfo);

                            // Add current users data to matrix
                            // If users are missing between current user ID
                            // and previous user ID, insert null into matrix
                            // for those missing user
                            ///for (int k = prevUserID + 1; k < currUserID; k++){
                            /*ratingsMatrix.add(k, null);
                            userMeans.add(k, null);
                            userSD.add(k, null);*/
                            //    ratingsMatrix.add(k, new UserInfo(k));
                            //}

                            //initialize for new user
                            userStats = new Stats();
                            currentUserRatings = new IntHashMap();
                            //index= 0;
                            //ratingsMatrix.add(currUserID, currentUserRatings);

                            prevUserID = currUserID;
                        }

                        int userID = rs.getInt(userFieldName);
                        int itemID = rs.getInt(itemFieldName);
                        float rating = rs.getFloat(ratingFieldName);

                        userStats.addNum(rating);
                        currentUserRatings.put(itemID, new ItemRating(userID, itemID, rating));
                        //currentUserRatings.setRating(itemID, rating);


                        //OG create new value in the items/users matrix
                        //Why doing this in a separate function?
                        //We can do it in one database access!?
                        // Create new value in the items/users matrix
                        if (!itemIndex.containsKey(itemID)) {
                            //add new ItemInfo to matrix
                            itemIndex.put(itemID, new ItemInfo(itemID));
                        }
                        ((ItemInfo) itemIndex.get(itemID)).addRating(new ItemRating(currUserID,
                                itemID, rating));
                    }
                    setNumUsersCached(myNumUsersCached);
                    rs.close();
                    conn.close();

                    /* DL Take care of statistics for the last user.
                    *    This is important, since we only add the mean and SD for
                    *    a user when we encounter a different user.  Since there
                    *    is no user after the last one, we must take care of it
                    *    specially.
                    */
                    if (currUserID > 0) {
                        /*userMeans.add(prevUserID, new Float(userStats.calcMean()));
                        userSD.add(prevUserID, new Float(userStats.calcSD()));*/
                        float[] ratings = new float[currentUserRatings.size()];
                        int[] itemIds = new int[currentUserRatings.size()];
                        Iterator iter = currentUserRatings.iterator();
                        int i = 0;
                        while (iter.hasNext()) {
                            ItemRating ir = (ItemRating) iter.next();
                            ratings[i] = ir.getRating();
                            itemIds[i] = ir.getItemID();
                            i++;
                        }
                        userInfo = new UserInfo(prevUserID, new RatingArrays(itemIds, ratings),
                                userStats.calcMean(), userStats.calcSD(), false);
                        ratingsMatrix.put(prevUserID, userInfo);
                    }
 //               }
//            }
//        }
        SystemUtils.printDebug("cached " + myNumUsersCached + " users");

        return currUserID;
    }

    /**
     Cache sampled ratings into memory.
     @return the last user's ID
     @throws SQLException
     */
    private int cacheSampledRatings() throws SQLException {

        //initialze the data structures
        initializeData();

        int currUserID = 0;
        int prevUserID = 0;

        Statement stmt;
        Connection conn = null;
        ResultSet rs = null;

        // find out how many ratings we're going to cache -- this assumes that
        // the NumRatings field is reasonably up to date, otherwise ratings
        // toward the lower end of Utility will be incomplete or uncached
        SystemUtils.printDebug("Computing number of ratings to obtain");
        String count_query = "SELECT " + userInfoNumRatingsFieldName + 
                " FROM " + getUserInfoTableName() +
                " ORDER BY " + userInfoSamplingFieldName + " DESC LIMIT " + maxSampledUsers;
        System.out.println(count_query);

        try {
          conn = getConnection();
          stmt = conn.createStatement();
          rs = stmt.executeQuery(count_query);
        } catch (Exception e) {
          try {
            conn = getConnection();
            stmt = conn.createStatement();
            rs = stmt.executeQuery(count_query);
          } catch (Exception ex) {
            System.err.println("Fatal DB error: " + ex);
            System.exit(1);
          }
        }

        System.out.println();

        int numRatingsCache = 0;
        while (rs != null && rs.next()) {
          numRatingsCache += rs.getInt(userInfoNumRatingsFieldName);
        }
        rs.close();
        // don't close connection yet
        SystemUtils.printDebug("Need at least " + numRatingsCache + " ratings");
        // allow for a small amount of slop in case the computed values are old
        numRatingsCache *= 1.05;
        SystemUtils.printDebug("Getting " + numRatingsCache + " ratings");

        // This *needs* a LIMIT clause to prevent silly amounts of memory usage
        // due to mysql_store_result.
        SystemUtils.printDebug("Caching Ratings for Users in "
                + getRatingDatabaseName());
        String query = "SELECT " + userFieldName + ", "
                + itemFieldName + ", "
                + ratingFieldName + ", "
                + userInfoSamplingFieldName +
                " FROM " + getRatingDatabaseName() + ", " + getUserInfoTableName() +
                " WHERE " + userFieldName + "=" + userInfoUserFieldName
                + " ORDER by 4 DESC, 2 LIMIT " + numRatingsCache;

        System.out.println(query);

        // Lock down the user cache datastructures to ensure they remain consistent.
//        synchronized (dbUsers) {
//            synchronized (itemIndex) {
//                synchronized (ratingsMatrix) {
                    try {
                        conn = getConnection();
                        stmt = conn.createStatement();
                        rs = stmt.executeQuery(query);
                    } catch (Exception e) {
                        try {
                            conn = getConnection();
                            stmt = conn.createStatement();
                            rs = stmt.executeQuery(query);
                        } catch (Exception ex) {
                            System.err.println("Fatal DB error: " + ex);
                            System.exit(1);
                        }
                    }

                    System.out.println();

                    /////////////////////////// Some Declaration ////////////////////////////////////////////
                    IntHashMap currentUserRatings = new IntHashMap();
                    Stats userStats = new Stats();
                    int myNumUsersCached = 0;
                    prevUserID = -1;
                    boolean first = true;
                    /////////////////////////////////////////////////////////////////////////////////////////

                    UserInfo userInfo;
                    while (rs != null && rs.next() && myNumUsersCached < maxSampledUsers) {
                        currUserID = rs.getInt(userFieldName);

                        if (currUserID != prevUserID) { // Got data for new user
                            // Add current user to dbUsers list FHH 3/17/02
                            dbUsers.add(new Integer(currUserID));
                            myNumUsersCached++;

                            // Add user stats
                            if (!first) {


                                float[] ratings = new float[currentUserRatings.size()];
                                int[] ratedItemIDs = new int[currentUserRatings.size()];
                                int i = 0;
                                Iterator iter = currentUserRatings.iterator();
                                while (iter.hasNext()) {
                                    ItemRating itemRating = (ItemRating) iter.next();
                                    ratings[i] = itemRating.getRating();
                                    ratedItemIDs[i] = itemRating.getItemID();
                                    i++;
                                }
                                userInfo = new UserInfo(prevUserID, new RatingArrays(ratedItemIDs, ratings),
                                        userStats.calcMean(), userStats.calcSD(), true);
                                ratingsMatrix.put(prevUserID, userInfo);
                                //insertionSort(userInfo.ratedItemIDs, userInfo.ratings);

                            } else {
                                first = false;

                                // Initialize for new user
                            }
                            userStats = new Stats();

                            currentUserRatings = new IntHashMap();
                            prevUserID = currUserID;
                        }

                        int itemID = rs.getInt(itemFieldName);
                        float rating = rs.getFloat(ratingFieldName);

                        userStats.addNum(rating);
                        currentUserRatings.put(itemID,
                                new ItemRating(currUserID, itemID, rating));

                        if (!itemIndex.containsKey(itemID)) {
                            //add new ItemInfo to matrix
                            itemIndex.put(itemID, new ItemInfo(itemID));
                        }
                        ((ItemInfo) itemIndex.get(itemID)).addRating(new ItemRating(currUserID,
                                itemID, rating));
                    }
                    setNumUsersCached(myNumUsersCached);
                    rs.close();
                    conn.close();

                    /*    Take care of statistics for the last user.
                    *    This is important, since we only add the mean and SD for
                    *    a user when we encounter a different user.  Since there
                    *    is no user after the last one, we must take care of it
                    *    specially.
                    */
                    if (currUserID > 0) {

                        float[] ratings = new float[currentUserRatings.size()];
                        int[] ratedItemIDs = new int[currentUserRatings.size()];

                        Iterator iter = currentUserRatings.iterator();
                        int i = 0;
                        while (iter.hasNext()) {
                            ItemRating itemRating = (ItemRating) iter.next();
                            ratings[i] = itemRating.getRating();
                            ratedItemIDs[i] = itemRating.getItemID();
                            i++;
                        }
                        userInfo = new UserInfo(currUserID, new RatingArrays(ratedItemIDs, ratings),
                                userStats.calcMean(), userStats.calcSD(), true);
                        ratingsMatrix.put(currUserID, userInfo);
                    }

                    SystemUtils.printDebug("Cached " + myNumUsersCached + " users");
                    SystemUtils.printDebug("Cached " + ratingsMatrix.size() + " users");

                    Runtime rt = Runtime.getRuntime();
                    System.out.println("Total memory = " + rt.totalMemory()
                            + ", free memory = " + rt.freeMemory());
//                }
//            }
//        }
        return currUserID;
    }

    /**
     *  Simple insertion sort.
     *  @param a an array of int items - the sort key.
     * @param b an array of parallel ratings of the same length as a
     **/
    public static void insertionSort(int[] a, float[] b) {
        for (int p = 1; p < a.length; p++) {
            int tmpa = a[p];
            float tmpb = b[p];
            int j = p;

            for (; (j > 0) && (tmpa < a[j - 1]); j--) {
                a[j] = a[j - 1];
                b[j] = b[j - 1];
            }
            a[j] = tmpa;
            b[j] = tmpb;
        }
    }

    /**
     *
     * Function called by the Resample Thread, at regular intervals of time.
     * Select the most relevant user data from the database, and insert (if necessary) into the cache.
     * After, this operation, remove the users that are no longer selected by the sampling method
     * and have their reference bit set to "0".
     * Note that this method will have no effect if you haven't rerun the calcUserInfo utility.
     * This method is synchronized to prevent it from being called by two different threads.
     */
    public void sampleUsers() {
        // Yikes! The following code was just deleting all of the datastructures.
        // This method doesn't seem to be the one to be called at startup, so I think
        // it's safe if we don't reinitialize all the datastructures.

//        try {
//            initializeData();
//        } catch (SQLException e) {
//            e.printStackTrace();
//            System.exit(1);
//        }

        int numUsersAdded = 0;
        int numUsersDiscarded = 0;
        SystemUtils.printDebug("Sampling Ratings for Users in "
                + getRatingDatabaseName());
        String query = "SELECT DISTINCT " + userFieldName + ", "
                + userInfoSamplingFieldName +
                " FROM " + getRatingDatabaseName() + ", " + getUserInfoTableName() +
                " WHERE " + userFieldName + "=" + userInfoUserFieldName
                + " ORDER by 2 DESC, 1";

        System.out.println(query);
        ResultSet rs = null;
        Statement stmt2 = null;
        Connection conn = null;
        // Get selected users
        try {
            conn = getConnection();
            stmt2 = conn.createStatement();
            rs = stmt2.executeQuery(query);
        } catch (Exception e) {
            try {
            	rs.close();
            	conn.close();
                conn = getConnection();
                stmt2 = conn.createStatement();
                rs = stmt2.executeQuery(query);
            } catch (Exception ex) {
                // Yun: if can not connect to DB now, skip sampleuser() and wait for the next time.
                System.err.println("Fatal DB error: "+ ex);
        //        ex.printStackTrace();
                System.err.println("Unable to resample.");
                return;
            }
        }

        Iterator iter = ratingsMatrix.iterator();
        UserInfo user;

        // We initially mark each user record as "not wanted" in the cache
        try {
            // We use try catch, so that we don't have to synchronize
            // the hasNext/next pair.
            while (true) {
                user = (UserInfo) iter.next();
                user.samplingFlag = false;
            }
        } catch (NoSuchElementException e) {
           // Ignore - just means we've reached the end of the list
        }

        int nUsers = 0;
        int currUserID;
        //long testStart = System.currentTimeMillis();

	/*	resample++;
		File outputFile = new File("sample_resultv" + resample);
        FileWriter fout = null;
        try {
	       fout = new FileWriter(outputFile);
	    } catch (java.io.IOException ex){
		   System.out.println("ERROR:out file not found");
	    } */

        // Load new sampled users, set sampling flag for users who were already here
        try {
      // 	BufferedWriter out = new BufferedWriter(fout);
      //  	out.write("Sampling result " + resample);
     // 		out.newLine();

            while (rs != null && rs.next() && nUsers < maxSampledUsers) {
                currUserID = rs.getInt(userFieldName);
                UserInfo ui = (UserInfo) ratingsMatrix.get(currUserID);
                if (ui != null) {
                    ui.samplingFlag = true;
      //              out.write(currUserID + "\tin cache" );

                } else {
                    ui = loadUser(currUserID, true);
    //                out.write(currUserID + "\tnewly loaded" );

                    ui.decUseCount();             // We are not current performing a CFEngine op on this user
                    numUsersAdded++;
                }
                nUsers++;
     //           out.newLine();
            }
            rs.close();
            conn.close();
     //       out.newLine();
     //       out.newLine();
     //       out.close();
        } catch (SQLException ex) {
            System.out.println("Fatal DB error: "+ ex);
            ex.printStackTrace();
        } /* catch(java.io.IOException ex){

            System.out.println("File write error:" + ex);

         }*/

        //long testStop = System.currentTimeMillis();
        //System.out.println("Time to go through result set=" + (testStop - testStart));

        iter = ratingsMatrix.iterator();

        // Clean cache. Remove ratings info for users no longer sampled and not referenced
        try {
            // We try/catch the NoSuchElementException, so that we don't have to
            // do a hasNext()/next() combo, which would require synchronization
            // of ratingsMatrix for the entire loop.
            while (true) {
                user = (UserInfo) iter.next();

                if (!user.samplingFlag && !user.isReferenced()) {
                    userUnload(user.userID);
                    numUsersDiscarded++;
                }
            }
        } catch (NoSuchElementException e) {
            // Ignore it - it means we reached the end of the list.
        }

        System.gc();

        System.out.println("Num users in cache = " + ratingsMatrix.size());
        System.out.println("Num items in cache = " + itemIndex.size());
        System.out.println("SampleUsers - Users added to cache: " + numUsersAdded);
        System.out.println("SampleUsers - Users discarded from cache: " + numUsersDiscarded);
    }

    /**
     Get an iterator through a list of all ItemRatings in the given table,
     ordered by UserID.  Used for experiments.
     @param whereClause
     @return Iterator of ratings
     @throws SQLException
     */
    public Iterator getRatingList(String whereClause) throws SQLException {
        LinkedList ratingList = new LinkedList();

        String query = "SELECT * FROM " + getRatingDatabaseName()
                + " WHERE (" + whereClause + ") "
                + " ORDER by " + userFieldName + ", " + itemFieldName;
        ResultSet rs = null;
        Statement stmt;

        Connection conn = null;
        try {
            conn = getConnection();
            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);
        } catch (Exception e) {
            try {
                conn = getConnection();
                stmt = conn.createStatement();
                rs = stmt.executeQuery(query);
            } catch (Exception ex) {
                System.err.println("Fatal DB error: " + ex);
                System.exit(1);
            }
        }

        while (rs != null && rs.next()) {
            int userID = rs.getInt(userFieldName);
            int itemID = rs.getInt(itemFieldName);
            float rating = rs.getFloat(ratingFieldName);

            ratingList.add(new ItemRating(userID, itemID, rating));
        }
        rs.close();
        conn.close();
        return ratingList.iterator();
    }

    /**
     Cache typed item list from database into memory - avoid too many disk IO access
     @throws SQLException
     */
    public void cacheTypedItemList() throws SQLException {
        String query = "SELECT MAX(" + typeFieldName + ") FROM " +
                getItemTypeInfoDatabaseName(); //typeInfoTableName;
        //  System.out.println(query);

        maxType = 0;
        ResultSet rs = null;
        Statement stmt;

        Connection conn = null;
        try {
            conn = getConnection();
            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);
        } catch (Exception e) {
            try {
                conn = getConnection();
                stmt = conn.createStatement();
                rs = stmt.executeQuery(query);
            } catch (Exception ex) {
                System.err.println("Fatal DB error: " + ex);
                System.exit(1);
            }
        }

        if (rs != null && rs.next()) {
            maxType = rs.getInt(1);
        }
        rs.close();
        conn.close();
        //System.out.println("maxType is " + maxType);

        TypedItemList = new List[maxType + 1];
        // TypedItemList[0] = new Vector();
        for (int i = 0; i <= maxType; i++) {
            List itemList = getTypedItemIDList(i);
            TypedItemList[i] = new Vector(itemList);
        }
    }

    /**
     Get the whole list of ItemID of specific type
     @param typeID
     @return Iterator of itemIDs
     @throws SQLException
     */
    public LinkedList getTypedItemIDList(int typeID) throws SQLException {

        LinkedList itemList = new LinkedList();

        String query = "SELECT " + itemFieldName + " FROM " +
                getItemTypeInfoDatabaseName()
                + " WHERE " + typeFieldName + " = " + typeID + " ORDER BY "
                + itemFieldName;

        ResultSet rs = null;
        Statement stmt;
        Connection conn = null;
        try {
            conn = getConnection();
            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);
        } catch (Exception e) {
            try {
                conn = getConnection();
                stmt = conn.createStatement();
                rs = stmt.executeQuery(query);
            } catch (Exception ex) {
                System.err.println("Fatal DB error: " + ex);
                System.exit(1);
            }
        }

        while (rs != null && rs.next()) {
            itemList.add(new Integer(rs.getInt(1)));
        }
        rs.close();
        conn.close();
       
        return itemList;
    }

    /**
     Get an iterator through a list of specific typed Item List.
     @param typeID
     @return Iterator of typed item list
     */
    // todo: sync - are lists in TypedItemList modified after starting server?
    public Iterator getCachedTypedItemLists(int typeID) {
        List l = TypedItemList[typeID];
        return l.iterator();
    }

    /**
     Get a vector of specific typed Item List.
     @param typeID
     @return Vector of typed item list
     */
    // todo: sync - are lists in TypedItemList modified after starting server?
    public Vector getCachedTypedItemList(int typeID) {

        Vector l = (Vector) TypedItemList[typeID];
        return l;
    }

    /**
     Get size of specific typed Item List.
     @param typeID
     @return size
     */
    public int getSizeOfTypedItemList(int typeID) {
        return TypedItemList[typeID].size();
    }

    /**
     Accessor, return itemTypeTableName
     @return itemTypeTableName
     */
    public String getItemTypeInfoDatabaseName() {
        return itemTypeTableName;
    }

    /**
     Accessor, return UserInfo
     @return UserInfo
     */
    public UserInfo getUser(int userID) {
        return (UserInfo) ratingsMatrix.get(userID);
    }

    /**
     Get an iterator through a list of all userIDs in the given table,
     ordered by UserID.
     @param tableName
     @return Iterator of userIDs
     @throws SQLException
     */
    public Iterator getUserList(String tableName) throws SQLException {
        LinkedList userList = new LinkedList();
        String query = "SELECT UserID FROM " + tableName + " ORDER by UserID";
        ResultSet rs = null;
        Statement stmt;
        Connection conn = null;
        try {
            conn = getConnection();
            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);
        } catch (Exception e) {
            try {
                conn = getConnection();
                stmt = conn.createStatement();
                rs = stmt.executeQuery(query);
            } catch (Exception ex) {
                System.err.println("Fatal DB error: " + ex);
                System.exit(1);
            }
        }
        while (rs != null && rs.next()) {
            userList.add(new Integer(rs.getInt(1)));
        }
        rs.close();
        conn.close();
        return userList.iterator();
    }

    /**
     This method initializes the data structures for caching with the sizes of the max user and item ID.
     @throws SQLException
     */
    public void initializeData() throws java.sql.SQLException {
        String query = "SELECT MAX(" + itemFieldName + ") FROM " +
                getRatingDatabaseName();
        ResultSet rs = null;
        Statement stmt = null;
        Connection conn = null;
        try {
            conn = getConnection();
            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);
        } catch (Exception e) {
            try {
                conn = getConnection();
                stmt = conn.createStatement();
                rs = stmt.executeQuery(query);
            } catch (Exception ex) {
                System.err.println("Fatal DB error: " + ex);
                System.exit(1);
            }
        }

        int myMaxItems = 0;
        if (rs != null && rs.next()) {
            myMaxItems = rs.getInt(1);
        } else {
            myMaxItems = 0;
        }
        rs.close();

        query = "SELECT COUNT(DISTINCT " + itemFieldName + ") FROM " +
                getRatingDatabaseName();
        try {
            rs = stmt.executeQuery(query);
        } catch (Exception e) {
            try {
                conn = getConnection();
                stmt = conn.createStatement();
                rs = stmt.executeQuery(query);
            } catch (Exception ex) {
                System.out.println("Fatal db error: " + ex);
                System.exit(1);
            }
        }

        int myNumItems = 0;
        if (rs != null && rs.next()) {
            myNumItems = rs.getInt(1);
        } else {
            myNumItems = 0;
        }
        setNumItems(myNumItems);

        rs.close();

        // Lock down the user cache datastructures to ensure they remain consistent.
//        synchronized (dbUsers) {
//            synchronized (itemIndex) {
//                synchronized (ratingsMatrix) {
                    // Initialize item/users vector
                    //itemIndex = new ArrayList(maxItems+1);
                    itemIndex = new SynIntHashMap(new IntHashMap());

                    /*for (int j=0; j <= maxItems; j++)
                    itemIndex.add(j, new IntHashMap());*/

                    // Number of users in database
                    query = "SELECT COUNT(DISTINCT " + userFieldName + ") FROM " +
                            getRatingDatabaseName();

                    try {
                        rs = stmt.executeQuery(query);
                    } catch (SQLException e) {
                        try {
                            conn = getConnection();
                            stmt = conn.createStatement();
                            rs = stmt.executeQuery(query);
                        } catch (Exception ex) {
                            System.out.println("fatal db error: " + ex);
                            System.exit(1);
                        }
                    }

                    int myNumUsers = 0;
                    if (rs != null && rs.next()) {
                        myNumUsers = rs.getInt(1);
                    } else {
                        myNumUsers = 0;
                    }
                    setNumUsers(myNumUsers);

                    rs.close();
                    //SystemUtils.printDebug("Num users = " + numUsers);

                    // Initialize users/items vector
                    // query = "SELECT MAX(UserID) FROM " + getUserDatabaseName();
                    query = "SELECT MAX(" + userFieldName + ") FROM " + getRatingDatabaseName();

                    try {
                        rs = stmt.executeQuery(query);
                    } catch (SQLException e) {
                        try {
                            conn = getConnection();
                            stmt = conn.createStatement();
                            rs = stmt.executeQuery(query);
                        } catch (Exception ex) {
                            System.out.println("Fatal db error: " + ex);
                            System.exit(1);
                        }
                    }

                    int myMaxUsers = 0;

                    if (rs != null && rs.next()) {
                        myMaxUsers = rs.getInt(1);
                    } else {
                        myMaxUsers = 0;
                    }
                    rs.close();
                    //System.out.println(maxUsers);
                    ratingsMatrix = new SynIntHashMap(new IntHashMap());

                    /*System.out.println("Max users= " + maxUsers);
                    System.out.println("Max items= " + maxItems);
                    System.out.println("Num users= " + numUsers);
                    System.out.println("Num items= " + numItems);*/

                    // Global user Id's list
                    dbUsers = new LinkedList();

                    //initialize idTable, see if max ids in ratingTable is
                    //bigger than the ones in idTable
                    query = "SELECT " + idHolder + " FROM " + idTableName + " WHERE "
                            + idName + " = '" + userFieldName + "'";
                    int u = 0;
                    try {
                        rs = stmt.executeQuery(query);
                    } catch (SQLException e) {
                        try {
                            conn = getConnection();
                            stmt = conn.createStatement();
                            rs = stmt.executeQuery(query);
                        } catch (Exception ex) {
                            System.out.println("Fatal db error: " + ex);
                            System.exit(1);
                        }
                    }

                    if (rs != null && rs.next()) {
                        u = rs.getInt(1);
                    }
                    rs.close();
                    if (u < myMaxUsers) {
                        query = "UPDATE " + idTableName + " SET " + idHolder + " = "
                                + myMaxUsers + " WHERE " + idName + " = '" + userFieldName + "'";

                        try {
                            stmt.executeQuery(query);
                        } catch (SQLException e) {
                            try {
                                conn = getConnection();
                                stmt = conn.createStatement();
                                stmt.executeQuery(query);
                            } catch (Exception ex) {
                                System.out.println("Fatal db error: " + ex);
                                System.exit(1);
                            }
                        }
                    }

                    query = "SELECT " + idHolder + " FROM " + idTableName + " WHERE "
                            + idName + " = '" + itemFieldName + "'";
                    int item = 0;
                    try {
                        rs = stmt.executeQuery(query);
                    } catch (SQLException e) {
                        try {
                            conn = getConnection();
                            stmt = conn.createStatement();
                            rs = stmt.executeQuery(query);
                        } catch (Exception ex) {
                            System.out.println("Fatal db error: " + ex);
                            System.exit(1);
                        }
                    }

                    if (rs != null && rs.next()) {
                        item = rs.getInt(1);
                    }
                    rs.close();
                    if (item < myMaxItems) {
                        query = "UPDATE " + idTableName + " SET " + idHolder + " = "
                                + myMaxItems + " WHERE " + idName + " = '" + itemFieldName + "'";
                        try {
                            stmt.executeQuery(query);
                        } catch (SQLException e) {
                            try {
                                conn = getConnection();
                                stmt = conn.createStatement();
                                stmt.executeQuery(query);
                            } catch (Exception ex) {
                                System.out.println("Fatal db error: " + ex);
                                System.exit(1);
                            }
                        }
                    }
                    conn.close();
                    setMaxUsers(myMaxUsers);
                    setMaxItems(myMaxItems);
//                }
//            }
//        }
    }

    /**
     Load users from an external file (found relatively to EXPERIMENTS_LOCATION)
     The file should be formatted as follow:
     - A line containing a user ID
     - Consecutive lines, each containing a tuple "item ID", "rating Value", "rating set number" separated by tab (the lattest being a ratings partition)
     - The previous pattern can be repeated as many times as there are users
     The second parameter indicates how many users from the begining of the files will be loaded.
     @param fileName
     @param numUsers
     @param ratingSetOut
     */
    public void loadRatingsFile(String fileName, int numUsers, int ratingSetOut) {
        //*******Declarations***************
        int user = 0;
        int item, ratingSet;
        float rating;
        //int l= 0;
        IntHashMap currentUserRatings = null;
        StringTokenizer token = null;
        String expDir = ""; //SystemConstants.EXPERIMENTS_LOCATION;
        FileReader fr = null;

        Stats userStats = new Stats();

        // Lock down the user cache datastructures to ensure they remain consistent.
        synchronized (dbUsers) {
            synchronized (itemIndex) {
                synchronized (ratingsMatrix) {
                    int nusers = 0;
                    //**********************************

                    //Open test set file and remaining ratings file
                    try {
                        fr = new FileReader(expDir + "/" + fileName);
                    } catch (java.io.FileNotFoundException ex) {
                        SystemUtils.printDebug("ERROR::Ratings file not found");
                        System.exit(1);
                    }

                    try {
                        //preparation for file reading
                        BufferedReader in = new BufferedReader(fr);
                        token = new StringTokenizer(in.readLine(), "\t");

                        boolean first = true;
                        boolean load = true;
                        UserInfo userInfo;




                        while (token != null && nusers < numUsers) {

                            if (token.countTokens() == 1) { //read a new user id
                                if (!first && load) {
                                    // place new matrix entry (for previous user)
                                    //userInfo = new UserInfo(user, currentUserRatings, userStats.calcMean(), userStats.calcSD(), true);
                                    float[] ratings = new float[currentUserRatings.size()];
                                    int[] itemIds = new int[currentUserRatings.size()];
                                    Iterator iter = currentUserRatings.iterator();
                                    int i = 0;
                                    while (iter.hasNext()) {
                                        ItemRating ir = (ItemRating) iter.next();
                                        ratings[i] = ir.getRating();
                                        itemIds[i] = ir.getItemID();
                                        i++;
                                    }
                                    userInfo = new UserInfo(user, new RatingArrays(itemIds, ratings),
                                            userStats.calcMean(), userStats.calcSD(), false);
                                    ratingsMatrix.put(user, userInfo);
                                    dbUsers.add(new Integer(user));
                                    nusers++;
                                } else {
                                    first = false;

                                    //initialize for new user
                                }
                                user = (new Integer(token.nextToken())).intValue();
                                //if (((UserInfo)ratingsMatrix.get(user)).userRatingsHash.size() == 0)
                                if (!ratingsMatrix.containsKey(user)) {
                                    load = true;
                                } else {
                                    load = false;

                                }

                                currentUserRatings = new IntHashMap();
                                userStats = new Stats();
                            } else {
                                //read rating data from file
                                item = (new Integer(token.nextToken())).intValue();
                                rating = (new Float(token.nextToken())).floatValue();
                                ratingSet = (new Integer(token.nextToken())).intValue();

                                if (ratingSet != ratingSetOut) { //add if we want to include ratings from this rating set (useful for experiments)
                                    currentUserRatings.put(item,
                                            new ItemRating(user, item, rating));
                                    userStats.addNum(rating);

                                    //create new value in the items/users matrix
                                    if (load) {
                                        // Create new value in the items/users matrix
                                        if (!itemIndex.containsKey(item)) {
                                            //add new ItemInfo to matrix
                                            itemIndex.put(item, new ItemInfo(item));
                                        }
                                        ((ItemInfo) itemIndex.get(item)).addRating(new ItemRating(user,
                                                item, rating));
                                    }
                                }
                            }

                            try {
                                token = new StringTokenizer(in.readLine(), "\t");
                            } catch (Exception ex) {
                                token = null;
                            }
                        }

                        // Add last user
                        //if (((UserInfo)ratingsMatrix.get(user)).userRatingsHash.size() == 0){
                        if (!ratingsMatrix.containsKey(user)) {
                            float[] ratings = new float[currentUserRatings.size()];
                            int[] itemIds = new int[currentUserRatings.size()];
                            Iterator iter = currentUserRatings.iterator();
                            int i = 0;
                            while (iter.hasNext()) {
                                ItemRating ir = (ItemRating) iter.next();
                                ratings[i] = ir.getRating();
                                itemIds[i] = ir.getItemID();
                                i++;
                            }
                            userInfo = new UserInfo(user, new RatingArrays(itemIds, ratings),
                                    userStats.calcMean(), userStats.calcSD(), false);
                            //userInfo = new UserInfo(user, currentUserRatings,userStats.calcMean(), userStats.calcSD(), true);
                            ratingsMatrix.put(user, userInfo);
                            dbUsers.add(new Integer(user));
                            nusers++;
                        }

                        //Just a check of the # of loaded users
                        int n = ratingsMatrix.size();

                        System.out.println("Users actually loaded= " + n);

                        n = itemIndex.size();

                        System.out.println("Items actually loaded= " + n);
                    } catch (java.io.IOException ex) {
                        ex.printStackTrace();
                        System.exit(1);
                    }
                }
            }
        }
    }

    /**
     Get the percentage of item entries that are not null
     i.e., items for which we can compute a prediction.
     @return percentage
     @throws SQLException
     */
    public float getCoverage() throws java.sql.SQLException, CFInternalErrorException{
        float coverage;
        int itemsNumber = getNumItems();

        String query = "SELECT COUNT(DISTINCT MovieID) FROM " +
                getRatingDatabaseName();
        ResultSet rs = null;
        Statement stmt;

        Connection conn = null;
        try {
            conn = getConnection();
            stmt = conn.createStatement();
            rs = stmt.executeQuery(query);
        } catch (Exception e) {
            try {
                conn = getConnection();
                stmt = conn.createStatement();
                rs = stmt.executeQuery(query);
            } catch (Exception ex) {
               throw new CFInternalErrorException("DB error in getCoverage()", ex);
            }
        }
        if (rs != null && rs.next()) {
            itemsNumber = rs.getInt(1);
        }
        rs.close();
        conn.close();

        //System.out.println("Item vector size= " + itemsNumber);
        int n = 0;
        IntHashMap hash;
        Iterator iter = itemIndex.iterator();

        // We do a try/catch to avoid having to synchronize
        // the hasNext()/getNext()
        try {
            while (true) {
                ItemInfo itemInfo = (ItemInfo) iter.next();
                hash = itemInfo.getItemRatingsHash();
                if (hash != null && hash.size() > 1) {
                    n++;
                }
            }
        } catch (NoSuchElementException e) {
            // Safely ignore this - means we just reached the end of the iterator
        }
        coverage = (float) n * 100 / itemsNumber;
        return coverage;
    }

    /**
     Accessor, return userTableName
     @return userTableName
     */
    /*  public String getUserDatabaseName() {
          return userTableName;
      }
     */
    /**
     Accessor, return itemTableName
     @return itemTableName
     */
    /*   public String getItemDatabaseName() {
           return itemTableName;
       }
     */
    /**
     Accessor, return ratingTableName
     @return ratingTableName
     */
    public String getRatingDatabaseName() {
        return ratingTableName;
    }

    /**
     Accessor, return MinRating
     @return MinRating
     */
    public float getMinRating() {
        return minRating;
    }

    /**
     Accessor, return MaxRating
     @return MaxRating
     */
    public float getMaxRating() {
        return maxRating;
    }

    /**
     Accessor, return error rating
     @return error rating
     */
    public float getErrRating() {
        return errRating;
    }

    /**
     Accessor, return max user ID
     @return max user ID
     */
    public int getMaxUser() {
        return maxUsers;
    }

    /**
     Accessor, return max item ID
     @return max item ID
     */
    public int getMaxItem() {
        return maxItems;
    }

    public int getMaxType() {
        return maxType;
    }

    /**
     Accessor, return userInfo table's name
     @return userInfo table name
     */
    public String getUserInfoTableName() {
        return userInfoTableName;
    }

    /**
     Accessor, return num of recommendations cached
     @return numTopNRecords
     */
    public int getNumTopNRecords() {
        return numTopNRecords;
    }

    /**
     Hiding the rating value for the item we are trying to predict for,
     re-compute the user SD and user average rating.
     Useful for MAE computation and other experiments.
     @param user
     @param item
     */
    public void hideRatingFromRecords(int user, int item) {

        RatingArrays ratings;
        UserInfo userInfo = (UserInfo) ratingsMatrix.get(user);
        if (userInfo == null) {
            ratings = new RatingArrays(20);
        } else {
            ratings = userInfo.getRatings();
        }

        //Iterator iter = ratings.iterator();
        Stats userStats = new Stats();
        int[] itemID = ratings.ids;
        float[] rating = ratings.ratings;
        for (int i = 0; i < ratings.getLength(); i++) {
            if (itemID[i] != item) {
                userStats.addNum(rating[i]);
            }
        }

        float mean = userStats.calcMean();
        float sd = userStats.calcSD();

        /*userMeans.set(user, new Float(userStats.calcMean()));
              userSD.set(user, new Float(userStats.calcSD()));*/

        userInfo.setUserMean(mean);
        userInfo.setUserSD(sd);
    }

    /**
     Returns the number of ratings for the corresponding user.
     @param userID
     @return number of ratings
     */
    public int getNumRatings(int userID) {
        UserInfo userInfo = (UserInfo) ratingsMatrix.get(userID);
        if (userInfo == null) {
            return 0;
        }
        RatingArrays ratings = userInfo.getRatings();
        return ratings.getLength();
    }

    /**
     Returns the number of users who have rated the corresponding item.
     @param itemID
     @return number of users
     */
    public int getNumRaters(int itemID) {
        if (itemID >= itemIndex.size()) {
            return 0;
        }

        int x = 0;
        ItemInfo itemInfo = (ItemInfo) itemIndex.get(itemID);
        if (itemInfo != null) {
            IntHashMap itemRatingsHash = itemInfo.getItemRatingsHash();
            x = itemRatingsHash.size();
        }
        return x;
    }

    /**
     Cache new user.
     Update data structures for caching when new user comes in. Also take care of
     updating userStats.
     When a user record is loaded, it's inUse variable is set to true. While the inUse variable
     is set, the user cannot be loaded or unloaded. You may need to immediate unset this variable.
     @param userID
     @return the new UserInfo structure that was created for this user
     */
    public UserInfo loadUser(int userID, boolean sampled) throws SQLException {

        int itemID;
        float rating;
        IntHashMap currentUserRatings = new IntHashMap();
        Stats userStats = new Stats();

       // String query = "SELECT * FROM " + getRatingDatabaseName() + " WHERE " +
        //        userFieldName + " = " + userID;

        ResultSet rs = null;
        Statement stmt;
        Connection conn = null;
        try {
            conn = getConnection();
        } catch (Exception e) {
           // while ( (conn = getConnection()) == null ) ;
            try {
                conn.close();
                conn = getConnection();
            } catch (Exception ex) {
               System.err.println("Fatal DB error: " + ex +" userID " + userID);
               ex.printStackTrace();
            }

        }
 
        ps1.setInt(1, userID);
    /*    try {
            rs = ps1.executeQuery();
        } catch (Exception e) {
            try {
				rs = ps1.executeQuery();
			} catch (Exception exx) {
				try {
				rs = ps1.executeQuery();
				} catch (Exception ex) {
               System.err.println("Fatal DB error: " + ex +" userID " + userID);
               ex.printStackTrace();
            	}
            }

        }*/

        UserInfo userInfo;

        // todo We could probably optimize this by locking out other load/unload users,
        // then creating most of the necessary structures outside of the loop, then doing
        // only minimal stuff within the loop (adding to itemIndex would have to happen in the loop)

        // We need to synchronize all of these to ensure that they remain
        // consistent.
        synchronized (dbUsers) {
            synchronized (itemIndex) {
                synchronized (ratingsMatrix) {
                    try {
            			rs = ps1.executeQuery();
       				 } catch (Exception e) {
            			try {
            				conn.close();
            				conn = getConnection();
							rs = ps1.executeQuery();
						} catch (Exception exx) {
							try {
								conn.close();
            					conn = getConnection();
								rs = ps1.executeQuery();
							} catch (Exception ex) {
               						System.err.println("Fatal DB error: " + ex +" userID " + userID);
               					ex.printStackTrace();
            				}
            			}

        			}
   
                    // Add the new user's ratings to the itemIndex
                    while (rs != null && rs.next()) {
                        itemID = rs.getInt(itemFieldName);
                        rating = rs.getFloat(ratingFieldName);

                        userStats.addNum(rating);
                        currentUserRatings.put(itemID,
                                new ItemRating(userID, itemID, rating));
                        if (!itemIndex.containsKey(itemID)) {
                            // It's a new item that isn't already in the cache!
                            itemIndex.put(itemID, new ItemInfo(itemID));
                        }
                        ItemInfo itemInfo = (ItemInfo) itemIndex.get(itemID);
                        if (!itemInfo.getWritersLock()) continue;
                        itemInfo.addRating(new ItemRating(userID, itemID, rating));
                        itemInfo.releaseWritersLock();
                    }
                    rs.close();
                    conn.close();

                    if (userID > getMaxUser()) {
                        // New user in the system
                        setMaxUsers(userID);
                        incNumUsers();
                    }

                    // Build the arrays of ids and ratings necessary for the
                    // speed optimized UserInfo structures
                    float[] ratings = new float[currentUserRatings.size()];
                    int[] itemIds = new int[currentUserRatings.size()];
                    Iterator iter = currentUserRatings.iterator();
                    int i = 0;
                    while (iter.hasNext()) {
                        ItemRating ir = (ItemRating) iter.next();
                        ratings[i] = ir.getRating();
                        itemIds[i] = ir.getItemID();
                        i++;
                    }
                    userInfo = new UserInfo(userID, new RatingArrays(itemIds, ratings),
                                                userStats.calcMean(), userStats.calcSD(), false);
                    ratingsMatrix.put(userID, userInfo);

                    // We can't just change dbUsers directly, because it will cause a
                    // ConcurrentModificationException to anybody who is using
                    // an iterator. So we do a swap instead
                    // todo Why is this? According to the JavaDoc of LinkedList, we should be able to call the add/remove
                    LinkedList db_users_clone = (LinkedList)dbUsers.clone();
                    db_users_clone.add(new Integer(userID));
                    dbUsers = db_users_clone;

                    // One more user data has been cached.
                    incCachedUsers();

                    // Notify the CacheUpdate thread if the cache just filled up
                    if ((getNumUsersCached() - maxSampledUsers) > maxCachedUsers
                            && !sampled) {
                        synchronized (cacheMonitor) {
                            cacheMonitor.notifyAll();
                        }
                    }
                }
            }
        }
        // System.out.println("loadUser(). user "+userID + " loaded");
        return userInfo;
    }

    /**
     Remove a user from the cache. This will not unload active users.
     If the UserInfo structure of the userID specified has
     a non-zero inUse field, then this method will *NOT* unload it the user. It will simply
     return.
     @param userID
     */
    public void userUnload(int userID) {
        RatingArrays ratings = null;

        UserInfo ui;
        ui = (UserInfo) ratingsMatrix.get(userID);
        if (ui == null) {
            return;   // No such user
        }

        // We synchronize on all, to make sure that dbUsers,
        // ratingsMatrix, and itemIndex are always consistent
        // Be wary of deadlock!
        synchronized (dbUsers) {
            synchronized (itemIndex) {
                synchronized (ratingsMatrix) {
                    if (ui.getUseCount() > 0) {
                        System.out.println("Refusing to unload user - user is active: " + userID);
                        return;
                    }
 //         System.out.println("unloadUser().synchronization, before ratingsMatrix.remove(). user " + userID);
                    ratingsMatrix.remove(userID);
//		System.out.println("unloadUser().synchronization, after ratingsMatrix.remove(). user " + userID);
                   // dbUsers.remove(new Integer(userID));

                    LinkedList db_users_clone = (LinkedList)dbUsers.clone();
                    db_users_clone.remove(new Integer(userID));
                    dbUsers = db_users_clone;

                    ratings = ui.getRatings();

                    // Remove this user's ratings from the itemIndex cache
                    int[] itemIDs = ratings.ids;
                    for (int i = 0; i < ratings.getLength(); i++) {
                        ItemInfo itemInfo = (ItemInfo) itemIndex.get(itemIDs[i]);

                        itemInfo.getWritersLock();
                        try {
                        	itemInfo.removeRating(userID);
                        } catch (Exception e){
                            System.out.println("Exception: user "+ userID + "item"+ itemIDs[i]+" e: "+ e);
                            e.printStackTrace();
                        }
                        itemInfo.releaseWritersLock();
                    }
                    // Decrement count of cached users
                    decCachedUsers();
                }
            }
        }
        System.out.println("Unloaded user: " + userID);
    }

    /**
     Check this user's ratings records and load into cache if not done yet
     Upon return from this method, the user's record is guaranteed to be in memory
     and will not be evicted until you call decUserCount() on the UserInfo structure.
     @param userID
     */
    public void checkCachedUser(int userID) throws SQLException {
        UserInfo user;

        // We need the synchronized block to ensure that an interrupt
        // between the get() and the incUseCount() doesn't unload the user
        synchronized (ratingsMatrix) {
            user = (UserInfo) ratingsMatrix.get(userID);

             if (user != null) {
        		user.incUseCount();
             }
        }



        if (user == null) {
            incCacheMisses();
            user = loadUser(userID, false);

            // Upon return from loadUser(), the inUse field of user is set to 1, so that user will not
            // be unloaded until it is cleared.
        } else {
            incCacheHits();
        }

        // A hint to the cache manager to not discard this user
        user.setReferenced(true);
    }


    /**
     Adds a new rating to both database and the cache.
     Adds this rating to the waiting queue for the database and places it in the cache
     (in both the ratings matrix and its transposed)
     These operations might be done concurrently by client Threads => synchronize data strucure objects.
     @param rating - an ItemRating
     */
    public void addRating(ItemRating rating) throws SQLException {
        int userID = rating.getUserID();
        int itemID = rating.getItemID();

                    // Add this rating to the waiting queue for the database

                DBList.insertOrReplaceRating(rating);


        // Also sets the inUse field, which ensures that userID's record is not
        // removed from the cache until we clear it.
        checkCachedUser(userID);

        RatingArrays ratingArrays = null;
        UserInfo userInfo = (UserInfo) ratingsMatrix.get(userID);

        ItemInfo itemInfo;
       // ratingArrays = userInfo.getRatings();


       	// Yun: itemIndex is syncronized in method. But if not put in synchronized block, how can we
       	//      garantee data's consistency. Some can not put into writer lock, which will cause
       	//      deadlock if we do.
 //       synchronized (itemIndex) {
            // Synchronized to ensure that the itemInfo associated with our item is not removed
            // While we are working.
            // Also ensures that containsKey() and put() are atomic.  We also want to make sure that no other
            // thread calling addRating() tries to add a rating for
            // this user until we have completed our update

            if (!itemIndex.containsKey(itemID)) {
            	itemIndex.put(itemID, new ItemInfo(itemID));

            }


            itemInfo = (ItemInfo) itemIndex.get(itemID);

 			if (!userInfo.getWritersLock()) return;

            if(!itemInfo.getWritersLock()){
                userInfo.releaseWritersLock();
                return;
            }
            itemInfo.addRating(rating);
 
			//  Yun: put this in the lock might be more appropriate
            ratingArrays = userInfo.getRatings();
            //synchronized (userInfo) {
                // sync on userinfo needed to keep ratings update and
                // stats update together.
                ratingArrays.addRating(itemID, rating.getRating());
   

                // Todo: These two release suppose to locate after userInfo.incUserAddedRatings().
                
                itemInfo.releaseWritersLock();
      //          userInfo.releaseWritersLock();
                // Make sure this operation is atomic with the rating insertion
                updateStats(userID);
 
                userInfo.incUserAddedRatings();

               //Yun: userInfo release lock here can garantee the data consistency
                userInfo.releaseWritersLock();

            //}
            // Possibility of a new item
            if (itemID > getMaxItem()) {
                setMaxItems(itemID);
                incNumItems();
            }
            if (userID > getMaxUser()) {
                setMaxUsers(userID);
                incNumUsers();
            }
 //       }
        // Done with the active user
        userInfo.decUseCount();

    }

    /**
     Add ratings to cache but not to the database.
     @param rating - an ItemRating
     */
    public void addCachedRating(ItemRating rating) { //Right now used by the AutoTester only
        int userID = rating.getUserID();
        int itemID = rating.getItemID();

        UserInfo userInfo = (UserInfo) ratingsMatrix.get(userID);
        userInfo.getRatings().addRating(itemID, rating.getRating());

        ItemInfo itemInfo = (ItemInfo) itemIndex.get(itemID);
        if (itemInfo != null)
        	itemInfo.addRating(rating);
    }

    /**
     Remove rating from our cache. Function is Thread safe.
     @param userID
     @param itemID
     */
    public void removeRating(int userID, int itemID) throws SQLException {
        // Add this rating to the waiting queue for the database
        DBList.removeRating(userID, itemID);

        // Also sets the inUse field, which ensures that userID's record is not
        // removed from the cache until we clear it
        checkCachedUser(userID);
  
        UserInfo userInfo = (UserInfo) ratingsMatrix.get(userID);

   //     synchronized (itemIndex) {


            // See addRating() for a description of why we synchronize on itemIndex
            //If rating not in cache, it doesn't exist, no need to remove it
            if (getRatingCached(userID, itemID) == getErrRating()) {
                decUseCount(userID);
                return;
            }

            RatingArrays ratingArrays = null;
            ItemInfo itemInfo = null;
            itemInfo = (ItemInfo) itemIndex.get(itemID);
   
            if(!userInfo.getWritersLock()) return;
            if(!itemInfo.getWritersLock()){
                userInfo.releaseWritersLock();
                return;
            }
  

            //synchronized (itemInfo) {
                // Synchronize to keep removeRation() and size() together
                itemInfo.removeRating(userID);

                // No rating left for that item
                if (itemInfo.getItemRatingsHash().size() == 0) {
                    decNumItems();
                    itemIndex.remove(itemID);
                    itemInfo.delete();
                }
            //}
            //synchronized (userInfo) {
                // We sync on userInfo, to keep getRatings, removeRating, and updateStats together
                // This way the ratings and the stats remain consistent
                ratingArrays = userInfo.getRatings();
                ratingArrays.removeRating(itemID);

                itemInfo.releaseWritersLock(); //Shuzhen: release the lock right after the work is done
       
                if(ratingArrays.getLength() == 0){
                    userInfo.delete();
                 //   userInfo.releaseWritersLock(); //Shuzhen: release the lock right after the work is done
                }
                else{
                  //  userInfo.releaseWritersLock(); //Shuzhen: release the lock right after the work is done
                                                   // should release the writers lock before updateStats,
                                                   // otherwise might lead to deadlock
                    updateStats(userID);
                }
           // }



            // No rating left for that user
//            if (ratingArrays.getLength() == 0) {
//                decNumUsers();
//                ratingsMatrix.remove(userID);
//            }
            // We let the CacheUpdate thread unload users.
   //     }
        userInfo.decUseCount();
        userInfo.releaseWritersLock();
   
    }

    /**
     Remove rating from the cache.
     param ItemRating
     */
    public void removeCachedRating(ItemRating rating) { // Right now used by the AutoTester only
        int userID = rating.getUserID();
        int itemID = rating.getItemID();

        UserInfo userInfo = (UserInfo) ratingsMatrix.get(userID);
        RatingArrays ratings = userInfo.getRatings();
        ratings.removeRating(itemID);

        ItemInfo itemInfo = (ItemInfo) itemIndex.get(itemID);
        itemInfo.removeRating(userID);
    }

    /**
     Remove useless data from the cache.
     This function is called by special Thread whose task is to update the cache. It is only run
     When the cache fills up. ((getNumUsersCached() - maxSampledUsers) > maxCachedUsers)
     First, it sets to "0" all the reference bits of the records in the cache, then sleeps for
     a certain amount of time. After waking up, it iterates through the cache records once again,
     clearing all records whose reference bit is not "1".
     */
    public void refreshCache() {

        // Disable the cache update thread (Normally the only one that is calling this method)
        // Awaken by the loadUser function, whenever the cache fills up.
        System.out.println("Start refresh cache");
        synchronized (cacheMonitor) {
            try {
                cacheMonitor.wait();
            } catch (InterruptedException ex) {
                ex.printStackTrace();
                // Interruptions shouldn't happen, but it's not a disaster.
                // We just end up checking the cache sooner than usual
            }
        }

        //System.out.println("Notified... Num users in cache= " + numUsersCached);

        //int numBits = 0;
        UserInfo user;

        Iterator iter;
        iter = ratingsMatrix.iterator();

        // We use a try/catch to avoid having to synchronize
        // the entire loop due to haveNext/next atomicity needs
        try {
            while (true) {
                user = (UserInfo) iter.next();

                // The users selected by a sampling method remain.
                // This part of the cache will be refreshed by another method.
                if (!user.samplingFlag) {
                    user.setReferenced(false);
                }
            }
        } catch (NoSuchElementException e) {
            // Ignore this - just means we've reached the end of the list
        }
        //System.out.println("cleared " + numBits + " bits, before going to bed. Num users in cache= " + numUsersCached);
        // Sleeps for some time
        // todo This needs to be externally configurable
        CacheUpdate.setIdle(5000);

   System.out.println("Woke up... Num users in cache= " + numUsersCached);

        // Second iteration to remove all the records that haven't been used during that time
        // (corresponding probably to logged off users)
        //int n = 0;

        iter = ratingsMatrix.iterator();

        // we use a try/catch to avoid synchronizing
        // the entirely loop due to hasNext/next atomicity needs
        try {
            while (true) {
                user = (UserInfo) iter.next();

                if (!user.samplingFlag && !user.isReferenced()) {
                    userUnload(user.userID);
                }
            }
        } catch (NoSuchElementException e) {
            // Ignore it - just means we've reached the end of the list.
        }
//    System.out.println("Removed " + n + " users ");
     System.out.println("Num users in cache= " + numUsersCached);

        // Run the garbage collector, to faster clean up unused discarded objects
        System.gc();
    }

    /**
     Check wether a user's TopN data is cached on the server side.
     @param userID
     @return boolean
     */
    public boolean checkCachedTopN(int userID) {
        UserInfo user;
        user = (UserInfo) ratingsMatrix.get(userID);
        if (user == null) {
            return false;
        }

        if (!user.getReadersLock(true)) return false;    // Shuzhen: we also need to synchronize this user
        if (user.userTopN.length != 0) {
            user.releaseReadersLock();
            return true;
        } else {
            user.releaseReadersLock();
            return false;
        }
    }

    /**
     Cache user TopN data computed by one of our algorithms.
     @param predictions
     */
    public void cacheUserTopN(int userID, ItemPrediction[] predictions) {
        // Admins or Developers may turn off caching of topns by setting
        // CFServer.mem.numTopNRecords to 0 or negative.
        UserInfo user;
        user = (UserInfo) ratingsMatrix.get(userID);
        if (user == null) {
            return;
        }

        // Initialize and fill TopN data structure
        int numToCache = Math.min(predictions.length, numTopNRecords);

        if(!user.getWritersLock()) return;    // Shuzhen: need to synchronize this user
        user.userTopN = new ItemPrediction[numToCache];

        for (int i = 0; i < numToCache; i++) {
            user.userTopN[i] = predictions[i];
        }

        user.releaseWritersLock();
    }


    /**
     Returns a user's top N recommendations, retrieved from the server side memory cache.
     Parameters specify the number of recommendations to retrieve and the offset
     @return The specified number recommendations, with "offset" starting at zero
     */
    public ItemPrediction[] getCachedTopN(int userID, int number, int offset) {
        UserInfo user;
        user = (UserInfo) ratingsMatrix.get(userID);

        // Yun: have to check it user is null
        if (user == null) {
            return new ItemPrediction[0];
        }

        ItemPrediction[] returnArray;

        // NOTE: also try to convert to List, then subList and eventually toArray.
        if (!user.getReadersLock(true)) return new ItemPrediction[0];    // Shuzhen: need to synchronize this user
        if (user != null && offset < (user.userTopN.length + 1)) { // If user info actually cached

            if ((number + offset) > user.userTopN.length) {
                number = user.userTopN.length - offset;
            } // If desired number does not exceed number of TopN records cached

            returnArray = new ItemPrediction[number];

            for (int i = 0; i < number; i++) {
                returnArray[i] = user.userTopN[offset++];
            }
        user.releaseReadersLock();
            return returnArray;
        } else {
            user.releaseReadersLock();
            return new ItemPrediction[0];
        }


    }

    /**
     * Returns a user's typed top N recommendations, No caching
     @return The specified number recommendations, with "offset" starting at zero
     *
     */
    public ItemPrediction[] getTopNByType(int userID, int number, int offset,
                                          int type, ItemPrediction[] pred) {
        UserInfo user;
        ItemPrediction[] returnArray;
        user = (UserInfo) ratingsMatrix.get(userID);

        if (number == 0 || user == null)
            return new ItemPrediction[0];

        if (offset >= pred.length) {
       //     System.out.println("Just " + pred.length + " recommendations available.");
            return (new ItemPrediction[0]);
        }

        if (pred.length > 1 && pred[offset].getItemID() > 0) { // If user info actually cached
            if ((number + offset) <= pred.length) { // If desired number does not exceed number of TopN records cached
                if (pred[number + offset - 1].getItemID() > 0) {
                    returnArray = new ItemPrediction[number];
                } else {
                    int n = getLengthAvailTopN(pred, offset, number);
                    returnArray = new ItemPrediction[n];
                }
            } else {
                if (pred[pred.length - 1].getItemID() > 0) {
                    returnArray = new ItemPrediction[pred.length - offset];
                } else {
                    int n = getLengthAvailTopN(pred, offset, number);
                    returnArray = new ItemPrediction[n];
                }
            }
            //   System.out.println(pred.length + "      " + returnArray.length);

            float maxrate = getMaxRating();
            float minrate = getMinRating();

            for (int i = 0; i < returnArray.length; i++, offset++) {
                if (pred[offset].getPrediction() > maxrate) {
                    returnArray[i] = new ItemPrediction(pred[offset].getUserID(),
                            pred[offset].getItemID(),
                            maxrate);
                } else if (pred[offset].getPrediction() < minrate) {
                    returnArray[i] = new ItemPrediction(pred[offset].getUserID(),
                            pred[offset].getItemID(),
                            minrate);
                } else {
                    returnArray[i] = pred[offset];
                }
            }

            return returnArray;
        } else {
            return new ItemPrediction[0];
        }
    }

    /**
     Retrieve the number of actual prediction, get rid of the Error rating.
     @return the number of the actual length of the ItemPrediction Array
     *
     */
    private int getLengthAvailTopN(ItemPrediction[] pred, int offset, int number) {
        if (pred[offset].getItemID() == 0) {
            return 0;
        }

        int length = 0;
        while (pred[offset].getItemID() > 0) {
            length++;
            offset++;
            if (length == number || length == pred.length) {
                break;
            }
        }

        return (length);

    }

    /**
     Reset a user's counter of added ratings.
     @param userID
     */
    public void resetUserCounter(int userID) {
        UserInfo user;
        user = (UserInfo) ratingsMatrix.get(userID);
        if (user == null) {
            return;
        }
        user.setUserAddedRatings(0);
    }

    /**
     Get the value of a user's counter of added ratings.
     @param userID
     */
    public int getAddedRatings(int userID) {
        UserInfo userInfo = (UserInfo) ratingsMatrix.get(userID);
        if (userInfo == null) {
            return 0;
        }
        return userInfo.getUserAddedRatings();
    }

    /*
     * Set of synchronized utility functions to process the global variables (of primitive type)
     */

    /*
     * Increments number of items.
     */
    public synchronized void incNumItems() {
        numItems++;
    }

    /*
     * Decrement number of items.
     */
    public synchronized void decNumItems() {
        numItems--;
    }

    /*
     * Increment number of users.
     */
    public synchronized void incNumUsers() {
        numUsers++;
    }

    /*
     * Decrement number of users.
     */
    public synchronized void decNumUsers() {
        numUsers--;
    }

    /*
     * Increment number of cached users.
     */
    public synchronized void incCachedUsers() {
        numUsersCached++;
    }

    /*
     * Decrement number of cached users.
     */
    public synchronized void decCachedUsers() {
        numUsersCached--;
    }

    /*
     * Set max user ID.
     */
    public synchronized void setMaxUsers(int userID) {
        maxUsers = userID;
    }

    /*
     * Decrement max user ID.
     */
    /*public synchronized void decMaxUsers(){
     maxUsers--;
      }*/

    /*
     * Set max item ID.
     */
    public synchronized void setMaxItems(int itemID) {
        maxItems = itemID;
    }

    /*
     * Decrement max item ID.
     */
    /*public synchronized void decMaxItems(){
     maxItems--;
      }*/

    /*
      Shuts the server down after proceeding pending work (like DB accesses).
     */
    public void shutdown() {

        System.out.println("Processing pending DB accesses...");
        DBList.flush();
        //System.out.println("Shutting down sever...");
        System.out.println("Thanks for using the CFEngine predictor.");
        System.out.println(
                "Copyright (c) 2002 Oregon State Universtity. All rights reserved.");
        System.exit(0);
    }


    /*
     * caches the group of ratings that is passed in. For use by autotester only.
     *
     * @param iter Iterator of the ratings to be cached
     */
    public void cacheRatingGroup(Iterator iter) {
        int userID = -999;
        int lastUserID = -999;

        if (iter != null) {
            while (iter.hasNext()) {
                ItemRating ir = (ItemRating) iter.next();
                if (ir != null) {
                    addCachedRating(ir);
                    userID = ir.getUserID();
                    if (userID != lastUserID && lastUserID != -999) {
                        updateStats(userID);
                    }
                    lastUserID = userID;
                }
            }
            if (userID != -999) {
                updateStats(userID);
            }
        }
    }

    /*
     * removes the group of ratings that is passed in from the cache
     * For testing only.
     *
     * @param iter Iterator of the ratings to be cached
     */
    public void removeCachedRatingGroup(Iterator iter) {

        int userID = -999;
        int lastUserID = -999;
        if (iter != null) {
//            try {
                while (iter.hasNext()) {
                    ItemRating ir = (ItemRating) iter.next();
                    removeCachedRating(ir);
                    userID = ir.getUserID();
                    if (userID != lastUserID && lastUserID != -999) {
                        updateStats(userID);
                    }
                    lastUserID = userID;
                }
                if (userID != -999) {
                    updateStats(userID);
                }
//            } catch (Exception e) {
//                System.out.println("Error removing cached rating group");
//                System.out.println(e);
//            }
        }
    }

    //for testing only
    /*
     * Reduces each users ratings to a defined amount
     *
     * @usage sparcifyData(iter, 25, 1) == reduced each user in iter so that they contain 25% of their total ratings in cache
     *
     * @param iter Iterator of the rating is the user group :: check Autotester for better understanding
     * @param NumParam Number component of options.
     * @param option  1 = %, 0 = given n.
     *
     * @return ll linkedlist containing linkedlists of the ratings removed from each user
     */
    public LinkedList sparcifyData(Iterator iter, int NumParam, int option) {

        int lastUserID = -1;

        LinkedList ll = new LinkedList();
        while (iter.hasNext()) {

            ItemRating currRating = (ItemRating) iter.next();

            int userID = currRating.getUserID();

            if (getNumRatings(userID) < 5) {
                continue;
            }

            if (userID != lastUserID && lastUserID != 0) {
                ll.add(givenNHide(userID, NumParam, option));
            }
            lastUserID = userID;
        }
        return ll;
    }


    /*
     * for testing only
     *
     * Reduces user's ratings to a defined amount
     *
     * @usage sparcifyData(20, 25, 1) == reduced each user in iter so that they contain 25% of their total ratings in cache
     *
     * @param userID id of the user to be changed
     * @param NumParam Number component of options.
     * @param option  1 = %, 0 = given n.
     *
     * @return ll linkedlist containing the ratings removed from user
     */
    public LinkedList givenNHide(int userID, int NumParam, int option) {
        int numRated = getNumRatings(userID);
        LinkedList ll = new LinkedList();
        ItemRating[] itemArray = getUserRatingList(userID);
        int testParam;

        if (option == 0) {
            testParam = NumParam;
        } else {
            testParam = (int) Math.round(.01 * NumParam * numRated);

        }

        if (numRated > testParam) {
            for (int i = numRated; i > testParam; i--) {
                boolean complete = false;
                //System.out.println("in for loop");
                //int count = 0;
                while (!complete) {
                    //count++;
                    //try {
                    //System.out.println("fight the while");
                    int itemNum = itemArray[((int) (java.lang.Math.random() * 10000)) %
                            numRated].getItemID();
                    float trat = getRatingCached(userID, itemNum);
                    if (trat != getErrRating()) {
                        ItemRating tempItem = new ItemRating(userID, itemNum, trat);
                        removeCachedRating(tempItem);
                        ll.add(tempItem);
                        complete = true;
                    }
                    /*}
                      catch(Exception e){
                                //System.out.println(e);
                      }*/

                }
                //System.out.println("back:" + count);
            }
        }
        updateStats(userID);
        return ll;
    }

    /**
     * Return a UserInfo object to methods that need to perfrom multiple computations on a single user
     * @param userID
     * @return  a single UserInfo stru
     */
    public UserInfo getUserInfo(int userID) {
        return (UserInfo) ratingsMatrix.get(userID);
    }

   /**
     * @return An unused integer itemid that is guaranteed to be unique
     */
    public int getNextItemId() throws SQLException, CFInternalErrorException{
       	//String query1 = "UPDATE ID_TABLE SET Id = last_insert_id(Id+1) WHERE name = 'itemid'";
        String query1 = "UPDATE " + idTableName + " SET " + idHolder + " = last_insert_id("
        				+ idHolder + "+1) WHERE " + idName + " = '" + itemFieldName + "'";

        String query2 = "select last_insert_id()";
        Connection conn = null;
        Statement stmt = null;
        ResultSet rs = null;
        int id = -1;

        try {
        	conn = getConnection();
        	stmt = conn.createStatement();
        	stmt.executeQuery(query1);
        	rs = stmt.executeQuery(query2);

        } catch (Exception e) {
      		try {
        	    conn = getConnection();
        	    stmt = conn.createStatement();
        	    stmt.executeQuery(query1);
        	    rs = stmt.executeQuery(query2);

            } catch (Exception ex) {
      	        throw new CFInternalErrorException("DB error in getNextItemId()", ex);
            }
    	}
       if (rs != null && rs.next()) {
             		id = rs.getInt(1);
        	}
        	rs.close();
        	conn.close();

        return id;

    }

    /**
     * @return An unused integer userid that is guaranteed to be unique
     */
    public int getNextUserId() throws SQLException, CFInternalErrorException {
         // suppose maxUserID is the biggest user ID existed

       // String query1 = "UPDATE ID_TABLE SET Id = last_insert_id(Id+1) WHERE name = 'userid'";
        String query1 = "UPDATE " + idTableName + " SET " + idHolder + " = last_insert_id("
        				+ idHolder + "+1) WHERE " + idName + " = '" + userFieldName + "'";

        String query2 = "select last_insert_id()";
        Connection conn = null;
        Statement stmt = null;
        ResultSet rs = null;
        int id = -1;

        try {
        	conn = getConnection();
        	stmt = conn.createStatement();
        	stmt.executeQuery(query1);
        	rs = stmt.executeQuery(query2);

        } catch (Exception e) {
      		try {
        	    conn = getConnection();
        	    stmt = conn.createStatement();
        	    stmt.executeQuery(query1);
        	    rs = stmt.executeQuery(query2);

            } catch (Exception ex) {
      		    throw new CFInternalErrorException("DB error in getNextUserId()", ex);
             }
    	}
        if (rs != null && rs.next()) {
             	id = rs.getInt(1);
        }
        rs.close();
        conn.close();

        return id;
    }

    public String getStatistics() {
        String stats = "";
        stats = stats.concat("User Cache misses: " + userCacheMisses + "\n");
        stats = stats.concat("User Cache hits: " + userCacheHits + "\n");
        stats = stats.concat("User Cache hit rate: " + (double) userCacheHits / (userCacheMisses + userCacheHits) + "\n");

        return stats;
    }


    public synchronized int getNumItems() {
        return numItems;
    }

    public synchronized void setNumItems(int numItems) {
        this.numItems = numItems;
    }

    public synchronized int getNumUsersCached() {
        return numUsersCached;
    }

    public synchronized void setNumUsersCached(int numUsersCached) {
        this.numUsersCached = numUsersCached;
    }

    public synchronized int getNumUsers() {
        return numUsers;
    }

    public synchronized void setNumUsers(int numUsers) {
        this.numUsers = numUsers;
    }

    public synchronized void incCacheHits() {
        userCacheHits++;
    }

    public synchronized void incCacheMisses() {
        userCacheMisses++;
    }

    /**
     * Sets the inUse field of the specified user.
     * @param userID
     */
    public void decUseCount(int userID) {
        UserInfo ui = (UserInfo) ratingsMatrix.get(userID);
        if (ui == null) {
            // No such user loaded
            return;
        }
        ui.decUseCount();
    }
}
