/* *
 * RatingArray.java
 *
 * Version 1.1
 * 23/03/2003
 *
 * @author  <A HREF="mailto:mclaughm@cs.orst.edu">Matt McLaughlin</A>
 * Copyright (c) 2003 Oregon State Universtity. All rights reserved.
 */

package org.recommender.server;

import java.lang.Math;
import java.util.LinkedList;

public class RatingArrays {

    /**
     * Array of ids
     */
    public int[] ids;

    /**
     * Array of ratings for ids in corresponding index in ids[]
     */
    public float[] ratings;

    /**
     * Index of last element in array
     */
    private int length;

    /**
     *  Amount to increment array by with it is resized
     */
    private int incrementSize;

    /**
     *   Constructor for RatingArrays
     * @param size initial number of ratings to be stored
     */
    public RatingArrays(int size) {
        this.length = 0;
        this.ids = new int[size + 10];
        this.ratings = new float[size + 10];
        this.incrementSize = 20;
    }

    /**
     * Constructor for RatingArrays
     * Note: arrays must be full of ids and ratings who's indexes correspond.
     *
     * ** If you a group of ratings, this is much faster than adding them individually
     *
     * @param ids array containing ids
     * @param ratings array of ratings corresponding to ids
     */
    public RatingArrays(int[] ids, float[] ratings) {
        this.ids = ids;
        this.ratings = ratings;
        this.length = ids.length;
        this.incrementSize = 20;
        quicksort(0, ids.length - 1);
    }

    /**
     *   Constructor for RatingArrays
     * @param size initial number of ratings to be stored
     * @param incrementSize amount to increment array size by when it passes threshhold
     */
    public RatingArrays(int size, int incrementSize) {
        this.length = 0;
        this.ids = new int[size + 10];
        this.ratings = new float[size + 10];
        this.incrementSize = incrementSize;
    }

    /**
     * Constructor for RatingArrays
     * Note: arrays must be full of ids and ratings who's indexes correspond.
     *
     * ** If you a group of ratings, this is much faster than adding them individually
     *
     * @param ids array containing ids
     * @param ratings array of ratings corresponding to ids
     * @param incrementSize amount to increment array size by when it passes threshhold
     */
    public RatingArrays(int[] ids, float[] ratings, int incrementSize) {
        this.ids = ids;
        this.ratings = ratings;
        this.length = ids.length;
        this.incrementSize = incrementSize;
        quicksort(0, ids.length - 1);
    }

    /**
     * Add rating to arrays
     * @param id
     * @param rating
     */
    public void addRating(int id, float rating) {
        int index = binarySearch(id, 0, length);
        if (index < length && ids[index] == id) {
            ratings[index] = rating;
        } else {
            checkSize();
            for (int i = length; i > index; i--) {
                int j = i - 1;
                ids[i] = ids[j];
                ratings[i] = ratings[j];
            }
            ids[index] = id;
            ratings[index] = rating;
            length++;
        }
    }

    /**
     * Remove rating from arrays
     * @param id
     */
    public void removeRating(int id) {
        int index = binarySearch(id, 0, length);
        if (index < length && ids[index] == id) {
	    length--;
            for (int i = index; i < length; i++) {
                int j = i + 1;
                ids[i] = ids[j];
                ratings[i] = ratings[j];
            }
        }
    }

    /**
     * Gets cached rating.  If id does not exist, returns -999.
     * @param id
     * @return
     */
    public float getRating(int id) {
        int index = binarySearch(id, 0, length);
        if (index < length && ids[index] == id) {
            return ratings[index];
        } else
            return -999;
    }

    /**
     * Resizes arrays if necessary
     */
    private void checkSize() {
        if (ids.length >= length) {
            int temp[] = new int[length + incrementSize];
            float tempD[] = new float[length + incrementSize];
            for (int i = 0; i < length; i++) {
                temp[i] = ids[i];
                tempD[i] = ratings[i];
            }
            ids = temp;
            ratings = tempD;
        }
    }

    /**
     * Searches ids for key
     * @param key
     * @param first
     * @param last
     * @return index of key or position where key would be inserted
     */
    private int binarySearch(int key, int first, int last) {
        int middle;
        int position;

        while (first != last) {
            middle = (first + last) / 2;
            if (key == ids[middle]) return middle;
            if (key < ids[middle])
                last = middle;
            else
                first = middle + 1;

        }
        return first;
    }

    /**
     * Standard quicksort.  Used to sort ids and ratings by ids
     * @param low
     * @param high
     */
    private void quicksort(int low, int high) {
        if (low < high) {
            int middle = partition(low, high);
            quicksort(low, middle - 1);
            quicksort(middle + 1, high);
        }
    }

    /**
     * Partitioning part of quicksort algorithm
     * @see private void quicksort
     * @param low
     * @param high
     * @return  pivotLocation
     */
    private int partition(int low, int high) {
        int pivot = ids[high];
        int smallSide = low;
        int largeSide = high - 1;

        while (smallSide < largeSide) {
            while (ids[smallSide] < pivot)
                smallSide++;
            while (largeSide >= 0 && pivot <= ids[largeSide])
                largeSide--;
            if (smallSide < largeSide) {
                int temp = ids[smallSide];
                ids[smallSide] = ids[largeSide];
                ids[largeSide] = temp;
                float tempD = ratings[smallSide];
                ratings[smallSide] = ratings[largeSide];
                ratings[largeSide] = tempD;
            }
        }

        int pivotLocation;
        if (pivot < ids[smallSide]) {
            int temp = ids[smallSide];
            ids[smallSide] = ids[high];
            ids[high] = temp;
            float tempD = ratings[smallSide];
            ratings[smallSide] = ratings[high];
            ratings[high] = tempD;
            pivotLocation = smallSide;
        } else {
            pivotLocation = high;
        }
        return pivotLocation;
    }

    /**
     * Array of ids.
     *
     * **Note:  to get the valid length of the array use RatingsArrays.length
     * @see public int getLength()
     * @return ids
     */
    public int[] getIds() {
        return ids;
    }

    /**
     * Array of ratings.
     *
     * **Note:  to get the valid length of the array use RatingsArrays.length
     * @see public int getLength()
     * @return ratings
     */
    public float[] getRatings() {
        return ratings;
    }

    /**
     * Returns the total number of valid elements in the array.
     * May be used similar to Array.length
     *
     * @return length
     */
    public int getLength() {
        return length;
    }

    /**
     * Size the arrays will increase once they reach threshold
     * @return
     */
    public int getIncrementSize() {
        return incrementSize;
    }

    /**
     * Size the arrays will increase once they reach threshold
     * @param incrementSize
     */
    public void setIncrementSize(int incrementSize) {
        this.incrementSize = incrementSize;
    }

    /**
     * Prints out the arrays.
     * @return
     */
    public String toString() {
        String retval = "Ratings array has the following ids:ratings\n";
        for (int i = 0; i < length; i++) {
            retval = retval + ids[i] + ":" + ratings[i] + "\n";
        }
        return retval + "\n";
    }

    /**
     * Builds a linkedlist of ItemRatings corresponding to values in the array
     * @return
     */
    public LinkedList getRatingsList() {
        LinkedList ratingsList = new LinkedList();
        for (int i = 0; i < length; i++) {
            ratingsList.add(new ItemRating(0, ids[i], ratings[i]));
        }
        return ratingsList;
    }

    /**
     * Returns an array of itemratings corresponding to the values cached;
     * @return ir
     */
    public ItemRating[] getItemRatingsArray() {
        ItemRating[] ir = new ItemRating[length];
        for (int i = 0; i < length; i++) {
            ir[i] = new ItemRating(0, ids[i], ratings[i]);
        }
        return ir;
    }

    /**
     * Returns if array has key
     * @param id
     * @return
     */
    public boolean containsKey(int id) {
        int index = binarySearch(id, 0, length);
        if (index >= length) return false;
        if (ids[index] == id) {
            return true;
        } else
            return false;
    }
}
