/*
 * UserInfo.java
 *
 * Version 1.1
 *
 * 15/08/2002
 *
 * Copyright (c) 2003 Oregon State Universtity. All rights reserved.
 */
package org.recommender.server;

/**
 * This class implements a user's records, including his/her rartings
 * list, mean rating and Standard deviation.
 * The class also provides some flags that are used by the
 * cache replacement algorithm.
 *
 * @author  <A HREF="mailto:godde@cs.orst.edu">Olivier Godde</A>
 */

public class UserInfo extends CFSynchronize{
    public int userID;

    // samplingFlag doesn't have to be synchronized. If we load or unload a user by mistake, it will
    // simply be fixed the next time the sample thread runs.
    public boolean samplingFlag;  // True means this user has been selected to reside into the cache by using a sampling methode

    // reference doesn't need to be synced - it's just a hint to the cache manager
    private boolean referenced;    // Set at each reference to this user data as an active user

    // doesn't need to be synced. Worst case, we rerun a topN computation
    public ItemPrediction[] userTopN;  // Array of TopN ItemPrediction for that user (Server side TopN cache)

    // The following fields are synchronized because we want
    // ratings to be in sync with userMean and userSD.
    private RatingArrays ratings;
    private float userMean;
    private float userSD;
    private int userAddedRatings;          // Counter of added ratings for this user.
    private int useCount = 1;       // # of CFEngine requests on this user is currently in progress.
    private float minRating = SystemProperties.getFloat("CFServer.db.minRating", 0);

//	private int readers = 0;
//	private int writers = 1;
	
	
    /**
     * Constructor with param of only userID.
     * @param ID
     */
    public UserInfo(int ID) {
        super();
        userID = ID;
        setRatings(new RatingArrays(20));
        setUserMean(minRating - 1000);
        setUserSD(minRating - 1000);
        samplingFlag = false;
        referenced = false;
        userTopN = new ItemPrediction[0];
        userAddedRatings = 0;
    }

    /**
     * Constructor with params of userID, ratings, mean and standard deviation.
     * @param ID
     * @param ratings
     * @param mean
     * @param SD
     */
    public UserInfo(int ID, RatingArrays ratings, float mean, float SD) {
        super();
        userID = ID;
        this.setRatings(ratings);
        setUserMean(mean);
        setUserSD(SD);
        samplingFlag = false;
        referenced = false;
        userTopN = new ItemPrediction[0];
        userAddedRatings = 0;
    }

    /**
     * Constructor with params of userID, ratings, mean and standard deviation.
     * @param ID
     * @param ratings
     * @param mean
     * @param SD
     * @param sampling
     */
    public UserInfo(int ID, RatingArrays ratings, float mean, float SD, boolean sampling) {
        super();
        userID = ID;
        this.setRatings(ratings);
        setUserMean(mean);
        setUserSD(SD);
        samplingFlag = sampling;
        referenced = false;
        userTopN = new ItemPrediction[0];
        userAddedRatings = 0;
    }

    public synchronized int getUseCount() {
        return useCount;
    }
    public synchronized void setUseCount(int useCount) {
        this.useCount = useCount;
    }
    public synchronized void incUseCount() {
        this.useCount++;
    }
    public synchronized void decUseCount() {
        this.useCount--;
    }

    public boolean isReferenced() {
        // reference is just a hint as to whether a user should remain
        // in the cache. So it wouldn't be a disaster if we read an out-of-date
        // value from the cache. No synchornization needed
        return referenced;
    }

    public void setReferenced(boolean referenced) {
        // reference is just a hint as to whether a user should remain
        // in the cache. So it wouldn't be a disaster if we read an out-of-date
        // value from the cache. No synchornization needed
        this.referenced = referenced;
    }

    public synchronized int getUserAddedRatings() {
        return userAddedRatings;
    }
    public synchronized void setUserAddedRatings(int userAddedRatings) {
        this.userAddedRatings = userAddedRatings;
    }
    public synchronized void incUserAddedRatings() {
        userAddedRatings++;
    }
    public synchronized void decUserAddedRatings() {
        userAddedRatings--;
    }

    public synchronized float getUserMean() {
        return userMean;
    }

    public synchronized void setUserMean(float userMean) {
        this.userMean = userMean;
    }

    public synchronized float getUserSD() {
        return userSD;
    }

    public synchronized void setUserSD(float userSD) {
        this.userSD = userSD;
    }

    public synchronized RatingArrays getRatings() {
        return ratings;
    }
    public synchronized void setRatings(RatingArrays ratings) {
        this.ratings = ratings;
    }
    
   
}