package org.recommender.test;

import org.recommender.server.*;

import java.net.*;
import java.rmi.*;
import java.lang.Thread;
import java.util.LinkedList;
import java.util.Iterator;

import org.recommender.test.*;

/**
 * Class to test the performance of the recommender engine
 */
public class PerformanceTest {
    private CFEngine cfengine;

    /**
     * Constructor - Initializes a connection to the cfengine, and spawns the necessary
     * threads.
     */
    public PerformanceTest() {
        // We could load the location of the RMI server from the properties files, but in
        // truth, clients shouldn't know about that.
        
        String installdir = System.getProperty("cf.path",".");
		String filename = installdir + "/lib/"
                        + "CFServer.properties";
        String serviceName = SystemProperties.getString("CFServer.rmi.objectname");
        String rmiNameserver = SystemProperties.getString("CFServer.rmi.nameserver");
        String url = "rmi://" + rmiNameserver + "/" + serviceName;
        try {
            cfengine = (CFEngine) Naming.lookup(url);
        } catch (NotBoundException e) {
            System.err.println("No such service " + serviceName + " registered with RMI on " + rmiNameserver);
            e.printStackTrace();
            fatalError();
        } catch (MalformedURLException e) {
            System.err.println("URL is malformed: " + url);
            e.printStackTrace();
            fatalError();
        } catch (RemoteException e) {
            e.printStackTrace();
            fatalError();
        }

        /**
         * Connection to engine exists, now create some client threads.
         */

        // Here we are assuming users 1-500 are loaded initially, and
        // then we request predictions for users above that.
        LinkedList threadList = new LinkedList();
        int numTopNsPerThread = 500;
        int numTopNThreads = 1;

        for (int i = 0; i < numTopNThreads; i++) {
            Thread t = new RandomTopNThread(cfengine, "TopN Thread" + i, 501, 940, numTopNsPerThread);
            threadList.add(t);
        }

        int numRatingsPerThread = 500;
        int numRatingsThreads = 0;
         for (int i = 0; i < numRatingsThreads; i++) {
            Thread t = new RandomRateThread(cfengine, "Rating Thread " + i, 1, 940, 1, 1000, numRatingsPerThread);
            threadList.add(t);
        }

        Iterator iter = threadList.iterator();

        long startTime = System.currentTimeMillis();

        // Starts the new thread running
        while (iter.hasNext()) {
            Thread thread = (Thread) iter.next();
            thread.start();
        }

        // Wait for threads to end, then exit.
        // Probably isn't really necessary now, but later if we want to
        // do some post-processing.
        boolean complete = false;
        do {
            try {
                iter = threadList.iterator();
                while (iter.hasNext()) {
                    Thread thread = (Thread) iter.next();
                    thread.join();
                }
                complete = true;
            } catch (InterruptedException e) {
                e.printStackTrace();
                fatalError();
            }
        } while (! complete);
        long finishTime = System.currentTimeMillis();

        System.out.println("TopN Rate Across All threads: "
                + (1000 * (numTopNsPerThread * numTopNThreads)) / (finishTime - startTime) + " per second");

        System.out.println("CFEngine Statistics");

        try {
            System.out.println(cfengine.getStatistics());
        } catch (RemoteException e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

    /**
     * Called whenever a fatal error occurs
     */
    private void fatalError() {
        System.exit(1);
    }

    /**
     * The main function of the PerformanceTest class. Starts the testing threads running.
     *
     * @param args Current ignored.
     */
    public static void main(String[] args) {
       PerformanceTest pt = new PerformanceTest();
    }
}
